/************************************************************************/
/* Module containing functions for manipulating distribution laws.	*/
/* Provides:								*/
/*	void	Write_Law( FILE *Stream, Law_Desc Loi, int Mode );	*/
/*	double	Mean( Law_Desc Loi );					*/
/*	double	Moment( Law_Desc Loi, int Order );			*/
/*	double	Laplace( Law_Desc Loi, double s );			*/
/*	double	DLaplace( Law_Desc Loi, double s );			*/
/*	Law_Desc New_Law( void );					*/
/*	Law_Desc Copy_Law( law_Desc Loi );				*/
/*	Law_Desc Rescale_Law( Law_Desc Loi, double Factor );		*/
/*      Law_Desc Parse_Law( char Nom, Liste_Reel Params, 		*/
/*			Liste_Reel Params2, Liste_Loi The_SubLaws );	*/
/*	Liste_Loi Append_Law( Liste_Loi	Law_List, Law_Desc New_Law );	*/
/*	void Free_Law( Law_Desc The_Law );				*/
/************************************************************************/
#define	TOOLNAME	"LawPackage"

#include <malloc.h>
#include <math.h>
#include "law.h"
#include "messages.h"
#ifdef SUN4SOL2
#define M_SQRT1_2   0.70710678118654752440  /* 1/sqrt(2) */
#endif
#ifdef SUNMP
#define M_SQRT1_2   0.70710678118654752440  /* 1/sqrt(2) */
#endif

#define SCAN(x,y)	for(x=y;x!=NULL;x=x->Next)
#define EXPAND(ptr,nb)	ptr = (double*) realloc( ptr, (size_t)(nb)*sizeof(double) )

/* external declarations */
double erfc(double x);
double rint(double x);
double lgamma(double x);

/* Constructor. The parameter table is initialized with room for one. */
Law_Desc New_Law(void) {

  Law_Desc The_Law;

  The_Law = (Law_Desc) malloc( sizeof( struct LAW_DESC ) );
  The_Law->Parameters = (double*) malloc( sizeof(double) );
  The_Law->Law_Parameters = (Law_Desc*) NULL;

  return The_Law;
}

/* Procedure to write the law. Accepts various styles.			*/
void Write_Law( FILE *OutFile, 
		Law_Desc The_Law, 
		int Mode )
{
  int	Nb_Params;
  int	i;


  switch( The_Law->Name ) {
  case 'E':
  case 'C':
  case 'B':
  case 'g':
  case 'G':
    fprintf( OutFile, "%c %8.4f ", The_Law->Name, The_Law->Parameters[0] );
    break;
  case 'U':
  case 'W':
  case 'N':
  case 'L':
  case 'p':
    fprintf( OutFile, "%c %8.4f %8.4f ", 
	     The_Law->Name,
	     The_Law->Parameters[0],
	     The_Law->Parameters[1] );
    break;
  case 'R':
    fprintf( OutFile, "%c %4.0f %8.4f ", The_Law->Name,
	     The_Law->Parameters[0],
	     The_Law->Parameters[1] );
    break;
  case 'O':
    fprintf( OutFile, "%c %8.4f ", The_Law->Name, The_Law->Parameters[0] );
    fprintf( OutFile, "%8.4f %8.4f ",
	     The_Law->Parameters[1],
	     The_Law->Parameters[2] );
    break;
  case 'H':
    fprintf( OutFile, "%c ", The_Law->Name );
    for ( i=0; i<The_Law->Parameters[0]; i+=2 ) {
      fprintf( OutFile, "%8.4f %8.4f",
	       The_Law->Parameters[i+1],
	       The_Law->Parameters[i+2] );
    }
    break;
  case 'D':
    fprintf( OutFile, "%c [ ", The_Law->Name );
    Nb_Params = (int)rint(The_Law->Parameters[0]);
    for ( i=1; i<= Nb_Params; i++ )
      fprintf( OutFile, "%8.4f ", The_Law->Parameters[i] );
    fprintf( OutFile, "] [ ");
    for ( i=1; i<= Nb_Params; i++ )
      fprintf( OutFile, "%8.4f ", The_Law->Parameters[i+Nb_Params] );
    fprintf( OutFile, "] ");
    break;
  case 'P':
    fprintf( OutFile, "IPP [ " );
    fprintf( OutFile, "%f, ", The_Law->Parameters[0] );
    Write_Law( OutFile, The_Law->Law_Parameters[0], Mode );
    fprintf( OutFile, ", ");
    Write_Law( OutFile, The_Law->Law_Parameters[1], Mode );
    fprintf( OutFile, " ] ");
    break;
  default:
    fprintf( stderr, "%s: error (Print): unexpected law id (%c)\n",
	     TOOLNAME, The_Law->Name );
  }
}

double Mean(  Law_Desc The_Law )
{	
  double	The_Mean;
  int		i;
  int		Nb_Params;
  double	Alpha;
  double       	Mean_On;
  double       	Mean_Off;

  switch( The_Law->Name ) {
  case 'C':
  case 'E':
  case 'B':
  case 'O':
  case 'N':
  case 'L':
    The_Mean = The_Law->Parameters[0];
    break;
  case 'l':
    The_Mean = 2 * The_Law->Parameters[0];
    break;
  case 'g':
    The_Mean = The_Law->Parameters[0] / ( 1.0 - The_Law->Parameters[0] );
    break;
  case 'G':
    The_Mean = 1.0 / ( 1.0 - The_Law->Parameters[0] );
    break;
  case 'U':
    The_Mean = ( The_Law->Parameters[1] + The_Law->Parameters[0] ) / 2.0;
    break;
  case 'R':
    The_Mean = The_Law->Parameters[1];
    break;
  case 'p':
    if ( The_Law->Parameters[1] > 1.0 )
      The_Mean = The_Law->Parameters[0]/(The_Law->Parameters[1] - 1.0);
    else
      The_Mean = 1.0/0.0;
    break;
  case 'W':
    Alpha = 1.0 / The_Law->Parameters[1];
    The_Mean = Alpha * exp( lgamma( Alpha ) 
			    - Alpha * log( The_Law->Parameters[0] ) );
    break;
  case 'H':
    The_Mean = 0.0;
    for ( i=0; (i<The_Law->Parameters[0]); i+=2 )
      The_Mean += The_Law->Parameters[i+1]*The_Law->Parameters[i+2];
    break;
  case 'D':
    The_Mean = 0.0;
    Nb_Params = (int) (The_Law->Parameters[0]);
    for ( i=1; i <= Nb_Params; i++ )
      The_Mean += The_Law->Parameters[i]*The_Law->Parameters[i+Nb_Params];
    break;
  case 'P':
    Mean_On = Mean( The_Law->Law_Parameters[0] );
    Mean_Off = Mean( The_Law->Law_Parameters[1] );
    The_Mean = The_Law->Parameters[0] * Mean_On / ( Mean_On + Mean_Off );
    break;
  default:
    fprintf( stderr, "%s: warning: cannot compute mean of law `%c`.",
	     TOOLNAME, The_Law->Name );
    fprintf( stderr, " Zero assumed.\n");
    The_Mean = 0.0;
  }

  return	The_Mean;
}

double Moment(
	      Law_Desc		The_Law,
	      int		Order )
{	
  double	The_Moment;
  int		i;
  int		Nb_Params;
  double	*Param;

  Param = The_Law->Parameters;

  if ( Order == 0 )
    return 1.0;
  else if ( Order == 1 )
    return Mean( The_Law );

  /* from here on, Order >= 2 */
  switch( The_Law->Name ) {
  case 'C':
    The_Moment = pow( Param[0], Order );
    break;
  case 'E':
    The_Moment = pow( Param[0], Order );
    for( i=1; i<=Order; i++ )
      The_Moment *= (double)i;
    break;
  case 'B':
    The_Moment = Param[0];
    break;
  case 'g':
    if ( Order == 2 )
      The_Moment = Param[0] * ( 1.0 + Param[0] ) 
	/ ( 1.0 - Param[0] ) / ( 1.0 - Param[0] );
    else {
      fprintf( stderr, "%s: warning: cannot compute moment of order > 1",
	       TOOLNAME );
      fprintf( stderr, "for law `%c`. 0.0 assumed.\n", The_Law->Name );
    }
    break;
  case 'G':
    if ( Order == 2 )
      The_Moment = ( 1.0 + Param[0] ) / ( 1.0 - Param[0] );
    else {
      fprintf( stderr, "%s: warning: cannot compute moment of order > 1",
	       TOOLNAME );
      fprintf( stderr, "for law `%c`. 0.0 assumed.\n", The_Law->Name );
    }
    break;
  case 'U':
    The_Moment = pow( Param[1], 1+Order ) - pow( Param[0], 1+Order );
    The_Moment /= ( Param[1] - Param[0] );
    The_Moment /= ( 1.0 + (double)Order );
    break;
  case 'R':
    The_Moment = pow( Param[1], Order );
    for( i=1; i<=Order-1; i++ )
      The_Moment *= (Param[0]+(double)i)/Param[0];
    break;
  case 'H':
    The_Moment = 0.0;
    for ( i=0; (i<Param[0]); i+=2 )
      The_Moment += Param[i+1] * pow( Param[i+2], Order );
    for( i=1; i<=Order; i++ )
      The_Moment *= (double)i;
    break;
  case 'D':
    The_Moment = 0.0;
    Nb_Params = (int) (Param[0]);
    for ( i=1; i <= Nb_Params; i++ )
      The_Moment += Param[i+Nb_Params] * pow(Param[i],Order);
    break;
  default:
    fprintf( stderr, "%s: warning: cannot compute %d-th moment of law `%c`.",
	     TOOLNAME, Order, The_Law->Name );
    fprintf( stderr, " Zero assumed.\n");
    The_Moment = 0.0;
  }

  return	The_Moment;
}

double Laplace( 
	       Law_Desc	The_Law,
	       double	s )
{	
  double	The_Value;
  int		i;
  int		Nb_Params;
  double	*Param;

  Param = The_Law->Parameters;

  switch( The_Law->Name ) {
  case 'C':
    The_Value = exp( -s * Param[0] );
    break;
  case 'E':
    The_Value =  1.0 / ( 1.0 + s * Param[0] );
    break;    
  case 'B':
    The_Value = 1.0 - ( 1.0 - exp( -s ) ) * Param[0];
    break;
  case 'g':
    The_Value = ( 1.0 - Param[0] ) / ( 1.0 - Param[0] * exp( -s ) );
    break;
  case 'G':
    The_Value = ( 1.0 - Param[0] ) * exp( -s )
      / ( 1.0 - Param[0] * exp( -s ) );
    break;
  case 'U':
    The_Value = ( s == 0.0 ? 1.0 :
		  ( exp( -s * Param[0] ) - exp( -s * Param[1] ) )
		  / s / ( Param[1] - Param[0] ) );
    break;
  case 'R':
    The_Value = Param[0] / ( s*Param[1] + Param[0] );
    The_Value = exp( Param[0] * log( The_Value ) );
    break;
  case 'H':
    The_Value =  0.0;
    for( i=0; i<Param[0]; i+=2 )
      The_Value += Param[i+1] / ( 1.0 + s * Param[i+2] );
    break;
  case 'D':
    The_Value = 0.0;
    Nb_Params = (int) (Param[0]);
    for ( i=1; i <= Nb_Params; i++ ) {
      The_Value += exp( -s * Param[i] )	* Param[i+Nb_Params];
    }
    break;
  default:
    fprintf( stderr, "%s: warning: cannot compute Laplace transform ",
	     TOOLNAME );
    fprintf( stderr, "of law `%c`. 1.0 assumed.\n", 
	     The_Law->Name );
    The_Value = 1.0;
  }

  return	The_Value;
}

double DLaplace( 
	       Law_Desc	The_Law,
	       double	s )
{	
  double	The_Value;
  int		i;
  int		Nb_Params;
  double	*Param;

  Param = The_Law->Parameters;

  switch( The_Law->Name ) {
  case 'C':
    The_Value = - Param[0] * exp( -s * Param[0] );
    break;
  case 'E':
    The_Value =  - Param[0] / ( 1.0 + s * Param[0] ) /
      ( 1.0 + s * Param[0] );
    break;    
  case 'B':
    The_Value = - exp( -s ) * Param[0];
    break;
  case 'g':
    The_Value = - ( 1.0 - Param[0] ) * Param[0] * exp( -s )
      / ( 1.0 - Param[0] * exp( -s ) )
      / ( 1.0 - Param[0] * exp( -s ) );
    break;
  case 'G':
    The_Value = - ( 1.0 - Param[0] ) * exp( -s )
      / ( 1.0 - Param[0] * exp( -s ) )
      / ( 1.0 - Param[0] * exp( -s ) );
    break;
  case 'U':
    The_Value = ( s == 0.0 ?
		  ( Param[1] - Param[0] ) / 2.0 :
		  ( 1.0 + s * Param[0] ) * exp( -s * Param[0] ) - 
		  ( 1.0 + s * Param[1] ) * exp( -s * Param[1] ) / s / s /
		  ( Param[1] - Param[0] ) );
    break;
  case 'R':
    The_Value =  Param[0] / ( s*Param[1] + Param[0] );
    The_Value = - Param[0]*Param[1] * exp( Param[0] * log( The_Value ) )
      / ( s*Param[1] + Param[0] );
    break;
  case 'H':
    The_Value =  0.0;
    for( i=0; i<Param[0]; i+=2 )
      The_Value -= Param[i+1]*Param[i+2]
	/ (1.0 + s * Param[i+2] ) / (1.0 + s * Param[i+2] );
    break;
  case 'D':
    The_Value = 0.0;
    Nb_Params = (int) (Param[0]);
    for ( i=1; i <= Nb_Params; i++ ) {
      The_Value -= Param[i] * exp( -s * Param[i] ) * Param[i+Nb_Params];
    }
    break;
  default:
    fprintf( stderr, "%s: warning: cannot compute Laplace transform ",
	     TOOLNAME );
    fprintf( stderr, "of law `%c`. 1.0 assumed.\n", 
	     The_Law->Name );
    The_Value = 1.0;
  }

  return	The_Value;
}

/* The CDF of a v.a. X is the function \bar F(x) = P( X > x ) */
double CDF( Law_Desc	The_Law,
	    double	x )
{	
  double	The_Value;
  int		i;
  int		Nb_Params;
  double	m;
  double	s2;
  double	*Param;

  Param = The_Law->Parameters;

  switch( The_Law->Name ) {
  case 'C':
    The_Value = ( x > Param[0] ? 0.0 : 1.0 );
    break;
  case 'E':
    if ( x < 0.0 )
      The_Value = 1.0;
    else {
      if ( Param[0] == 0.0 ) 
	The_Value = 0.0;
      else 
	The_Value = exp( -x / Param[0] );
    }
    break;    
    /* lazy... will do it later ...
       case 'B':
       The_Value = 1.0 - ( 1.0 - exp( -s ) ) * Param[0];
       break;
       case 'g':
       The_Value = ( 1.0 - Param[0] ) / ( 1.0 - Param[0] * exp( -s ) );
       break;
       case 'G':
       The_Value = ( 1.0 - Param[0] ) * exp( -s )
       / ( 1.0 - Param[0] * exp( -s ) );
       break;
    */
  case 'U':
    if ( x <= Param[0] )
      The_Value = 1.0;
    else 
      The_Value = ( x > Param[1] ? 0.0 :
		    ( Param[1] - x ) / ( Param[1] - Param[0] ) );
    break;
    /* laaaaazy
       case 'R':
       The_Value = Param[0] / ( s*Param[1] + Param[0] );
       The_Value = exp( Param[0] * log( The_Value ) );
       break;
    */
  case 'H':
    if ( x < 0.0 )
      The_Value = 1.0;
    else {
      The_Value =  0.0;
      for( i=0; i<Param[0]; i+=2 ) {
	if ( Param[i+2] > 0.0 ) 
	  The_Value += Param[i+1] * exp( - x * Param[i+2] );
      }
    }
    break;
  case 'D':
    The_Value = 0.0;
    Nb_Params = (int) (Param[0]);
    for ( i=1; i <= Nb_Params; i++ ) {
      if ( x > Param[i] )
	The_Value += Param[i+Nb_Params];
    }
    break;
  case 'W':
    if ( x < 0.0 )
      The_Value = 1.0;
    else 
      The_Value = exp( - Param[0] * pow( x, Param[1] ) );
    break;
  case 'p':
    if ( x < 0.0 )
      The_Value = 1.0;
    else
      The_Value = pow( Param[0]/(Param[0]+x), Param[1] );
    break;
  case 'N':
    The_Value = erfc( ( x - Param[0] ) / Param[1] * M_SQRT1_2 ) / 2.0;
    break;
  case 'L':
    if ( x <= 0.0 )
      The_Value = 1.0;
    else {
      s2 = log( 1.0 + ( Param[1] / Param[0] )*( Param[1] / Param[0] ) );
      m = log( Param[0] ) - s2 / 2;
      The_Value = erfc( ( log(x) - m ) / sqrt( 2 * s2 ) ) / 2.0;
    }
    break;
  default:
    fprintf( stderr, "%s: warning: cannot compute CDF ",
	     TOOLNAME );
    fprintf( stderr, "of law `%c`. 0.0 assumed.\n", 
	     The_Law->Name );
    The_Value = 1.0;
  }

  return	The_Value;
}

/* A procedure to rescale, a law. Creates a new one.		*/ 
Law_Desc Rescale_Law( Law_Desc The_Law, double Factor )
{	
  int		i;
  int		Nb_Params;
  Law_Desc	Res;

  Res = New_Law();

  if ( NULL != Res ) {
    Res->Name = The_Law->Name;

    /* a first pass to allocate space */
    switch( The_Law->Name ) {
    case 'C':
    case 'E':
    case 'B':
    case 'G':
    case 'g':
      /* No need to alloc space for 1 parameter. */
      break;
    case 'R':
    case 'U':
    case 'W':
    case 'N':
    case 'L':
    case 'p':
      EXPAND( Res->Parameters, 2 );
      break;
    case 'O':
      EXPAND( Res->Parameters, 3 );
      break;
    case 'H':
      Nb_Params = (int)rint(The_Law->Parameters[0]);
      EXPAND( Res->Parameters, 1+Nb_Params );
      break;
    case 'D':
      Nb_Params = (int)rint(The_Law->Parameters[0]);
      EXPAND( Res->Parameters, 1+2*Nb_Params );
      break;
    case 'P':
      EXPAND( Res->Parameters, 2 );
      Res->Law_Parameters = (Law_Desc*) calloc( 2, sizeof( Law_Desc ) );
      break;
    default:
      fprintf( stderr, "%s: error: unknown law `%c`. No rescaling.\n",
	       TOOLNAME, The_Law->Name );
      free( Res );
      Res = (Law_Desc)NULL;
    }
  }
  else {
    fprintf( stderr, "%s: error: could not allocate memory for law.\n",
	     TOOLNAME );
    Res = (Law_Desc)NULL;
  }

  if ( Res != NULL ) {
    /* second pass to fill in parameters */
    switch( The_Law->Name ) {
    case 'C':
    case 'E':
      Res->Parameters[0] = Factor * The_Law->Parameters[0];
      break;
    case 'O':
      Res->Parameters[0] = Factor * The_Law->Parameters[0];
      Res->Parameters[1] = Factor * The_Law->Parameters[1];
      Res->Parameters[2] = Factor * The_Law->Parameters[2];
      break;
    case 'g':
    case 'G':
      fprintf( stderr, "%s: warning: geometric distributions cannot",
	       TOOLNAME );
      fprintf( stderr, "be rescaled. No change.\n");
      Res->Parameters[0] = The_Law->Parameters[0];
      break;
    case 'U':
      Res->Parameters[0] = Factor * The_Law->Parameters[0];
      Res->Parameters[1] = Factor * The_Law->Parameters[1];
      break;
    case 'p':
      Res->Parameters[0] = Factor * The_Law->Parameters[0];
      Res->Parameters[1] = The_Law->Parameters[1];
      break;
    case 'R':
      Res->Parameters[0] = The_Law->Parameters[0];
      Res->Parameters[1] = Factor * The_Law->Parameters[1];
      break;
    case 'W':
      Res->Parameters[0] = The_Law->Parameters[0] 
	* pow( Factor, The_Law->Parameters[1] );
      Res->Parameters[1] = The_Law->Parameters[1];
      break;
    case 'H':
      Res->Parameters[0] = The_Law->Parameters[0];
      for ( i=0; (i<The_Law->Parameters[0]); i+=2 ) {
	Res->Parameters[i+1] = The_Law->Parameters[i+1];
	Res->Parameters[i+2] = Factor * The_Law->Parameters[i+2];
      }
      break;
    case 'D':
      Res->Parameters[0] = The_Law->Parameters[0];
      Nb_Params = (int)rint(The_Law->Parameters[0]);
      for ( i=1; i <= Nb_Params; i++ ) {
	Res->Parameters[i] = Factor * The_Law->Parameters[i];
	Res->Parameters[i+Nb_Params] = The_Law->Parameters[i+Nb_Params];
      }
      break;
    case 'P':
      Res->Parameters[0] = Factor * The_Law->Parameters[0];
      Res->Law_Parameters[0] = Rescale_Law( The_Law->Law_Parameters[0],
					    Factor );
      Res->Law_Parameters[1] = Rescale_Law( The_Law->Law_Parameters[1],
					    Factor );
      break;
    default:
      fprintf( stderr, "%s: error: don't know how to rescale law `%c`.\n",
	       TOOLNAME, The_Law->Name );
      free( Res );
      Res = (Law_Desc)NULL;
    }
  }

  return Res;
}

/* A procedure to copy a law. Uses "rescale" with factor 1.0!		*/ 
Law_Desc Copy_Law( Law_Desc The_Law )
{	
  Law_Desc Res;

  /* Checking that the law is "registered". This is only to preempt	*/
  /* the warning message of Rescale_Law, which would sound cryptic...	*/
  switch( The_Law->Name ) {
  case 'C':
  case 'E':
  case 'B':
  case 'G':
  case 'g':
  case 'p':
  case 'R':
  case 'U':
  case 'W':
  case 'O':
  case 'H':
  case 'D':
  case 'P':
  case 'N':
  case 'L':
    Res = Rescale_Law( The_Law, 1.0 );
    break;
  default:
    fprintf( stderr, "%s: error: unknown law `%c`. No copy.\n",
	     TOOLNAME, The_Law->Name );
    free( Res );
    Res = (Law_Desc)NULL;
  }

  return Res;
}

Law_Desc Parse_Law( 
		   char		Nom, 
		   Liste_Reel 	The_Params,
		   Liste_Reel 	The_Params2,
		   Liste_Loi	The_SubLaws )
{	
  int		Nb_Params;
  int		i;
  Law_Desc	The_Law;
  Liste_Reel	R_Scan;
  Liste_Loi	LL_Scan;

  The_Law = New_Law();

  if ( The_Law != NULL )
    switch( The_Law->Name = Nom ) {
    case 'E':
    case 'C':
    case 'B':
    case 'g':
    case 'G':
      /* Space for one parameter is already available from New_Law() */
      if ( The_Params != NULL ) {
	The_Law->Parameters[0] = The_Params->Val;
	if ( The_Params->Next != NULL )
	  W_EXT( The_Law->Name );
      }
      else {
	W_MIS( The_Law->Name );
	The_Law->Parameters[0] = 0.0;
      }
      break;
    case 'U':
    case 'W':
    case 'L':
    case 'p':
    case 'h':
      /* 'h' is the "old" QNAP way of defining hyperexponentials by	*/
      /* mean and variance or coefficient of variation. Inactive at the	*/
      /* moment.							*/
      EXPAND( The_Law->Parameters, 2 );
      if ( ( The_Params != NULL ) && ( The_Params->Next != NULL ) ) {
	The_Law->Parameters[1] = The_Params->Val;
	The_Law->Parameters[0] = The_Params->Next->Val;
	if ( The_Params->Next->Next != NULL )
	  W_EXT( The_Law->Name );
      }
      else {
	W_MIS( The_Law->Name );
	The_Law->Parameters[1] = The_Law->Parameters[0] = 0.0;
      }
      break;
    case 'H':
      /* first count the parameters */
      Nb_Params = 0;
      SCAN( R_Scan, The_Params )
	Nb_Params++;
      
      if ( Nb_Params % 2 != 0 ) {
	printf( "Warning: odd number of parameters for HEXP. Last ignored.\n");
	Nb_Params--;
      }
      if ( Nb_Params == 0 ) {
	printf( "Error: not enough parameters for HEXP.\n");
      }
      else {
	/* one has to add a cell to store nb of parameters */
	EXPAND( The_Law->Parameters, 1+Nb_Params );
	The_Law->Parameters[0] = Nb_Params;
	/* store the parameters, reversing the list! */
	SCAN( R_Scan, The_Params )
	  The_Law->Parameters[Nb_Params--] = R_Scan->Val;
      }
      break;
    case 'O':
      EXPAND( The_Law->Parameters, 3 );
      if ( ( The_Params != NULL ) 
	   && ( The_Params->Next != NULL )
	   && ( The_Params->Next->Next != NULL ) ) {
	The_Law->Parameters[2] = The_Params->Val;		/* 1/beta */
	The_Law->Parameters[1] = The_Params->Next->Val;		/* 1/alpha */
	The_Law->Parameters[0] = The_Params->Next->Next->Val;	/* 1/lambda */
	if ( The_Params->Next->Next->Next != NULL )
	  W_EXT( The_Law->Name );
      }
      else {
	W_MIS( The_Law->Name );
      }
      break;
    case 'R': /* differe de U dans l'ordre d'empilement des parametres */
      EXPAND( The_Law->Parameters, 2 );
      if ( ( The_Params != NULL ) && ( The_Params->Next != NULL ) ) {
	The_Law->Parameters[0] = The_Params->Val;
	The_Law->Parameters[1] = The_Params->Next->Val;
	if ( The_Params->Next->Next != NULL )
	  W_EXT( The_Law->Name );
      }
      else {
	W_MIS( The_Law->Name );
	The_Law->Parameters[1] = The_Law->Parameters[0] = 0.0;
      }
      break;

    case 'D':
      /* find the length of the (first) parameter list */
      Nb_Params = 0;
      SCAN( R_Scan, The_Params )
	Nb_Params++;
      EXPAND( The_Law->Parameters, 1 + 2*Nb_Params );
      The_Law->Parameters[0] = Nb_Params;

      /* Store values of the discrete distribution */
      i = 0;
      SCAN( R_Scan, The_Params ) {
	i++;
	The_Law->Parameters[i] = R_Scan->Val;
      }
      /* Store probas of the distribution */
      for ( R_Scan = The_Params2; ( i < 2*Nb_Params ) && ( R_Scan != NULL );
	    R_Scan = R_Scan->Next ) {
	i++;
	The_Law->Parameters[i] = R_Scan->Val;
      }
      if ( R_Scan != NULL ) {
	printf("Warning! Probability list too long (ignored)\n" );
      }
      else if ( i < 2*Nb_Params ) {
	printf("Warning! Probability list too short. 0.0 assumed\n" );
	for( ; i <= 2*Nb_Params; i++ )
	  The_Law->Parameters[i] = 0.0;
      }

    break;
    case 'P':
      The_Law->Parameters[0] = The_Params->Val;
      Nb_Params = 0;
      SCAN( LL_Scan, The_SubLaws )
	Nb_Params++;
      The_Law->Law_Parameters = (Law_Desc*) calloc( (unsigned)Nb_Params,
						   sizeof( Law_Desc ) );
      i = 0;
      SCAN( LL_Scan, The_SubLaws ) {
	The_Law->Law_Parameters[i] = LL_Scan->Val;
	i++;
      }
      break;

    default:
      printf("Warning! Law '%c' ID not recognized\n", The_Law->Name);
    }

  Free_Real_List( The_Params );
  Free_Real_List( The_Params2 );

  return The_Law;
}

Liste_Loi Append_Law( Liste_Loi	Law_List, Law_Desc New_Law )
{
  Liste_Loi The_List;

  The_List = (Liste_Loi) malloc( sizeof( struct LAW_LIST ) );

  if ( The_List != NULL ) {
    The_List->Val = New_Law;
    The_List->Next = Law_List;
  }
  else {
    E_MEM( "Law list" );
  }

  return The_List;
}

void Free_Law( Law_Desc The_Law )
{
  if ( The_Law->Parameters != NULL )
    free( The_Law->Parameters );
  
  /* number of Law parameters depends on the type */
  if ( The_Law->Name == 'P' ) {
    Free_Law( The_Law->Law_Parameters[0] );
    Free_Law( The_Law->Law_Parameters[1] );
    free( The_Law->Law_Parameters );
  }
    
  free( The_Law );
}
