/* Module providing functions for the topological analysis of an Event	*/
/* graph, and the classification of its transitions in levels.		*/

#include "eg.h"

#define SCAN(x,y)	for(x=y;x!=NULL;x=x->Next)

#define max(x,y)        ( (x>y) ? x : y )

#ifdef USE_SP_MAT
typedef struct SP_BOOL_MAT {
  int Size;
  int *Nb_Elts;		/* For each i, nb of "outputs" */
  int **Elts;		/* For each i, table of the j's connected. */
} *SP_Bool_Mat;

#define BOOL_MAT	SP_Bool_Mat
#define ADJ( A, i, j, n )  Is_Adj( A, i, j )

#else

#define BOOL_MAT	boolean*
#define ADJ( A, i, j, n )	A[i*n+j]

#endif

#ifdef USE_SP_MAT
boolean Is_Adj( BOOL_MAT A, int From, int To ) 
{
  static int i;
  static boolean Res;

  Res = FALSE;
  for (i=0; (i<A->Nb_Elts[From]) && !Res; i++)
    Res = (A->Elts[From][i] == To);

  return Res;
}
#endif

void dumpe_adj(BOOL_MAT	A,int	n)
{
  int i;
  int j;

  for (i=0; i<n; i++) {
    for (j=0; j<n; j++)
      if ( ADJ(A,i,j,n) )
	printf("1");
      else
	printf("0");
    printf("\n");
  }
}

void print_cycle(int *Cycle)
{
  int i;

  for ( i=1 ; i <= Cycle[0]; i++ )
    printf("%1d -> ",Cycle[i]+1);

  printf("%1d\n", Cycle[1]+1 );
}

/* finds a cycle in graph A, knowing A is strongly connected	*/
/* Skips "recycles".						*/
/* returns a pointer to the cycle as an array of integers, the	*/
/*   first one being the size of the cycle			*/
int *Find_Cycle(BOOL_MAT	A,
		int		The_Size,
		boolean 	Skip_Recycles )
{
  int 		i;
  int		j;
  int		k;
  boolean	*Visited;
  boolean	*Present;
  int		*Outdeg;	/* matrix of outdegrees	*/
  int		Start;
  int		*Cycle;
  int		*Chemin;
  int		Lg;
  int		Dep;

  Visited = (boolean*) calloc( (unsigned)The_Size, sizeof( boolean ) );
  Present = (boolean*) calloc( (unsigned)The_Size, sizeof( boolean ) );
  Outdeg = (int*) calloc( (unsigned)The_Size, sizeof( int ) );
  Chemin = (int*) calloc( (unsigned)The_Size, sizeof( int ) );

  /* initialization. Mark nodes not visited, compute outdegree	*/
  /* and mark present if outdegree <> 0				*/
  for (i=0; i<The_Size; i++) {
    Visited[i] = FALSE;
    Outdeg[i] = 0;
    for (j=0; j<The_Size; j++)
      if ( ADJ(A,i,j,The_Size)
	   && ( ( i <= j ) || Present[j] )
	   && ( ( i != j ) || !Skip_Recycles ) )
	Outdeg[i]++;
    Present[i] = ( Outdeg[i] != 0 );
    }

  /* first pass: follow a path until visit to a previous node 	*/
  /* first find initial node with outdegree <> 0 			*/
  Dep = 0;
  Lg = 0;
  while ( Lg == 0 ) {	/* repeat until cycle was found */
    /* find next node with outdegree <> 0 */
    for ( ; Present[Dep] && ( Outdeg[Dep] == 0 ) 
	    && ( Dep < The_Size ); Dep++ );
    if ( Dep == The_Size ) {
      printf("Error: no cycle ?!?\n");
      Lg = -1;	/* to exit the loop */
    }

    i = Dep;

    while ( ! Visited[i] ) {
      Visited[i] = TRUE;
      Chemin[Lg] = i;
      Lg ++;
      for ( j = 0;
	    ( !Present[j] || ( Outdeg[j] == 0 ) ||
	      (Skip_Recycles && (j==i)) || !ADJ(A,i,j,The_Size) )
	      && ( j < The_Size );
	    j++ );
      if ( j == The_Size ) {	/* dead end: backtrack	*/
	Present[i] = FALSE;
	i = Chemin[Lg-1];
	Lg --;
      }
      else
	i=j;
    }
    if ( i == Dep ) {	/* we have explored all below Dep */
      Dep++;		/* we have to try another one */
    }
  }
  /* second pass: store the cycle					*/
  Start = i;
  Cycle	= (int*) calloc( (size_t)(Lg+1), sizeof( int ) ); 
  /* find the start of the cycle in the path	*/
  for ( i = 0; Chemin[i] != Start; i++);
  Cycle[1] = Start;
  k = i;
  for ( i++ ; Chemin[i] != Start; i++ ) {
    Cycle[i-k+1] = Chemin[i];
  }
  Cycle[0] = Lg-k;

  free( Visited );
  free( Present );
  free( Chemin );
  free( Outdeg );

  return Cycle;
}

/* extracts a layer structure from an adjacency matrix Adj, of size	*/
/* The_Size, and stores it in the structure Level.			*/ 
/* The space needed for the layer structure must have been allocated	*/
/* The number of layers found, or 0 if a cycle exists, is stored into	*/
/* the network structure.						*/
void Find_Layer_Structure(EventGraph	*Reseau,
			  BOOL_MAT	Adj,
			  int		The_Size )
{
  int		Layer;
  int		Left_Nodes;
  int		InDegree;
  boolean	Found_One;
  boolean	*Present;	/* matrice de booleens */
  Layer_Desc	Niveau;
  int		i;
  int		j;

  Present = (boolean*) calloc( (size_t)The_Size, sizeof( boolean ) );

  for (i=0; i<The_Size; i++) {
    Present[i] = TRUE;
    Reseau->Level[i] = NULL;
  }
	
  Left_Nodes = The_Size;
  Found_One = TRUE;
  for ( Layer = 0; (Left_Nodes > 0) && Found_One ; Layer++ ) {
    Reseau->Level[Layer] = (Layer_Desc) malloc( sizeof( struct LAYER_DESC ) );
    Niveau = Reseau->Level[Layer];
    Niveau->Nb_Nodes = 0;
    Niveau->Node_Id = (int*) calloc( (size_t)Left_Nodes, sizeof (int) );

    /* look for nodes without fathers */
    Found_One = FALSE;
    for (j=0; j<The_Size; j++) {
      if ( Present[j] ) {
	InDegree = 0;
	for (i=0; i<The_Size; i++)
	  if ( Present[i] && ADJ( Adj, i, j, The_Size ) )
	    InDegree++;
	if (InDegree == 0) {
	  /* node j has no fathers... */
	  Niveau->Node_Id[Niveau->Nb_Nodes] = j;
	  Niveau->Nb_Nodes++;
	  Found_One = TRUE;
	}
      }
    }

    if ( Found_One ) {
      for (i=0; i<Niveau->Nb_Nodes; i++)
	Present[Niveau->Node_Id[i]] = FALSE;
      Left_Nodes -= Niveau->Nb_Nodes;
    }
  }

  if ( Found_One )
    Reseau->Nb_Layers = Layer;
  else
    Reseau->Nb_Layers = 0;

  free( Present );
}

/****************************************************************/
/* Procedure to fill in the list of predecessors.		*/
/****************************************************************/
void	Make_Inputs_List( EventGraph	*Reseau )
{
  int		i;
  int		Dest_Id;
  Liste_Place	P_Scan;
  Liste_Place 	Pred_List;
    
  for( i=0; i<Reseau->Size; i++ )
    Reseau->T[i]->Indegree = 0;
    
  /* scan transition lists in search of places and store them	*/
  for( i=0; i<Reseau->Size; i++ ) {
	
    SCAN( P_Scan, Reseau->T[i]->Outputs ) {
      Dest_Id = P_Scan->Val->Trans_Id-1;

      Pred_List = (Liste_Place) malloc( sizeof( struct PLACE_LIST ) );
      Pred_List->Val = P_Scan->Val;
      Pred_List->Next = Reseau->T[Dest_Id]->Inputs;
      Reseau->T[Dest_Id]->Inputs = Pred_List;
      Reseau->T[Dest_Id]->Indegree++;
    }
  }
}

BOOL_MAT Make_Adjacency_Matrix(EventGraph *Reseau,
			       boolean ZeroWeight)
{
  int		i;
  int		j;
  BOOL_MAT	Adj;
  Liste_Place	P_Scan;

  /* Allocate matrix */
#ifdef USE_SP_MAT
  Adj = (SP_Bool_Mat) malloc( sizeof( struct SP_BOOL_MAT ) );
  Adj->Size = Reseau->Size;
  Adj->Nb_Elts = (int*) calloc( (size_t)Reseau->Size, sizeof(int) );
  Adj->Elts = (int**) calloc( (size_t)Reseau->Size, sizeof(int*) );
#else
  Adj = (boolean*) calloc ( Reseau->Size*Reseau->Size, sizeof( boolean ) );
#endif

  /* Fill matrix in */
  for (i=0; i<Reseau->Size; i++) {

#ifndef USE_SP_MAT
    /* initialize to false */
    for (j=0; j<Reseau->Size; j++) {
      Adj[i*Reseau->Size+j] = FALSE;
    }
#else
    /* initialize neighbor list */
    Adj->Nb_Elts[i] = 0;
    Adj->Elts[i] = (int*)NULL;
#endif

    SCAN( P_Scan, Reseau->T[i]->Outputs ) {
      if ( !ZeroWeight || (P_Scan->Val->Init_Pop == 0) ) {
	j = P_Scan->Val->Trans_Id - 1;
#ifdef USE_SP_MAT
	Adj->Nb_Elts[i] ++;
        Adj->Elts[i] = (int*)realloc( Adj->Elts[i], 
				      ((size_t)Adj->Nb_Elts[i])*sizeof(int));
	Adj->Elts[i][Adj->Nb_Elts[i]-1] = j; 
#else
	Adj[i*Reseau->Size+j] = TRUE;
#endif
      }
    }
  }

  return Adj;
}

void Free_Adjacency_Matrix(BOOL_MAT Adj)
{
#ifdef USE_SP_MAT
  int		i;

  for (i=0; i<Adj->Size; i++) {
    free( Adj->Elts[i] );
  }
  free( Adj->Nb_Elts );
  free( Adj->Elts );
#endif

  free( Adj );
}

boolean topological_analysis(EventGraph *Reseau)
{
  BOOL_MAT	Adj0;		/* Adjacency matrix for zero weight arcs */
  int		*Cycle;		/* Cycle description 			 */
  boolean	Res;

  Adj0 = Make_Adjacency_Matrix( Reseau, /* zero weight ? */ TRUE  );

  if ( Debug ) {
    printf("Adj0 ====\n");
    dumpe_adj(Adj0, Reseau->Size);
  }

  Reseau->Level = (Layer_Desc*) calloc ( Reseau->Size, sizeof( Layer_Desc ) );

  /* Computing the layer structure with respect to empty places.	*/
  /* Computing also the precedence structure. It may be needed although	*/
  /* the network is not live.						*/
  Find_Layer_Structure( Reseau, Adj0, Reseau->Size);
  Make_Inputs_List( Reseau );

  if ( 0 != Reseau->Nb_Layers ) {
    Res =  TRUE;
  }
  else {
    printf("Network not live!!\n");
    Cycle = Find_Cycle(Adj0,Reseau->Size,FALSE);
    print_cycle( Cycle );
    free( Cycle);

    Res = FALSE;
  }

  /* Cleanup of temporary structures */
  Free_Adjacency_Matrix( Adj0 );

  /* return result */
  return Res;
}
