#include <math.h>

#include "eg.h"

#define TOOLNAME	"eg_crit_path"

#define PRECISION	1e-9
#define SCAN(x,y)       for(x=y;x!=NULL;x=x->Next)

typedef struct PATH_DESC {
  int		Length;
  int		*From;
  int		*By;
  int		*To;
  double	*Value;
} *Path_Desc;

/* external declarations */
extern EventGraph	*Read_Network(char*);
extern boolean		topological_analysis(EventGraph *Reseau);
extern void		dumpe_network(EventGraph *Reseau);
extern void		Free_Event_Graph(EventGraph *Reseau);
extern boolean		is_fifo(EventGraph *Reseau );
extern double		Sample(Law_Desc);

/* forward declarations */
boolean	is_fifo( EventGraph* );
void	Find_Critical_Paths( EventGraph*, FILE* );

/* main program */
int main(int argc, char *argv[])
{
  static char	In_File_Name[128] = "";
  static char	Out_File_Name[128] = "";
  static char	Arc_File_Name[128] = "";
  int		i;
  boolean	Help = FALSE;
  boolean	Args_Ok = TRUE;
  boolean	File_Ok = TRUE;
  FILE		*Arcs;
  EventGraph	*Le_Reseau;

  i = 1;

  if (argc>1)
    do {

      if GET_STRING( "-f", In_File_Name )
      else if GET_STRING( "-o", Out_File_Name )
      else if GET_STRING( "-arcs", Arc_File_Name )

      else if SWITCH( "-debug", Debug, TRUE )

      else if IS_VERSION

      else if ( ( 0 == strcmp( "-h", argv[i] ) )
		|| ( 0 == strcmp( "-?", argv[i] ) ) ) {
	printf("Usage: %s [-f <file>] ", TOOLNAME );
	printf("[-debug] [-o <file>] -arcs <file>\n");
	Args_Ok = FALSE;
	Help = TRUE;
	i++;
      }

      else
	i++; /* skip unknown argument */
    }
    while ( Args_Ok && ( i < argc ) );

  if (Args_Ok) {

    Le_Reseau = Read_Network(In_File_Name);

    if ( Le_Reseau != NULL ) {
      
      if ( topological_analysis( Le_Reseau ) ) {
	
	if ( is_fifo( Le_Reseau ) ) {
	  if ( Le_Reseau->Max_Pop_Init <= 1 ) {

	    /* Ok with net. Launch Analysis */
	    /* redirecting the output if necessary */
	    if ( strlen(Out_File_Name) != 0 )
	      File_Ok = ( NULL != freopen( Out_File_Name, "w", stdout ) );
	    
	    if ( !File_Ok ) {
	      fprintf( stderr, "%s: Error opening file %s.\n", 
		       TOOLNAME, Out_File_Name );
	      return -1;
	    }
	    else {
	      if ( strlen( Arc_File_Name ) == 0 ) {
		fprintf( stderr, "%s: Error: missing arc file.\n", TOOLNAME );
		return -1;
	      }
	      else {
		File_Ok = ( NULL != ( Arcs = fopen( Arc_File_Name, "r" ) ) );

		if ( !File_Ok ) {
		  fprintf( stderr, "%s: Error opening file %s.\n", 
			   TOOLNAME, Arc_File_Name );
		  return -1;
		}
		
		/* Seems ok so far */
		Find_Critical_Paths( Le_Reseau, Arcs );

		/* Analysis successful */
		return 0;
	      }
	    }
	  }
	  else { /* not flat */
	    printf("%s: error: Network not flat.\n", TOOLNAME );
	    return -4;
	  }
	}
	else {     /* not fifo */
	  printf("%s: error: Network not FIFO.\n", TOOLNAME );
	  return -3;
	}
      }
      else {  /* not live */
	printf("%s: error: Network not live.\n", TOOLNAME );
	return -2;
      }
    }
    else { /* not good */
      fprintf( stderr, "%s: Error reading network. No analysis.\n", 
	       TOOLNAME );
      return -1;
    }
  }
  else if (!Help)
    return -1;
  else
    return 0;
}

void Print_Path( Path_Desc The_Path )
{
  int		i;

  for ( i=The_Path->Length-1; i > 0; i-- )
    printf( "%2d -- %2d -> %2d v %f m 0\n", 
	    The_Path->From[i], The_Path->By[i],
	    The_Path->To[i], The_Path->Value[i] );

  printf( "%2d -- %2d -> %2d v %f m 1\n", 
	  The_Path->From[0], The_Path->By[0],
	  The_Path->To[0], The_Path->Value[0] );
}

void Find_Empty_Paths( EventGraph *Le_Reseau, 
		       int From, int To, double Value,
		       Path_Desc Path )
{
  int		k;
  int		Mu;
  double	x;
  double	Phi;
  Trans_Desc	The_Trans;
  Liste_Place	P_Scan;
  Place_Desc	The_Place;

  if ( ( fabs( Value ) <= PRECISION ) && ( From == To ) ) {
    Print_Path( Path );
  }
  else if ( Value > 0.0 ) {
    SCAN( P_Scan, Le_Reseau->T[From]->Outputs ) {
      The_Place = P_Scan->Val;
      k = The_Place->Trans_Id-1;
      The_Trans = Le_Reseau->T[k];
      Mu = The_Place->Init_Pop;
      x = Sample( The_Place->Place_Law ) 
	+ Sample( Le_Reseau->T[To]->Trans_Law );

      if ( Mu == 0 ) {
	Path->From[Path->Length] = From;
	Path->By[Path->Length] = The_Place->Place_Id;
	Path->To[Path->Length] = k;
	Path->Value[Path->Length] = x;
	Path->Length++;
	Find_Empty_Paths( Le_Reseau, k, To, Value - x, Path );
	Path->Length--;
      }
    }
  }
  else if ( Debug ) {
    printf("Failed path (%d,%d) at value %e\n", From, To, Value );
    Print_Path( Path );
  }
}

void Find_Critical_Paths( EventGraph *Le_Reseau, FILE* Arcs )
{
  int		From;
  int		To;
  int		k;
  int		Mu;
  double	x;
  double	Phi;
  double	Value;
  boolean	Read_Ok = TRUE;
  Trans_Desc	The_Trans;
  Liste_Place	P_Scan;
  Place_Desc	The_Place;
  Path_Desc	Path;

  Path = (Path_Desc) malloc( sizeof( struct PATH_DESC ) );
  Path->From = (int*) calloc( Le_Reseau->Size, sizeof( int ) );
  Path->By   = (int*) calloc( Le_Reseau->Size, sizeof( int ) );
  Path->To   = (int*) calloc( Le_Reseau->Size, sizeof( int ) );
  Path->Value = (double*) calloc( Le_Reseau->Size, sizeof( double ) );

  do {
    Read_Ok = ( 3 == fscanf( Arcs, "%d %d %lf", &From, &To, &Value ) );

    if ( Read_Ok ) {
      if ( Debug ) {
	printf("Looking for path %d -> %d with value %f\n", From, To, Value );
      }

      SCAN( P_Scan, Le_Reseau->T[From]->Outputs ) {
	The_Place = P_Scan->Val;
	k = The_Place->Trans_Id-1;
	The_Trans = Le_Reseau->T[k];
	Mu = The_Place->Init_Pop;
	x = Sample( The_Place->Place_Law ) 
	  + Sample( Le_Reseau->T[From]->Trans_Law );

	if ( Mu == 1 ) {
	  Path->Length = 1;
	  Path->From[0] = From;
	  Path->By[0] = The_Place->Place_Id;
	  Path->To[0] = k;
	  Path->Value[0] = x;
	  Find_Empty_Paths( Le_Reseau, k, To, Value - x, Path );
	}
      }
    }
  } while( Read_Ok );

}
