/* program/module for the equational simulation of an Event Graph	*/

#include <stddef.h>
#include <math.h>		/* for random number generation	*/
#include <sys/time.h>
#include <time.h>

#include "eg.h"

#define INFINI		1e12

#define MAX_RANDOM_VALUE          0x7FFFFFFF

#ifndef ALPHA
extern long random(void);
#endif

#define max(x,y)        ( (x>y) ? x : y )
#define min(x,y)        ( (x<y) ? x : y )

#define SCAN(x,y)       for(x=y;x!=NULL;x=x->Next)

boolean *T_Dont_Do_It;		/* for saving time when drawing tables */
boolean **P_Dont_Do_It;		/* for saving time when drawing tables */

/****************************************************************/
/* Drawing of random numbers according to several classical	*/
/* distributions.						*/
/****************************************************************/
double U_0_1(void)
{
  return (double)random()/MAX_RANDOM_VALUE;
}

double Uniform(double a, double b)
{
  return a + (b-a)*U_0_1();
}

double Exponential(double Mean)
{
  return -Mean*log(U_0_1());
}

double Bernoulli(double Proba)
{
  if (U_0_1() < Proba)
    return 1.0;
  else
    return 0.0;
}

double Geometric(double Proba)
{
  if ( Proba == 0.0 )
    return 0.0;
  else if ( Proba < 1 )
    return floor( Exponential( -1/log(Proba)) );
  else {
    fprintf(stderr, "Error: Geometric with proba=1. 0.0 assumed.\n");
    return 0.0;
  }
}

double Erlang(double N, double Mean)
/* N is kept double to avoid conversions */
{
  register int		i;
  register double	U;

  U = U_0_1();
	
  for (i=1; i<N; i++)	/* even if N not integer, this should	*/
			/* be the correct number of iterations	*/
    U *= U_0_1();

  return -Mean*log(U)/N;
}

double HyperExpo(double *Parameters)
{
  register double	U;
  register int    	i;

  U = U_0_1();

  for (i=0; (i<Parameters[0]) && (U>Parameters[i+1]); i+=2 )
    U -=Parameters[i+1];

  return Exponential( Parameters[i+2] );
}

double Discrete(double *Parameters)
{
  register double	U;
  register int    	i;
  register int		NbVals;

  NbVals = (int) Parameters[0];

  U = U_0_1();

  for ( i=NbVals+1; (i<=2*NbVals) && (U>Parameters[i]); i++ )
    U -=Parameters[i];

  return Parameters[i-NbVals];
}

/* Note: the generation of Weibull sambles wouls be faster if log(a)	*/
/* and 1/nu were stored instead of a and nu.				*/
double	Weibull(double a, double nu)
{
  register double	U;

  U = U_0_1();
  
  return pow(-log(U)/log(a),1.0/nu);
}

/* Generation of an exponential using the Polar Method (Knuth v2 p. 117) */
double Normal(double Mean, double StdDev )
{
  register double	v1;
  register double	v2;
  register double	S;

  /* Generate uniform on the unit circle */
  do { 
    v1 = 2 * U_0_1() - 1;
    v2 = 2 * U_0_1() - 1;
   } while ( 1 < ( S = v1*v1+v2*v2 ) );

  return Mean + StdDev * v1 * sqrt( -2.0 * log(S)/S );
}

/* Lognormal = exp(Normal). Adjusting the mean and var requires 	*/
/* computing parameters of the Normal. Would be faster if pre-computed.	*/
double LogNormal(double Mean, double StdDev )
{
  register double	m;
  register double	s2;

  s2 = log( 1.0 + (StdDev / Mean)*(StdDev / Mean) );
  m = log( Mean ) - s2 / 2;

  return exp( Normal( m, sqrt(s2) ) );
}

double Pareto( double a, double Alpha )
{
  return a * ( pow( U_0_1(), -1.0/Alpha ) - 1.0 );
}

/****************************************************************/
/* Procedure:		draw_iid_table				*/
/* Fill a table with random values drawn from a given law.	*/
/* Parameters:							*/
/*	The_Val:	array of double, returned filled	*/
/*	The_Size:	its size				*/
/*	The_Law:	the probability law according to which	*/
/*			the random numbers are drawn		*/
/****************************************************************/
void draw_iid_table(double	*The_Val,
		    int		The_Size,
		    Law_Desc	The_Law )
{	
  int	i;

  switch( The_Law->Name ) {
  case 'B':
    for (i=0; i<The_Size; i++)
      The_Val[i] = Bernoulli( The_Law->Parameters[0] );
    break;
  case 'C':
    for (i=0; i<The_Size; i++)
      The_Val[i] = The_Law->Parameters[0];
    break;
  case 'D':
    for (i=0; i<The_Size; i++)
      The_Val[i] = Discrete( The_Law->Parameters );
    break;
  case 'E':
    for (i=0; i<The_Size; i++)
      The_Val[i] = Exponential( The_Law->Parameters[0] );
    break;
  case 'g':
    for (i=0; i<The_Size; i++)
      The_Val[i] = Geometric( The_Law->Parameters[0] );
    break;
  case 'G':
    for (i=0; i<The_Size; i++)
      The_Val[i] = 1.0 + Geometric( The_Law->Parameters[0] );
    break;
  case 'H':
    for (i=0; i<The_Size; i++)
      The_Val[i] = HyperExpo( The_Law->Parameters );
    break;
  case 'L':
    for (i=0; i<The_Size; i++)
      The_Val[i] = LogNormal( The_Law->Parameters[0],
			      The_Law->Parameters[1] );
    break;
  case 'N':
    for (i=0; i<The_Size; i++)
      The_Val[i] = Normal( The_Law->Parameters[0],
			   The_Law->Parameters[1] );
    break;
  case 'p':
    for (i=0; i<The_Size; i++)
      The_Val[i] = Pareto( The_Law->Parameters[0],
			   The_Law->Parameters[1] );
    break;
  case 'R':
    for (i=0; i<The_Size; i++)
      The_Val[i] = Erlang( The_Law->Parameters[0],
			   The_Law->Parameters[1] );
    break;
  case 'U':
    for (i=0; i<The_Size; i++)
      The_Val[i] = Uniform( The_Law->Parameters[0],
			    The_Law->Parameters[1] );
    break;
  case 'W':
    for (i=0; i<The_Size; i++)
      The_Val[i] = Weibull( The_Law->Parameters[0],
			    The_Law->Parameters[1] );
  default:
    printf("Warning: unsupported law: %c. 0 values assumed.\n",
	   The_Law->Name );
    for (i=0; i<The_Size; i++)
      The_Val[i] = 0.0;		
  }
}

void Fill_Tables(EventGraph	*Reseau,
		 double		**Fire_Times,
		 double		***Hold_Times,
		 int		Buf_Size)
{
  int		i,j;
  Place_Desc	The_Place;
  Liste_Place	P_Scan;

  for (i=0; i<Reseau->Size; i++) {
    if ( !T_Dont_Do_It[i] )
      draw_iid_table(Fire_Times[i], Buf_Size, Reseau->T[i]->Trans_Law );

    SCAN( P_Scan, Reseau->T[i]->Outputs ) {
      The_Place = P_Scan->Val;
      j = The_Place->Trans_Id -1;
      if ( !P_Dont_Do_It[i][j] )
	draw_iid_table( &( Hold_Times[i][j][0] ),
			Buf_Size, The_Place->Place_Law );
    }
  }
}

boolean init_simulation(EventGraph 	*Reseau,
			double		**X,
			double		**Fire_Times,
			double		***Hold_Times,
			int		Buf_Size)
{
  int		i;
  int		j;
  int		The_Size;
  Place_Desc	The_Place;
  boolean	Alloc_Ok = TRUE;
  Liste_Place	P_Scan;

  The_Size = Reseau->Size;

  T_Dont_Do_It = (boolean*) calloc( The_Size, sizeof( boolean ) );
  P_Dont_Do_It = (boolean**) calloc( The_Size, sizeof( boolean* ) );

  /* Initialize date tables */
  for (i=0; (i<Reseau->Size) && Alloc_Ok ; i++) {

    Fire_Times[i] = (double*) calloc( Buf_Size, sizeof( double ) );
    Hold_Times[i] = (double**) calloc( The_Size, sizeof( double* ) );

    Alloc_Ok = ( NULL != Fire_Times[i] ) && ( NULL != Hold_Times[i] );
    P_Dont_Do_It[i] = (boolean*) calloc( The_Size, sizeof( boolean ) );

    if ( Alloc_Ok ) {
      X[i] = (double*) calloc( Buf_Size,sizeof( double ) );
      Alloc_Ok = ( NULL != X[i] );
    }

    SCAN( P_Scan, Reseau->T[i]->Outputs ) {
      The_Place = P_Scan->Val;
      j = The_Place->Trans_Id -1;
      Alloc_Ok = ( NULL !=
		   ( Hold_Times[i][j] = (double*) 
		     calloc( Buf_Size, sizeof( double ) ) ) );
    }
  }

  if ( Alloc_Ok ) {
    for ( i=0; i<The_Size; i++ )
      T_Dont_Do_It[i] = FALSE;

    Fill_Tables( Reseau, Fire_Times, Hold_Times, Buf_Size );

    /* position swith not to redraw constant times anymore */
    for ( i=0; i<The_Size; i++ ) {
      T_Dont_Do_It[i] = ( Reseau->T[i]->Trans_Law->Name == 'C' );
      if ( Debug && T_Dont_Do_It[i] )
	printf( "Won't draw tables again for trans. %d\n", i );

      SCAN( P_Scan, Reseau->T[i]->Outputs ) {
	The_Place = P_Scan->Val;
	j = The_Place->Trans_Id -1;
	P_Dont_Do_It[i][j] = ( The_Place->Place_Law->Name == 'C' );

	if ( Debug && P_Dont_Do_It[i][j] ) {
	  printf( "Won't draw tables again for Place" );
	  printf(" (%1d->%1d).\n", i, j );
	}
      }
    }

    return TRUE;
  }
  else
    return FALSE;
    
}

/************************************************************************/
/* Simulate the evolution of the network by computing sucessively	*/
/* the instants at which firings occur. The equations used are that 	*/
/* "in Y" (end of firings).						*/
/* The function returns a Size x Buf_Size table containing the last	*/
/* part of the trajectory.						*/
/************************************************************************/
double **simulate(EventGraph 	*Reseau,
		  int		Sim_Length,
		  int		Buf_Size)
{	
  int		i;
  int		j;
  int		Time;
  int		L_Time = 0;	/* "local" time in the circular buffer */
  int		The_Size;
  double	x;
  int		Trans_Id;
  int		Pred_Id;
  int		Max_Init_Pop;
  int		Mu;
  Trans_Desc	The_Trans;
  double	**Fire_Times;
  double	**YY;
  double	***Hold_Times;
  int		*Idx;		/* table to store precomputed modulos */
  Liste_Place	P_Scan;
  Place_Desc	The_Place;

  The_Size = Reseau->Size;
	
  Fire_Times = (double**) calloc( The_Size, sizeof (double*) );
  Hold_Times = (double***) calloc( The_Size, sizeof (double**) );
  YY = (double**) calloc( The_Size, sizeof (double*) );

  Max_Init_Pop = Reseau->Max_Pop_Init;

  if (! init_simulation(Reseau, YY, Fire_Times, Hold_Times, Buf_Size ) ) {
    printf("Could not allocate space. Simulation not launched.\n");
    YY = (double**) NULL;
  }
  else {
    /* compute modulos */
    Idx = (int*) calloc( (unsigned int)Buf_Size, sizeof(int) );
    for ( i=0; i<Buf_Size; i++ )
      Idx[i] = Buf_Size - i;
			     
    /* compute Max_Init_Pop first firing dates. No modulos, as Buf_Size */
    /* is larger than Max_Init_Pop					    */
    for (Time=0; (Time<Sim_Length) && (Time<Max_Init_Pop) ; Time++) {
      for ( i=0; i<Reseau->Nb_Layers; i++)
	for ( j=0; j<Reseau->Level[i]->Nb_Nodes; j++) {
	  Trans_Id = Reseau->Level[i]->Node_Id[j];
	  x = - INFINI;
	  SCAN( P_Scan, Reseau->T[Trans_Id]->Inputs ) {
	    The_Place = P_Scan->Val;
	    Pred_Id = The_Place->Input_Id-1;
	    The_Trans = Reseau->T[Pred_Id];
	    Mu = The_Place->Init_Pop;
	    if ( Mu <= Time )
	      /* the fact that times are iid is used here 	*/
	      /* to shift the index in Fire_Times		*/
	      x = max( x, 
		       YY[Pred_Id][Time-Mu] +
		       Hold_Times[Pred_Id][Trans_Id][Time] );
	    else 
	      x = max( x, 
		       The_Place->Lag_Times[Time] +
		       Hold_Times[Pred_Id][Trans_Id][Time] );
	  }
	  YY[Trans_Id][Time] = x + Fire_Times[Trans_Id][Time];
	}
      L_Time++;
    }

    /* compute the rest... */
    for ( ; Time<Sim_Length; Time++) {
      for ( i=0; i<Reseau->Nb_Layers; i++)
	for ( j=0; j<Reseau->Level[i]->Nb_Nodes; j++) {
	  Trans_Id = Reseau->Level[i]->Node_Id[j];
	  x = - INFINI;
	  SCAN( P_Scan, Reseau->T[Trans_Id]->Inputs ) {
	    The_Place = P_Scan->Val;
	    Pred_Id = The_Place->Input_Id-1;
	    The_Trans = Reseau->T[Pred_Id];
	    Mu = The_Place->Init_Pop;
	    /* the fact that times are iid is used here 	*/
	    /* to shift the index in Fire_Times		*/
	    x = max( x, 
		     YY[Pred_Id][(L_Time-Mu+Buf_Size) % Buf_Size ] +
		     Hold_Times[Pred_Id][Trans_Id][L_Time] );
	  }
	  YY[Trans_Id][L_Time] = x + Fire_Times[Trans_Id][L_Time];
	}
      L_Time++;
      if ( L_Time == Buf_Size ) {
	/* time to draw new variables */
	Fill_Tables( Reseau, Fire_Times, Hold_Times, Buf_Size );
	L_Time = 0;
      }
    }
  } /* "else" branch of !init_simulation() */

  return YY;
}

void dumpe_simulation(EventGraph	*Reseau,
		      double		**Y,
		      int		Sim_Length,
		      int		Buf_Size)
{
  int		i;
  int		j;
  double	val;
  int		Pop_Max;

  Pop_Max = Reseau->Max_Pop_Init;

  for (i=0; i<Reseau->Size; i++) {
    printf("Transition %d ...\n",i);
    for (j=Sim_Length-Pop_Max-1; j<Sim_Length; j++) {
      val = Y[i][j % Buf_Size ];
      if ( val > 100000.0 )
	printf(" eps");
      else
	printf("%8.3f", val );
      printf("\n");
    }
  }
}

/****************************************************************/
/* Procedure de calcul du temps ecoule                          */
/* (AJM | 16/11/92 )                                            */
/****************************************************************/
double elapsed_time(struct timeval t_start,
		    struct timeval t_end)
{
  return (double)((t_end.tv_sec - t_start.tv_sec) * 1000000.0
		  + t_end.tv_usec - t_start.tv_usec)/1000000.0;
}

void write_stat_results(EventGraph 	*Reseau,
			double		**Y,
			int		Sim_Length,
			int		Buf_Size,
			struct timeval	t_debut,
			struct timeval	t_fin)
{
  int	i;

  for (i=0; i<Reseau->Size; i++) {
    printf("Y_%1d[%1d] = %f    ",
           i, Sim_Length,
           Y[i][(Sim_Length-1) % Buf_Size] );
    if ( Sim_Length > 0 ) {
      printf("Cycle time: %f\n",
             Y[i][(Sim_Length-1) % Buf_Size] / Sim_Length );
    }
    else {
      printf("Cycle time: impossible to estimate\n");
    }
  }

  printf("Total execution time: %f", 
	 elapsed_time( t_debut, t_fin ) );
  printf("\t\t(%f evt/sec)\n", (double)(Sim_Length*Reseau->Size)
	 / elapsed_time( t_debut, t_fin ) );
}
