/* Module containing procedures for computing the (max,+) matrices	*/
/* corresponding to the evolution of an Event Graph.			*/
/* The EG is assumed to be "flat" i.e. its maximal marking = 1.		*/
/* The convention is to consider the __end__ of firings Y(n), and to	*/
/* perform multiplication on the right.					*/
/* Then: Y(n) = A1(n) A0*(n) Y(n-1).					*/
/*									*/
/* int	Get_Event_Graph(file), 						*/
/*		dont l'effet est de lire un graphe d'evenement en 	*/
/*		langage EGL de l'entree standard, et de retourner sa 	*/
/*		taille, N, en nombre de					*/
/*		transitions (y compris, si necessaire, les transitions	*/
/*		fictives qui resultent de la mise "a plat" du graphe)	*/
/*		Dans le cas ou le reseau n'est pas FIFO, ou pas plat,	*/
/*		un nombre negatif est retourne.				*/
/*			-1: lecture echouee: reseau incorrect		*/
/*			-2: analyse topologique echouee: pas vivant	*/
/*			-3: pas FIFO.					*/
/*			-4: pas plat.					*/
/*									*/
/* int	Get_Simulation_Length(),					*/
/*		qui retourne la "duree" de la simulation, c'est a dire	*/
/*		le parametre "Duree<-" du fichier EGL.			*/
/*									*/
/* void	Fill_Matrix(M),							*/
/*		qui prend en entree une matrice de reels de taille N x N*/
/*		et la remplit avec les temporisations aleatoires ad hoc.*/
/*									*/
/* void	Fill_Slice(M, H),						*/
/*		qui prend en entree une matrice de reels de taille N x N*/
/*		et la remplit avec les chemins critiques d'une tranche  */
/*              de graphe de hauteur H.                                 */
/*									*/
/* void Fill_Initial_Vector(Y),						*/
/*		prend en entree un vecteur de reels de taille N et le	*/
/*		remplit avec le vecteur Y(0).				*/
/*									*/
/* void Fill_Initial_Vector_Flat(Y,Dirac),				*/
/*		prend en entree un vecteur de reels de taille N et le	*/
/*		remplit avec le vecteur Y(0). Cas special "flat"	*/
/*									*/
/* boolean Fill_Vector(Y, H),						*/
/*		prend en entree un vecteur de reels de taille N et 	*/
/*		comme point de depart pour calculer le vecteur Y(H).	*/
/*									*/
/* boolean Fill_Vector_Backwards(Y, H),					*/
/*		la mme, mais en partant du "bas".			*/
/*									*/
/* Le code de ce module est largement repris de celui de "simulate_eg.c"*/
/*									*/
/************************************************************************/

#include <stddef.h>
#include <math.h>		/* for random number generation	*/
#include <stdio.h>

#include "eg.h"

#define MAX_RANDOM_VALUE          0x7FFFFFFF

#define EPSILON		-1	/* convention de Sylvain */


#define SCAN(x,y)       for(x=y;x!=NULL;x=x->Next)

#ifndef ALPHA
extern long		random(void);
#endif

extern EventGraph	*Read_Network(char*);
extern int		topological_analysis(EventGraph*);
EventGraph		*Le_Reseau;

/* declaration of min as a function because it is used with random numbers */
double max( double x, double y )
{
  return ( x > y ? x : y );
}

/****************************************************************/
/* Drawing of random numbers according to several classical	*/
/* distributions.						*/
/****************************************************************/
double U_0_1(void)
{
  return (double)random()/MAX_RANDOM_VALUE;
}

double Uniform(double a, double b)
{
  return a + (b-a)*U_0_1();
}

double Exponential(double Mean)
{
  return -Mean*log(U_0_1());
}

double Bernoulli(double Proba)
{
  if (U_0_1() < Proba)
    return 1.0;
  else
    return 0.0;
}

double Geometric(double Proba)
{
  if ( Proba == 0.0 )
    return 0.0;
  else if ( Proba < 1 )
    return floor( Exponential( -1/log(Proba)) );
  else {
    fprintf(stderr, "Error: Geometric with proba=1. 0.0 assumed.\n");
    return 0.0;
  }
}

double Erlang(double N, double Mean)
/* N is kept double to avoid conversions */
{
  register int		i;
  register double	U;

  U = U_0_1();
	
  for (i=1; i<N; i++)	/* even if N not integer, this should	*/
			/* be the correct number of iterations	*/
    U *= U_0_1();

  return -Mean*log(U)/N;
}

double HyperExpo(double *Parameters)
{
  register double	U;
  register int    	i;

  U = U_0_1();

  for (i=0; (i<Parameters[0]) && (U>Parameters[i+1]); i+=2 )
    U -=Parameters[i+1];

  return Exponential( Parameters[i+2] );
}

double Discrete(double *Parameters)
{
  register double	U;
  register int    	i;
  register int		NbVals;

  NbVals = (int) Parameters[0];

  U = U_0_1();

  for ( i=NbVals+1; (i<=2*NbVals) && (U>Parameters[i]); i++ )
    U -=Parameters[i];

  return Parameters[i-NbVals];
}

/* Note: the generation of Weibull sambles wouls be faster if log(a)	*/
/* and 1/nu were stored instead of a and nu.				*/
double	Weibull(double a, double nu)
{
  register double	U;

  U = U_0_1();
  
  return pow(-log(U)/log(a),1.0/nu);
}

/* Generation of an exponential using the Polar Method (Knuth v2 p. 117) */
double Normal(double Mean, double StdDev )
{
  register double	v1;
  register double	v2;
  register double	S;

  /* Generate uniform on the unit circle */
  do { 
    v1 = 2 * U_0_1() - 1;
    v2 = 2 * U_0_1() - 1;
   } while ( 1 < ( S = v1*v1+v2*v2 ) );

  return Mean + StdDev * v1 * sqrt( -2.0 * log(S)/S );
}

/* Lognormal = exp(Normal). Adjusting the mean and var requires 	*/
/* computing parameters of the Normal. Would be faster if pre-computed.	*/
double LogNormal(double Mean, double StdDev )
{
  register double	m;
  register double	s2;

  s2 = log( 1.0 + (StdDev / Mean)*(StdDev / Mean) );
  m = log( Mean ) - s2 / 2;

  return exp( Normal( m, sqrt(s2) ) );
}

double Pareto( double a, double Alpha )
{
  return a * ( pow( U_0_1(), -1.0/Alpha ) - 1.0 );
}

double Sample(Law_Desc	The_Law)
{
  double	The_Val;

  switch( The_Law->Name ) {
  case 'C':
    The_Val = The_Law->Parameters[0];
    break;
  case 'E':
    The_Val = Exponential( The_Law->Parameters[0] );
    break;
  case 'B':
    The_Val = Bernoulli( The_Law->Parameters[0] );
    break;
  case 'G':
    The_Val = Geometric( The_Law->Parameters[0] );
    break;
  case 'U':
    The_Val = Uniform( The_Law->Parameters[0],
		       The_Law->Parameters[1] );
    break;
  case 'R':
    The_Val = Erlang( The_Law->Parameters[0],
		      The_Law->Parameters[1] );
    break;
  case 'H':
    The_Val = HyperExpo( The_Law->Parameters );
    break;
  }

  return The_Val;
}

void draw_iid_table(double	*The_Val,
		    int		The_Size,
		    Law_Desc	The_Law )
{	
  int	i;

  switch( The_Law->Name ) {
  case 'B':
    for (i=0; i<The_Size; i++)
      The_Val[i] = Bernoulli( The_Law->Parameters[0] );
    break;
  case 'C':
    for (i=0; i<The_Size; i++)
      The_Val[i] = The_Law->Parameters[0];
    break;
  case 'E':
    for (i=0; i<The_Size; i++)
      The_Val[i] = Exponential( The_Law->Parameters[0] );
    break;
  case 'g':
    for (i=0; i<The_Size; i++)
      The_Val[i] = Geometric( The_Law->Parameters[0] );
    break;
  case 'G':
    for (i=0; i<The_Size; i++)
      The_Val[i] = 1.0 + Geometric( The_Law->Parameters[0] );
    break;
  case 'H':
    for (i=0; i<The_Size; i++)
      The_Val[i] = HyperExpo( The_Law->Parameters );
    break;
  case 'L':
    for (i=0; i<The_Size; i++)
      The_Val[i] = LogNormal( The_Law->Parameters[0],
			      The_Law->Parameters[1] );
    break;
  case 'N':
    for (i=0; i<The_Size; i++)
      The_Val[i] = Normal( The_Law->Parameters[0],
			   The_Law->Parameters[1] );
    break;
  case 'p':
    for (i=0; i<The_Size; i++)
      The_Val[i] = Pareto( The_Law->Parameters[0],
			   The_Law->Parameters[1] );
    break;
  case 'R':
    for (i=0; i<The_Size; i++)
      The_Val[i] = Erlang( The_Law->Parameters[0],
			   The_Law->Parameters[1] );
    break;
  case 'U':
    for (i=0; i<The_Size; i++)
      The_Val[i] = Uniform( The_Law->Parameters[0],
			    The_Law->Parameters[1] );
    break;
  case 'W':
    for (i=0; i<The_Size; i++)
      The_Val[i] = Weibull( The_Law->Parameters[0],
			    The_Law->Parameters[1] );
  default:
    printf("Warning: unsupported law: %c. 0 values assumed.\n",
	   The_Law->Name );
    for (i=0; i<The_Size; i++)
      The_Val[i] = 0.0;		
  }
}

int Get_Simulation_Length(void)
{
  fprintf(stderr,"MatrixModule: Error: call to Get_Simulation_Length() obsolete.\n");
  fprintf(stderr,"Please update your code. Default value 1000 returned.\n");

  return 1000;
}

void Fill_Matrix( double *A )
{	
  int		i;
  int		j;
  int		k;
  int		l;
  int		m;
  int		The_Size;
  double	x;
  double	Phi;
  int		Mu;
  Trans_Desc	The_Trans;
  Liste_Place	P_Scan;
  Place_Desc	The_Place;

  The_Size = Le_Reseau->Size;

  for( l=0; l<Le_Reseau->Nb_Layers; l++) {
    for( m=0; m<Le_Reseau->Level[l]->Nb_Nodes; m++) {

      i = Le_Reseau->Level[l]->Node_Id[m];
      Phi = Sample( Le_Reseau->T[i]->Trans_Law );

      for ( j=0; j<The_Size; j++ ) {

	x = EPSILON;

	SCAN( P_Scan, Le_Reseau->T[i]->Inputs ) {
	  The_Place = P_Scan->Val;
	  k = The_Place->Input_Id-1;
	  The_Trans = Le_Reseau->T[k];
	  Mu = The_Place->Init_Pop;

	  if ( Mu == 1 ) {
	    if ( j == k )
	      x = max( x, Sample( The_Place->Place_Law ) );
	  }
	  else {	
	    /* Then Mu = 0: the network is assumed to be flat */
	    x = max( x, A[k*The_Size+j] + Sample( The_Place->Place_Law ) );
	  }
	  A[i*The_Size+j] = ( x > EPSILON ? x + Phi : EPSILON );
	}
      }
    }
  }
}

/************************************************************************/
/* Code copi de la procdure "simulate" du module "simulate_eg.c" 	*/
/* mais adapt au calcul de la matrice de transfer.                	*/
/* Deux fonctions boolenes:                          		   	*/
/*	Fill_Slice():	remplit une matrice avec les lg des chemins	*/
/*			critiques. Retourne faux si probleme		*/
/*	Init_Times():	initialise les tableaux de temps et d'autres.	*/
/*			Retourne faux si probleme			*/
/************************************************************************/
double	**Fire_Times;
double	***Hold_Times;
boolean *T_Dont_Do_It;
boolean **P_Dont_Do_It;
/* tableaux utilises dans Fill_Slice 		*/
double  **Lg_0;               /* les Y[n-1] 	*/
double  **Lg_1;               /* les Y[n]   	*/

boolean Init_Times( int		Height)
{
  int		i,j;
  int		The_Size;
  Place_Desc	The_Place;
  boolean	Alloc_Ok = TRUE;
  Liste_Place	P_Scan;

  The_Size = Le_Reseau->Size;

  T_Dont_Do_It = (boolean*) calloc( The_Size, sizeof( boolean ) );
  P_Dont_Do_It = (boolean**) calloc( The_Size, sizeof( boolean* ) );

  Fire_Times = (double**) calloc( The_Size, sizeof( double* ) );
  Hold_Times = (double***) calloc( The_Size, sizeof( double** ) );

  /* creer aussi les tableaux locaux */
  Lg_0 = (double**) calloc( The_Size, sizeof( double* ) );
  Lg_1 = (double**) calloc( The_Size, sizeof( double* ) );

  Alloc_Ok = ( ( Lg_0 != NULL ) && ( Lg_1 != NULL )
	       && ( Fire_Times != NULL ) && ( Hold_Times != NULL )
	       && ( T_Dont_Do_It != NULL ) && ( P_Dont_Do_It != NULL ) );

  /* Initialize date tables */
  for (i=0; (i<The_Size) && Alloc_Ok ; i++) {

    Fire_Times[i] = (double*) calloc( Height, sizeof( double ) );
    Hold_Times[i] = (double**) calloc( The_Size, sizeof( double* ) );

    P_Dont_Do_It[i] = (boolean*) calloc( The_Size, sizeof( boolean ) );
    
    Lg_0[i] = (double*) calloc( The_Size, sizeof( double ) );
    Lg_1[i] = (double*) calloc( The_Size, sizeof( double ) );

    Alloc_Ok = ( ( NULL != Fire_Times[i] )
		 && ( NULL != Hold_Times[i] )
		 && ( NULL != P_Dont_Do_It[i] )
		 && ( NULL != Lg_0[i] ) && ( NULL != Lg_1[i] ) );

    if (Alloc_Ok) {
      /* positionne le drapeau de "retirage" pour la transition #i */
      T_Dont_Do_It[i] = ( Le_Reseau->T[i]->Trans_Law->Name == 'C' );
      /* fait le tirage pour les temps constants */
      if ( T_Dont_Do_It[i] )
	draw_iid_table(Fire_Times[i], Height, 
		       Le_Reseau->T[i]->Trans_Law );

    }

    for ( P_Scan=Le_Reseau->T[i]->Outputs; ( P_Scan != NULL ) && Alloc_Ok;
	  P_Scan = P_Scan->Next ) {
      The_Place = P_Scan->Val;
      j = The_Place->Trans_Id -1;

      Alloc_Ok = ( NULL !=
		   ( Hold_Times[i][j] = (double*) 
		     calloc( Height, sizeof( double ) ) ) );
      P_Dont_Do_It[i][j] = ( The_Place->Place_Law->Name == 'C' );

      if (Alloc_Ok && P_Dont_Do_It[i][j] )
	draw_iid_table( &( Hold_Times[i][j][0] ), Height, 
			The_Place->Place_Law );

    }
  }

  return Alloc_Ok;
}

/************************************************************************/
/* procedure pour re-allouer les tableaux quant la taille de "tranche"  */
/* a change.								*/
/************************************************************************/
boolean Re_Init_Times( int Height )
{
  int		i,j;
  int		The_Size;
  Place_Desc	The_Place;
  boolean	Alloc_Ok = TRUE;
  Liste_Place	P_Scan;

  The_Size = Le_Reseau->Size;

  /* Initialize date tables */
  for (i=0; (i<The_Size) && Alloc_Ok ; i++) {

    Fire_Times[i] = (double*) realloc( Fire_Times[i], Height*sizeof( double ) );
    
    Alloc_Ok = ( NULL != Fire_Times[i] );

    if (Alloc_Ok && T_Dont_Do_It[i] )
      draw_iid_table(Fire_Times[i], Height, Le_Reseau->T[i]->Trans_Law );

    for ( P_Scan=Le_Reseau->T[i]->Outputs; ( P_Scan != NULL ) && Alloc_Ok;
	  P_Scan = P_Scan->Next ) {
      The_Place = P_Scan->Val;
      j = The_Place->Trans_Id -1;

      Alloc_Ok = ( NULL !=
		   ( Hold_Times[i][j] = (double*) 
		     realloc( Hold_Times[i][j], Height*sizeof( double ) ) ) );

      if (Alloc_Ok && P_Dont_Do_It[i][j] )
	draw_iid_table( &( Hold_Times[i][j][0] ), Height, 
			The_Place->Place_Law );

    }
  }

  return Alloc_Ok;
}

void Fill_Tables( 
EventGraph	*Reseau,
double		**Fire_Times,
double		***Hold_Times,
int		Height )
{
  int		i,j;
    Place_Desc	The_Place;
    Liste_Place	P_Scan;

    for (i=0; i<Reseau->Size; i++) {
	if ( !T_Dont_Do_It[i] )
	    draw_iid_table(Fire_Times[i], Height, Reseau->T[i]->Trans_Law );

	SCAN( P_Scan, Reseau->T[i]->Outputs ) {
	    The_Place = P_Scan->Val;
	    j = The_Place->Trans_Id -1;
	    if ( !P_Dont_Do_It[i][j] )
		draw_iid_table( &( Hold_Times[i][j][0] ),
				Height, The_Place->Place_Law );
	}
    }
}

boolean Fill_Slice(double	*A,
		   int     	H )
{	
  int		i;
  int		j;
  int		l;
  int		m;
  int		Time;              /* Depth in the task graph */
  int		The_Size;
  double	y;
  double        Date;
  int		Orig;
  int		Dest;
  int		Pred;
  int		Mu;
  Trans_Desc	The_Trans;
  double	**Tmp_Tab;
  boolean       Ok = TRUE;
  Liste_Place	P_Scan;
  Place_Desc	The_Place;

  The_Size = Le_Reseau->Size;

  /* Reinitialize Lg_0 to the identity matrix */
  for ( i=0; ( i<The_Size) && Ok; i++ ) {
    for ( j=0; j<The_Size; j++ )
      Lg_0[i][j] = EPSILON;
    Lg_0[i][i] = 0.0;
  }

  Fill_Tables( Le_Reseau, Fire_Times, Hold_Times, H );

  /* Computation of the slide */
  for ( Time=0; Time<H; Time++) {
    for( l=0; l<Le_Reseau->Nb_Layers; l++) {
      for( m=0; m<Le_Reseau->Level[l]->Nb_Nodes; m++) {
	Dest = Le_Reseau->Level[l]->Node_Id[m];

	for ( Orig=0; Orig<The_Size; Orig++ ) {

	  y = EPSILON;
	  SCAN( P_Scan, Le_Reseau->T[Dest]->Inputs ) {
	    The_Place = P_Scan->Val;
	    Pred = The_Place->Input_Id-1;
	    The_Trans = Le_Reseau->T[Pred];
	    Mu = The_Place->Init_Pop;
				    
	    /* Here, I know that Mu is either 0 or 1 */
	    Date = ( Mu == 1 ? Lg_0[Orig][Pred] : Lg_1[Orig][Pred] );

	    y = ( Date > EPSILON ?
		  max( y, Date + Hold_Times[Pred][Dest][Time] ) :
		  y );
					    
	  }
	  Lg_1[Orig][Dest] = ( y > EPSILON ?
			       y + Fire_Times[Dest][Time] :
			       EPSILON );
	}
      }
    }
    Tmp_Tab = Lg_0;
    Lg_0 = Lg_1;
    Lg_1 = Tmp_Tab;
  }

  /* linearize the resulting matric */
  for ( i=0; i<The_Size; i++ )
    for ( j=0; j<The_Size; j++ )
      A[i*The_Size+j] = Lg_0[j][i];
    
  return TRUE;

}

boolean Fill_Vector(
double	*V,
int     H )
{
  int		i;
  int		l;
  int		m;
  int		Time;              /* la profondeur dans le graphe */
  int		The_Size;
  double	y;
  double        Date;
  int		Dest;
  int		Pred;
  int		Mu;
  Trans_Desc	The_Trans;
  double	**Tmp_Tab;
  boolean       Ok;
  Liste_Place	P_Scan;
  Place_Desc	The_Place;

  The_Size = Le_Reseau->Size;

  /* initialiser Lg_0 au vecteur V */
  for ( i=0; ( i<The_Size) && Ok; i++ )
      Lg_0[i][0] = V[i];

  Fill_Tables( Le_Reseau, Fire_Times, Hold_Times, H );

  /* calcul de la tranche */
  for ( Time=0; Time<H; Time++) {
    for( l=0; l<Le_Reseau->Nb_Layers; l++) {
      for( m=0; m<Le_Reseau->Level[l]->Nb_Nodes; m++) {
	Dest = Le_Reseau->Level[l]->Node_Id[m];

	y = EPSILON;
	SCAN( P_Scan, Le_Reseau->T[Dest]->Inputs ) {
	  The_Place = P_Scan->Val;
	  Pred = The_Place->Input_Id-1;
	  The_Trans = Le_Reseau->T[Pred];
	  Mu = The_Place->Init_Pop;
				    
	  /* Here, I know that Mu is either 0 or 1 */
	  Date = ( Mu == 1 ? Lg_0[Pred][0] : Lg_1[Pred][0] );

	  y = ( Date > EPSILON ?
		max( y, Date + Hold_Times[Pred][Dest][Time] ) :
		y );
					    
	}
	Lg_1[Dest][0] = ( y > EPSILON ?
			  y + Fire_Times[Dest][Time] :
			  EPSILON );
      }
    }

    Tmp_Tab = Lg_0;
    Lg_0 = Lg_1;
    Lg_1 = Tmp_Tab;
  }

  /* Copy the result into V */
  for ( i=0; i<The_Size; i++ )
    V[i] = Lg_0[i][0];
	
  return TRUE;

}

boolean Fill_Vector_Backwards(
double	*V,
int     H )
{
  int		i;
  int		l;
  int		m;
  int		Time;              /* la profondeur dans le graphe */
  int		The_Size;
  double	y;
  double        Date;
  int		Dest, Pred;
  int		Mu;
  Trans_Desc	The_Trans;
  double	**Tmp_Tab;
  boolean       Ok;
  Liste_Place	P_Scan;
  Place_Desc	The_Place;

  The_Size = Le_Reseau->Size;

  /* initialiser Lg_0 au vecteur V */
  for ( i=0; ( i<The_Size) && Ok; i++ )
      Lg_0[i][0] = V[i];

  Fill_Tables( Le_Reseau, Fire_Times, Hold_Times, H );

  /* calcul de la tranche */
  for ( Time=0; Time<H; Time++) {
    for( l=0; l<Le_Reseau->Nb_Layers; l++)
      for( m=0; m<Le_Reseau->Level[l]->Nb_Nodes; m++) {
	Dest = Le_Reseau->Level[l]->Node_Id[m];

	y = EPSILON;
	SCAN( P_Scan, Le_Reseau->T[Dest]->Inputs ) {
	  The_Place = P_Scan->Val;
	  Pred = The_Place->Input_Id-1;
	  The_Trans = Le_Reseau->T[Pred];
	  Mu = The_Place->Init_Pop;
				    
	  /* Here, I know that Mu is either 0 or 1 */
	  Date = ( Mu == 1 ? Lg_0[Pred][0] : Lg_1[Pred][0] );

	  y = ( Date > EPSILON ?
		max( y, Date + Hold_Times[Pred][Dest][Time] ) :
		y );
					    
	}
	Lg_1[Dest][0] = ( y > EPSILON ?
			  y + Fire_Times[Dest][Time] :
			  EPSILON );
      }

    Tmp_Tab = Lg_0;
    Lg_0 = Lg_1;
    Lg_1 = Tmp_Tab;
  }

  /* Copy the result into V */
  for ( i=0; i<The_Size; i++ )
    V[i] = Lg_0[i][0];
	
  return TRUE;

}

void	Fill_Initial_Vector( double *Y )
{
  int		i;
  int		m;
  double	Last_Lag;
  Liste_Place	P_Scan;
  Place_Desc	The_Place;

  for( m=0; m<Le_Reseau->Size; m++ )
    Y[m] = EPSILON;

  for( m=0; m<Le_Reseau->Level[0]->Nb_Nodes; m++) {

    i = Le_Reseau->Level[0]->Node_Id[m];

    /* Compute maximum lag of marks in the input places if Tr_i */
    Last_Lag = EPSILON;
    SCAN( P_Scan, Le_Reseau->T[i]->Inputs ) {
      The_Place = P_Scan->Val;
      Last_Lag = max( Last_Lag, The_Place->Lag_Times[0] );
    }

    Y[i] = ( Last_Lag == EPSILON ? 
	     EPSILON :
	     Last_Lag + Sample( Le_Reseau->T[i]->Trans_Law ) );
  }
}

/************************************************************************/
/* This version of "Fill_Initial_Vector" takes better into account the	*/
/* fact that the "transition" matrix A is designed to change ***ENDS*** */
/* of firings at level N into ends at level N+1. Therefore, horizontal  */
/* paths must be taken into consideration in order to compute the time	*/
/* from bottom to the END of the nodes of the first row.	        */
/************************************************************************/
void	Fill_Initial_Vector_Flat( double *Y, int Dirac )
{
  int		i;
  int		k;
  int		l;
  int		m;
  double	X;
  int		Mu;
  Liste_Place	P_Scan;
  Place_Desc	The_Place;

  if ( ( Dirac < 0 ) || ( Dirac >= Le_Reseau->Size ) ) {

    for( l=0; l<Le_Reseau->Nb_Layers; l++) {
      for( m=0; m<Le_Reseau->Level[l]->Nb_Nodes; m++) {

	i = Le_Reseau->Level[l]->Node_Id[m];
	
	X = EPSILON;
	SCAN( P_Scan, Le_Reseau->T[i]->Inputs ) {
	  The_Place = P_Scan->Val;
	  Mu = The_Place->Init_Pop;

	  if ( Mu == 1 )
	    X = max( X, The_Place->Lag_Times[0] );
	  else 
	    X = max( Y[k], X );

	  Y[i] = ( X == EPSILON ? 
		   EPSILON :
		   X + Sample( Le_Reseau->T[i]->Trans_Law ) );
	}
      }
    }
  }
  else {
    /* initialization assuming one single link from bottom at Dirac */
    for( l=0; l<Le_Reseau->Nb_Layers; l++) {
      for( m=0; m<Le_Reseau->Level[l]->Nb_Nodes; m++) {

	i = Le_Reseau->Level[l]->Node_Id[m];

	if ( i == Dirac )
	  X = 0.0;
	else 
	  X = EPSILON;

	SCAN( P_Scan, Le_Reseau->T[i]->Inputs ) {
	  The_Place = P_Scan->Val;
	  Mu = The_Place->Init_Pop;

	  if ( Mu == 0 )
	    X = max( Y[k], X );
	}

	Y[i] = ( X == EPSILON ? 
		 EPSILON :
		 X + Sample( Le_Reseau->T[i]->Trans_Law ) );
      }
    }
  }
}

/************************************************************************/
/* Chargement et verification du graphe					*/
/************************************************************************/
boolean is_constant( Law_Desc	The_Law )
{
  return ( The_Law->Name == 'C' );
}

boolean is_recycled( int		Trans_Id,
		     EventGraph 	*Reseau )
{
  boolean	Is_Recycled = FALSE;
  Liste_Place	P_Scan;
  
  SCAN( P_Scan, Reseau->T[Trans_Id]->Outputs ) {
    if ( P_Scan->Val->Trans_Id == (Trans_Id+1) )
      Is_Recycled = TRUE;
  }

  return Is_Recycled;
}

boolean	is_fifo( EventGraph	*Reseau )
{
  int		i;
  boolean	Is_Fifo = TRUE;
  Liste_Place	P_Scan;
  Place_Desc	The_Place;

  for (i=0; (i<Reseau->Size) && Is_Fifo; i++) {
    Is_Fifo = Is_Fifo 
      && ( is_recycled( i, Reseau ) 
	   || is_constant( Reseau->T[i]->Trans_Law ) );

    SCAN( P_Scan, Reseau->T[i]->Inputs ) {
      The_Place = P_Scan->Val;
      Is_Fifo = Is_Fifo && is_constant( The_Place->Place_Law );
    }
  }

  return Is_Fifo;
}

int	Get_Event_Graph(char* File_Name)
{
  Le_Reseau = Read_Network(File_Name);

  if ( Le_Reseau != NULL ) {

    if ( topological_analysis( Le_Reseau ) ) {
    
      if ( is_fifo( Le_Reseau ) ) {
	if ( Le_Reseau->Max_Pop_Init <= 1 ) 
	  return	
	    Le_Reseau->Size;
	else	/* pas plat */
	  return -4;
      }
      else      /* pas fifo */
	return -3;
    }
    else  /* pas vivant */
      return -2;
  }
  else /* pas bon */
    return -1;
}

