/* Implementation of Methods for Random Bernoulli Products of 		*/
/* (max,+) matrices.							*/

#include <malloc.h>
#include "bernoulli.h"

#define TOOLNAME "Bernoulli"

/* Contructor and Destructor */
Bernoulli_Product New_Bernoulli_Product( int Nb_Matrices, int Matrix_Size )
{
  boolean		Alloc_Ok;
  int			i;
  Bernoulli_Product 	Res;
  
  Res = (Bernoulli_Product) malloc( sizeof( struct BERNOULLI_PRODUCT ) );

  if ( (Bernoulli_Product)NULL != Res ) {
    Res->Nb_Matrices = Nb_Matrices;
    Res->Size = Matrix_Size;

    Res->Weight = (double*) calloc( (size_t)Nb_Matrices, sizeof(double) );
    Res->Proba  = (double*) calloc( (size_t)Nb_Matrices, sizeof(double) );
    Res->Pr_Cum = (double*) calloc( (size_t)Nb_Matrices, sizeof(double) );
    Res->Permut = (int*)    calloc( (size_t)Nb_Matrices, sizeof(int) );
    Res->A = (MaxPlus_Matrix*) calloc( (size_t)Nb_Matrices, 
				       sizeof(MaxPlus_Matrix) );

    Alloc_Ok = ( ( NULL != Res->Weight ) 
		 && ( NULL != Res->Proba ) 
		 && ( NULL != Res->Pr_Cum ) 
		 && ( NULL != Res->Permut ) 
		 && ( (MaxPlus_Matrix*)NULL != Res->A ) );    
 
    if ( Alloc_Ok ) {
      for ( i=0; Alloc_Ok && (i<Nb_Matrices); i++ ) {
	Res->A[i] = New_MaxPlus_Matrix( Matrix_Size, Matrix_Size, ZERO );
	Alloc_Ok = ( (MaxPlus_Matrix)NULL != Res->A[i] );
      }
    }

    if ( !Alloc_Ok ) {
      fprintf(stderr,"%s: memory error. Not enough space.\n", TOOLNAME );
      Free_Bernoulli_Product( Res );
      Res = (Bernoulli_Product)NULL;
    }
  }

  return Res;
}

/* garbage collecting procedure. Probably useless in the present state. */
#define FREE(ptr,type) if ( (type*)NULL != (ptr) ) { free( (void*)ptr ); }

void Free_Bernoulli_Product( Bernoulli_Product Problem ) 
{
  int	i;

  if ( (Bernoulli_Product) NULL != Problem ) {
    FREE( Problem->Weight, double );
    FREE( Problem->Proba, double );
    FREE( Problem->Pr_Cum, double );
    FREE( Problem->Permut, int );
    if ( (MaxPlus_Matrix*)NULL != Problem->A ) {
      for ( i=0; i<Problem->Nb_Matrices; i++ ) {
	Free_MaxPlus_Matrix( Problem->A[i] );
     }
      free( Problem->A );
    }
    
    free( Problem );
  }
}

Bernoulli_Product Read_Bernoulli_Product( FILE *Input )
{
  int		i;
  int		k;
  int		Nb_Mat = 0;
  int		Size = 0;
  boolean	Ok = TRUE;
  double	Total;
  Bernoulli_Product 	Problem = (Bernoulli_Product)NULL;
  Bernoulli_Product 	Res;

  Ok = ( 1 == fscanf( Input, "%d", &Size ) );
  Ok = Ok && ( Size > 0 ) && ( Size < 10000 );

  Ok = Ok && ( 1 == fscanf( Input, "%d", &Nb_Mat ) );
  Ok = Ok && ( Nb_Mat > 0 ) && ( Nb_Mat < 10000 );

  if ( Ok ) {
    Problem = New_Bernoulli_Product( Nb_Mat, Size );
    Ok = ( (Bernoulli_Product)NULL != Problem );
  }
  else {
    printf("%s: error reading dimension (%d) or nb of matrices (%d).\n", 
	   TOOLNAME, Size, Nb_Mat );
    Ok = FALSE;
  }

  if ( Ok ) {
    /* Allocations ok. Proceeding with reading matrix data. */

    /* Proba contains normalized weights. */
    /* First read values as weights */
    Total = 0.0;
    for ( i=0; i<Nb_Mat; i++ ) {
      Ok = Ok && ( 1 == fscanf( Input, "%lf",  &(Problem->Weight[i]) ) );
      Total += Problem->Weight[i];
    }
    if ( Ok ) {
      /* Next normalize them to one, while computing cumulants */
      for ( i=0; i<Nb_Mat; i++ ) {
	Problem->Proba[i] = Problem->Weight[i] / Total;
	if ( i == 0 )
	  Problem->Pr_Cum[i] = Problem->Proba[i];
	else if ( i < Nb_Mat - 1 )
	  Problem->Pr_Cum[i] = Problem->Pr_Cum[i-1] + Problem->Proba[i];
	else 
	  /* putting a safe "guard" just in case of roundoff errors */
	  Problem->Pr_Cum[i] = 1.0;
      }
    }

    for ( k=0; (k<Nb_Mat) && Ok; k++ ) {
      if ( !Read_MaxPlus_Matrix( Input, Problem->A[k] ) ) {
	printf("%s: error: reading matrix #%d. Abort.\n", TOOLNAME, k );
	Ok = FALSE;
      }
    }
  }

  if ( Ok ) {
    Res = Problem;
  }
  else {
    Free_Bernoulli_Product( Problem );
    Res = (Bernoulli_Product) NULL;
  }
    
  return Res;
}

void Write_Bernoulli_Product(Bernoulli_Product	P,
			     FILE 		*Output,
			     boolean		Format,
			     char		*Name)
{
  int		i;
  static char	FullName[256];

  if ( Format == FORMAT_SCILAB ) {
    fprintf( Output, "%s_Size = %d\n", Name, P->Size );
    fprintf( Output, "%s_Number = %d\n", Name, P->Nb_Matrices );
  }
  else {
    fprintf( Output, "%d\n", P->Size );
    fprintf( Output, "%d\n", P->Nb_Matrices );
  }

  /* write out weights. If they are integer, shorter print */
  if ( Format == FORMAT_SCILAB ) {
    fprintf( Output, "%s_Weights = [", Name );
  }
  for ( i=0; i<P->Nb_Matrices; i++ ) {
    if ( (int)(P->Weight[i]) == (P->Weight[i]) ) {
      fprintf( Output, " %d", (int)(P->Weight[i]) );
    }
    else {
      fprintf( Output, " %f", P->Weight[i] );
    }
  }
  if ( Format == FORMAT_SCILAB ) {
    fprintf( Output, "]" );
  }
  fprintf( Output, "\n");

  /* write out matrices */
  for ( i=0; i<P->Nb_Matrices; i++ ) {
    sprintf( FullName, "%s_%d", Name, i ); 
    Write_MaxPlus_Matrix( P->A[i], Output, Format, FullName );
    fprintf( Output, "\n");
  }
}
