/* modules pour generer des graphes reguliers dans le format de gr_sim	*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <time.h>

#define boolean int
#define FALSE 0
#define TRUE  1

#define TOOLNAME	"gen_gr"
#define VERSION		"v1.1"

#define	Line		0
#define Cycle		1
#define Star		2
#define Random		3
#define RRandom		4
#define Complete	5
#define BiComplete	6

#define MAX_RANDOM_VALUE          0x7FFFFFFF

static int NON[2] = { 0, 1 };
static int OUI[2] = { 1, 0 };

typedef int Net_Type;

#ifndef ALPHA
extern void srandom( unsigned seed );
extern long random( void );
#endif

void write_entry(FILE		*OutFile,
		 int 		i,
		 int 		j,
		 int 		Val,
		 boolean	Full_Matrix )
{
  if ( Val == 1 ) {
    if ( Full_Matrix )
      fprintf(OutFile, "%d ", Val );
    else
      fprintf(OutFile, "%d %d\n", i, j );
  }
  else if ( Full_Matrix )
    fprintf(OutFile, "%d ", Val );
}

void gen_size_weight( FILE 	*OutFile, 
		      int 	Net_Size,
		      boolean	Full_Matrix )
{
  int	i;

  if ( !Full_Matrix ) 
    fprintf(OutFile, "sp\n" );

  fprintf(OutFile, "%d\n", Net_Size );

  for (i=0; i<Net_Size; i++)
    fprintf( OutFile, " 1" );
  fprintf(OutFile,"\n");
}

void gen_random(
		FILE		*OutFile,
		int		Net_Size,
		double		Proba,
		boolean		Full_Matrix,
		boolean		NoLoop,
		boolean		Reverse )
{
  int		i;
  int		j;
  double	U;

  for (i=0; i<Net_Size; i++) {
    for (j=0; j<Net_Size; j++ ) {
      if ( ( i == j ) && !NoLoop ) {
	write_entry( OutFile, i, j, OUI[Reverse], Full_Matrix );
      }
      else {
	U = (double)random()/(double)MAX_RANDOM_VALUE;

	if ( U < Proba )
	  write_entry( OutFile, i, j, OUI[Reverse], Full_Matrix );
	else
	  write_entry( OutFile, i, j, NON[Reverse], Full_Matrix );
      }
    }
    if ( Full_Matrix )
      fprintf(OutFile, "\n");
  }
}

void gen_series_open(
		     FILE		*OutFile,
		     int		Net_Size,
		     boolean		Full_Matrix,
		     boolean		NoLoop,
		     boolean		Reverse )
{
  int	i;
  int	j;

  for (i=0; i<Net_Size; i++) {
    for (j=0; j<Net_Size; j++ ) {
      if ( ( ( i == j ) && !NoLoop ) || ( i+1 == j ) )
	write_entry( OutFile, i, j, OUI[Reverse], Full_Matrix );
      else
	write_entry( OutFile, i, j, NON[Reverse], Full_Matrix );
    }
    if ( Full_Matrix )
      fprintf(OutFile, "\n");
  }
}

void gen_cycle(
	       FILE		*OutFile,
	       int		Net_Size,
	       boolean		Full_Matrix,
	       boolean		NoLoop,
	       boolean		Reverse )
{
  int	i;
  int	j;

  for (i=0; i<Net_Size; i++) {
    for (j=0; j<Net_Size; j++ ) {
      if ( ( ( i == j ) && !NoLoop ) || ( (i+1) % Net_Size == j ) )
	write_entry( OutFile, i, j, OUI[Reverse], Full_Matrix );
      else
	write_entry( OutFile, i, j, NON[Reverse], Full_Matrix );
    }
    if ( Full_Matrix )
      fprintf(OutFile, "\n");
  }
}


void gen_complete(FILE		*OutFile,
		  int		Net_Size,
		  boolean	Full_Matrix,
		  boolean	Reverse )
{
  int	i;
  int	j;

  for (i=0; i<Net_Size; i++) {
    for (j=0; j<Net_Size; j++ ) {
      write_entry( OutFile, i, j, OUI[Reverse], Full_Matrix );
    }
    if ( Full_Matrix )
      fprintf(OutFile, "\n");
  }
}

void gen_complete_bipartite(FILE	*OutFile,
			    int		Net_Size,
			    int		Net_Size2,
			    boolean	Full_Matrix,
			    boolean	NoLoop,
			    boolean	Reverse )
{
  int	i;
  int	j;

  for (i=0; i<Net_Size; i++) {
    for (j=0; j<Net_Size; j++ ) {
      if ( ( i == j ) && !NoLoop ) {
	write_entry( OutFile, i, j, OUI[Reverse], Full_Matrix );
      }
      else {
	write_entry( OutFile, i, j, NON[Reverse], Full_Matrix );
      }
    }
    for (j=Net_Size; j<Net_Size+Net_Size2; j++ )
      write_entry( OutFile, i, j, OUI[Reverse], Full_Matrix );

    if ( Full_Matrix )
      fprintf(OutFile, "\n");
  }
  for (i=Net_Size; i<Net_Size+Net_Size2; i++) {
    for (j=0; j<Net_Size; j++ )
      write_entry( OutFile, i, j, OUI[Reverse], Full_Matrix );
    for (j=Net_Size; j<Net_Size+Net_Size2; j++ ) {
      if ( ( i == j ) && !NoLoop ) {
	write_entry( OutFile, i, j, OUI[Reverse], Full_Matrix );
      }
      else {
	write_entry( OutFile, i, j, NON[Reverse], Full_Matrix );
      }
    }
    if ( Full_Matrix )
      fprintf(OutFile, "\n");
  }
}

void gen_star(
	      FILE		*OutFile,
	      int		Net_Size,
	      boolean		Full_Matrix,
	      boolean		NoLoop,
	      boolean		Reverse )
{
  int	i;
  int	j;

  for (i=0; i<Net_Size-1; i++) {
    for (j=0; j<Net_Size-1; j++ ) {
      if ( ( i == j ) && !NoLoop )
	write_entry( OutFile, i, j, OUI[Reverse], Full_Matrix );
      else
	write_entry( OutFile, i, j, NON[Reverse], Full_Matrix );
    }
    write_entry( OutFile, i, j, OUI[Reverse], Full_Matrix );
    if ( Full_Matrix )
      fprintf(OutFile, "\n");
  }

  for ( j=0; j<Net_Size-1; j++ )
    write_entry( OutFile, i, j, OUI[Reverse], Full_Matrix );
  if ( !NoLoop )
    write_entry( OutFile, Net_Size-1, Net_Size-1, OUI[Reverse], Full_Matrix );
  else
    write_entry( OutFile, Net_Size-1, Net_Size-1, NON[Reverse], Full_Matrix );

  if ( Full_Matrix )
    fprintf(OutFile, "\n");

}
			
void display_help(void)
{
  printf("Usage: %s [-t <type>] [-s <size>] ", TOOLNAME );
  printf("[-reverse] [-p proba] [-o <file>]");
  printf("\n\t[-full] [-h] [-?]\n");
  printf("  <type> : cycle (def) line star random Random complete komplete\n");
}

int main(int argc, char *argv[] )
{
  Net_Type	Graph_Type = Cycle;
  static char	File_Name[128] = "";
  int		Net_Size = 2;
  int		Net_Size2 = 2;
  int		i;
  int		Res;
  double	Proba = 0.5;
  FILE*		OutFile;
  boolean	Full_Matrix = FALSE;
  boolean	Args_Ok = TRUE;
  boolean	Reverse = FALSE;
  boolean	NoLoop = FALSE;
  boolean	Help = FALSE;

  i = 1;
  if (argc>1) {
    do if ( 0 == strcmp( "-t", argv[i] ) ) {
      i++;
      if ( 0 == strcmp( "line", argv[i] ) )
	Graph_Type = Line;
      else if ( 0 == strcmp( "cycle", argv[i] ) )
	Graph_Type = Cycle;
      else if ( 0 == strcmp( "star", argv[i] ) )
	Graph_Type = Star;
      else if ( 0 == strcmp( "complete", argv[i] ) )
 	Graph_Type = Complete;
      else if ( 0 == strcmp( "komplete", argv[i] ) )
	Graph_Type = BiComplete;
      else if ( 0 == strcmp( "random", argv[i] ) )
	Graph_Type = Random;
      else if ( 0 == strcmp( "Random", argv[i] ) )
	Graph_Type = RRandom;
      else printf("%s: unrecognized graph type\n", TOOLNAME);
      i++;
    }
    else if ( 0 == strcmp( "-s", argv[i] ) ) {
      i++;
      if ( Graph_Type == BiComplete ) {
	if ( ( i+1 >= argc ) 
	     || ( 1 != sscanf( argv[i], "%d", &Net_Size ) )
	     || ( 1 != sscanf( argv[i+1], "%d", &Net_Size2 ) ) ) {
	  printf("%s: two size expected after %s when complete bipartite\n", 
		 TOOLNAME, argv[i-1] );
	  Args_Ok = FALSE;
	}
	else {
	  i++;
	  i++;
	}
      }
      else {
	if ( ( i >= argc ) 
	     || ( 1 != sscanf( argv[i], "%d", &Net_Size ) ) ) {
	  printf("%s: size expected after %s\n", TOOLNAME, argv[i-1] );
	  Args_Ok = FALSE;
	}
	else
	  i++;
      }
    }
    else if ( 0 == strcmp( "-p", argv[i] ) ) {
      i++;
      Res = sscanf( argv[i], "%lf", &Proba );
      if ( ( Res != 1 ) || ( Proba < 0.0 ) || ( Proba > 1.0 ) ) {
	printf("%s: probability expected after %s\n", 
	       TOOLNAME, argv[i-1] );
	Args_Ok = FALSE;
      }
      else
	i++;
    }
    else if ( 0 == strcmp( "-o", argv[i] ) ) {
      i++;
      if ( i < argc )
	strcpy( File_Name, argv[i] );
      else
	printf("%s: filename expected\n", TOOLNAME);
      i++;
    }
    else if ( 0 == strcmp( "-full", argv[i] ) ) {
      i++;
      Full_Matrix = TRUE;
    }
    else if ( ( 0 == strcmp( "-h", argv[i] ) )
		|| ( 0 == strcmp( "-?", argv[i] ) ) ) {
      i++;
      display_help();
      Args_Ok = FALSE;
      Help = TRUE;
    }
    else if ( 0 == strcmp( "-reverse", argv[i] ) ) {
      i++;
      Reverse = TRUE;
    }
    else if ( 0 == strcmp( "-noloop", argv[i] ) ) {
      i++;
      NoLoop = TRUE;
    }
    else if ( 0 == strcmp( "-v", argv[i] ) ) {
      i++;
      printf("%s (Version %s)\n", TOOLNAME, VERSION);
      Args_Ok = FALSE;
      Help = TRUE;
    }
    else 
      i++;	/* skip argument */
    while ( ( i < argc ) && Args_Ok );
  }

  if (Args_Ok) {
    if ( strlen(File_Name) != 0 )
      OutFile = fopen( File_Name, "w" );
    else
      OutFile = stdout;
    
    gen_size_weight( OutFile, 
		     ( Graph_Type == BiComplete ? 
		       Net_Size+Net_Size2 :
		       Net_Size ), 
		     Full_Matrix );

    switch( Graph_Type ) {
    case Line:
      gen_series_open( OutFile, Net_Size, Full_Matrix, NoLoop, Reverse );
      break;
    case Cycle:
      gen_cycle( OutFile, Net_Size, Full_Matrix, NoLoop, Reverse );
      break;
    case Star:
      gen_star( OutFile, Net_Size, Full_Matrix, NoLoop, Reverse );
      break;
    case Complete:
      gen_complete( OutFile, Net_Size, Full_Matrix, Reverse );
      break;
    case BiComplete:
      gen_complete_bipartite( OutFile, Net_Size, Net_Size2,
			      Full_Matrix, NoLoop, Reverse );
      break;
    case RRandom:
      srandom( (unsigned)time( (time_t)NULL ) );
    case Random:
      gen_random( OutFile, Net_Size, Proba, Full_Matrix, NoLoop, Reverse );
      break;
    }
    return 0;
  }
  else if ( Help )
    return 0;
  else
    return -1;
}
