/* program/module for generating a "Tetris" set from an interaction	*/
/* graph.								*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "boolean.h"
#include "parseargs.h"
#include "i_graph.h"
#include "tetris.h"

#define	TOOLNAME	"gr2tetris"
#define VERSION		"1.1"

#define max(x,y)        ( (x>y) ? x : y )
#define min(x,y)        ( (x<y) ? x : y )

#define MAX_NAME_LENGTH	256

typedef struct EDGE {
  int	Id;
  int	Orig;
  int	Dest;
} *Edge;

typedef struct CLIQUE {
  int	Cardinal;
  int	*Member;
} *Clique;

/* Global declarations */
extern char	*todays_date( void );

void Gen_Tetris_Code( char File_Name[MAX_NAME_LENGTH], Tetris The_Game );
void Find_Clique_Cover( Boolean_Matrix A, int Max_Size, boolean Optimize,
			boolean Greedy,
			int *Nb_Columns, Piece_Desc **Piece_List );
Edge	*Make_Edge_List( Boolean_Matrix, int Nb_Nodes );
void 	Set_Attributes( Tetris The_Game ) ;

boolean	Debug = FALSE;

int main(int argc, char *argv[] )
{
  static char	In_File_Name[MAX_NAME_LENGTH] = "";
  static char	Out_File_Name[MAX_NAME_LENGTH] = "";

  int		i;
  int		Max_Clique_Size = 2;
  boolean	Optimize = FALSE;
  boolean	Greedy = FALSE;
  boolean	Help = FALSE;
  boolean	Args_Ok = TRUE;
  boolean	File_Ok = TRUE;
  Tetris	The_Game;
  Interaction_Graph G;
  Piece_Desc	*Piece_List;
  int		Nb_Columns;

  i = 1;

  if (argc>1)
    do {
      if GET_STRING( "-f", In_File_Name )

      else if GET_STRING( "-o", Out_File_Name )

      else if GET_INT( "-maxsize", Max_Clique_Size )

      else if SWITCH( "-optimize", Optimize, TRUE )		       
      else if SWITCH( "-greedy", Greedy, TRUE )		       
      else if SWITCH( "-debug", Debug, TRUE )		       

      else if IS_VERSION

      else if ( ( 0 == strcmp( "-h", argv[i] ) )
		|| ( 0 == strcmp( "-?", argv[i] ) ) ) {
	printf("Usage: %s [-f <file>] [-o <file>]", TOOLNAME );
	printf(" [-maxsize <int>] [-optimize] [-greedy]");
	printf("\n\t[-debug] [-h] [-?]\n");
	Args_Ok = FALSE;
	Help = TRUE;
	i++;
      }
      else
	i++; /* skip unknown argument */
    } while ( Args_Ok && ( i < argc ) );

  if (Args_Ok) {
    if ( strlen(In_File_Name) != 0 )
      File_Ok = ( (FILE*)NULL != freopen( In_File_Name, "r", stdin ) );
    if ( !File_Ok ) {
      fprintf( stderr, "%s: error opening file %s\n",
	       TOOLNAME, In_File_Name );
      return -1;
    }
    else {
      if ( strlen(Out_File_Name) != 0 )
	File_Ok = ( (FILE*)NULL != freopen( Out_File_Name, "w", stdout ) );

      if ( !File_Ok ) {
	fprintf( stderr, "%s: error opening file %s\n",
		 TOOLNAME, Out_File_Name );
	return -1;
      }
      else {

	G = Read_Interaction_Graph( stdin );

	if ( (Interaction_Graph)NULL != G ) {

	  if ( Is_Uniform( G ) ) {

	    Find_Clique_Cover( G->A, Max_Clique_Size, 
			       Optimize, Greedy,
			       &Nb_Columns,
			       &Piece_List );

	    if ( NULL != Piece_List ) {

	      The_Game = New_Tetris( Nb_Columns, G->Size, Piece_List );

	      Set_Colors( The_Game );
	      Write_Tetris( The_Game, stdout );

	      /* successful! */
	      Free_Tetris( The_Game );

	      return 0;
	    }
	    else {
	      fprintf( stderr, 
		       "%s: could not find clique cover. No translation.\n", 
		       TOOLNAME );
	      return -5;
	    }
	  }
	  else {
	    fprintf( stderr, 
		     "%s: weights are not uniform. Cannot translate.\n", 
		     TOOLNAME );
	    return -4;
	  }
	}
	else {
	  fprintf( stderr, "%s: bad input data. No translation.\n", 
		   TOOLNAME );
	  return -2;
	}
      }
    }
  }
  else if ( Help )
    return 0;
  else
    return -1;
}

/* Finds a clique covering of a graph given by an adjacency matrix A */
/* Warning! The procedure modifies the contents of A in the process. */
void Find_Clique_Cover( Boolean_Matrix A, int Max_Size, 
			boolean Optimize, boolean Greedy,
			int *Nb_Columns, Piece_Desc **Piece_List )
{
  int		Edge_Count;
  int		i;
  int		j;
  int		k;
  boolean	*Used;
  Piece_Desc	*Res;
  int		Clique_Nb;
  Clique	*Clique_List;
  Clique	The_Clique;

  /* Count edges in A */
  Edge_Count = 0;
  for ( i=0; i<A->Size; i++ )
    for ( j=i+1; j<A->Size; j++ )
      if ( Is_Adj( A, i, j ) )
	Edge_Count++;

  Used = (boolean*) calloc( (size_t)Edge_Count, sizeof(boolean) );
  for ( k=0; k<Edge_Count; k++ )
    Used[k] = FALSE;

  Clique_Nb = 0;
  Clique_List = (Clique*)NULL;

  if ( !Optimize && ( Max_Size == 2 ) ) {
    /* the standard request/resource set. See Brilman, p. 26 */

    for ( i=0; i<A->Size; i++ )
      for ( j=i+1; j<A->Size; j++ ) 
	if ( Is_Adj( A, i, j ) ) {
	  Clique_Nb++;
	  Clique_List = realloc( Clique_List, 
				 (unsigned)Clique_Nb*sizeof(Clique) );
	  Clique_List[Clique_Nb-1] = malloc( sizeof(struct CLIQUE) );
	  The_Clique = Clique_List[Clique_Nb-1];
	  The_Clique->Cardinal = 2;
	  The_Clique->Member = (int*)malloc( 2 * sizeof(int) );
	  The_Clique->Member[0] = i;
	  The_Clique->Member[1] = j;

	  if ( Debug ) {
	    printf("Cl #%d is edge (%d,%d)\n", Clique_Nb-1, i, j );
	  }
	}
  }
  else if ( !Optimize && ( Max_Size == 3 ) ) {

    /* first look for triangles */

    for ( i=0; i<A->Size; i++ )
      for ( j=i+1; j<A->Size; j++ )
	if ( Is_Adj( A, i, j ) ) {
	  for ( k=j+1; k<A->Size; k++ ) {
	    /* note: have to re-check that A[i][j] is still there */
	    if ( ( Is_Adj( A, i, j ) ) && ( Is_Adj( A, i, k ) )
		 && ( Is_Adj( A, j, k ) ) ) {
	      Clique_Nb++;
	      Clique_List = realloc( Clique_List, 
				     (unsigned)Clique_Nb*sizeof(Clique) );
	      Clique_List[Clique_Nb-1] = malloc( sizeof(struct CLIQUE) );
	      The_Clique = Clique_List[Clique_Nb-1];
	      The_Clique->Cardinal = 3;
	      The_Clique->Member = (int*)malloc( 3 * sizeof(int) );
	      The_Clique->Member[0] = i;
	      The_Clique->Member[1] = j;
	      The_Clique->Member[2] = k;

	      if ( Debug ) {
		printf("Cl #%d is triangle (%d,%d,%d)\n", 
		       Clique_Nb-1, i, j, k );
	      }

	      /* indicate that the edges have been used */
	      Set_Adj( A, j, i, 0 );
	      Set_Adj( A, k, i, 0 );
	      Set_Adj( A, k, j, 0 );
	      if ( Greedy ) {
		/* do not look for them again */
		Set_Adj( A, i, j, 0 );
		Set_Adj( A, i, k, 0 );
		Set_Adj( A, j, k, 0 );
	      }
	    }
	  }
	}
    /* now collect edges not in triangles */
    for ( i=0; i<A->Size; i++ )
      for ( j=0; j<i; j++ ) 
	if ( Is_Adj( A, i, j ) ) {
	  Clique_Nb++;
	  Clique_List = realloc( Clique_List, 
				 (unsigned)Clique_Nb*sizeof(Clique) );
	  Clique_List[Clique_Nb-1] = malloc( sizeof(struct CLIQUE) );
	  The_Clique = Clique_List[Clique_Nb-1];
	  The_Clique->Cardinal = 2;
	  The_Clique->Member = (int*)malloc( 2 * sizeof(int) );
	  The_Clique->Member[0] = i;
	  The_Clique->Member[1] = j;

	  if ( Debug ) {
	    printf("Cl #%d is edge (%d,%d)\n", Clique_Nb-1, i, j );
 	  }
	}
  }
  else {
    printf("%s: error: cannot compute covering with cliques of size > 3.\n",
	   TOOLNAME );
    Res = (Piece_Desc*) NULL;
  }

  (*Nb_Columns) = Clique_Nb;

  if ( Clique_Nb > 0 ) {
    /* prepare the pieces */
    Res = (Piece_Desc*) calloc( A->Size, sizeof(Piece_Desc) );
    for ( i=0; i<A->Size; i++ ) {
      Res[i] = New_Piece( Clique_Nb );
      Res[i]->Number = i;
    }

    /* scan cliques and set piece profiles up */
    for ( k=0; k<Clique_Nb; k++ ) {
      The_Clique = Clique_List[k];
      for ( i=0; i<The_Clique->Cardinal; i++ ) {
	Res[The_Clique->Member[i]]->LoProf[k] = 0;
	Res[The_Clique->Member[i]]->HiProf[k] = 1;
      }
    }
  }

  (*Piece_List) = Res;

  free( Used );
}

Edge	*Make_Edge_List( Boolean_Matrix A, int Nb_Nodes )
{
  int	i;
  int	j;
  int	k;
  int	Edge_Count;
  Edge	*Res;

  /* Count edges in A */
  Edge_Count = 0;
  for ( i=0; i<Nb_Nodes; i++ )
    for ( j=i+1; j<Nb_Nodes; j++ )
      if ( Is_Adj ( A, i, j ) )
	Edge_Count++;

  Res = (Edge*) calloc( Edge_Count, sizeof(Edge) );

  k = 0;
  for ( i=0; i<Nb_Nodes; i++ )
    for ( j=i+1; j<Nb_Nodes; j++ ) 
      if ( Is_Adj ( A, i, j ) ) {
	Res[k] = (Edge) malloc( sizeof( struct EDGE ) );
	Res[k]->Id = k;
	Res[k]->Orig = i;
	Res[k]->Dest = j;
	if ( Debug ) {
	  printf("ED %d -> %d\n", i, j );
	}
	k++;
      }

  return Res;
}





