/**************************************************/
/* simulation de calcul de gamma dans des graphes */
/**************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <math.h>
#include <string.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include "boolean.h"
#include "parseargs.h"
#include "maxpluscore.h"
#include "i_graph.h"

#define TOOLNAME	"gr_sim"
#define VERSION		"1.0"

#define MAX_RANDOM_VALUE        0x7FFFFFFF
#define infty 			-10000

#define MAX_NAME_LENGTH 256

boolean Debug = FALSE;

#ifndef ALPHA
extern long random(void);
#endif

#define PRODUCT(A,X,Y) MV_Mult(A,X,Y)

/* forward declarations */
boolean	Compile_Graph_And_Run( char In_File_Name[MAX_NAME_LENGTH],
			       int Sim_Length,
			       boolean Keep,
			       char *argp[] );
void Simulate ( Bernoulli_Product 	Problem,
		int 			Sim_Length,
		int 			Nb_Replicates,
		boolean			Stats );
        

/*************************************************************************/
/*           Procedures elementaires                                     */
/*************************************************************************/
int choice(double *Pr_Cum)
{
  int 		i;
  double 	ra;

  ra = (double)random()/MAX_RANDOM_VALUE;
  i = 0;
  while ( ra > Pr_Cum[i])
    i++;

  return i;
}

/***********************************************************************/
/*                Programme Principal                                  */
/***********************************************************************/

int main(int argc, char *argv[], char *argp[] )
{
  static char	In_File_Name[MAX_NAME_LENGTH] = "";
  static char	Out_File_Name[MAX_NAME_LENGTH] = "";

  int 		i;
  int 		Sim_Length = 100000;     /* length of the simulation */
  int		Nb_Replicates = 1;
  boolean	Stats = FALSE;
  boolean	Compile = FALSE;
  boolean	Keep = FALSE;
  boolean	Help = FALSE;
  boolean	Args_Ok = TRUE;
  boolean	File_Ok = TRUE;
  Interaction_Graph	G;
  Bernoulli_Product	BP;

  /******************* Parsing command line args *************************/
  i = 1;

  if (argc>1)
    do {

      if GET_STRING( "-f", In_File_Name )
      else if GET_STRING( "-o", Out_File_Name )
      else if GET_INT( "-nevt", Sim_Length )
      else if GET_INT( "-nrep", Nb_Replicates )
      else if SWITCH( "-stats", Stats, TRUE )
      else if SWITCH( "-compile", Compile, TRUE )
      else if SWITCH( "-keep", Keep, TRUE )
      else if SWITCH( "-debug", Debug, TRUE )
      else if IS_VERSION

      else if (( 0 == strcmp( "-h", argv[i] )) || 
	       ( 0 == strcmp( "-?", argv[i] )) ||
	       ( 0 == strcmp( "-help", argv[i] ))) {	
	i++;
	fprintf(stdout,"usage: %s [-f inputfile] [-o outputfile] ",
		TOOLNAME );
	fprintf(stdout,"[-nevt <int>] [-nrep <int>]\n\t[-stats] ");
	fprintf(stdout," [-compile] [-keep]\n\t[-v] [-h] [-help] [-?]\n");
	Args_Ok = FALSE;
	Help = TRUE;
      }
      else 
	/* skip unknown argument */
	i++;
    } while( Args_Ok && ( i < argc ) );

  if (Args_Ok) {

    if ( strlen(In_File_Name) != 0 )
      File_Ok = ( NULL != freopen( In_File_Name, "r", stdin ) );

    if ( !File_Ok ) {
      fprintf(stderr,"%s: could not open file %s\n", TOOLNAME, In_File_Name );
    }

    if ( File_Ok ) {
      if ( strlen(Out_File_Name) != 0 )
	File_Ok = (NULL != freopen( Out_File_Name, "w", stdout ) );

      if ( !File_Ok ) {
	fprintf(stderr,"%s: could not open file %s\n", 
		TOOLNAME, Out_File_Name);
	File_Ok = FALSE;
      }
    }
  } else if ( Help ) {
    return 0;
  } else {
    return -1;
  }

  /************* start of the program if everything is Ok *******************/

  if ( Args_Ok && File_Ok ) {

    if ( Compile ) {
      if ( Compile_Graph_And_Run( In_File_Name, Sim_Length, Keep, argp ) )
	return 0;
      else 
	return -3;
    }
    else {

      /************* read the incidence matrix + prob vector ****************/
      G = Read_Interaction_Graph( stdin );
      /***************** compute synchronization matrices *****************/

      if ( (Interaction_Graph)NULL != G ) {

	BP = IG_To_BP( G );

	if ( Debug ) {
	  Write_Bernoulli_Product( BP, stdout, FORMAT_FULL, "A" );
	}
	/***************** compute gamma *****************/
	if ( (Bernoulli_Product)NULL != BP ) {

	  Simulate( BP, Sim_Length, Nb_Replicates, Stats );

	  /* simulation successful */
	  Free_Interaction_Graph( G );
	  Free_Bernoulli_Product( BP );

	  return 0;
	}
	else {
	  printf("%s: error: could not convert to (max,+) matrices.\n", 
		 TOOLNAME );
	  return -4;
	}
      }
      else {
	fprintf( stderr, "%s: bad input data. No simulation.\n", 
		 TOOLNAME );
	return -2;
      }
    }
  }
  else if ( !Help ) {
    return -1;
  }
  else
    return 0;
}

#define TEMP_FILE "/tmp/gr_sim.graph"
#define TEMP_C_FILE "/tmp/gr_sim.c"
#define TEMP_SIM "/tmp/gr_sim"

boolean	Compile_Graph_And_Run( char 	In_File_Name[MAX_NAME_LENGTH],
			       int	Sim_Length,
			       boolean	Keep,
			       char 	*argp[] )
{
  char		gr2c_Input[MAX_NAME_LENGTH];
  char		command[2*MAX_NAME_LENGTH];
  char		arg[MAX_NAME_LENGTH];
  boolean	All_Ok = TRUE;
  int		Son_PID;
  int		Res;
  FILE		*Out;
  char		c;
    
  if ( strlen( In_File_Name ) > 0 ) {
    strncpy( gr2c_Input, In_File_Name, MAX_NAME_LENGTH );
  }
  else {
    /* pipe the standard input to a temporary file */
    All_Ok = ( NULL != ( Out = fopen( TEMP_FILE, "w" ) ) );
    if ( !All_Ok ) {
      printf("%s: error: could not open file %s.\n", 
	     TOOLNAME, TEMP_FILE );
      return FALSE;
    }
    else {
      do {
	c = getchar();
	if ( c != EOF )
	  fprintf( Out, "%c", c );
      } while ( c != EOF );
    }
    fclose( Out );

    strncpy( gr2c_Input, TEMP_FILE, MAX_NAME_LENGTH );
  }

  if ( All_Ok ) {

    sprintf( command, "gr2c -f %s -o %s", gr2c_Input, TEMP_C_FILE );
    printf("Executing: %s\n", command );
    if ( 0 == ( Son_PID = fork() ) ) {
      Res = execlp( "gr2c", "gr2c", "-f", gr2c_Input, "-o", TEMP_C_FILE,
		    (char*)NULL, argp );
      printf("Res = %d, Err = %d\n", Res, errno );
    }
    else
      waitpid( Son_PID, NULL, 0 );

    if ( 0 == ( Son_PID = fork() ) ) {
      sprintf( command, "gcc %s -o %s -lm -O2", TEMP_C_FILE, TEMP_SIM );
      printf("Executing: %s\n", command );
      
      Res = execlp( "gcc", "gcc", TEMP_C_FILE, "-o", TEMP_SIM, 
		   "-lm", "-O2", (char*)NULL, argp );
      printf("Res = %d, Err = %d\n", Res, errno );
    }
    else
      waitpid( Son_PID, NULL, 0 );
	
    if ( 0 == ( Son_PID = fork() ) ) {

      sprintf( command, "%s -nevt %d", TEMP_SIM, Sim_Length );
      printf("Executing: %s\n", command );
      sprintf( arg, "%d", Sim_Length );
      Res = execl( TEMP_SIM, TEMP_SIM, "-nevt", arg, (char*)NULL );
      printf("Res = %d, Err = %d\n", Res, errno );
    }
    else
      waitpid( Son_PID, NULL, 0 );

    if ( !Keep ) {
      unlink( TEMP_FILE );
      unlink( TEMP_C_FILE );
      unlink( TEMP_SIM );
    }

    return TRUE;
  }
  else
    return FALSE;
}

/* Bernoulli product Simulation procedure from mat_sim.c */
void Simulate ( Bernoulli_Product 	Problem,
		int 			Sim_Length,
		int 			Nb_Replicates,
		boolean			Stats )
{
  int		i;
  int		l;
  int		n;
  int		r;
  MaxPlus_Vector X;
  MaxPlus_Vector Y;
  MaxPlus_Vector Tmp;
  double	G;		/* Temporary convenience variable */
  double	*Gamma;		/* estimate for average */
  double	*Gamma2;	/* estimate for variance */
  double	Sigma_Hat;

  X =      New_MaxPlus_Vector( Problem->Size, NONE );
  Y =      New_MaxPlus_Vector( Problem->Size, NONE );
  Gamma =  (double *) calloc( (size_t)Problem->Size, sizeof(double));
  Gamma2 = (double *) calloc( (size_t)Problem->Size, sizeof(double));

  for ( i = 0; i < Problem->Size; i++ ) {
    Gamma[i] = 0.0;
    Gamma2[i] = 0.0;
  }

  for ( n = 0; n < Nb_Replicates; n++ ) {
    /* (re) init recurrence vector */
    for ( i = 0; i < Problem->Size; i++ ) {
#ifdef USE_FULL_MP_MATRICES      
      X->Vals[i][0] = 0.0;
#else
      X->Vals[i] = 0.0;
#endif
    }

    for (l = 0; l < Sim_Length; l++) {
      r = choice( Problem->Pr_Cum );
      PRODUCT( Problem->A[r], X, Y );
      Tmp = X; X = Y; Y = Tmp;
    }

    /* accumulate stats */
    for ( i = 0; i < Problem->Size; i++ ) {
#ifdef USE_FULL_MP_MATRICES  
      G = X->Vals[i][0]/(double)Sim_Length;
#else
      G = X->Vals[i]/(double)Sim_Length;
#endif
      Gamma[i] += G;
      Gamma2[i] += G*G;
    }     
  }

  /* Compute stats and output result */
  if ( Sim_Length > 0 ) {
    for (i=0;i<Problem->Size;i++) {

      printf("gamma[%d] = ", i );
      if ( Gamma[i] == M_INFINITY )
	printf("eps\n");
      else
	printf("%f", (double) Gamma[i] / Nb_Replicates );

      if ( Stats && ( Nb_Replicates > 1 ) ) {
	Sigma_Hat = sqrt( ( Gamma2[i] - Gamma[i]*Gamma[i]/Nb_Replicates ) 
			  / (Nb_Replicates-1) );
	printf("  +/- %f (conf. 95%%)", 
	       Sigma_Hat * 1.96 / sqrt(Nb_Replicates) );
      }
      printf("\n");
    }
  }

  Free_MaxPlus_Vector( X );
  Free_MaxPlus_Vector( Y );
  free( Gamma );
  free( Gamma2 );
}
