/* Implementation of Methods for Interaction Graphs	*/

#include <malloc.h>
#include "i_graph.h"
#include "bernoulli.h"

#define TOOLNAME "I_Graph"

extern boolean Debug;

/* Contructor and Destructor */
Interaction_Graph New_Interaction_Graph( int Matrix_Size )
{
  boolean		Alloc_Ok;
  Interaction_Graph 	Res;
  
  Res = (Interaction_Graph) malloc( sizeof( struct INTERACTION_GRAPH ) );

  if ( (Interaction_Graph)NULL != Res ) {
    Res->Size = Matrix_Size;

    Res->Weight = (double*) calloc( (size_t)Matrix_Size, sizeof(double) );
    Res->Proba  = (double*) calloc( (size_t)Matrix_Size, sizeof(double) );
    Res->Pr_Cum = (double*) calloc( (size_t)Matrix_Size, sizeof(double) );
    Res->Permut = (int*)    calloc( (size_t)Matrix_Size, sizeof(int) );
    Res->A = New_Boolean_Matrix( Matrix_Size );

    Alloc_Ok = ( ( NULL != Res->Weight ) 
		 && ( NULL != Res->Proba ) 
		 && ( NULL != Res->Pr_Cum ) 
		 && ( NULL != Res->Permut ) 
		 && ( (Boolean_Matrix)NULL != Res->A ) );    
 
    if ( !Alloc_Ok ) {
      fprintf(stderr,"%s: memory error. Not enough space.\n", TOOLNAME );
      Free_Interaction_Graph( Res );
      Res = (Interaction_Graph)NULL;
    }
  }

  return Res;
}

/* garbage collecting procedure. Probably useless in the present state. */
#define FREE(ptr,type) if ( (type*)NULL != (ptr) ) { free( (void*)ptr ); }

void Free_Interaction_Graph( Interaction_Graph Problem ) 
{
  if ( (Interaction_Graph) NULL != Problem ) {
    FREE( Problem->Weight, double );
    FREE( Problem->Proba, double );
    FREE( Problem->Pr_Cum, double );
    FREE( Problem->Permut, int );
    if ( (Boolean_Matrix)NULL != Problem->A ) {
      Free_Boolean_Matrix( Problem->A );
    }
    free( Problem );
  }
}

/* Properties */
boolean Is_Uniform( Interaction_Graph G )
{
  int 		i;
  boolean	Res = TRUE;

  for ( i=1; i<G->Size; i++ ) {
    Res = Res && ( G->Weight[i] == G->Weight[0] );
  }

  return Res;
}

/* I/O */
Interaction_Graph Read_Interaction_Graph( FILE *Input )
{
  int		i;
  int		j;
  int		Nb_Nodes;
  int		Val;
  double	Sum;
  boolean	Ok = TRUE;
  boolean	Continue;
  boolean	Sparse_Input;
  char		Tok[32];
  Interaction_Graph	Res;

  /* look for first token. */
  fscanf( Input, "%32s", Tok );
  if ( !strcmp( Tok, "sp" ) ) {
    Sparse_Input = TRUE;
    Ok = ( 1 == scanf( "%d", &Nb_Nodes ) );
  }
  else {
    Sparse_Input = FALSE;
    Ok = ( 1 == sscanf( Tok, "%d", &Nb_Nodes ) );
  }

  if ( Ok ) {
    Res = New_Interaction_Graph( Nb_Nodes );
    Ok = ( (Interaction_Graph)NULL != Res );
  }

  if ( Ok ) {
    /* reading probas */
    Sum = 0.0;
    for ( i=0; i<Res->Size; i++ ) {
      Ok = Ok && ( 1 == scanf( "%lf",  &(Res->Weight[i]) ) );
      Sum += Res->Weight[i];
    }
    /* Next normalize them to one, while computing cumulants */
    for ( i=0; i<Res->Size; i++ ) {
      Res->Proba[i] = Res->Weight[i] / Sum;
      if ( i == 0 )
	Res->Pr_Cum[i] = Res->Proba[i];
      else if ( i < Res->Size - 1 )
	Res->Pr_Cum[i] = Res->Pr_Cum[i-1] + Res->Proba[i];
      else 
	/* putting a safe "guard" just in case of roundoff errors */
	Res->Pr_Cum[i] = 1.0;
    }

    /* now reading the adjacency matrix */
    if ( Ok ) {
      if ( Sparse_Input ) {
	Continue = TRUE;
	while( Continue ) {
	  Continue = ( 2 == fscanf( Input, "%d %d", &i, &j ) );
	  if ( ( i < 0 ) || ( i >= Res->Size ) ) {
	    printf("Error: out of range i value found (%d). Abort.\n", i );
	    Continue = FALSE;
	    Ok = FALSE;
	  } 
	  else if ( ( j < 0 ) || ( j >= Res->Size ) ) {
	    printf("Error: out of range j value found (%d). Abort.\n", j );
	    Continue = FALSE;
	    Ok = FALSE;
	  }
	  if ( Continue ) {
	    Set_Adj( Res->A, i, j, 1 );
#ifdef USE_SP_BOOL_MATRICES
	    if ( Debug ) {
	      printf("recording arc %d->%d\n", i, j );
	    }
#endif
	  }
	}
      }
      else if ( Ok ) {
	for ( i=0; (i<Res->Size) && Ok; i++ ) {
	  for ( j=0; (j<Res->Size) && Ok; j++ ) {
	    scanf( "%d",  &Val );
#ifdef USE_SP_BOOL_MATRICES
	    if ( Val == 1 ) {
	      Res->A->Nb_Elts[i] ++;
	      Res->A->Elts[i] = 
		(int*) realloc( Res->A->Elts[i], 
				(size_t)Res->A->Nb_Elts[i]*sizeof(int) );
	      Res->A->Elts[i][Res->A->Nb_Elts[i]-1] = j;
	    }
#else
	    Res->A->Vals[i][j] = Val;
#endif
	  }
	}
      }
    }
  }

  if ( !Ok ) {
    Res = (Interaction_Graph)NULL;
  }

  return Res;
}

/* Canonical conversion into a Bernoulli product */
Bernoulli_Product IG_To_BP( Interaction_Graph G )
{
  int			i;
  int			k;
  Bernoulli_Product 	Res;

  Res = New_Bernoulli_Product( G->Size, G->Size );

  if ( (Bernoulli_Product)NULL != Res ) {
    for ( k=0; k<G->Size; k++ ) {
      Res->Weight[k] = G->Weight[k];
      Res->Proba[k] = G->Proba[k];
      Res->Pr_Cum[k] = G->Pr_Cum[k];
      /* (max,+) matrices have been initialized to eps by New_B_P() */
      /* Init_MaxPlus_Matrix( Res->A[k], ZERO ); */

      /* Fill in Matrix values. Adapted from Bruno's old code. */
      for ( i=0; i<G->Size; i++ ) {
	Set_Matrix_Entry( Res->A[k], i, i, 0.0 );
	if ( Is_Adj( G->A, k, i ) ) {
	  Set_Matrix_Entry( Res->A[k], i, k, 1.0 );
	}
      }
      Set_Matrix_Entry( Res->A[k], k, k, 1.0 );
    }
  }

  return Res;
}
