/************************************************************************/
/* Bounds on the growth rate of a random product of (max,+) matrices	*/
/************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <math.h>
#include <string.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include "boolean.h"
#include "parseargs.h"
#include "maxplusmat.h"
#include "bernoulli.h"

#define TOOLNAME	"mat_bound"
#ifdef USE_FULL_MATRICES
#define VERSION		"1.2 (full matrices)"
#else
#define VERSION		"1.2 (sparse matrices)"
#endif

#define MAX_RANDOM_VALUE          0x7FFFFFFF
#define MAX_NAME_LENGTH 256

#ifndef ALPHA
extern long random(void);
#endif

#define GNUPLOT_STYLE 	0
#define MAPLE_STYLE	1

boolean Debug = FALSE;
char Tmp_String[32];
/************************************************************************/
/* Structure de description d'un probleme de type "produit de matrices" */
/************************************************************************/
typedef struct RANDOM_PRODUCT {
  int		Size;		/* Dimension of the problem */
  int		Nb_Matrices;	/* Number of different matrices */
  double	***A;		/* table of matrices */
  double	*Proba;		/* table of probabilities */
  double	*Pr_Cum;	/* table of cumulative probabilities */
  int		*Permut;	/* Permutation to optimize drawing of probas */
} *Random_Product;

/************************************************************************/
/* Structure pour faciliter le passage de parametres dans le calcul     */
/* de fonctions...                                                      */
/************************************************************************/
typedef struct {
  double        Start;
  double        Stop;
  double        Step;
  boolean       Queue;
  boolean       Log;
  boolean       Log10;
} RANGE_INFO;

/*************************************************************************/
/*           Procedures elementaires                                     */
/*************************************************************************/

/****************************/
#define MAX(a,b) ((a>b)?a:b)

/*************************/
int choice( Bernoulli_Product Problem )
{
  int 		i;
  double 	ra;

  ra = (double)random()/MAX_RANDOM_VALUE;
  i = 0;
  while ( ra > Problem->Pr_Cum[i])
    i++;

  return i;
}


/* I_Binomial returns the binomial coefficient in integer form.		*/
/* I_Binomial( n, p ) = 0 if n < 0.					*/
int I_Binomial( int n, int p )
{
  int Res;
  int i;

  Res = 1;
  for ( i = 1; i <= p; i++ ) {
    Res = ( Res * (n-i+1) ) / i;
  }

  return Res;
}

double Binomial( int n, int p )
{
  double Res;
  int i;

  Res = 1.0;
  for ( i = 1; i <= p; i++ ) {
    Res = ( Res * (double)(n-i+1) ) / (double)i;
  }

  return Res;
}

void	sort_probas( Bernoulli_Product Problem )
{
  int	i;

  /* to be implemented... */
  for ( i=0; i<Problem->Nb_Matrices; i++ )
    Problem->Permut[i] = i;
}

void Print_Term( double Binom_Coeff, double Coeff, int p, int q )
{
  boolean	Has_Coeff;

  Has_Coeff = FALSE;

  if ( q != 0 )
    printf("+ ");
  if ( Binom_Coeff != 1.0 ) {
    printf("%.0f", Binom_Coeff );
    Has_Coeff = TRUE;
  }
  if ( Coeff != 1.0 ) {
    if ( Has_Coeff )
      printf("*");
    printf("%12.8f", Coeff );
    Has_Coeff = TRUE;
  }
  if ( Has_Coeff )
    printf("*");
  if ( p == 1 )
    printf("(1-p)");
  else if ( p > 1 )
    printf("(1-p)**%d", p );
  if ( Has_Coeff && ( p != 0 ) && ( q != 0 ) )
    printf("*");
  if ( q == 1 )
    printf("p");
  else if ( q > 1 )
    printf("p**%d", q );
}

void Compute_Bounds( Bernoulli_Product Problem,
		    int N_Max, boolean Numeric, 
		    RANGE_INFO P_Range, int Style ) 
{
  int		i;
  int		j;
  int		k;
  int		l;
  int		m;
  int		n;
  int		p;
  int		Nb_Points;
  double	Val;
  double	MaxVal;
  double	Coeff;
  double	alpha;
  Full_MaxPlus_Matrix	A;
  Full_MaxPlus_Matrix	B;
  Full_MaxPlus_Matrix	*ApowB;		/* Table of matrices A^m B */
  Full_MaxPlus_Matrix	**M;		/* 2d-Table of vectors */
  double	**U;
  double	*Norm_ApowB;
  double	*Sum;
  double	UB;

  M = (Full_MaxPlus_Matrix**) calloc( (unsigned)(1+N_Max), 
				      sizeof( Full_MaxPlus_Matrix* ) );
  ApowB = (Full_MaxPlus_Matrix*) calloc( (unsigned)(1+N_Max), 
					  sizeof( Full_MaxPlus_Matrix ) );
  U = (double**) calloc( (unsigned)(1+N_Max), sizeof( double* ) );
  Norm_ApowB = (double*) calloc( (unsigned)(1+N_Max), sizeof( double) );

  for ( n=0; n<=N_Max; n++) {
    M[n] = (Full_MaxPlus_Matrix*) calloc( (unsigned)(1+N_Max), 
					  sizeof(Full_MaxPlus_Matrix) );
    for ( k=0; k<=n; k++ ) {
      /* creating column vectors initialized with (real) zeroes */
      /* because M[0,0] is X(0), taken as (0,...,0) by default. */
      M[n][k] = New_Full_MaxPlus_Matrix( Problem->Size, 1, ONE );
    }
    ApowB[n] = New_Full_MaxPlus_Matrix( Problem->Size, Problem->Size,
					NONE );
    U[n] = (double*) calloc( (unsigned)(1+N_Max), sizeof(double) );
  }

  /* initializations */
  A = Convert_MaxPlus_Matrix_To_Full( Problem->A[0] );
  B = Convert_MaxPlus_Matrix_To_Full( Problem->A[1] );
  U[0][0] = 0.0;
  Norm_ApowB[0] = M_Norm( Problem->A[1] );

  /* ApowB[n] contains A^n B, M[n,0] is A^n*X0 and M[n,n] is B^n*X0 */
  for ( i=0; i<Problem->Size; i++ )
    for (j=0; j<Problem->Size; j++ )
      ApowB[0]->Vals[i][j] = B->Vals[i][j];
  for ( n=1; n<=N_Max; n++ ) {
    Full_MM_Mult( A, ApowB[n-1], ApowB[n] );
    Full_MM_Mult( A, M[n-1][0], M[n][0] );
    Full_MM_Mult( B, M[n-1][n-1], M[n][n] );
    U[n][0] = Full_Matrix_Norm( M[n][0] );
    U[n][n] = Full_Matrix_Norm( M[n][n] );
    Norm_ApowB[n] = Full_Matrix_Norm( ApowB[n] );
  }

  /* recursive computation of the vectors M, and the scalars U */
  for ( k=1; k<=N_Max; k++ ) {
    for ( n=k+1; n<=N_Max; n++ ) {

      for ( i=0; i<Problem->Size; i++ ) {

	/* M[n]->Vals[k][i] has been initialized to 0 above */
	alpha = (double) k / (double)n;

	for ( m=0; m<=n-k; m++ ) {

	  MaxVal = M_INFINITY;

	  for ( l=0; l<Problem->Size; l++ ) {
	    if ( ( ApowB[m]->Vals[i][l] != M_INFINITY ) 
		 && ( M[n-m-1][k-1]->Vals[l][0] != M_INFINITY ) ) {
	      /*
		MaxVal = MAX( MaxVal, Binomial(n-m-1,k-1)*ApowB[m][i][l] 
		+ M[n-m-1][k-1][l] );
		*/
	      MaxVal = max( MaxVal, 
			    ApowB[m]->Vals[i][l] + M[n-m-1][k-1]->Vals[l][0] );
	    }
	  }
	  M[n][k]->Vals[i][0] += MaxVal * alpha;

	  alpha *= (double) (n-m-k) / (double) (n-m-1);
	}

	/* computing the upper bound */
	U[n][k] = 0.0;
	alpha = (double) k / (double)n;

	for ( m=0; m<=n-k; m++ ) {
	  U[n][k] += alpha * ( Norm_ApowB[m] + U[n-m-1][k-1] );
	  alpha *= (double) (n-m-k) / (double) (n-m-1);
	}
      }
    }
  }

  /* computation of the bound from the matrices M and scalars U */
  if ( Numeric ) {
    Sum = (double*) calloc( Problem->Size, sizeof(double) );

    /* taking care of degenerate cases */
    if ( P_Range.Start > P_Range.Stop ) {
      fprintf( stderr, "%s: bad range\n", TOOLNAME );
    }
    else if ( P_Range.Step < 0.0 ) {
      fprintf( stderr, "%s: bad step value\n", TOOLNAME );
    }
    else {
      printf("# Bounds bound on the Expectation of X_i(%d) vs p, ", N_Max );
      printf("0 <= i <= %d,\n", Problem->Size-1 );
      printf("# and p in range [%f,%f]\n", P_Range.Start, P_Range.Stop );
      printf("#\n");
      
      Nb_Points = ( P_Range.Stop == P_Range.Start ?
		1 :
		1 + floor( ( P_Range.Stop - P_Range.Start ) / P_Range.Step ) );

      for ( p=0; p<Nb_Points; p++ ) {

	Val = P_Range.Start + p*P_Range.Step;

	if ( Val == 0.0 ) {
	  for ( i=0; i<Problem->Size; i++ )
	    Sum[i] = M[N_Max][N_Max]->Vals[i][0];
	  UB = U[N_Max][N_Max];
	}
	else if ( Val == 1.0 ) {
	  for ( i=0; i<Problem->Size; i++ )
	    Sum[i] = M[N_Max][0]->Vals[i][0];
	  UB = U[N_Max][0];
	}
	else {
	  for ( i=0; i<Problem->Size; i++ )
	    Sum[i] = 0.0;

	  UB = 0.0;
      	  Coeff = pow( Val, (double)N_Max );

	  for ( k=0; k<=N_Max; k++ ) {
	    for ( i=0; i<Problem->Size; i++ )
	      Sum[i] += Coeff * M[N_Max][k]->Vals[i][0];
	    UB += Coeff * U[N_Max][k];
	    Coeff *= (1.0-Val)/Val * (double)(N_Max-k)/(double)(k+1);
	  }
	}
    
	/* output result */
	printf("%8.6f ", Val );

	for (i=0;i<Problem->Size;i++) {
	  /* printf("gamma[%d] = ", i ); */
	  printf("%f ", Sum[i]/(double)N_Max);
	}
	printf("%f", UB/(double)N_Max );
	printf("\n");
      }
    }
  }
  else {
    for ( i=0; i<Problem->Size; i++ ) {
      printf("gamma%d", i );
      switch( Style ) {
	case GNUPLOT_STYLE:
	  printf("(p) = " );
	  break;
      case MAPLE_STYLE:
	  printf(" := " );
	  break;
      }
      for ( k=0; k<=N_Max; k++ ) {
	Print_Term( Binomial(N_Max,k), M[N_Max][k]->Vals[i][0] / N_Max,
		    N_Max-k, k );
      }
      switch( Style ) {
	case GNUPLOT_STYLE:
	  printf(";\n" );
	  break;
      case MAPLE_STYLE:
	  printf(":\n" );
	  break;
      }
    }
    printf("up_bound");
    switch( Style ) {
    case GNUPLOT_STYLE:
      printf("(p) = " );
      break;
    case MAPLE_STYLE:
      printf(" := " );
      break;
    }
    for ( k=0; k<=N_Max; k++ ) {
      Print_Term( Binomial(N_Max,k), U[N_Max][k] / N_Max,
		  N_Max-k, k );
    }
    switch( Style ) {
    case GNUPLOT_STYLE:
      printf(";\n" );
      break;
    case MAPLE_STYLE:
      printf(":\n" );
      break;
    }
  }

  /* cleanup */
  for ( n=0; n<=N_Max; n++) {
    free( U[n] );
    for ( k=0; k<=n; k++ )
      Free_Full_MaxPlus_Matrix( M[n][k] );
    Free_Full_MaxPlus_Matrix( ApowB[n] );
    free( Norm_ApowB );
  }
}

void Tree_Gen_Products( int Current_Order, int Order,
			int Current_Total, int Total,
			int Current_Index,
			Full_MaxPlus_Matrix Current_Mat,
			Full_MaxPlus_Matrix *Tmp_Mat, 
			Full_MaxPlus_Matrix *Tmp_X,
			Full_MaxPlus_Matrix *Am_Multi, double *Alpha,
			Full_MaxPlus_Matrix *Am )
{
  int		i;
  int		n;
  int		Idx;

  for ( i=0; i <= Total-Current_Total; i++ ) {
    n = Current_Total + i;
    Idx = Current_Index + I_Binomial( n+Current_Order-1, Current_Order );

    Full_MM_Mult( Am[i], Current_Mat, Tmp_Mat[Current_Order] );

    if ( Current_Order < Order ) {
      Tree_Gen_Products( Current_Order+1, Order, n, Total, Idx,
			 Tmp_Mat[Current_Order], Tmp_Mat, Tmp_X,
			 Am_Multi, Alpha, Am );
    }
    else {
      /* Storage of the final matrix */
      Am_Multi[Idx] = Copy_Full_MaxPlus_Matrix( Tmp_Mat[Order] );
      /* Accumulation of the Alpha quantity for LB */
      Alpha[n] += Full_Matrix_Norm( Tmp_Mat[Order] );

      if ( Debug ) {
	printf("[%d|%d|%f] ", Idx, n, Full_Matrix_Norm( Tmp_Mat[Order] ) );
      }
    }
  }
}
			
void Tree_Gen_M( int Current_Order, int Order,
		 int Current_Total, int Total,
		 Full_MaxPlus_Matrix Current_Mat,
		 Full_MaxPlus_Matrix *Tmp_Mat, Full_MaxPlus_Matrix *Tmp_X,
		 Full_MaxPlus_Matrix *Am, Full_MaxPlus_Matrix **M )
{
  int		i;
  int		m;

  for ( i=0; i <= Total-(Current_Total+1); i++ ) {
    /* computing the total number of matrices: adding the i A's 1 B. */
    m = Current_Total + + i + 1;
    
    if ( Debug ) {
      printf("Gen M[%d][%d]\n", m, Current_Order );
    }

    Full_MM_Mult( Am[i], Current_Mat, Tmp_X[Current_Order] );

    /* accumulate vector in all M[n][k], n = current number of matrices */
    /* EXCEPT if n=k (already computed).				*/
    if ( m != Current_Order ) {
      Full_MM_Standard_Sum( Tmp_X[Current_Order], 
			    M[m][Current_Order], 
			    M[m][Current_Order] );

      if ( Debug ) {
	printf("Am[%d] =\n", i );
	Write_Full_MaxPlus_Matrix( Am[i], stdout, FORMAT_DEBUG,
				   "Am" );
	printf("fois\n");
	Write_Full_MaxPlus_Matrix( Current_Mat, stdout, FORMAT_DEBUG, NULL );
	printf("egal\n");
	Write_Full_MaxPlus_Matrix( Tmp_X[Current_Order], 
				   stdout, FORMAT_DEBUG, NULL  );
	printf("== adding to M[%d][%d]\n", m, Current_Order );
	Write_Full_MaxPlus_Matrix( M[m][Current_Order],
				   stdout, FORMAT_DEBUG, NULL );
      }
    }

    if ( Current_Order < Order ) {
      Tree_Gen_M( Current_Order+1, Order, m, Total,
		  Tmp_X[Current_Order], 
		  Tmp_Mat, Tmp_X,
		  Am, M ); 
    }
  }
}

void Tree_Aggregate_M( int Current_Order, int Order,
		       int Current_Total, int Total,
		       int Cur_Idx,
		       Full_MaxPlus_Matrix *Am_Multi,
		       Full_MaxPlus_Matrix Multiplicand,
		       Full_MaxPlus_Matrix Target, double Coefficient,
		       Full_MaxPlus_Matrix Tmp_X )
{
  int		i;
  int		j;
  int		m;
  int		Idx;

  if ( Current_Order < Order ) {
    for ( i=0; i <= Total-Current_Total; i++ ) {
      m = Current_Total + i;
      Idx = Cur_Idx + I_Binomial( m+Current_Order-1, Current_Order );

      Tree_Aggregate_M( Current_Order+1, Order, m, Total,
			Idx,
			Am_Multi, 
			Multiplicand, 
			Target, Coefficient, 
			Tmp_X );
    }
  }
  else {
    /* complete index mK to exact total and determine index of the 	*/
    /* matrix Am1 ... AmK						*/
    m = Total - Current_Total;
    Idx = Cur_Idx + I_Binomial( m+Order-1, Order );

    Full_MM_Mult( Am_Multi[Idx], Multiplicand, Tmp_X );
    for ( j=0; j<Tmp_X->Rows; j++ )
      Tmp_X->Vals[j][0] *= Coefficient;
    Full_MM_Standard_Sum( Tmp_X, Target, Target );

    if ( Debug ) {
      printf("Amm[%d] =\n", Idx );
      Write_Full_MaxPlus_Matrix( Am_Multi[Idx], stdout, FORMAT_DEBUG, "Amm" );
      printf("fois\n");
      Write_Full_MaxPlus_Matrix( Multiplicand, stdout, FORMAT_DEBUG, NULL  );
      printf("egal, multiplied by Coeff=%f\n", Coefficient );
      Write_Full_MaxPlus_Matrix( Tmp_X, stdout, FORMAT_DEBUG, NULL );
      printf("== adding to Target gives\n");
      Write_Full_MaxPlus_Matrix( Target, stdout, FORMAT_DEBUG, NULL );
    }
  }
}
			
void Compute_Bounds_Order( Bernoulli_Product Problem,
			   int Order, int N_Max, 
			   boolean Numeric, 
			   RANGE_INFO P_Range, int Style ) 
{
  int		i;
  int		k;
  int		m;
  int		n;
  int		p;
  int		Nb_Points;
  double	Val;
  double	Coeff;
  double	Rho;
  double	Eta;
  double	*Alpha;
  Full_MaxPlus_Matrix	A;
  Full_MaxPlus_Matrix	B;
  Full_MaxPlus_Matrix	Id;
  Full_MaxPlus_Matrix	*Am;		/* Table of matrices A^m B */
  Full_MaxPlus_Matrix	*Am_Zero;	/* Table of vectors  A^m X0 */
  Full_MaxPlus_Matrix	*Am_Multi;	/* Table of matrices Am1..AmK */
  Full_MaxPlus_Matrix	**M;		/* Table of (vector) lower bounds */
  double		**U;		/* Table of (scalar) upper bounds */
  double		*Sum;
  double		UB;
  boolean		Alloc_Ok = TRUE;
  Full_MaxPlus_Matrix	*Tmp_Mat;
  Full_MaxPlus_Matrix	*Tmp_X;

  /* Allocation of various arrays */
  Am = (Full_MaxPlus_Matrix*) 
    calloc( (unsigned)(1+N_Max), sizeof(Full_MaxPlus_Matrix) );
  Am_Zero = (Full_MaxPlus_Matrix*) 
    calloc( (unsigned)(1+N_Max), sizeof(Full_MaxPlus_Matrix) );
  Am_Multi = (Full_MaxPlus_Matrix*) 
    calloc( (unsigned)I_Binomial(N_Max,Order), sizeof(Full_MaxPlus_Matrix) );
  M = (Full_MaxPlus_Matrix **) 
    calloc( (unsigned)(1+N_Max), sizeof(Full_MaxPlus_Matrix*) );
  U = (double**) calloc( (unsigned)(1+N_Max), sizeof( double* ) );
  Alpha = (double*) calloc( (unsigned)(1+N_Max), sizeof( double ) );
  Tmp_Mat = (Full_MaxPlus_Matrix*)
    calloc( (unsigned)(1+Order), sizeof(Full_MaxPlus_Matrix) );
  Tmp_X = (Full_MaxPlus_Matrix*)
    calloc( (unsigned)(1+Order), sizeof(Full_MaxPlus_Matrix) );

  Alloc_Ok = ( NULL != Am ) && ( NULL != Am_Zero ) 
    && ( NULL != Am_Multi )
    && ( NULL != M ) && ( NULL != U )
    && ( NULL != Alpha ) && ( NULL != Tmp_Mat ) && ( NULL != Tmp_X );

  for ( n=0; Alloc_Ok && (n<=N_Max); n++) {
    Am[n] = New_Full_MaxPlus_Matrix( Problem->Size, Problem->Size, NONE );
    Am_Zero[n] = New_Full_MaxPlus_Matrix( Problem->Size, 1, NONE );
    M[n] = (Full_MaxPlus_Matrix*)
      calloc( (unsigned)(1+N_Max), sizeof(Full_MaxPlus_Matrix) );
    U[n] = (double*) calloc( (unsigned)(1+N_Max), sizeof(double) );

    Alloc_Ok = Alloc_Ok && ( NULL != Am[n] ) && ( NULL != Am_Zero[n] )
      && ( NULL != M[n] ) && ( NULL != U[n] );

    for ( k=0; Alloc_Ok && (k<=n); k++ ) {
      M[n][k] = New_Full_MaxPlus_Matrix( Problem->Size, 1, ONE );
      Alloc_Ok = Alloc_Ok && ( NULL != M[n][k] );
    }
  }

  if ( !Alloc_Ok ) {
    printf("%s: error: could not allocate space. No analysis.\n",
	   TOOLNAME );
    return;
  }

  /* Conversion of the problem matrices into (Max,+) structures */
  A = Convert_MaxPlus_Matrix_To_Full( Problem->A[0] );
  B = Convert_MaxPlus_Matrix_To_Full( Problem->A[1] );
  if ( Debug ) {
    Write_Full_MaxPlus_Matrix( A, stdout, FORMAT_DEBUG, "A" );
    Write_Full_MaxPlus_Matrix( B, stdout, FORMAT_DEBUG, "B" );
  }
  /* convenience identity matrix, and temporary structures */
  Id = New_Full_MaxPlus_Matrix( Problem->Size, Problem->Size, IDENTITY );
  for ( i=1; i<=Order; i++ ) {
    Tmp_Mat[i] = New_Full_MaxPlus_Matrix( Problem->Size, Problem->Size, NONE );
    Tmp_X[i] = New_Full_MaxPlus_Matrix( Problem->Size, 1, NONE );
  }
  if ( Debug ) {
    Write_Full_MaxPlus_Matrix( Id, stdout, FORMAT_DEBUG, "Id" );
  }
  
  /* M[0,0] is X(0), taken as (0,...,0) by default */
  U[0][0] = 0.0;

  /* Initialization of matrix and vector arrays.			*/
  /* Am[n] 	contains A^n B, 					*/
  /* Am_Zero[n] contains A^n X0,					*/
  /* M[n,0] 	is A^n*X0 = Am_Zero					*/
  /* M[n,n] 	is B^n*X0 						*/
  Am[0] = Copy_Full_MaxPlus_Matrix( B );
  Am_Zero[0] = Copy_Full_MaxPlus_Matrix( M[0][0] );

  for ( n=1; n<=N_Max; n++ ) {
    Full_MM_Mult( A, Am[n-1], Am[n] );
    if ( Debug ) {
      sprintf( Tmp_String, "Am[%d]", n );
      Write_Full_MaxPlus_Matrix( Am[n], stdout, FORMAT_DEBUG, Tmp_String );
    }
    Full_MM_Mult( A, Am_Zero[n-1], Am_Zero[n] );
    Full_MM_Mult( A, M[n-1][0], M[n][0] );
    Full_MM_Mult( B, M[n-1][n-1], M[n][n] );
  }

  /* Preliminary computation of the multidimentional products Am1...AmK	*/
  /* and of the scalars alpha.						*/
  /* recursive computation of the vectors M, and the scalars U */
  Tree_Gen_Products( 1, Order, 0, N_Max-Order, 0,
		     Id, Tmp_Mat, Tmp_X,
		     Am_Multi, Alpha, Am );
  if ( Debug ) {
    for ( n=0; n<I_Binomial(N_Max,Order); n++ ) {
      sprintf( Tmp_String, "Amm[%d] =\n", n );
      Write_Full_MaxPlus_Matrix( Am_Multi[n], stdout, FORMAT_DEBUG, 
				 Tmp_String );
    }
  }

  /* looking at the result */
  if ( Debug ) {
    for ( n=0; n<=N_Max; n++ ) {
      for ( k=0; k<=n; k++ ) {
	printf("n=%d k=%d: U = %f\n", n, k, U[n][k] );
	printf("M(1)=[");
	for ( i=0; i<M[0][0]->Rows; i++ )
	  if ( M_INFINITY == M[n][k]->Vals[i][0] )
	    printf(" eps" );
	  else 
	    printf(" %f", M[n][k]->Vals[i][0] );
	printf("]\n");
      }
    }
  }
  /* Generation of the initial values of M[n][k], for 1<=k<Order and 	*/
  /* 0 < n < k.	M[k][n] has been initialized to 0 above. 		*/
  if ( Order > 1 ) {
    for ( n=0; n<=N_Max-1; n++ ) {
      /* n represents the number of trailing A's */
      Tree_Gen_M( 1, Order-1, n, N_Max, Am_Zero[n], 
		  Tmp_Mat, Tmp_X, Am, M );
    }
  }
  /* computing the "reduced" form M[n][k] / binom( n, k ) */
  for ( k=1; k < Order; k++ ) {
    Rho = 1.0 / (double) (k+1);
    for ( n=k+1; n<=N_Max; n++ ) {
      Full_Standard_Scalar_Product( M[n][k], M[n][k], Rho );
      Rho *= (double)(n+1-k) / (double)(n+1);
    }
  }
  /* initializing the U[n][k], 0<=k<Order 				*/
  for ( k=0; k<Order; k++ )
    for ( n=k; n<=N_Max; n++ )
      U[n][k] = Full_Matrix_Norm( M[n][k] );
  for ( n=1; n<=N_Max; n++ )
    U[n][n] = Full_Matrix_Norm( M[n][n] );

  /* looking at the result */
  if ( Debug ) {
    for ( n=0; n<=N_Max; n++ ) {
      for ( k=0; k<=n; k++ ) {
	printf("n=%d k=%d: U = %f\n", n, k, U[n][k] );
	printf("M(2)=[");
	for ( i=0; i<M[0][0]->Rows; i++ )
	  if ( M_INFINITY == M[n][k]->Vals[i][0] )
	    printf(" eps" );
	  else 
	    printf(" %f", M[n][k]->Vals[i][0] );
	printf("]\n");
      }
    }
  }
  for ( k=Order; k<=N_Max; k++ ) {
    for ( n=k+1; n<=N_Max; n++ ) {

      U[n][k] = 0.0;

      Rho = 1.0 / Binomial( n, k );
      Eta = (double) k / (double)n;
      for ( i=k-1; i>=Order; i-- )
	Eta *= (double)i / (double)(n-k+i);

      for ( m=k; m<=n; m++ ) {
	if ( Debug ) {
	  printf("### [n=%d,k=%d,m=%d] Aggregating |%d| to M[%d][%d]\n", 
		 n, k, m,
		 n-m, m-Order, k-Order );
	}
	Tree_Aggregate_M( 1, Order, 
			  0, n-m, 
			  0,
			  Am_Multi, 
			  M[m-Order][k-Order], 
			  M[n][k], Rho, 
			  Tmp_X[1] );

	U[n][k] += Rho * Alpha[m] + Eta * U[m-Order][k-Order];
	Rho *= (double) (m+1-Order) / (double) (m+1-k);
	Eta *= (double) (n-m) / (double) (n-m+Order-1);
      }
    }
  }

  /* looking at the result */
  if ( Debug ) {
    for ( n=0; n<=N_Max; n++ ) {
      for ( k=0; k<=n; k++ ) {
	printf("n=%d k=%d: U = %f\n", n, k, U[n][k] );
	printf("M(3)=[");
	for ( i=0; i<M[0][0]->Rows; i++ )
	  if ( M_INFINITY == M[n][k]->Vals[i][0] )
	    printf(" eps" );
	  else 
	    printf(" %f", M[n][k]->Vals[i][0] );
	printf("]\n");
      }
    }
  }

  /* computation of the bound from the matrices M and scalars U */
  if ( Numeric ) {
    Sum = (double*) calloc( Problem->Size, sizeof(double) );

    /* taking care of degenerate cases */
    if ( P_Range.Start > P_Range.Stop ) {
      fprintf( stderr, "%s: bad range\n", TOOLNAME );
    }
    else if ( P_Range.Step < 0.0 ) {
      fprintf( stderr, "%s: bad step value\n", TOOLNAME );
    }
    else {
      printf("# Bounds bound on the Expectation of X_i(%d) vs p, ", N_Max );
      printf("0 <= i <= %d,\n", Problem->Size-1 );
      printf("# and p in range [%f,%f]\n", P_Range.Start, P_Range.Stop );
      printf("#\n");
      
      Nb_Points = ( P_Range.Stop == P_Range.Start ?
		1 :
		1 + floor( ( P_Range.Stop - P_Range.Start ) / P_Range.Step ) );

      for ( p=0; p<Nb_Points; p++ ) {

	Val = P_Range.Start + p*P_Range.Step;

	if ( Val == 0.0 ) {
	  for ( i=0; i<Problem->Size; i++ )
	    Sum[i] = M[N_Max][N_Max]->Vals[i][0];
	  UB = U[N_Max][N_Max];
	}
	else if ( Val == 1.0 ) {
	  for ( i=0; i<Problem->Size; i++ )
	    Sum[i] = M[N_Max][0]->Vals[i][0];
	  UB = U[N_Max][0];
	}
	else {
	  for ( i=0; i<Problem->Size; i++ )
	    Sum[i] = 0.0;

	  UB = 0.0;
      	  Coeff = pow( Val, (double)N_Max );

	  for ( k=0; k<=N_Max; k++ ) {
	    for ( i=0; i<Problem->Size; i++ )
	      Sum[i] += Coeff * M[N_Max][k]->Vals[i][0];
	    UB += Coeff * U[N_Max][k];
	    Coeff *= (1.0-Val)/Val * (double)(N_Max-k)/(double)(k+1);
	  }
	}
    
	/* output result */
	printf("%12.8f ", Val );

	for (i=0;i<Problem->Size;i++) {
	  /* printf("gamma[%d] = ", i ); */
	  printf("%f ", Sum[i]/(double)N_Max);
	}
	printf("%f", UB/(double)N_Max );
	printf("\n");
      }
    }
  }
  else {
    for ( i=0; i<Problem->Size; i++ ) {
      printf("gamma%d", i );
      switch( Style ) {
	case GNUPLOT_STYLE:
	  printf("(p) = " );
	  break;
      case MAPLE_STYLE:
	  printf(" := " );
	  break;
      }
      for ( k=0; k<=N_Max; k++ ) {
	Print_Term( Binomial(N_Max,k), M[N_Max][k]->Vals[i][0] / N_Max,
		    N_Max-k, k );
      }
      switch( Style ) {
	case GNUPLOT_STYLE:
	  printf(";\n" );
	  break;
      case MAPLE_STYLE:
	  printf(":\n" );
	  break;
      }
    }
    printf("up_bound");
    switch( Style ) {
    case GNUPLOT_STYLE:
      printf("(p) = " );
      break;
    case MAPLE_STYLE:
      printf(" := " );
      break;
    }
    for ( k=0; k<=N_Max; k++ ) {
      Print_Term( Binomial(N_Max,k), U[N_Max][k] / N_Max,
		  N_Max-k, k );
    }
    switch( Style ) {
    case GNUPLOT_STYLE:
      printf(";\n" );
      break;
    case MAPLE_STYLE:
      printf(":\n" );
      break;
    }
  }
}

/************************************************************************/
/* Procedures Compute_Exact(), Tree_Mult() and Compute_Stat()		*/
/*									*/
/* Compute exacts bound on gamma by making all possible products of	*/
/* matrices of a certain given length N_Max.				*/
/* The computation is made by exploring the "prefix tree" depth first	*/
/* via recursive calls to Tree_Mult().					*/
/* The final computation of E(max(X(n)) and E(min(X(n)) is gathered	*/
/* on leaved of the tree by Compute_Stat().				*/
/*									*/
/* In these procedures, matrices may be full or sparse. The prevailing	*/
/* implementation is used: Full matrices are not enforced.		*/
/************************************************************************/
void Compute_Stat( MaxPlus_Vector X,
		   double *The_Max, double *The_Min, double Proba )
{
  double	X_Max;
  double	X_Min;

  if ( Debug ) {
    printf("X=");
    Write_MaxPlus_Vector( X, stdout, /* ignored format!*/FORMAT_DEBUG );
    printf(" (p=%e)\n", Proba );
  }

  /* Assert: X is column vector. */
  X_Max = V_Norm( X );
  X_Min = V_Min( X );

  (*The_Max) += Proba * X_Max;
  (*The_Min) += Proba * X_Min;
  
}

void Tree_Mult( Bernoulli_Product Problem,
		int Level, int Max_Level, 
		MaxPlus_Vector *M, double Proba,
		double *The_Max, double *The_Min )
{
  int		i;

  for ( i=0; i < Problem->Nb_Matrices; i++ ) {
    MV_Mult( Problem->A[i], M[Level], M[Level+1] );
    if ( Level+1 == Max_Level ) {
      Compute_Stat( M[Max_Level], The_Max, The_Min, 
		    Problem->Proba[i]*Proba );
    }
    else {
      Tree_Mult( Problem, Level+1, Max_Level, 
		 M, Problem->Proba[i]*Proba, 
		 The_Max, The_Min );
    }
  }
}

void Compute_Exact( Bernoulli_Product Problem,
		    int N_Max,
		    RANGE_INFO P_Range, int Style ) 
{
  int		n;
  int		p;
  int		Nb_Points;
  double	Val;
  double	The_Max;
  double	The_Min;
  MaxPlus_Vector *M;		/* M[i] will be the vector containing 	*/
  				/* the current product of length i	*/
  				/* M[0] is X(0) = (0,...,0).		*/

  M = (MaxPlus_Vector*) calloc( (unsigned)(1+N_Max), 
				sizeof( MaxPlus_Vector ) );

  for ( n=0; n<=N_Max; n++) {
    M[n] = New_MaxPlus_Vector( Problem->Size, ONE );
  }

  /* Range computations are possible only for 2 matrices */
  if ( Problem->Nb_Matrices != 2 ) {
    /* recursive computation of ALL the vectors X(n) */
    The_Max = The_Min = 0.0;
    Tree_Mult( Problem, 0, N_Max, M, 1.0, &The_Max, &The_Min );

    /* output result */
    printf("%f %f\n", The_Min/N_Max, The_Max/N_Max );
  }
  else {
    /* two matrices. Taking range info into account */
    /* taking care of degenerate cases */
    if ( P_Range.Start > P_Range.Stop ) {
      fprintf( stderr, "%s: bad range\n", TOOLNAME );
    }
    else if ( P_Range.Step < 0.0 ) {
      fprintf( stderr, "%s: bad step value\n", TOOLNAME );
    }
    else {
      printf("# Bounds on gamma computed from Emin/max(X_i(%d)),\n", N_Max );
      printf("# Range of p: [%f,%f]\n", P_Range.Start, P_Range.Stop );
      printf("#\n");
	
      Nb_Points = ( P_Range.Stop == P_Range.Start ?
		    1 :
		    1 + floor( ( P_Range.Stop - P_Range.Start ) 
			       / P_Range.Step ) );

      for ( p=0; p<Nb_Points; p++ ) {

	Val = P_Range.Start + p*P_Range.Step;
	Problem->Proba[0] = Val;
	Problem->Proba[1] = 1.0-Val;

	The_Max = The_Min = 0.0;

	Tree_Mult( Problem, 0, N_Max, M, 1.0, &The_Max, &The_Min );

	/* NOTE: there could be some optimization if p=0 or 1... */

	/* output result */
	printf("%8.6f %f %f\n", Val, The_Min/N_Max, The_Max/N_Max );
      }
    }
  }

  /* cleanup */
  for ( n=0; n<=N_Max; n++) {
    Free_MaxPlus_Vector( M[n] );
  }
  free( M );

}

/***********************************************************************/
/*                Programme Principal                                  */
/***********************************************************************/

int main(int argc, char *argv[], char *argp[] )
{
  static char		In_File_Name[MAX_NAME_LENGTH] = "";
  static char		Out_File_Name[MAX_NAME_LENGTH] = "";

  int 			i;
  int			N_Max = 10;
  int			Order = 1;
  boolean		Numeric = FALSE;
  boolean		Exact = FALSE;
  boolean		Value = FALSE;
  int			Style = GNUPLOT_STYLE;
  boolean		Help = FALSE;
  boolean		Args_Ok = TRUE;
  boolean		File_Ok = TRUE;
  Bernoulli_Product	Problem;
  RANGE_INFO 		P_Range = { 0.0, 1.0, 0.05, TRUE, FALSE, FALSE };

  /************************* Parsing *****************************/
  i = 1;

  if (argc>1)
    do {

      if GET_STRING( "-f", In_File_Name )
      else if GET_STRING( "-o", Out_File_Name )
      else if SWITCH( "-debug", Debug, TRUE )
      else if SWITCH( "-numeric", Numeric, TRUE )
      else if SWITCH( "-exact", Exact, TRUE )
      else if GET_INT( "-nmax", N_Max )
      else if GET_INT( "-order", Order )
      else if GET_REAL_SWITCH( "-pval",  P_Range.Start, Value, TRUE )
      else if GET_REAL_SWITCH( "-pfrom", P_Range.Start, Numeric, TRUE )
      else if GET_REAL_SWITCH( "-pto",   P_Range.Stop, Numeric, TRUE )
      else if GET_REAL_SWITCH( "-pstep", P_Range.Step, Numeric, TRUE )
      else if IS_VERSION

      else if (( 0 == strcmp( "-h", argv[i] )) || 
	       ( 0 == strcmp( "-?", argv[i] )) ||
	       ( 0 == strcmp( "-help", argv[i] ))) {
	i++;
	fprintf(stdout,"usage: %s [-f inputfile] [-o outputfile] [-nmax int]",
		TOOLNAME );
	fprintf(stdout," [-order int]");
	fprintf(stdout,"\n\t[-exact] [-numeric]");
	fprintf(stdout,"\n\t[-pval <real>] [-pfrom <real>]");
	fprintf(stdout," [-pto <real>] [-pstep <real>]");
	fprintf(stdout,"\n\t[-maple] [-v] [-debug] [-h] [-help] [-?]\n");
	Args_Ok = FALSE;
	Help = TRUE;
      }
      else if ( 0 == strcmp( "-maple", argv[i] ) ) {
	Style = MAPLE_STYLE;
	i++;
      }
      else 
	i++;
    } while( Args_Ok && ( i < argc ) );

  if (Args_Ok) {

    if ( strlen(In_File_Name) != 0 )
      File_Ok = ( NULL != freopen( In_File_Name, "r", stdin ) );

    if ( !File_Ok ) {
      fprintf(stderr,"%s: could not open file %s\n", TOOLNAME, In_File_Name );
    }

    if ( File_Ok ) {
      if ( strlen(Out_File_Name) != 0 )
	File_Ok = (NULL != freopen( Out_File_Name, "w", stdout ) );

      if ( !File_Ok ) {
	fprintf(stderr,"%s: could not open file %s\n", 
		TOOLNAME, Out_File_Name);
	File_Ok = FALSE;
      }
    }
  }


  /************* start of the program if everything is Ok *******************/

  if ( Args_Ok && File_Ok ) {

    Problem = Read_Bernoulli_Product( stdin );

    if ( (Bernoulli_Product)NULL != Problem ) {

      if ( Value ) {
	Numeric = TRUE;
	P_Range.Stop = P_Range.Start;
      }

      if ( Exact ) {
	Compute_Exact( Problem, N_Max, P_Range, Style );
      }
      else if ( Problem->Nb_Matrices == 2 ) {
/* 	Compute_Bounds( Problem, N_Max, Numeric, P_Range, Style ); */
	Compute_Bounds_Order( Problem, Order, N_Max, 
			      Numeric, P_Range, Style );
	if ( Order == 1 ) 
	  Compute_Bounds( Problem, N_Max, 
			  Numeric, P_Range, Style );
      }
      else {
	printf("%s: error: can compute only for two matrices. ", TOOLNAME );
	printf("%d encountered.\n", Problem->Nb_Matrices );

	return -4;
      }

      /* computation successful */
      Free_Bernoulli_Product( Problem );

      return 0;
    }
    else
      return -2;
  }
  else if ( Help ) {
    return 0;
  }
  else {
    return -1;
  }
}
