/************************************************************************/
/* File mat_sim.c in ERS						*/
/*									*/
/* Simulation of a (random) product of (max,+) matrices with gradient 	*/
/* analysis.								*/
/* Based on a random matrix product code by B. Gaujal, revised by      	*/
/* A. Jean-Marie, and the Robins-Monroe-type algorithm by 		*/
/* B. Heidergott.							*/
/*									*/
/* July/August 1998.							*/
/************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <math.h>
#include <string.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include "boolean.h"
#include "parseargs.h"
#include "maxplusmat.h"
#include "bernoulli.h"

#define TOOLNAME	"mat_sim"
#ifdef USE_FULL_MATRICES
#define VERSION		"0.5 (full matrices)"
#else
#define VERSION		"0.5 (sparse matrices)"
#endif

#define MAX_RANDOM_VALUE	0x7FFFFFFF

#define MAX_NAME_LENGTH 256

#ifndef ALPHA
extern long random();
#endif

#define PRODUCT(A,X,Y) MV_Mult(A,X,Y)

typedef struct DOUBLE_ARRAY {
  int		Size;
  double	*Val;
} *Double_Array;

/* forward declarations */
void Free_Bernoulli_Product( Bernoulli_Product Problem );
boolean	Compile_Matrix_And_Run( char In_File_Name[MAX_NAME_LENGTH],
			       int Sim_Length, int Nb_Replicates,
			       boolean Stats, boolean Keep, 
			       char *argp[] );
Double_Array Parse_Direction( char *String );
void Simulate_Sequence( Bernoulli_Product Problem,
			char [MAX_NAME_LENGTH], int Format, char[128] );
        
boolean Debug = FALSE;

/*************************************************************************/
double U_0_1() 
{
  return (double)random()/MAX_RANDOM_VALUE;
}

/************************************************************************/
/* Perform a Bernoulli choice based on a vector of cumulative 		*/
/* probabilities. The size of the vector is not specified but the 	*/
/* vector is requested to be "guarded" by a 1.0.			*/
/************************************************************************/
int choice(double *Pr_Cum)
{
  int 		i;
  double 	ra;

  ra = U_0_1();
  i = 0;
  while ( ra > Pr_Cum[i])
    i++;

  return i;
}

void	sort_probas( Bernoulli_Product Problem )
{
  int	i;

  /* to be implemented... */
  for ( i=0; i<Problem->Nb_Matrices; i++ )
    Problem->Permut[i] = i;
}

/* #define TRACE( t, s, x1, x2, x3 ) Trace( t, s, x1, x2, x3 ) */
#define TRACE( t, s, x1, x2, x3 ) 

void Trace( int time, int Size, double *X, double *X_Plus, double *X_Minus ) {
  int i;

  for ( i = 0; i < Size; i++ ) {
    printf("t= %d   |  X-[%d] = %f    X[%d] = %f    X+[%d] = %f\n",
	   time, i, X_Minus[i], i, X[i], i, X_Plus[i] );
  }
}

/* Creating weak derivative while checking the validity of the direction. */
/* Return NULL if anything goes wrong.					*/
Weak_Derivative Split_Distribution( Double_Array Direction ) {
  int			i;
  double 		G = 0.0;
  double		C = 0.0;
  double		Cum_Plus;
  double		Cum_Minus;
  Weak_Derivative	Res;

  for ( i = 0; i < Direction->Size; i++ ) {
    if ( Debug ) {
      printf("Dir[%d] = %f\n", i, Direction->Val[i] );
    }
    G += Direction->Val[i];
    if ( Direction->Val[i] > 0.0 )
      C += Direction->Val[i];
  }
  if ( fabs( G ) > 1e-6 ) {
      printf("%s: error: bad direction vector (does not sum up to zero).\n",
	     TOOLNAME );
      Res = (Weak_Derivative)NULL;
  } else {
    if ( fabs( C ) < 1e-6 ) {
      printf("%s: error: bad direction vector (could not normalize).\n",
	     TOOLNAME );
      Res = (Weak_Derivative)NULL;
    }
    else {
      /* ok. Creating it */
      Res = (Weak_Derivative) malloc( sizeof( struct WEAK_DERIVATIVE ) );
      if ( (Weak_Derivative)NULL != Res ) {
	Res->Size = Direction->Size;
	Res->C = C;
	Res->Pr_Cum_Plus = (double*) calloc( (size_t)Direction->Size, 
					     sizeof( double ) );
	Res->Pr_Cum_Minus = (double*) calloc( (size_t)Direction->Size, 
					      sizeof( double ) );
	if ( ( (double*)NULL == Res->Pr_Cum_Plus ) 
	     || ( (double*)NULL == Res->Pr_Cum_Minus ) ) {
	  printf("%s: error: courd not allocate space for distribution.\n",
		 TOOLNAME );
	  free( Res );
	  Res = (Weak_Derivative)NULL;
	}
	else {
	  Cum_Plus = 0.0;
	  Cum_Minus = 0.0;
	  for ( i = 0; i < Direction->Size; i++ ) {
	    if ( Direction->Val[i] >= 0 ) {
	      Cum_Plus += Direction->Val[i] / C;
	    }
	    else { 
	      Cum_Minus -= Direction->Val[i] / C;
	    }
	    Res->Pr_Cum_Plus[i] = Cum_Plus;
	    Res->Pr_Cum_Minus[i] = Cum_Minus;
	  }
	  /* setting the guard */
	  Res->Pr_Cum_Plus[Direction->Size-1] = 1.0;
	  Res->Pr_Cum_Minus[Direction->Size-1] = 1.0;
	}
      }
    }
  }

  return Res;
}

/************************************************************************/
/* Simulation with gradient estimate. 					*/
/************************************************************************/
void Simulate_With_Gradient ( Bernoulli_Product 	Problem,
			      Weak_Derivative		Mu,
			      int 			Sim_Length,
			      int 			Nb_Replicates,
			      boolean			Stats )
{
  int		i;
  int		l;
  int		n;
  int		r;
  MaxPlus_Vector X;
  MaxPlus_Vector X_Plus;
  MaxPlus_Vector X_Minus;
  MaxPlus_Vector Y;
  MaxPlus_Vector Tmp;
  int		sigma;		/* random point of perturbation 	*/
  double	G;		/* Temporary convenience variable 	*/
  double	DG;		/* Temporary convenience variable 	*/
  double	*Gamma;		/* estimate for average of gamma 	*/
  double	*Gamma2;	/* estimate for variance of gamma 	*/
  double	*DGamma;	/* estimate for derivative of gamma 	*/
  double	*DGamma2;	/* estimate for variance of derivative 	*/
  double	Sigma_Hat;

  X =        New_MaxPlus_Vector( Problem->Size, NONE );
  X_Minus =  New_MaxPlus_Vector( Problem->Size, NONE );
  X_Plus =   New_MaxPlus_Vector( Problem->Size, NONE );
  Y =        New_MaxPlus_Vector( Problem->Size, NONE );
  Gamma =    (double *) calloc( (size_t)Problem->Size, sizeof(double));
  DGamma =   (double *) calloc( (size_t)Problem->Size, sizeof(double));
  Gamma2 =   (double *) calloc( (size_t)Problem->Size, sizeof(double));
  DGamma2 =  (double *) calloc( (size_t)Problem->Size, sizeof(double));

  for ( i = 0; i < Problem->Size; i++ ) {
    Gamma[i] = 0.0;
    DGamma[i] = 0.0;
    Gamma2[i] = 0.0;
    DGamma2[i] = 0.0;
  }

  for ( n = 0; n < Nb_Replicates; n++ ) {
    /* (re) init recurrence vector */
    for ( i = 0; i < Problem->Size; i++ ) {
#ifdef USE_FULL_MP_MATRICES
      X->Vals[i][0] = 0.0;
      X_Plus->Vals[i][0] = 0.0;
      X_Minus->Vals[i][0] = 0.0;
#else
      X->Vals[i] = 0.0;
      X_Plus->Vals[i] = 0.0;
      X_Minus->Vals[i] = 0.0;
#endif
    }

    /* determine point of perturbation */
    sigma = (int) ( Sim_Length * U_0_1() );

    /* compute common part */
    for (l = 0; l < sigma; l++) {
      r = choice( Problem->Pr_Cum );
      PRODUCT( Problem->A[r], X, Y );
      Tmp = X; X = Y; Y = Tmp;
      TRACE( l, Problem->Size, X, X_Plus, X_Minus );
    }

    /* make perturbation. */
    r = choice( Mu->Pr_Cum_Plus );
    PRODUCT( Problem->A[r], X, X_Plus );
    if ( Debug ) {
      printf( "Rep #%d: Ch+ = %d   ", n, r );
    }
    r = choice( Mu->Pr_Cum_Minus );
    PRODUCT( Problem->A[r], X, X_Minus );
    if ( Debug ) {
      printf( "Ch- = %d\n", r );
    }
    r = choice( Problem->Pr_Cum );
    PRODUCT( Problem->A[r], X, Y );
    Tmp = X; X = Y; Y = Tmp;
    TRACE( l, Problem->Size, X, X_Plus, X_Minus );

    /* compute final part for X, X_Plus and X_Minus */
    for ( l++; l < Sim_Length; l++) {
      r = choice( Problem->Pr_Cum );
      PRODUCT( Problem->A[r], X_Plus, Y );
      Tmp = X_Plus; X_Plus = Y; Y = Tmp;
      PRODUCT( Problem->A[r], X_Minus, Y );
      Tmp = X_Minus; X_Minus = Y; Y = Tmp;
      PRODUCT( Problem->A[r], X, Y );
      Tmp = X; X = Y; Y = Tmp;
      TRACE( l, Problem->Size, X, X_Plus, X_Minus );
    }

    /* accumulate stats. Care is taken because X[i] may be eps */
    for ( i = 0; i < Problem->Size; i++ ) {
#ifdef USE_FULL_MP_MATRICES
      if ( X->Vals[i][0] != M_INFINITY ) {
	G = X->Vals[i][0]/(double)Sim_Length;
	Gamma[i] = plus( G, Gamma[i] );
	Gamma2[i] += G*G;
      }
      else {
	Gamma[i] = M_INFINITY;
      }
      if ( ( X_Minus->Vals[i][0] != M_INFINITY ) 
	   && ( X_Plus->Vals[i][0] != M_INFINITY ) ) {
	DG = X_Plus->Vals[i][0] - X_Minus->Vals[i][0];
	DGamma[i] = plus( DG, DGamma[i] );
	DGamma2[i] += DG*DG;
      }
      else {
	DGamma[i] = M_INFINITY;
      }
#else
      if ( X->Vals[i] != M_INFINITY ) {
	G = X->Vals[i]/(double)Sim_Length;
	Gamma[i] = plus( G, Gamma[i] );
	Gamma2[i] += G*G;
      }
      else {
	Gamma[i] = M_INFINITY;
      }
      if ( ( X_Minus->Vals[i] != M_INFINITY ) 
	   && ( X_Plus->Vals[i] != M_INFINITY ) ) {
	DG = X_Plus->Vals[i] - X_Minus->Vals[i];
	DGamma[i] = plus( DG, DGamma[i] );
	DGamma2[i] += DG*DG;
      }
      else {
	DGamma[i] = M_INFINITY;
      }
#endif
    }   
  }

  /* output result */
  if ( Sim_Length > 0 ) {
    for (i=0;i<Problem->Size;i++) {

      printf("gamma[%d] = ", i );
      if ( Gamma[i] == M_INFINITY )
	printf("eps\n");
      else
	printf("%f", (double) Gamma[i] / Nb_Replicates );

      if ( Stats && ( Nb_Replicates > 1 ) ) {
	Sigma_Hat = sqrt( ( Gamma2[i] - Gamma[i]*Gamma[i]/Nb_Replicates ) 
			  / (Nb_Replicates-1) );
	printf("  +/- %f ", 
	       Sigma_Hat * 1.96 / sqrt(Nb_Replicates) );
      }
      printf("  grad: ");
      if ( DGamma[i] == M_INFINITY )
	printf("eps\n");
      else
	printf("%f", (double) DGamma[i] * Mu->C / Nb_Replicates );

      if ( Stats && ( Nb_Replicates > 1 ) ) {
	Sigma_Hat = sqrt( ( DGamma2[i] - DGamma[i]*DGamma[i]/Nb_Replicates ) 
			  / (Nb_Replicates-1) );
	printf("  +/- %f (conf. 95%%)", 
	       Sigma_Hat * 1.96 * Mu->C / sqrt(Nb_Replicates) );
      }
      printf("\n");
    }
  }

  /* cleanup */
  Free_MaxPlus_Vector( X ); 
  Free_MaxPlus_Vector( X_Plus );
  Free_MaxPlus_Vector( X_Minus );
  Free_MaxPlus_Vector( Y ); 
  free( Gamma );
  free( DGamma );
  free( Gamma2 );
  free( DGamma2);
}

/* Simulation without gradient estimate */
void Simulate ( Bernoulli_Product 	Problem,
		int 			Sim_Length,
		int 			Nb_Replicates,
		boolean			Stats )
{
  int		i;
  int		l;
  int		n;
  int		r;
  MaxPlus_Vector X;
  MaxPlus_Vector Y;
  MaxPlus_Vector Tmp;
  double	G;		/* Temporary convenience variable */
  double	*Gamma;		/* estimate for average */
  double	*Gamma2;	/* estimate for variance */
  double	Sigma_Hat;

  X =      New_MaxPlus_Vector( Problem->Size, NONE );
  Y =      New_MaxPlus_Vector( Problem->Size, NONE );
  Gamma =  (double *) calloc( (size_t)Problem->Size, sizeof(double));
  Gamma2 = (double *) calloc( (size_t)Problem->Size, sizeof(double));

  for ( i = 0; i < Problem->Size; i++ ) {
    Gamma[i] = 0.0;
    Gamma2[i] = 0.0;
  }

  for ( n = 0; n < Nb_Replicates; n++ ) {
    /* (re) init recurrence vector */
    for ( i = 0; i < Problem->Size; i++ ) {
#ifdef USE_FULL_MP_MATRICES      
      X->Vals[i][0] = 0.0;
#else
      X->Vals[i] = 0.0;
#endif
    }

    for (l = 0; l < Sim_Length; l++) {
      r = choice( Problem->Pr_Cum );
      PRODUCT( Problem->A[r], X, Y );
      Tmp = X; X = Y; Y = Tmp;
    }

    /* accumulate stats */
    for ( i = 0; i < Problem->Size; i++ ) {
#ifdef USE_FULL_MP_MATRICES  
      G = X->Vals[i][0]/(double)Sim_Length;
#else
      G = X->Vals[i]/(double)Sim_Length;
#endif
      Gamma[i] += G;
      Gamma2[i] += G*G;
    }     
  }

  /* Compute stats and output result */
  if ( Sim_Length > 0 ) {
    for (i=0;i<Problem->Size;i++) {

      printf("gamma[%d] = ", i );
      if ( Gamma[i] == M_INFINITY )
	printf("eps\n");
      else
	printf("%f", (double) Gamma[i] / Nb_Replicates );

      if ( Stats && ( Nb_Replicates > 1 ) ) {
	Sigma_Hat = sqrt( ( Gamma2[i] - Gamma[i]*Gamma[i]/Nb_Replicates ) 
			  / (Nb_Replicates-1) );
	printf("  +/- %f (conf. 95%%)", 
	       Sigma_Hat * 1.96 / sqrt(Nb_Replicates) );
      }
      printf("\n");
    }
  }

  Free_MaxPlus_Vector( X );
  Free_MaxPlus_Vector( Y );
  free( Gamma );
  free( Gamma2 );
}

/***********************************************************************/
/*                Programme Principal                                  */
/***********************************************************************/

int main(int argc, char *argv[], char *argp[] )
{
  static char	In_File_Name[MAX_NAME_LENGTH] = "";
  static char	Out_File_Name[MAX_NAME_LENGTH] = "";
  static char	Seq_File_Name[MAX_NAME_LENGTH] = "";

  int 		i;
  int		Sim_Length = 100000;
  int		Nb_Replicates = 1;
  boolean	Stats = FALSE;
  boolean	Gradient = FALSE;
  boolean	Compile = FALSE;
  boolean	Sequence = FALSE;
  boolean	Keep = FALSE;
  boolean	Help = FALSE;
  boolean	Args_Ok = TRUE;
  boolean	File_Ok = TRUE;
  int		Format = FORMAT_FULL;
  static char	Mat_Prefix[128] = "A";
  Bernoulli_Product	Problem;
  Double_Array		Direction;
  Weak_Derivative 	Mu;

  Direction = (Double_Array)NULL;

  /************************* Parsing *****************************/
  i = 1;

  if (argc>1)
    do {

      if GET_STRING( "-f", In_File_Name )
      else if GET_STRING( "-o", Out_File_Name )
      else if GET_STRING_SWITCH( "-seq", Seq_File_Name, Sequence, TRUE )
      else if GET_STRING( "-prefix", Mat_Prefix )
      else if GET_INT( "-nevt", Sim_Length )
      else if GET_INT( "-nrep", Nb_Replicates )
      else if SWITCH( "-stats", Stats, TRUE )
      else if SWITCH( "-gradient", Gradient, TRUE )
      else if SWITCH( "-compile", Compile, TRUE )
      else if SWITCH( "-keep", Keep, TRUE )
      else if SWITCH( "-debug", Debug, TRUE )
      else if IS_VERSION

      else if (( 0 == strcmp( "-h", argv[i] )) || 
	       ( 0 == strcmp( "-?", argv[i] )) ||
	       ( 0 == strcmp( "-help", argv[i] ))) {	
	i++;
	fprintf(stdout,"usage: %s [-f <file>] [-o <file>] [-seq <file>]",
		TOOLNAME );
	fprintf(stdout,"\n\t[-scilab] [-sparse] [-prefix <string>]");
	fprintf(stdout,"\n\t[-nevt <int>] [-nrep <int>] ");
	fprintf(stdout,"[-stats] [-gradient] [-dir <string>]\n\t" );
	fprintf(stdout,"[-compile] [-keep] [-v] [-h] [-help] [-?]\n");
	Args_Ok = FALSE;
	Help = TRUE;
      }
      else if ( ( 0 == strcmp( "-scilab", argv[i] ) ) ) {
	i++;
	Format = FORMAT_SCILAB;
      }
      else if ( ( 0 == strcmp( "-sparse", argv[i] ) ) ) {
	i++;
	Format = FORMAT_SPARSE;
      }
      else if ( 0 == strcmp( "-dir", argv[i] ) ) {
	i++;
	if ( i < argc ) {
	  Direction = Parse_Direction( argv[i] );
	  if ( (Double_Array)NULL != Direction ) {
	    /* presumed successful... */
	    i++;
	  }
	}
	else {
	  fprintf(stdout,"%s: error: string expected after %s.\n",
		  TOOLNAME,
		  argv[i-1]);
	  Args_Ok = FALSE;
	}
      }
      else 
	i++;
    } while( Args_Ok && ( i < argc ) );

  if (Args_Ok) {

    if ( strlen(In_File_Name) != 0 )
      File_Ok = ( NULL != freopen( In_File_Name, "r", stdin ) );

    if ( !File_Ok ) {
      fprintf(stderr,"%s: could not open file %s\n", TOOLNAME, In_File_Name );
    }

    if ( File_Ok ) {
      if ( strlen(Out_File_Name) != 0 )
	File_Ok = (NULL != freopen( Out_File_Name, "w", stdout ) );

      if ( !File_Ok ) {
	fprintf(stderr,"%s: could not open file %s\n", 
		TOOLNAME, Out_File_Name);
	File_Ok = FALSE;
      }
    }
  }

  /************* start of the program if everything is Ok *******************/
  if ( Args_Ok && File_Ok ) {

    if ( !Sequence && Compile ) {
      if ( !Gradient ) {
	if ( Compile_Matrix_And_Run( In_File_Name, 
				     Sim_Length, Nb_Replicates,
				     Stats, Keep, argp ) )
	  return 0;
	else 
	  return -3;
      }
      else {
	fprintf(stderr,"%s: cannot compile when gradient is asked.\n",
		TOOLNAME );
	return -1;
      }
    }
    else {

      Problem = Read_Bernoulli_Product( stdin );
      if ( (Bernoulli_Product) NULL != Problem ) {

	if ( Sequence ) {
	  Simulate_Sequence( Problem, Seq_File_Name, Format, Mat_Prefix );
	}
	else {
	  /* stochastic simulation of Bernoulli product */
	
	  sort_probas( Problem );

	  if ( Gradient ) {
	    /* if no direction, make default one */
	    if ( (Double_Array)NULL == Direction ) {
	      Direction = (Double_Array) malloc( sizeof( struct DOUBLE_ARRAY ) );
	      Direction->Size = Problem->Nb_Matrices;
	      Direction->Val = (double*) calloc( (size_t)Direction->Size, 
						 sizeof( double ) );
	      Direction->Val[0] = 1.0;
	      Direction->Val[1] = -1.0;
	    }
	    else {
	      /* Direction is there. Check consistency */
	      if ( Problem->Nb_Matrices < Direction->Size ) {
		printf("%s: error: inconsistent gradient direction. Nb matrices (%d) < max index (%d).\n",
		       TOOLNAME, Problem->Nb_Matrices, Direction->Size );
		return -3;
	      }
	      else if ( Problem->Nb_Matrices > Direction->Size ) {
		Direction->Size = Problem->Nb_Matrices;
		Direction->Val = 
		  realloc( Direction->Val, 
			   (size_t)Direction->Size*sizeof( double ) );
	      }
	    }

	    /* check assumptions */
	    Mu = Split_Distribution( Direction );
	    if ( (Weak_Derivative)NULL != Mu ) {
	      Simulate_With_Gradient( Problem, Mu,
				      Sim_Length, Nb_Replicates, Stats );
	    }
	    else {
	      /* bad assumptions. Error message already issued. */
	      return -4;
	    }
	  }
	  else {
	    Simulate( Problem, Sim_Length, Nb_Replicates, Stats );
	  }
	}

	/* simulation successful */
	Free_Bernoulli_Product( Problem );

	return 0;
      }
      else
	return -2;
    }
  }
  else if ( !Help ) {
    return -1;
  }
  else
    return 0;
}

#define TEMP_FILE "/tmp/mat_sim.mat"
#define TEMP_C_FILE "/tmp/mat_sim.c"
#define TEMP_SIM "/tmp/mat_sim"

boolean	Compile_Matrix_And_Run( char 	In_File_Name[MAX_NAME_LENGTH],
				int	Sim_Length, 
				int	Nb_Replicates,
				boolean Stats,
				boolean	Keep,
				char 	*argp[] )
{
  char		mat2c_Input[MAX_NAME_LENGTH];
  char		command[2*MAX_NAME_LENGTH];
  char		arg_evt1[MAX_NAME_LENGTH];
  char		arg_evt2[MAX_NAME_LENGTH];
  char		arg_rep1[MAX_NAME_LENGTH];
  char		arg_rep2[MAX_NAME_LENGTH];
  char		arg_stat[MAX_NAME_LENGTH];
  char		arg_grad[MAX_NAME_LENGTH];
  boolean	All_Ok = TRUE;
  int		Son_PID;
  int		Res;
  FILE		*Out;
  char		c;
    
  if ( 0 != strlen( In_File_Name ) ) {
    strncpy( mat2c_Input, In_File_Name, MAX_NAME_LENGTH );
  }
  else {
    /* pipe the standard input to a temporary file */
    All_Ok = ( NULL != ( Out = fopen( TEMP_FILE, "w" ) ) );
    if ( !All_Ok ) {
      printf("%s: error: could not open file %s.\n", 
	     TOOLNAME, TEMP_FILE );
      return FALSE;
    }
    else {
      do {
	c = getchar();
	if ( c != EOF )
	  fprintf( Out, "%c", c );
      } while ( c != EOF );
    }
    fclose( Out );

    strncpy( mat2c_Input, TEMP_FILE, MAX_NAME_LENGTH );
  }

  if ( All_Ok ) {

    sprintf( command, "mat2c -f %s -o %s", mat2c_Input, TEMP_C_FILE );
    printf("Executing: %s\n", command );
    if ( 0 == ( Son_PID = fork() ) ) {
      Res = execlp( "mat2c", "mat2c", "-f", mat2c_Input, "-o", TEMP_C_FILE,
		    (char*)NULL, argp );
      printf("Res = %d, Err = %d\n", Res, errno );
    }
    else
      waitpid( Son_PID, NULL, 0 );

    if ( 0 == ( Son_PID = fork() ) ) {
      sprintf( command, "gcc %s -o %s -lm -O2", TEMP_C_FILE, TEMP_SIM );
      printf("Executing: %s\n", command );
      
      Res = execlp( "gcc", "gcc", TEMP_C_FILE, "-o", TEMP_SIM, 
		   "-lm", "-O2", (char*)NULL, argp );
      printf("Res = %d, Err = %d\n", Res, errno );
    }
    else
      waitpid( Son_PID, NULL, 0 );
	
    if ( 0 == ( Son_PID = fork() ) ) {

      /* preparing command line */
      if ( Sim_Length != -1 ) {
	sprintf( arg_evt1, "%s", "-nevt" );
	sprintf( arg_evt2, "%d", Sim_Length );
      }
      else {
	sprintf( arg_evt1, "%s", "" );
	sprintf( arg_evt2, "%s", "" );
      }

      if ( Stats ) {
	sprintf( arg_stat, "%s", "-stats" );
      }
      else {
	sprintf( arg_stat, "%s", "" );
      }

      if ( Nb_Replicates > 1 ) {
	sprintf( arg_rep1, "%s", "-nrep" );
	sprintf( arg_rep2, "%d", Nb_Replicates );
      }
      else {
	sprintf( arg_rep1, "%s", "" );
	sprintf( arg_rep2, "%s", "" );
      }

      printf("Executing: %s %s %s %s %s %s\n", 
	     TEMP_SIM, 
	     arg_evt1, arg_evt2,
	     arg_rep1, arg_rep2,
	     arg_stat );

      Res =  execl( TEMP_SIM, TEMP_SIM,
		    arg_evt1, arg_evt2,
		    arg_rep1, arg_rep2,
		    arg_stat,
		    (char*)NULL );

      printf("Res = %d, Err = %d\n", Res, errno );
    }
    else
      waitpid( Son_PID, NULL, 0 );

    if ( !Keep ) {
      unlink( TEMP_FILE );
      unlink( TEMP_C_FILE );
      unlink( TEMP_SIM );
    }

    return TRUE;
  }
  else
    return FALSE;
}


#define ST_FAIL		0
#define ST_STOP		1
#define ST_BEG_TERM	2
#define ST_END_TERM	3
#define ST_ID		4
char *State_Name[5] = { "Fail", "Stop", "Begin", "End", "Id" };

Double_Array Parse_Direction( char *String ) {
  int		Cur_Idx;
  int		Nb_Read;
  int		i;
  int		Length;
  int		Cur_Id;
  int		Cur_Dec;
  double	Cur_Val;
  double	Cur_Sign;
  int		State;
  char		*S;
  Double_Array	Res;

  Res = (Double_Array) malloc( sizeof( struct DOUBLE_ARRAY ) );
  Res->Size = 0;
  Res->Val = (double*)NULL;

  Length = strlen(String);
  S = (char*) calloc( (size_t)(1+Length), sizeof( char ) );
  S = strcpy( S, String );

  Cur_Idx = 0;
  Cur_Val = 0.0;
  Cur_Dec = 0;
  Cur_Sign = 1.0;
  if ( S[0] == '-' ) {
    Cur_Sign = -1.0;
    S[0] = ' ';
    Cur_Idx++;
  }
  State = ST_BEG_TERM;

  do {
    if ( Debug ) {
      printf("ParseDir: Idx = %2d  St = %6s  (S = %s)\n", 
	     Cur_Idx, State_Name[State], S );
    }
    switch ( State ) {
    case ST_END_TERM:
      if ( S[Cur_Idx] == '+' ) {
	Cur_Sign = 1.0;
	S[Cur_Idx] = ' ';
	Cur_Idx++;
	State = ST_BEG_TERM;
      }
      else if ( S[Cur_Idx] == '-' ) {
	Cur_Sign = -1.0;
	S[Cur_Idx] = ' ';
	Cur_Idx++;
	State = ST_BEG_TERM;
      }
      else {
	State = ST_FAIL;
      }
      break;
    case ST_BEG_TERM:
      if ( 1 == sscanf( S, "%lf p%n", &Cur_Val, &Nb_Read ) ) {
	if ( Debug ) {
	  printf("ParseDir: Read = %d\n", Nb_Read );
	}
	for ( i = 0; i < Nb_Read; i++ ) {
	  S[i] = ' ';
	}
	Cur_Idx = Nb_Read;
	State = ST_ID;
      }
      else if ( S[Cur_Idx] == 'p' ) {
	S[Cur_Idx] = ' ';
	Cur_Idx ++;
	Cur_Val = 1.0;
	State = ST_ID;
      }
      else {
	State = ST_FAIL;
      }
      break;
    case ST_ID:
      if ( 1 == sscanf( S, "%d%n", &Cur_Id, &Nb_Read ) ) {
	if ( Debug ) {
	  printf("ParseDir: Read = %d\n", Nb_Read );
	}
	for ( i = 0; i < Nb_Read; i++ ) {
	  S[i] = ' ';
	}
	Cur_Idx = Nb_Read;
	if ( Cur_Id > Res->Size ) {
	  Res->Size = Cur_Id;
	  Res->Val = (double*) realloc( Res->Val, Cur_Id*sizeof(double) );
	}
	else if ( Res->Val[Cur_Id-1] != 0.0 ) {
	  printf("%s: warning: duplicate p%d in direction.\n",
		 TOOLNAME, Cur_Id );
	}
	Res->Val[Cur_Id-1] = Cur_Val * Cur_Sign;
/* 	printf("D[%d] <- %f\n", Cur_Id-1, Res->Val[Cur_Id-1] ); */
	if ( Cur_Idx == Length )
	  State = ST_STOP;
	else
	  State = ST_END_TERM;
      }
      else {
	State = ST_FAIL;
      }
    }
  } while ( ( State != ST_STOP ) && ( State != ST_FAIL ) );

  if ( State == ST_FAIL ) {
    printf("%s: syntax error in direction at char. #%d (%c).\n",
	   TOOLNAME, Cur_Idx, S[Cur_Idx] );
    if ( (double*)NULL != Res->Val )
      free( Res->Val );
    free( Res );
    Res = (Double_Array)NULL;
  }

  return Res;
}

/* Simulate a finite product of matrices given by a sequence 	*/
/* The result is a (max,+) matrix.				*/
/* Full matrices are used because products usually are.		*/
void Simulate_Sequence( Bernoulli_Product 	Problem,
			char 			Seq_File_Name[MAX_NAME_LENGTH],
			int			Format,
			char			Prefix[128] )
{
  int		Num;
  int		idx;
  int		i;
  int		j;
  int		k;
  MaxPlus_Matrix A;
  Full_MaxPlus_Matrix Prod;
  Full_MaxPlus_Matrix Res;
  Full_MaxPlus_Matrix Tmp;
  FILE		*Input;
  boolean	Ok;

  if ( NULL == ( Input = fopen( Seq_File_Name, "r" ) ) ) {
    printf("%s: error: could not open file %s for reading.\n", 
	   TOOLNAME, Seq_File_Name );
    return;
  } 

  /* Initialize matrix product */
  Prod = New_Full_MaxPlus_Matrix( Problem->Size, Problem->Size, IDENTITY );
  Res = New_Full_MaxPlus_Matrix( Problem->Size, Problem->Size, NONE );

  if ( (Full_MaxPlus_Matrix)NULL != Res ) {

    idx = 0;

    do {
      Ok = ( 1 == fscanf( Input, "%d", &Num ) );
      if ( Ok ) {
	idx ++;
	if ( ( Num >= 1 ) && ( Num <= Problem->Nb_Matrices ) ) {
#ifdef USE_FULL_MATRICES
	  Full_MM_Mult( Problem->A[Num-1], Prod, Res );
#else
	  /* performing the hybrid sparse/full multiplication */
	  A = Problem->A[Num-1];
	  for ( i=0; i<Problem->Size; i++ ) {
	    for ( j=0; j<Problem->Size; j++ ) {
	      Res->Vals[i][j] = M_INFINITY;
	      for( k=0; k<A->Nb_Elts[i]; k++ ) {
		Res->Vals[i][j] = max( Res->Vals[i][j], 
				       plus( A->Vals[i][k], 
					     Prod->Vals[A->Elts[i][k]][j] ) );
	      }
	    }
	  }
#endif
	  Tmp = Res; Res = Prod; Prod = Tmp;
	}
	else {
	  fprintf( stdout, "%s: Warning: matrix id %d out of bounds at sequence index %d.\n",
		   TOOLNAME, Num, idx );
	}
      }
    } while ( Ok );

    if ( Format != FORMAT_SCILAB ) {
      fprintf( stdout, "%d\n", Problem->Size );
    }
    Write_Full_MaxPlus_Matrix( Prod, stdout, Format, Prefix );

  }
  else {
    fprintf( stderr, "%s: error: not enough memory. Abort.\n", TOOLNAME );
  }

  fclose( Input );

}
