/****************************************************************/
/* Partie relative a lex/yacc					*/
/****************************************************************/

#include <malloc.h>
#include "tetris.h"
#include "tetris.tab.h"

#define MAX(x,y) ( x > y ? x : y )

#define	E_MEM(x) fprintf(stderr,"Warning: not enough memory: %s not created\n",x);
#define W_EXT(x) fprintf(stderr,"Warning: extra parameters for %c (ignored)\n", x );
#define W_MIS(x) fprintf(stderr,"Warning: missing parameters for %c. Default 0.0.\n", x );

#define SCAN(x,y)	for(x=y;x!=NULL;x=x->Next)

extern FILE *yyin;
extern int yydebug;
extern char yytext[];
#ifdef YY_FLEX_LEX_COMPAT
extern int yylineno;
#endif
extern int yyparse();
extern int yylex();
extern int Tetris_lineno;

/* extern Tetris methods */
extern Tetris New_Tetris(int,int,Piece_Desc *);
extern void Free_Tetris(Tetris);

/* forced external declaration */
extern char *strdup(const char *);

/* forward declaration */
void Free_Piece_List(Piece_List);

boolean	Debug = FALSE;
Tetris The_Tetris;
int    Nb_Of_Columns;
#define PARSE_RESULT The_Tetris

Tetris Read_Tetris(char *File_Name)
{
  FILE	*Input_File;
  FILE	*Current_Input;
  int	Res;

  if ( strlen( File_Name ) > 0 ) {
    Input_File = fopen( File_Name, "r" );
  }
  else {
    Input_File = stdin;
  }

  if ( NULL != Input_File ) {

    Current_Input = yyin;
    yyin = Input_File;

    yydebug = 0;
    Res = yyparse();

    yyin = Current_Input;

    fclose( Input_File );

    if ( Res == 0 )
      return PARSE_RESULT;
    else
      return (Tetris)NULL;
  }
  else {
    printf("Error: could not open file %s for reading.\n", File_Name );

    return (Tetris)NULL;
  }
}

void yyerror(char *s)
{
#ifdef YY_LEX_FLEX_COMPAT
  printf("---%s---[%s] at line %d\n",s,yytext,yylineno);
#else
  printf("---%s---[%s] at line %d\n",s,yytext,Tetris_lineno);
#endif
}

Tetris Parse_Tetris(int Nb_Columns, 
		    Piece_List The_Pieces )
{
  int			idx;
  int			Nb_Pieces;
  Piece_List	 	P_Scan;
  Piece_Desc		*Piece_Table;
  Tetris		Result;

  /* count nb of pieces */
  Nb_Pieces = 0;
  SCAN( P_Scan, The_Pieces ) {
    Nb_Pieces++;
  }

  Piece_Table = (Piece_Desc*) 
    calloc( (size_t)Nb_Pieces, sizeof( Piece_Desc ) );

  /* Scan the list of pieces and put them in the table (in reverse as 	*/
  /* the parsin produces a "pile".					*/
  idx = Nb_Pieces-1;
  SCAN( P_Scan, The_Pieces ) {
    Piece_Table[idx] = P_Scan->Val;
    idx--;
  }

  Result = New_Tetris( Nb_Columns, Nb_Pieces, Piece_Table );

  /* collect colors. */
  for ( idx=0; idx<Nb_Pieces; idx++ ) {
    if ( (char*)NULL !=  Piece_Table[idx]->Color ) 
      sprintf( Result->Color[idx], Piece_Table[idx]->Color );
  }

  Free_Piece_List( The_Pieces );

  return Result;
}

Piece_Desc Parse_Piece(int		Piece_Num,
		       double		Weight,
		       char		*Color,
		       Sparse_Int_Array	Lo_Profile,
		       Sparse_Int_Array	Hi_Profile )
{
  int		i;
  int		low;
  Piece_Desc	The_Piece;

  The_Piece = New_Piece( Nb_Of_Columns );

  /* trace profiles
  printf("Lo Pro -> ");
  for ( i=0; i < Lo_Profile->Nb_Indexes; i++ ) {
    printf("(%d,%d) ", Lo_Profile->Indexes[i], Lo_Profile->Indexes[i] );
  }
  printf("\n");
  printf("Hi Pro -> ");
  for ( i=0; i < Hi_Profile->Nb_Indexes; i++ ) {
    printf("(%d,%d) ", Hi_Profile->Indexes[i], Hi_Profile->Indexes[i] );
  }
  printf("\n");
  */

  if ( The_Piece != NULL ) {
    The_Piece->Number = Piece_Num;
    The_Piece->Nb_Columns = Nb_Of_Columns;
    The_Piece->LoProf 
      = (int*) calloc( (size_t)The_Piece->Nb_Columns, sizeof(int) );
    The_Piece->HiProf 
      = (int*) calloc( (size_t)The_Piece->Nb_Columns, sizeof(int) );
    The_Piece->Color = (char*)NULL;

    if ( ( (int*)NULL == The_Piece->LoProf )
	 || ( (int*)NULL == The_Piece->LoProf ) ) {
      E_MEM("piece");
      free( The_Piece );
      The_Piece = (Piece_Desc)NULL;
    }
    /* Now fill profiles in while performing checks and computing normal
      form. First, initialize. */
    The_Piece->Color = Color;

    for ( i=0; i<The_Piece->Nb_Columns; i++ ) {
      The_Piece->HiProf[i] = -1;	/* -1 means epsilon because in */
      The_Piece->LoProf[i] = -1;	/* normal form, all entries are >=0 */
    }

    if ( Lo_Profile->Nb_Indexes == 0 ) {
      fprintf(stderr,"Warning: piece %d is empty.\n", The_Piece->Number );
    }
    else {
      /* Computing lowest point in low profile */
      low = Lo_Profile->Values[0];
      for ( i=1; i<Lo_Profile->Nb_Indexes; i++ )
	low = ( Lo_Profile->Values[i] < low ? Lo_Profile->Values[i] : low );

      /* storing, shifting with respect to "low" */
      for ( i=0; i<Lo_Profile->Nb_Indexes; i++ ) {
	The_Piece->LoProf[Lo_Profile->Indexes[i]] 
	  = Lo_Profile->Values[i] - low;
      }
      for ( i=0; i<Hi_Profile->Nb_Indexes; i++ ) {
	The_Piece->HiProf[Hi_Profile->Indexes[i]] 
	  = Hi_Profile->Values[i] - low;
      }
      
      /* checking consistency */
      for ( i=0; i<The_Piece->Nb_Columns; i++ ) {
	if ( ( ( The_Piece->LoProf[i] == -1 )
	       && ( The_Piece->HiProf[i] != -1 ) )
	     || ( ( The_Piece->LoProf[i] != -1 )
		  && ( The_Piece->HiProf[i] == -1 ) )
	     || ( ( The_Piece->LoProf[i] != -1 )
		  && ( The_Piece->HiProf[i] < The_Piece->LoProf[i] ) )
	     ) {
	  fprintf(stderr,"Warning: inconsistency in profiles of piece %d", 
		  The_Piece->Number );
	  fprintf(stderr," at column %d (%d|%d). Column ignored.\n", 
		  The_Piece->LoProf[i], The_Piece->HiProf[i], i );
	  The_Piece->LoProf[i] = -1;
	  The_Piece->HiProf[i] = -1;
	}
      }
      if ( Debug ) {
	/* trace piece */
	printf("Got Piece %d: ", The_Piece->Number );
	for ( i=0; i<The_Piece->Nb_Columns; i++ ) {
	  if ( The_Piece->LoProf[i] != -1 ) {
	    printf("%d[%d-%d] ", i, The_Piece->LoProf[i], 
		   The_Piece->HiProf[i] );
	  }
	}
	printf("\n");
      }
    }
  }

  /* cleanup */
  free( Hi_Profile->Values );
  free( Hi_Profile->Indexes );
  free( Hi_Profile );
  free( Lo_Profile->Values );
  free( Lo_Profile->Indexes );
  free( Lo_Profile );

  return The_Piece;
}

Piece_List Append_Piece(Piece_List	Cur_List,
			Piece_Desc	New_Piece )
{
  Piece_List The_List;

  The_List = (Piece_List) malloc( sizeof( struct PIECE_LIST ) );

  if ( The_List != NULL ) {
    The_List->Val = New_Piece;
    The_List->Next = Cur_List;
  }
  else {
    E_MEM("Piece");
  }

  return The_List;
}

Sparse_Int_Array Append_Pair( Sparse_Int_Array 	The_Profile,
			      BiPoint		The_Pair )
{
  Sparse_Int_Array	New_Profile;

  if ( The_Pair.Height == EPSILON ) {
    return The_Profile;
  }
  else {
    if ( (Sparse_Int_Array)NULL == The_Profile ) {
      New_Profile = 
	(Sparse_Int_Array) malloc( sizeof( struct SPARSE_INT_ARRAY ) );
      if ( (Sparse_Int_Array)NULL == New_Profile ) {
	E_MEM("Profile");
      }
      else {
	New_Profile->Nb_Indexes = 0;
	New_Profile->Indexes = (int*)NULL;
	New_Profile->Values = (int*)NULL;
      }
    }
    else {
      New_Profile = The_Profile;
    }

    New_Profile->Nb_Indexes++;
    New_Profile->Indexes = (int*)realloc( New_Profile->Indexes,
					  (size_t)New_Profile->Nb_Indexes
					  *sizeof(int));
    New_Profile->Values = (int*)realloc( New_Profile->Values,
					 (size_t)New_Profile->Nb_Indexes
					 *sizeof(int));
    New_Profile->Indexes[New_Profile->Nb_Indexes-1] = The_Pair.Col;
    New_Profile->Values[New_Profile->Nb_Indexes-1] = The_Pair.Height;

    return New_Profile;
  }
}

/****************************************************************/
/* Cleanup procedures						*/
/****************************************************************/
void Free_Piece_List( Piece_List L )
{
  Piece_List	M;
  Piece_List	N;

  for ( M = L; M != NULL; ) {
    N = M;
    M = N->Next;
    free( N );
  }
  
}
