/* Various Sparse matrix manipulation procedures */

#include <stdlib.h>
#include <stdio.h>
#include "boolean.h"
#include "maxpluscore.h"
#include "maxplusmat.h"
#include "spmat.h"

#define TOOLNAME		"SpMat"

/* Constructors and Destructors */
SP_MaxPlus_Matrix New_SP_MaxPlus_Matrix(int	Rows, 
					int	Cols,
					int 	Init_Mode )
{
  SP_MaxPlus_Matrix	Res;

  Res = (SP_MaxPlus_Matrix) malloc( sizeof( struct SP_MAXPLUS_MATRIX ) );

  if ( (SP_MaxPlus_Matrix)NULL != Res ) {
    Res->Rows = Rows;
    Res->Cols = Cols;

    Res->Nb_Elts = (int*) calloc( (size_t)Rows, sizeof(int) );
    Res->Elts = (int**) calloc( (size_t)Rows, sizeof(int*) );
    Res->Vals = (double**) calloc( (size_t)Rows, sizeof(double*) );

    if ( ( (int*)NULL != Res->Nb_Elts )
	 && ( (int**)NULL != Res->Elts )
	 && ( (double**)NULL != Res->Vals ) ) {
      Init_SP_MaxPlus_Matrix( Res, Init_Mode );
    } 
    else {
      /* abort... */
      Free_SP_MaxPlus_Matrix( Res );
      Res = (SP_MaxPlus_Matrix)NULL;
    }
  }

  return Res;
}

void Free_SP_MaxPlus_Matrix(SP_MaxPlus_Matrix A)
{
  int	i;

  if ( (SP_MaxPlus_Matrix)NULL != A ) {
    free( A->Nb_Elts );
    if ( (int**)NULL != A->Elts ) {
      for ( i=0; i<A->Rows; i++ )
	free( A->Elts[i] );
      free( A->Elts );
    }
    if ( (double**)NULL != A->Vals ) {
      for ( i=0; i<A->Rows; i++ )
	free( A->Vals[i] );
      free( A->Vals );
    }
    free( A );
  }
}

SP_MaxPlus_Vector New_SP_MaxPlus_Vector(int Size, int Init_Mode)
{
  int			i;
  SP_MaxPlus_Vector	Res;

  Res = (SP_MaxPlus_Vector) malloc( sizeof( struct SP_MAXPLUS_VECTOR ) );

  if ( (SP_MaxPlus_Vector)NULL != Res ) {
    Res->Size = Size;

    Res->Vals = (double*) calloc( (size_t)Size, sizeof(double) );

    if ( (double*)NULL != Res->Vals ) {
      /* initialization */
      switch( Init_Mode ) {
      case ZERO:
	for ( i=0; i<Res->Size; i++ ) {
	  Res->Vals[i] = M_INFINITY;
	}
	break;
      case ONE:
	for ( i=0; i<Res->Size; i++ ) {
	  Res->Vals[i] = 0.0;
	}
      case NONE:
	/* none ! */
	break;
      default:
	printf("%s: error: illegal init mode (%d) for vectors.\n",
	       TOOLNAME, Init_Mode );
	/* abort... */
	Free_SP_MaxPlus_Vector( Res );
	Res = (SP_MaxPlus_Vector)NULL;
      }
    }
    else {
      /* abort... */
      Free_SP_MaxPlus_Vector( Res );
      Res = (SP_MaxPlus_Vector)NULL;
    }
  }

  return Res;
}

void Free_SP_MaxPlus_Vector(SP_MaxPlus_Vector V)
{
  if ( (SP_MaxPlus_Vector)NULL != V ) {
    free( V->Vals );
    free( V );
  }
}

boolean Init_SP_MaxPlus_Matrix(SP_MaxPlus_Matrix 	A,
			       int			Init_Mode)
{
  int		i;
  int		j;
  boolean 	Ok = TRUE;

  /* initialization */
  switch( Init_Mode ) {
  case ZERO:
    for ( i=0; i<A->Rows; i++ ) {
      A->Nb_Elts[i] = 0;
      A->Elts[i] = (int*)NULL;
      A->Vals[i] = (double*)NULL;
    }
    break;
  case ONE:
    for ( i=0; i<A->Rows; i++ ) {
      A->Nb_Elts[i] = A->Rows;
      A->Elts[i] = (int*) calloc( (size_t)A->Rows, sizeof(int) );
      A->Vals[i] = (double*) calloc( (size_t)A->Rows, sizeof(double) );
      for ( j=0; j<A->Rows; j++ ) {
	A->Elts[i][j] = j;
	A->Vals[i][j] = 0.0;
      }
    }
    break;
  case IDENTITY:
    for ( i=0; i<A->Rows; i++ ) {
      A->Nb_Elts[i] = 1;
      A->Elts[i] = (int*) calloc( 1, sizeof(int) );
      A->Vals[i] = (double*) calloc( 1, sizeof(double) );
      A->Elts[i][0] = i;
      A->Vals[i][0] = 0.0;
    }
    break;
  case NONE:
    /* none ! */
    break;
  default:
    printf("%s: error: illegal init mode (%d) for matrices.\n",
	   TOOLNAME, Init_Mode );
    Ok = FALSE;
  }

  return Ok;
}

SP_MaxPlus_Matrix Make_SP_MaxPlus_Matrix(int		Rows,
					 int		Cols,
					 double		**Vals )
{
  int		i;
  int		j;
  int		Nb_Elts;
  boolean 	Ok = TRUE;
  SP_MaxPlus_Matrix	Res;

  Res = New_SP_MaxPlus_Matrix( Rows, Cols, NONE );
  
  Ok = ( (SP_MaxPlus_Matrix)NULL != Res );

  for ( i=0; (i<Rows) && Ok; i++ ) {
    /* counting the number of non-eps elements */
    Nb_Elts = 0;
    for ( j=0; j<Cols; i++ ) {
      if ( Vals[i][j] != M_INFINITY )
	Nb_Elts++;
    }
    /* allocating */
    Res->Nb_Elts[i] = Nb_Elts;
    Res->Elts[i] = (int*) calloc( (size_t)Nb_Elts, sizeof(int) );
    Res->Vals[i] = (double*) calloc( (size_t)Nb_Elts, sizeof(double) );
    Ok = Ok && ( (int*)NULL != Res->Elts[i] ) 
      && ( (double*)NULL != Res->Vals[i] );

    if ( Ok ) {
      /* setting values */
      Nb_Elts = 0;
      for ( j=0; j<Cols; i++ ) {
	if ( Vals[i][j] != M_INFINITY ) {
	  Res->Elts[i][Nb_Elts] = j;
	  Res->Vals[i][Nb_Elts] = Vals[i][j];
	  Nb_Elts++;
	}
      }
    }
  }

  if ( ! Ok ) {
    Free_SP_MaxPlus_Matrix( Res );
    Res = (SP_MaxPlus_Matrix)NULL;
  }

  return Res;
}

Full_MaxPlus_Matrix Convert_SP_Matrix_To_Full(SP_MaxPlus_Matrix A)
{
  int		i;
  int		j;
  Full_MaxPlus_Matrix	Res;

  Res = New_Full_MaxPlus_Matrix( A->Rows, A->Cols, ZERO );
  
  if ( (Full_MaxPlus_Matrix)NULL != Res ) {
    for ( i=0; i<A->Rows; i++ ) {
      for ( j=0; j<A->Cols; j++ ) {
	Res->Vals[i][A->Elts[i][j]] = A->Vals[i][j];
      }
    }
  }
  else {
    Free_Full_MaxPlus_Matrix( Res );
    Res = (Full_MaxPlus_Matrix)NULL;
  }

  return Res;
}

/* Setting an element in the matrix */
boolean Set_SP_Matrix_Entry(SP_MaxPlus_Matrix	Matrix, 
			    int			Row,
			    int			Col,
			    double		Val)
{
  int		i;
  int		j;

  if ( ( 0 <= Row ) && ( Row < Matrix->Rows ) 
       && ( 0 <= Col ) && ( Col < Matrix->Cols ) ) {

    /* lookup if element already there */
    for ( i=0; 
	  ( i<Matrix->Nb_Elts[Row] ) && ( Matrix->Elts[Row][i]!=Col ); 
	  i++ );
    if ( i<Matrix->Nb_Elts[Row] ) {
      /* found it. */
      if ( Val != M_INFINITY ) {
	/* Just replacing value */
	Matrix->Vals[Row][i] = Val;
      }
      else {
	/* removing value */
	for ( j=i+1; j<Matrix->Nb_Elts[Row]; j++ ) {
	  Matrix->Elts[Row][j-1] = Matrix->Elts[Row][j];
	  Matrix->Vals[Row][j-1] = Matrix->Vals[Row][j];
	}
	Matrix->Nb_Elts[Row]--;
      }
      return TRUE;
    }
    else {
      /* not there */
      if ( Val != M_INFINITY ) {
	/* Adding it to the structure */
	Matrix->Nb_Elts[Row]++;
	Matrix->Elts[Row] = 
	  (int*)realloc( Matrix->Elts[Row],
			 (size_t)Matrix->Nb_Elts[Row]*sizeof(int) );
	Matrix->Vals[Row] = 
	  (double*)realloc( Matrix->Vals[Row],
			    (size_t)Matrix->Nb_Elts[Row]*sizeof(double) );
	
	if ( ( (int*)NULL == Matrix->Elts[Row] )
	     || ( (double*)NULL == Matrix->Vals[Row] ) ) {
	  return FALSE;
	}
	else {
	  Matrix->Elts[Row][Matrix->Nb_Elts[Row]-1] = Col;
	  Matrix->Vals[Row][Matrix->Nb_Elts[Row]-1] = Val;
	  return TRUE;
	}
      }
      else {
	/* adding an eps always successful ! */
	return TRUE;
      }
    }
  }
  else {
    printf("Warning: setting value to (%d,%d) in a %d x %d matrix.",
	   Row, Col, Matrix->Rows, Matrix->Cols );
    printf(" Ignored.\n");
    return FALSE;
  }
}

/* Reading from a file, and adding an element to the matrix */
boolean SP_Read_Full_Matrix_Entry(FILE		*Input, 
				  SP_MaxPlus_Matrix	Matrix, 
				  int		Row,
				  int		Col)
{
  char		Entry[256];
  double	Val;

  fscanf( Input, "%s", Entry );
  if ( 1 == sscanf( Entry, "%lf", &Val ) ) {
    if ( Set_SP_Matrix_Entry( Matrix, Row, Col, Val ) )
      return TRUE;
    else {
      printf("Warning: failed to set entry (%d,%d) to %f.\n",
	     Row, Col, Val );
      return FALSE;
    }
  }
  else if ( !strcmp( Entry, "eps" ) ) {
    return TRUE;
  }
  else {
    printf("%s: illegal matrix entry %s.\n", TOOLNAME, Entry );
    return FALSE;
  }
}

/* Reading from a file, and adding an element to the matrix */
boolean SP_Read_SP_Matrix_Entry(FILE			*Input, 
				SP_MaxPlus_Matrix	Matrix)
{
  double	Val;
  int		Row;
  int		Col;

  if ( 3 == fscanf( Input, "%d %d %lf", &Row, &Col, &Val ) ) {
    if ( Set_SP_Matrix_Entry( Matrix, Row, Col, Val ) )
      return TRUE;
    else {
      printf("Warning: failed to set entry (%d,%d) to %f.\n", Row, Col, Val );
      return FALSE;
    }
  }
  else {
    return FALSE;
  }
}

boolean SP_Read_Full_MaxPlus_Matrix( FILE		*Input,
				     SP_MaxPlus_Matrix 	A )
{
  int	i;
  int	j;
  boolean Ok;

  Ok = TRUE;

  for ( i=0; (i<A->Rows) && Ok; i++ ) {
    for ( j=0; (j<A->Cols) && Ok; j++ ) {

      if ( !SP_Read_Full_Matrix_Entry( Input, A, i, j ) ) {
	printf("%s: error: unexpected value (%d,%d) for matrix. Abort.\n",
	       TOOLNAME, i, j );
	Ok = FALSE;
      }
    }
  }
  return Ok;
}

boolean SP_Read_SP_MaxPlus_Matrix( FILE		*Input,
				   SP_MaxPlus_Matrix A )
{
  boolean Ok;
  char	Token[256];

  Ok = TRUE;

  do {
    Ok = SP_Read_SP_Matrix_Entry( Input, A );
  } while ( Ok );
  /* skip stopping mark */
  fscanf( Input, "%s", Token );

  return TRUE;
}

boolean SP_Read_MaxPlus_Matrix( FILE			*Input,
				SP_MaxPlus_Matrix	A )
{
  int		i;
  int		j;
  boolean 	Ok;
  double	Val;
  char		Token[128];

  /* I write one single procedure with reading the two types of matrices */
  /* because I could not solve the problem of the first token, which	*/
  /* may be either the specifyer "sp" or an entry of the matrix.	*/

  /* Reading the first token to determine if the matrix is sparse. */
  fscanf( Input, "%128s", Token );
  if ( !strcmp( Token, "sp" ) ) {
    do {
      Ok = SP_Read_SP_Matrix_Entry( Input, A );
    } while ( Ok );
    /* skip stopping mark */
    fscanf( Input, "%s", Token );
    Ok = TRUE;
  }
  else {
    /* Must be full. Proceed with the reading, but skip the first read. */
    Ok = TRUE;
    for ( i=0; Ok && (i<A->Cols); i++ ) {
      for ( j=0; Ok && (j<A->Rows); j++ ) {
	if ( ( i != 0 ) || ( j != 0 ) ) 
	  fscanf( Input, "%s", Token );

	/* convert value and store it */
	if ( 1 == sscanf( Token, "%lf", &Val ) ) {
	  if ( Val < M_INFINITY ) {
	    fprintf(stderr,"warning: matrix entry (%e) less than eps.", Val );
	    fprintf(stderr," eps assumed.\n");
	    Val = M_INFINITY;
	  }
	}
	else if ( !strcmp( Token, "eps" ) ) {
	  Val = M_INFINITY;
	}
	else {
	  printf("%s: illegal matrix entry (%d,%d): %s.\n", 
		 TOOLNAME, i, j, Token );
	  Ok = FALSE;
	}
	
	if ( Ok ) {
	  if ( !Set_SP_Matrix_Entry( A, i, j, Val ) ) {
	    printf("Warning: failed to set entry (%d,%d) to %f.\n", 
		   i, j, Val );
	    Ok = FALSE;
	  }
	}
      }
    }
  }

  return Ok;
}
  
/* Retrieving an entry from the matrix. This is not optimized at all... */
double Get_SP_Matrix_Entry(SP_MaxPlus_Matrix	Matrix, 
			   int			Row,
			   int			Col)
{
  int		i;
  double	Res;

  Res = M_INFINITY;
  for( i=0; i<Matrix->Nb_Elts[Row]; i++ ) {
    if ( Matrix->Elts[Row][i] == Col ) {
      Res = Matrix->Vals[Row][i];
    }
  }

  return Res;
}

/* Linear algebra... */
void SP_VM_Mult(SP_MaxPlus_Vector 	X,
		SP_MaxPlus_Matrix	Matrix, 
		SP_MaxPlus_Vector	Res )
{
  int		i;
  int		j;
  int		k;

  if ( ( X->Size != Matrix->Rows ) 
       || ( X->Size != Res->Size ) ) {
    fprintf( stderr, "Error: incompatible matrix/vector sizes in VM_Mult.\n" );
  }
  else {
    /* (re) initialization */
    for ( i=0; i<Matrix->Rows; i++ ) 
      Res->Vals[i] = M_INFINITY;

    for ( i=0; i<Matrix->Rows; i++ ) {
      if ( X->Vals[i] != M_INFINITY ) {
	for ( j=0; j<Matrix->Nb_Elts[i]; j++ ) {
	  k = Matrix->Elts[i][j];
	  Res->Vals[k] = max( Res->Vals[k], X->Vals[i] + Matrix->Vals[i][j] );
	}
      }
    }
  }
}

void SP_MV_Mult(SP_MaxPlus_Matrix	Matrix, 
		SP_MaxPlus_Vector 	X,
		SP_MaxPlus_Vector	Res )
{
  int		i;
  int		j;
  int		k;

  if ( ( X->Size != Matrix->Cols ) 
       || ( X->Size != Res->Size ) ) {
    fprintf( stderr, "Error: incompatible matrix/vector sizes in VM_Mult.\n" );
  }
  else {
    /* (re) initialization */
    for ( i=0; i<Matrix->Rows; i++ ) 
      Res->Vals[i] = M_INFINITY;

    for ( i=0; i<Matrix->Rows; i++ ) {
      for ( j=0; j<Matrix->Nb_Elts[i]; j++ ) {
	k = Matrix->Elts[i][j];
	if ( X->Vals[k] != M_INFINITY ) {
	  Res->Vals[i] = max( Res->Vals[i], X->Vals[k] + Matrix->Vals[i][j] );
	}
      }
    }
  }
}

/******* The "norm" of a matrix is the maximal value of its elements */
double SP_Matrix_Norm(SP_MaxPlus_Matrix A)
{
  int i;
  int j;
  double Res;

  Res = M_INFINITY;

  for ( i = 0; i < A->Rows; i++ ) {
    for ( j = 0; j < A->Nb_Elts[i]; j++ ) {
      Res = max( Res, A->Elts[i][j] );
    }
  }

  return Res;
}

double SP_Vector_Norm(SP_MaxPlus_Vector V)
{
  int i;
  double Res;

  Res = M_INFINITY;

  for ( i = 0; i < V->Size; i++ )
    Res = max( Res, V->Vals[i] );

  return Res;
}

double SP_Vector_Min(SP_MaxPlus_Vector V)
{
  int i;
  double Res;

  Res = V->Vals[0];

  for ( i = 1; i < V->Size; i++ )
    Res = min( Res, V->Vals[i] );

  return Res;
}

/* Writing it out... */
void SP_Write_Matrix(SP_MaxPlus_Matrix	Matrix,
		     FILE 		*Output,
		     int		Format,
		     char		*Name)
{
  int		i;
  int		j;
  double	Val;

  if ( (SP_MaxPlus_Matrix)NULL == Matrix ) {
    fprintf( stderr, "%s: error: cannot print null matrix.\n", TOOLNAME );
    return;
  }

  if ( Format == FORMAT_SPARSE )
    fprintf( Output, "sp\n" );
  else if ( Format == FORMAT_SCILAB ) 
    fprintf( Output, "%s=#([\n\t", Name );
  else if ( Format == FORMAT_DEBUG ) 
    fprintf( Output, "Matrix %s\n", 
	     ( Name == (char*)NULL ? "<no name>" : Name ) );

  for ( i=0; i<Matrix->Rows; i++ ) {
    if ( Format == FORMAT_DEBUG )
      fprintf( Output, "[%2d]", Matrix->Nb_Elts[i] );

    switch ( Format ) {
    case FORMAT_FULL:
      for ( j=0; j<Matrix->Cols; j++ ) {
	Val = Get_SP_Matrix_Entry( Matrix, i, j );
	if ( Val == M_INFINITY ) 
	  fprintf( Output, "   eps   ");
	else
	  fprintf( Output, " %8.4f", Val );
	if ( (j+1) % 7 == 0 )
	  fprintf( Output, "\n");
       }
      fprintf( Output, "\n");
      break;
    case FORMAT_SCILAB:
      for ( j=0; j<Matrix->Cols; j++ ) {
	Val = Get_SP_Matrix_Entry( Matrix, i, j );
	if ( Val == M_INFINITY ) 
	  fprintf( Output, "   %%0    ");
	else
	  fprintf( Output, " %8.4f", Val );
      }
      if ( i < Matrix->Rows-1 )
	fprintf( Output, ";\n\t");
      else
	fprintf( Output, "])\n");
      break;
    case FORMAT_SPARSE:
      for ( j=0; j<Matrix->Nb_Elts[i]; j++ ) {
	fprintf( Output, " %2d %2d %8.4f\n", 
		 i, Matrix->Elts[i][j], Matrix->Vals[i][j] );
      }
      break;
    case FORMAT_DEBUG:
      for ( j=0; j<Matrix->Nb_Elts[i]; j++ ) {
	fprintf( Output, " %d:%8.4f", 
		 Matrix->Elts[i][j], Matrix->Vals[i][j] );
      }
      fprintf( Output, "\n");
      break;
    }
  }

  if ( Format == FORMAT_SPARSE )
    fprintf( Output, "end\n" );
}

void SP_Write_Vector(SP_MaxPlus_Vector	V,
		     FILE 		*Output,
		     int		Format )
{
  int		i;
  double	Val;
  
  for ( i=0; i<V->Size; i++ ) {
    Val = V->Vals[i];
    if ( Val == M_INFINITY )
      fprintf( Output, "   eps   ");
    else
      fprintf( Output, " %8.4f", Val );
  }
  fprintf( Output, "\n");
}
