/* Implementation of Methods for Tetris	*/

#include <stdio.h>
#include <malloc.h>
#include <math.h>
#include "tetris.h"

#define TOOLNAME "Tetris"

extern boolean Debug;

/* Contructor and Destructor */
Tetris New_Tetris( int Nb_Columns, int Nb_Pieces, Piece_Desc *The_List )
{
  boolean	Alloc_Ok;
  int		i;
  Tetris 	Res;
  
  Res = (Tetris) malloc( sizeof( struct TETRIS ) );

  if ( (Tetris)NULL != Res ) {
    Res->Nb_Columns = Nb_Columns;
    Res->Nb_Pieces = Nb_Pieces;

    if ( ( Nb_Pieces != 0 ) && (Piece_Desc*)NULL == The_List ) {
      fprintf( stderr,
	       "%s: warning error. No piece list supplied. Zero pieces assumed.\n", TOOLNAME );
      Res->Nb_Pieces = 0;
    }

    Res->Weight = (double*) calloc( (size_t)Res->Nb_Pieces, sizeof(double) );
    Res->Color = (char**) calloc( (size_t)Res->Nb_Pieces, sizeof(char*) );

    Alloc_Ok = ( ( NULL != Res->Weight ) 
		 && ( NULL != Res->Color ) );

    if ( Res->Nb_Pieces > 0 ) {
      Res->Piece = 
	(Piece_Desc*) calloc( (size_t)Nb_Pieces, sizeof(Piece_Desc) );
      Alloc_Ok = Alloc_Ok && ( (Piece_Desc*)NULL != Res->Piece );
    }
    else {
      Res->Piece = (Piece_Desc*)NULL;
    }
 
    if ( Alloc_Ok ) {
      for ( i=0; i<Res->Nb_Pieces; i++ ) {
	Res->Weight[i] = 1.0;
	Res->Color[i] = (char*) malloc( 64*sizeof(char) );
	/* no protection for color allocation! That was a bad mistake! */
	sprintf( Res->Color[i], "#dfdfdf" );

	Res->Piece[i] = Copy_Piece( The_List[i] );
      }
    }
    else {
      fprintf(stderr,"%s: memory error. Not enough space.\n", TOOLNAME );
      Free_Tetris( Res );
      Res = (Tetris)NULL;
    }
  }

  return Res;
}

Piece_Desc New_Piece( int Nb_Columns )
{
  int		k;
  static 	Piece_Desc 	Res;
  static	boolean Alloc_Ok;

  Res = (Piece_Desc) malloc( sizeof(struct PIECE_DESC) );

  Alloc_Ok = ( (Piece_Desc)NULL != Res );

  if ( Alloc_Ok ) {
    Res->Number = -1;	/* no id */
    Res->Nb_Columns = Nb_Columns;
    Res->LoProf = (int*) calloc( (size_t)Nb_Columns, sizeof(int) );
    Res->HiProf = (int*) calloc( (size_t)Nb_Columns, sizeof(int) );

    Alloc_Ok = ( ( (int*)NULL != Res->LoProf ) 
		 && ( (int*)NULL != Res->HiProf ) );

    if ( Alloc_Ok ) {
      /* initialize piece to ... nothing. */
      for ( k=0; k<Res->Nb_Columns; k++ ) {
	Res->LoProf[k] = -1;
	Res->HiProf[k] = -1;
      }
      Res->Color = (char*)NULL;
    }
  }

  if ( !Alloc_Ok ) {
    fprintf(stderr,"%s: memory error. Not enough space.\n", TOOLNAME );
    Free_Piece( Res );
    Res = (Piece_Desc)NULL;
  }

  return Res;
}

Piece_Desc Copy_Piece( Piece_Desc Orig )
{
  static 	Piece_Desc 	Res;
  static	boolean Alloc_Ok;
  int		i;

  Res = New_Piece( Orig->Nb_Columns );

  Alloc_Ok = ( (Piece_Desc)NULL != Res );

  if ( Alloc_Ok ) {
    for ( i=0; i<Res->Nb_Columns; i++ ) {
      Res->LoProf[i] = Orig->LoProf[i];
      Res->HiProf[i] = Orig->HiProf[i];
    }
  }
  else {
    fprintf(stderr,"%s: memory error. Not enough space for copy.\n", 
	    TOOLNAME );
    Free_Piece( Res );
    Res = (Piece_Desc)NULL;
  }

  return Res;
}

/* garbage collecting procedure. Probably useless in the present state. */
#define FREE(ptr,type) if ( (type*)NULL != (ptr) ) { free( (void*)ptr ); }

void Free_Tetris( Tetris Game ) 
{
  int i;
  if ( (Tetris) NULL != Game ) {
    FREE( Game->Weight, double );
    FREE( Game->Color, char* );
    if ( (Piece_Desc*)NULL != Game->Piece ) {
      for ( i=0; i<Game->Nb_Pieces; i++ ) {
	Free_Piece( Game->Piece[i] );
      }
      free( Game->Piece );
    }
    free( Game );
  }
}

void Free_Piece( Piece_Desc P )
{
  FREE( P->LoProf, int );
  FREE( P->HiProf, int );
  free( P );
}

/************************************************************************/
/* Procedures manipulating Tetris games.				*/
/************************************************************************/
/* Normalize the weights to probabilities. In situ. */
void Normalize_Weights( Tetris Game ) 
{
  int		i;
  double	Total_Weight = 0.0;

  for ( i=0; i<Game->Nb_Pieces; i++ )
    Total_Weight += Game->Weight[i];
  for ( i=0; i<Game->Nb_Pieces; i++ )
    Game->Weight[i] /= Total_Weight;
}

/************************************************************************/
/* Procedures for the output of a Tetris game in the right format.	*/
/************************************************************************/
/* Converts a number between 0 and 255 into a color code */
void Num2Col( double Alpha, char Res[2] )
{
  int	Val;

  Val = (int)floor(256*Alpha);

  if ( Val == 256 ) {
    sprintf( Res, "ff" );
  }
  else {
    sprintf( Res, "%02x", Val );
  }
  if ( Debug ) {
    printf( "Col: %f -> %s\n", Alpha, Res );
  }
}

/* Find colors for the pieces */
void Set_Colors( Tetris The_Game ) 
{
  int		i;
  double	Alpha;
  char		red[3];
  char		green[3];
  char		blue[3];

  for ( i=0; i<The_Game->Nb_Pieces; i++ ) {
    Alpha = 6.0*i/(double)The_Game->Nb_Pieces;

    if ( Alpha < 1.0 ) {
      sprintf( red,   "ff" );
      sprintf( green, "00" );
      Num2Col( 1.0-Alpha, blue );
    } 
    else if ( Alpha < 2.0 ) {
      Alpha -= 1.0;
      sprintf( red,  "ff" );
      Num2Col( Alpha, green );
      sprintf( blue,  "00" );
    }
    else if ( Alpha < 3.0 ) {
      Alpha -= 2.0;
      Num2Col( 1.0-Alpha, red );
      sprintf( green,  "ff" );
      sprintf( blue,  "00" );
    }
    else if ( Alpha < 4.0 ) {
      Alpha -= 3.0;
      sprintf( red,  "00" );
      sprintf( green,  "ff" );
      Num2Col( Alpha, blue );
    } 
    else if ( Alpha < 5.0 ) {
      Alpha -= 4.0;
      sprintf( red,  "00" );
      Num2Col( 1.0-Alpha, green );
      sprintf( blue,  "ff" );
    } else {
      Alpha -= 5.0;
      Num2Col( Alpha, red );
      sprintf( green,  "00" );
      sprintf( blue,  "ff" );
    }

    if ( Debug ) {
      printf("Color #%d is r=%s g=%s b=%s\n", i, red, green, blue );
    }
    The_Game->Color[i] = (char*) realloc( The_Game->Color[i],
					  64*sizeof(char) );
    sprintf( The_Game->Color[i], "#%2s%2s%2s", red, green, blue );
  }
}

/* Input/Output */
#define E_TOKEN(t,m) printf("Tetris Parse Error: expecting token %s and found %s.", t, m );
#define NEXT_TOKEN do { fscanf( Input, "%s", buf ); } while ( buf[0] == '#' )

extern Tetris Read_Tetris(char *Input_Name);

#define CODE(x)		fprintf(Output,x);
#define CODE1(x,y)	fprintf(Output,x,y);
#define CODE2(x,y,z)	fprintf(Output,x,y,z);
#define CODE3(x,y,z,t)	fprintf(Output,x,y,z,t);
#define CODELN		CODE("\n")

void Write_Tetris(Tetris 	The_Game,
		  FILE* 	Output )
{
  int 		i;

  CODE1("Columns %d\n\n", The_Game->Nb_Columns );

  for ( i=0; i<The_Game->Nb_Pieces; i++ ) {
    CODE1("Piece %d\n", i+1 );
    if ( floor(The_Game->Weight[i]) == The_Game->Weight[i] ) {
      CODE1("Weight %d\n", (int)The_Game->Weight[i] );
    }
    else
      CODE1("Weight %f\n", The_Game->Weight[i] );
    CODE1("Color %s\n", The_Game->Color[i] );
    Write_Piece( The_Game->Piece[i], Output );
  }
}

void	Write_Piece(Piece_Desc	The_Piece,
		    FILE	*Output )
{
  int		k;

  CODE("LoProf ");
  for ( k=0; k<The_Piece->Nb_Columns; k++ ) {
    /* using "sparse" convention */
    if ( The_Piece->LoProf[k] != -1 ) {
      CODE2(" %d %d", k, The_Piece->LoProf[k] );
    }
    /*
      else {
      CODE(" %d eps", k );
      }
    */
  }
  CODELN;
  CODE("HiProf ");
  for ( k=0; k<The_Piece->Nb_Columns; k++ ) {
    /* using "sparse" convention */
    if ( The_Piece->HiProf[k] != -1 ) {
      CODE2(" %d %d", k, The_Piece->HiProf[k] );
    }
    /*
      else {
      CODE(" %d eps", k );
      }
    */
  }
  CODELN;
  CODELN;
}
