/* A program for simulating a trajectory of some Markov chain */

#define MAX_RANDOM_VALUE        0x7FFFFFFF
#define MAXREAL			1e38

#include <math.h>
#include "mc.h"

#define TOOLNAME	"mc_sim"

/* Declaration of Debug, normally in the parser... */
int Debug = FALSE;

/* External declarations */
extern Markov_Chain New_Markov_Chain(int,int);

/* Forward declaration */
Markov_Chain Read_Chain( char *File_Name );
void	     Sim_Chain( Markov_Chain, double, boolean, boolean, boolean );
void Sim_Cont_Time_Chain( Markov_Chain Chain, double T_Max,
			  boolean Stats, boolean Traj,
			  boolean Increments );
void Sim_Disc_Time_Chain( Markov_Chain Chain, int T_Max,
			  boolean Stats, boolean Traj );

/*======================================================*/
/* Statistiques                                         */
/*======================================================*/
double U_0_1( void )
{
  return (double)random()/MAX_RANDOM_VALUE;
}

double Exponential( double Rate )
{
  return ( Rate == 0.0 ? MAXREAL : -log(U_0_1())/Rate );
}


main(
int	argc,
char	*argv[] )
{
  static char		File_Name[128] = "";
  int			i;
  boolean		Stats = FALSE;
  boolean		Traj = TRUE;
  boolean		Increments = FALSE;
  boolean		Help = FALSE;
  boolean		Args_Ok = TRUE;
  boolean		File_Ok = TRUE;
  Markov_Chain		Chain;
  double		T_Max = 100;

  i = 1;

  if (argc>1)
    do {

      if GET_STRING( "-f", File_Name )

      else if GET_REAL( "-tmax", T_Max )

      else if SWITCH( "-stats", Stats, TRUE )

      else if SWITCH( "-no_traj", Traj, FALSE )

      else if SWITCH( "-increments", Increments, TRUE )

      else if SWITCH( "-debug", Debug, TRUE )

      else if IS_VERSION

      else if ( ( 0 == strcmp( "-h", argv[i] ) )
		|| ( 0 == strcmp( "-?", argv[i] ) ) ) {
	printf("Usage: %s [-f <file>] [-tmax <real>] [-stats] ", TOOLNAME );
	printf("[-no_traj]\n\t[-increments] [-?] [-h] \n" );
	Args_Ok = FALSE;
	Help = TRUE;
	i++;
      }
      else
	i++; /* skip unknown argument */
    }
    while ( Args_Ok && ( i < argc ) );


  if ( Args_Ok ) {

    Chain = Read_Chain( File_Name );

    if ( Chain != NULL ) {

      if ( Debug ) 
	Write_Matrix( stdout, Chain->Generator, /* full? */ FALSE );

      Sim_Chain( Chain, T_Max, Stats, Traj, Increments );

      return 0;
    }
    else {
      fprintf( stderr, "%s: Error reading Markov chain.\n", 
	       TOOLNAME );
      return -1;
    }
  }
  else if (!Help)
    return -1;
  else
    return 0;
}


void Sim_Disc_Time_Chain( Markov_Chain Chain, int T_Max,
			  boolean Stats, boolean Traj )
{
  int		Cur_State;
  int		Time;
  int		j;
  int		Size;
  double	U;
  boolean	Reward;
  int		*Cum_Time;
  double	Tot_Reward = 0;

  Size = Chain->State_Space_Size;
  Cur_State = Chain->Init_State;
  
  if ( Stats ) {
    Cum_Time = (int*) calloc( (size_t)Size, sizeof( int ) );
  }
  /* is this a Markov Reward Process? */
  Reward = ( (SP_Real_Matrix)NULL != Chain->Reward );

  for ( Time=0; Time < T_Max; Time++ ) {
    if ( Traj )
      printf("%6d %4d\n", Time, Cur_State );
    if ( Stats )
      Cum_Time[ Cur_State ]++;

    U = U_0_1();
    for ( j=0; ( j < Chain->Generator->Nb_Elts[ Cur_State ] ) 
	    && ( 0 < ( U -= Chain->Generator->Vals[ Cur_State ][j] ) );
	  j++ );
    Cur_State = Chain->Generator->Elts[ Cur_State ][j];
    if ( Reward ) {
      Tot_Reward += Chain->Reward->Vals[ Cur_State ][j];
    }
  }
  /* adjust end of simulation */
  if ( Traj )
    printf("%6d %4d\n", T_Max, Cur_State );

  if ( Stats && ( T_Max > 0 ) ) {
    for ( j=0; j<Size; j++ ) {
      printf("State %3d:   cum time = %5d   ( %8.4f%%)\n",
	     j, Cum_Time[j], 100.0*(double)Cum_Time[j]/(double)T_Max );
    }
    if ( Reward ) {
      printf("Average reward: %f\n", Tot_Reward/(double)T_Max );
    }
    free( Cum_Time );
  }
}

void Sim_Cont_Time_Chain( Markov_Chain Chain, double T_Max,
			  boolean Stats, boolean Traj,
			  boolean Increments )
{
  int		Cur_State;
  double	Time;
  int		j;
  int		Size;
  double	U;
  double	Nu;
  double	Adv;
  double	*Cum_Time;

  Size = Chain->State_Space_Size;
  Cur_State = Chain->Init_State;
  
  if ( Stats ) {
    Cum_Time = (double*) calloc( (size_t)Size, sizeof( double ) );
  }

  Time = 0.0;
  do {

    Nu = Get_Sparse_Matrix_Entry( Chain->Generator, Cur_State, Cur_State );
    Adv = Exponential( -Nu );
    U = - U_0_1() * Nu;
    for ( j=0; ( j < Chain->Generator->Nb_Elts[Cur_State] ) && ( 0 <= U ); j++ ) 
      if ( Chain->Generator->Elts[Cur_State][j] != Cur_State )
	U -= Chain->Generator->Vals[ Cur_State ][j];
    if ( Stats )
      Cum_Time[ Cur_State ] += ( Time+Adv >= T_Max ? T_Max - Time : Adv );

    if ( Traj ) {
      printf("%12.6f %4d", Time, Cur_State );
      if ( Increments )
	printf(" %10.6f", Adv );

      printf("\n");
    }

    Time += Adv;
    Cur_State = Chain->Generator->Elts[Cur_State][j-1];
  } while ( Time < T_Max );
  
  /* adjust end of simulation */
  if ( Traj ) {
    printf("%12.6f %4d", T_Max, Cur_State );
    if ( Increments )
      printf(" %10.6f", Adv );

    printf("\n");
  }

  if ( Stats ) {
    for ( j=0; j<Size; j++ ) {
      printf("State %3d:   cum time = %5f   ( %8.4f%% )\n",
	     j, Cum_Time[j], 100.0*Cum_Time[j]/T_Max );
    }
    free( Cum_Time );
  }

}

void Sim_Chain( Markov_Chain Chain, double T_Max,
		boolean Stats, boolean Traj, boolean Increments )
{
  switch( Chain->Time_Type ) {
  case DISCRETE:
    Sim_Disc_Time_Chain( Chain, (int) rint(T_Max), Stats, Traj );
    break;

  case CONTINUOUS:
    Sim_Cont_Time_Chain( Chain, T_Max, Stats, Traj, Increments );
    break;

  default:
    printf("%s: error: unknown Chain type (%d). No simulation.\n",
	   TOOLNAME, Chain->Time_Type );
  }
}

