/* A program for simulating a trajectory of some Markov chain */

#include <math.h>
#include "mc.h"

#define TOOLNAME	"mc_solve"

/* Declaration of Debug, normally in the parser... */
int Debug = FALSE;

/* Forward declaration */
Markov_Chain Read_Chain( char *File_Name );
void	     Solve_Chain( Markov_Chain Chain, boolean Reward, boolean Progress );
void Solve_Cont_Time_Chain( Markov_Chain Chain, boolean Reward, boolean Progress );
void Solve_Disc_Time_Chain( Markov_Chain Chain, boolean Reward, boolean Progress );

int main(
int	argc,
char	*argv[] )
{
  static char		File_Name[128] = "";
  int			i;
  boolean		Progress = FALSE;
  boolean		Reward = FALSE;
  boolean		Help = FALSE;
  boolean		Args_Ok = TRUE;
  boolean		File_Ok = TRUE;
  Markov_Chain		Chain;

  i = 1;

  if (argc>1)
    do {

      if GET_STRING( "-f", File_Name )

      else if SWITCH( "-progress", Progress, TRUE )
      else if SWITCH( "-reward", Reward, TRUE )

      else if SWITCH( "-debug", Debug, TRUE )

      else if IS_VERSION

      else if ( ( 0 == strcmp( "-h", argv[i] ) )
		|| ( 0 == strcmp( "-?", argv[i] ) ) ) {
	printf("Usage: %s [-f <file>] [-reward] [-progress]", TOOLNAME );
	printf(" [-?] [-h] \n" );
	Args_Ok = FALSE;
	Help = TRUE;
	i++;
      }
      else
	i++; /* skip unknown argument */
    }
    while ( Args_Ok && ( i < argc ) );


  if ( Args_Ok ) {

    Chain = Read_Chain( File_Name );

    if ( Chain != (Markov_Chain)NULL ) {

      if ( Debug ) 
	Write_Matrix( stdout, Chain->Generator, /* full? */ FALSE );

      Solve_Chain( Chain, Reward, Progress );

      return 0;
    }
    else {
      fprintf( stderr, "%s: Error reading Markov chain.\n", 
	       TOOLNAME );
      return -1;
    }
  }
  else if (!Help)
    return -1;
  else
    return 0;
}

double Dist( double *v1, double *v2, int Size )
{
  static int	i;
  static double Res;

  Res = 0.0;
  for ( i=0; i<Size; i++ ) {
    Res += fabs( v1[i] - v2[i] );
  }
  return Res;
}

void Solve_Disc_Time_Chain( Markov_Chain Chain, 
			    boolean Reward,
			    boolean Progress )
{
  int		i;
  int		j;
  int		Size;
  int		Time;
  boolean	Continue;
  double	*Pi;
  double	*Res;
  double	*Tmp;
  double	err;
  double	Expect;
  int		T_Max = 1000;

  Size = Chain->State_Space_Size;

  Pi = (double*) calloc( (size_t)Size, sizeof( double ) );
  Res = (double*) calloc( (size_t)Size, sizeof( double ) );

  /* initialize Pi */
  Pi[0] = 1.0;

  Continue = TRUE;
  for ( Time=0; (Time<T_Max) && Continue; Time++ ) {
    VM_Mult( Pi, Chain->Generator, Res );
    
    if ( Debug ) {
      printf("# Pi(%d) = ", Time );
      for ( i=0; i<Size; i++ ) {
	printf("%8.6f ", Res[i] );
      }
      printf("\n");
    }

    err = Dist( Pi, Res, Size );
    if ( Progress ) {
      fprintf( stderr, "Iteration #%3d: distance %10e\n", Time, err );
      fflush( stderr );
    }
    if ( err < 1e-7 ) {
      Continue = FALSE;
    } 
    else {
      Tmp = Res; Res = Pi; Pi = Tmp;
    }
  }
  if ( Time >= T_Max ) {
    fprintf( stderr, "Warning: Maximum number of iterations reached. ");
    fprintf( stderr, "Distribution may not be correct.\n");
  }

  if ( Reward ) {
    /* compute expected reward */
    Expect = 0.0;
    for ( i=0; i<Size; i++ ) {
      for ( j=0; j<Chain->Generator->Nb_Elts[i]; j++ ) {
	Expect += Pi[i] * Chain->Generator->Vals[i][j]
	  * Chain->Reward->Vals[i][j];
      }
    }
    printf("# Expected reward\n");
    printf("%f\n", Expect );
  }
  else {
    /* write stationary distribution out */
    printf("# Stationary distribution\n");
    for ( i=0; i<Size; i++ ) {
      printf("%3d %8.6f\n", i, Pi[i] );
    }
  }

  /* cleanup */
  free( Pi );
  free( Res );

}

void Solve_Cont_Time_Chain( Markov_Chain Chain, 
			    boolean Reward,
			    boolean Progress )
{
  int		i;
  int		j;
  int		k;
  int		Size;
  double	Lambda;
  double	Nu;

  Size = Chain->State_Space_Size;

  /* uniformize matrix */
  Nu = 0.0;
  for ( i=0; i<Size; i++ ) {
    Lambda = -Get_Sparse_Matrix_Entry( Chain->Generator, i, i );
    Nu = ( Lambda > Nu ? Lambda : Nu );
  }
  if ( Debug ) {
    printf("Uniformizing with value nu = %f\n", Nu );
  }
  for ( i=0; i<Size; i++ ) {
    for ( j=0; j<Chain->Generator->Nb_Elts[i]; j++ ) {
      k = Chain->Generator->Elts[i][j];
      if ( k != i )
	Chain->Generator->Vals[i][j] /= Nu;
      else
	Chain->Generator->Vals[i][j] = 1.0 + Chain->Generator->Vals[i][j] / Nu;
    }
  }

  Solve_Disc_Time_Chain( Chain, Reward, Progress );
}

void Solve_Chain( Markov_Chain Chain,
		  boolean Reward,
		  boolean Progress )
{
  switch( Chain->Time_Type ) {
  case DISCRETE:
    Solve_Disc_Time_Chain( Chain, Reward, Progress );
    break;

  case CONTINUOUS:
    Solve_Cont_Time_Chain( Chain, Reward, Progress );
    break;

  default:
    printf("%s: error: unknown Chain type (%d). No solution.\n",
	   TOOLNAME, Chain->Time_Type );
  }
}

