/* Procedure to read a Markov chain from a file. Serves as "parser"	*/
/* for the time being.							*/

#include <stdio.h>
#include <math.h>
#include "mc.h"

#define TOOLNAME	"Parse_MC"
#define ABORT(x) { printf("%s: error: %s.\n", TOOLNAME, x ); if ( (Markov_Chain)NULL != Res ) { free( Res ); } return (Markov_Chain)NULL; }

/* Constructors */

Markov_Chain New_Markov_Chain(int	Size,
			      int	Time_Type)
{
  Markov_Chain	Res;

  Res = (Markov_Chain) malloc( sizeof( struct MARKOV_CHAIN ) );

  if ( (Markov_Chain)NULL != Res ) {
    Res->State_Space_Size = Size;
    Res->Time_Type = Time_Type;
    Res->Init_State = 0;		/* arbitrary default value */

    Res->Generator = New_SP_Real_Matrix( Size );
    Res->Reward = New_SP_Real_Matrix( Size );

    if ( (SP_Real_Matrix)NULL == Res->Generator ) {
      /* abandon... */
      free( Res );
      Res = (Markov_Chain)NULL;
    }
  }

  return Res;
}

/* "parsing" procedure */

Markov_Chain Read_Chain( char *File_Name )
{
  int		i;
  int		j;
  FILE		*Input;
  Markov_Chain	Res = (Markov_Chain)NULL;
  int		Time_Type;
  int		Size;
  int		Matrix_Type;
  int		Row;
  int		Col;
  boolean	Reward = FALSE;
  boolean	Read_Ok = TRUE;
  boolean	Alloc_Ok = TRUE;
  char		Message[128];
  double	Epsilon = 1e-6;
  double	Sum;
  double	Value;
  char		Token[128];

  if ( strlen( File_Name ) == 0 )
    Input = stdin;
  else if ( NULL == ( Input = fopen( File_Name, "r" ) ) ) {
    sprintf( Message, "could not open input file %s", File_Name );
    ABORT( Message );
  }

  /* determine chain type */
  fscanf( Input, "%128s", Token );
  if ( ( !strcmp( Token, "discrete" ) )
       || ( !strcmp( Token, "1" ) ) ) {
    Time_Type = DISCRETE;
  } 
  else if ( ( !strcmp( Token, "continuous" ) )
       || ( !strcmp( Token, "2" ) ) ) {
    Time_Type = CONTINUOUS;
  }
  else {
    ABORT("could not read chain type");
  }

  if ( Debug ) 
    printf("Found time type = %d\n", Time_Type );

  /* find out if reward or not */
  fscanf( Input, "%128s", Token );
  if ( !strcmp( Token, "reward" ) ) {
    Reward = TRUE;
    /* read next token */
    fscanf( Input, "%128s", Token );
  }

  /* determine size, or sparse indicator */
  if ( !strcmp( Token, "sparse" ) ) {
    Matrix_Type = SPARSE;
    Read_Ok = ( 1 == fscanf( Input, "%d", &(Size) ) );
  }
  else {
    Matrix_Type = FULL;
    Read_Ok = ( 1 == sscanf( Token, "%d", &(Size) ) );
  }
  
  if ( !Read_Ok )
    ABORT("could not read chain size");

  if ( Reward && ( Matrix_Type == FULL ) )
    ABORT("full matrices for reward MC not supported");

  Res = New_Markov_Chain( Size, Time_Type );

  Alloc_Ok = ( NULL != Res );

  if ( !Alloc_Ok )
    ABORT("could not allocate memory for matrix");

  if ( Matrix_Type == FULL ) {
    for ( i=0; Read_Ok && ( i < Res->State_Space_Size ); i++ ) {
      for ( j=0; Read_Ok && ( j < Res->State_Space_Size ); j++ ) {

	if ( ( Res->Time_Type == DISCRETE )
	     || ( ( Res->Time_Type == CONTINUOUS ) && ( i != j ) ) ) {
	  
	  Read_Ok = Read_Full_Matrix_Entry( Input, Res->Generator, i, j );
	
	  if ( !Read_Ok ) {
	    sprintf( Message, "failed reading matrix element (%d,%d)", i, j );
	    ABORT( Message );
	  } else if (Debug ) {
	    printf("Got entry (%d,%d) = %f\n",
		   i, j, Get_Sparse_Matrix_Entry( Res->Generator, i, j ) );
	  }
	}
      }
    }
  }
  else if ( Matrix_Type == SPARSE ) { 
    do {
      /* have to copy the code of "Read_Sparse..." because I need the
	 values read. Read_Sparse_... does not say what it has read 8-( */
      if ( 3 == fscanf( Input, "%d %d %lf", &Row, &Col, &Value ) ) {
	Set_Sparse_Matrix_Entry( Res->Generator, Row, Col, Value );
      } 
      else
	Read_Ok = FALSE;
	
      if ( Read_Ok && Reward ) {
	/* get "wr" token, then read reward */
	fscanf( Input, "%s", Token );
	if ( strcmp( Token, "wr" ) ) {
	  fprintf( stderr, "%s: warning: expecting \"wr\" before reward but got %s.\n",
		   TOOLNAME, Token );
	  /* fprintf( stderr, ", line %d.\n", Line_Number ); */
	  /* Reading reward from this token... */
	  if ( 1 != sscanf( Token, "%lf", &Value ) ) {
	    sprintf( Message, "%s: error: unexpected token %s",
		     TOOLNAME, Token );
	    ABORT( Message );
	  }
	}
	else if ( 1 != fscanf( Input, "%lf", &Value ) ) {
	    sprintf( Message, "%s: error: could not read reward", TOOLNAME );
	    ABORT( Message );
	}

	Set_Sparse_Matrix_Entry( Res->Reward, Row, Col, Value );
      }
    } while ( Read_Ok );
    /* skip stopping mark */
    fscanf( Input, "%s", Token );
  }
    
  Read_Ok = ( 1 == fscanf( Input, "%d", &(Res->Init_State) ) );
  if ( !Read_Ok )
    ABORT("could not read initial state");
    
  /* now checking/adjusting the generator */
  for ( i=0; Read_Ok && ( i < Res->State_Space_Size ); i++ ) {
    Sum = Row_Sum( Res->Generator, i );
    
    switch( Res->Time_Type ) {
    case DISCRETE:
      if ( fabs( Sum - 1.0 ) > Epsilon ) {
	printf("%s: warning: row %d does not sum up to 1.0. Dif = %f.\n",
	       TOOLNAME, i, Sum - 1.0 );
      }
      break;
    case CONTINUOUS:
      Set_Sparse_Matrix_Entry( Res->Generator, i, i, -Sum );
    }
  }

  return Res;

}


void Write_Markov_Chain( FILE *Out, Markov_Chain MC )
{
  int		i;
  int		j;
  double	Val;
  SP_Real_Matrix Matrix;

  Matrix = MC->Generator;

  fprintf( Out, "discrete sparse\n" );
  fprintf( Out, "%d\n", Matrix->Size );

  for ( i=0; i<Matrix->Size; i++ ) {
    for ( j=0; j<Matrix->Nb_Elts[i]; j++ ) {
      fprintf( Out, "%4d %4d %8.4f", 
	       i, Matrix->Elts[i][j], Matrix->Vals[i][j] );
      if ( (SP_Real_Matrix)NULL != MC->Reward ) {
	fprintf( Out, " wr %f\n", MC->Reward->Vals[i][j] );
      }
    }
  }

  fprintf( Out, "stop\n" );
}
