/* Computation of a stochastic lower bound for the actual workload	*/
/* distribution.							*/

#include <stdio.h>
#include <malloc.h>
#include <values.h>
#include <math.h>
#include <strings.h>
#include "boolean.h"
#include "parseargs.h"
#include "law.h"

#define	TOOLNAME	"gi_gi_1"
#define VERSION 	"v1.0"

/************************************************************************/
/* Structure pour faciliter le passage de parametres dans le calcul	*/
/* de fonctions...							*/
/************************************************************************/
typedef struct {
  double	Start;
  double	Stop;
  double	Step;
  boolean	Queue;
  boolean	Log;
  boolean	Log10;
} RANGE_INFO;
#define DEFAULT_RANGE { 0.0, 10.0, 0.5, FALSE, FALSE, FALSE }

/* Structure useful for avoiding some global variables */
typedef struct DISTRIB {
  double	*Proba;
  double	*CDF;
} DISTRIB;

/* Global variables */
boolean	Debug = FALSE;
boolean	ShowDists = FALSE;
boolean	NewLine = TRUE;

/* Parsing procedure */
int Parse_Law_From_Args( char **argv, int index, int argc, Law_Desc The_Law );
/* Procedures for analysis */
double	Load_Factor( Law_Desc, Law_Desc );
DISTRIB	Prepare_Distribs( Law_Desc, Law_Desc, RANGE_INFO, RANGE_INFO );
double  *Compute_Distrib( Law_Desc, Law_Desc, RANGE_INFO, RANGE_INFO, int );
int	Nb_Pts(RANGE_INFO);
void	Print_Range(RANGE_INFO, int, RANGE_INFO, double* );
double	Compute_Moment( double *Dist, RANGE_INFO Range, int Order, boolean );

int main(int argc, char *argv[])
{
  static char		Result_File_Name[256] = "";
  int			i;
  int			Cust_Nb = -2;
  boolean		Help = FALSE;
  boolean		Args_Ok = TRUE;
  boolean		File_Ok = TRUE;
  boolean		Value = FALSE;
  boolean		Moment = FALSE;
  RANGE_INFO		S_Range = DEFAULT_RANGE; /* Range for computing S/U */
  RANGE_INFO		W_Range = DEFAULT_RANGE; /* Range for computing W */
  RANGE_INFO		P_Range = DEFAULT_RANGE; /* Range for printing */
  double		x_Val;
  double		Rho;
  double		*Dist;
  int			Order = 1;	/* for the moments	*/
  Law_Desc		Arrival_Law;
  Law_Desc		Service_Law;

  /* Various initializations before reading arguments */
  Arrival_Law = New_Law();
  Service_Law = New_Law();
  Service_Law->Name = 'E';
  Arrival_Law->Name = 'E';
  Service_Law->Parameters[0] = 1.0;
  Arrival_Law->Parameters[0] = 2.0;

  i = 1;

  if (argc>1) {
    do {
      if GET_STRING( "-o", Result_File_Name )

      else if GET_INT( "-customer", Cust_Nb )
      else if GET_INT_SWITCH( "-moment", Order, Moment, TRUE )

      else if GET_REAL( "-xfrom", P_Range.Start )
      else if GET_REAL( "-xto", P_Range.Stop )
      else if GET_REAL( "-xstep", P_Range.Step )
      else if GET_REAL_SWITCH( "-xval", x_Val, Value, TRUE )
      else if GET_REAL( "-Smax", S_Range.Stop )
      else if GET_REAL( "-Wmax", W_Range.Stop )
      else if GET_REAL( "-dt", W_Range.Step )
			       
      else if SWITCH( "-queue", P_Range.Queue, TRUE )
      else if SWITCH( "-log", P_Range.Log, TRUE )
      else if SWITCH( "-log10", P_Range.Log10, TRUE )
      else if SWITCH( "-nonewline", NewLine, FALSE )
      else if SWITCH( "-debug", Debug, TRUE )
      else if SWITCH( "-showdists", ShowDists, TRUE )
      else if IS_VERSION		     

      else if ( 0 == strcmp( "-arrival_law", argv[i] ) ) {
	i++;
	i = Parse_Law_From_Args( argv, i, argc, Arrival_Law );
      }
      else if ( 0 == strcmp( "-service_law", argv[i] ) ) {
	i++;
	i = Parse_Law_From_Args( argv, i, argc, Service_Law );
      }
      else if ( ( 0 == strcmp( "-h", argv[i] ) )
		|| ( 0 == strcmp( "-?", argv[i] ) ) ) {
	printf("Usage: %s [-o <file>]\n\t", TOOLNAME );
	printf("[-arrival_law <law>] [-service_law <law>] ");
	printf("[-customer <int>]\n\t");
	printf("[-xfrom <real>] [-xto <real>] [-xval <real>] ");
	printf("[-xstep <real>]\n\t");
	printf("[-Smax <real>] [-Wmax <real>] [-dt <real>]\n\t");
	printf("[-queue] [-log] [-nonewline] [-moment <int>] ");
	printf("[-debug] [-showdists] [-h] [-?]\n");
	Args_Ok = FALSE;
	Help = TRUE;
	i++;
      }
      else
	i++; /* skip unknown argument */
    } while ( Args_Ok && ( i < argc ) );
  }

  if (Args_Ok) {

    /* Making sure discretizations still coincide */
    S_Range.Step = W_Range.Step;

    if ( Value ) {
      P_Range.Start = x_Val;
      P_Range.Stop = x_Val;
    }

    if ( strlen(Result_File_Name) != 0 )
      File_Ok = ( NULL != freopen( Result_File_Name, "w", stdout ) );

    if ( !File_Ok ) {
      printf("%s: error: could not open file %s\n", 
	     TOOLNAME, Result_File_Name );
      return -1;
    }
    else { 

      Rho = Load_Factor( Arrival_Law, Service_Law );

      if ( Rho < 1.0 ) {

	Dist = Compute_Distrib( Arrival_Law, Service_Law, 
				W_Range, S_Range,
				Cust_Nb );

	if ( (double*)NULL != Dist ) {

	  if ( !Moment ) {
	    printf("# Lower bound on the waiting time distribution ");
	    printf("in the GI/GI/1\n");
	    printf("# Arrivals: ");
	    Write_Law( stdout, Arrival_Law, 0 );
	    printf("\n");
	    printf("# Services: ");
	    Write_Law( stdout, Service_Law, 0 );
	    printf("\n");
	    printf("# Load Factor Rho = %f\n", Rho );
	    if ( Cust_Nb > 0 )
	      printf("# Transient: %d-th customer.\n", Cust_Nb );
	    else
	      printf("# Stationary version.\n");
	    printf("# Printing Range: [%6.3f,%8.3f]\n", 
		   P_Range.Start, P_Range.Stop );
	    printf("# EW = %f\n", 
		   Compute_Moment( Dist, W_Range, 1, /* CDF? */ TRUE ) );

	    Print_Range( P_Range, Cust_Nb, W_Range, Dist );
	  }
	  else {
	    if ( Debug ) {
	      printf("# Moment of order %d of the stationary waiting time ", 
		     Order );
	      printf("in the GI/GI/1\n");
	      printf("#\n");
	    }
	    printf("%f\n", 
		   Compute_Moment( Dist, W_Range, Order, /* CDF? */ TRUE ) );
	  }
	  return 0;
	}
	else {
	  printf("%s: error computing W distribution.\n",
		 TOOLNAME );
	  return -4;
	}
      }
      else {
	printf("%s: Unstable queue (rho=%8.4f). No stationary distribution.\n",
	       TOOLNAME, Rho );
	return -2;
      }
    }
  }

  else if (Help)
    return 0;
  else
    return 1;
}


/************************************************************************/
/* Computation of Discretized laws for Sigma, Tau and U			*/
/* If Cumulative is true, then the result is P( X >= k )		*/
/************************************************************************/
double	*Discretize_Law( Law_Desc 	The_Law, 
			 RANGE_INFO	Range,
			 boolean	Upwards,
			 boolean	Cumulative ) 
{
  int		i;
  int		Nb_Points;
  double	x;
  double	v;
  double	v0;
  double	v1;
  double	delta;
  double	*Res;

  if ( Range.Step == 0.0 ) {
    fprintf( stderr, "%s: error: zero discretization step.\n", 
	     TOOLNAME);
    return (double*) NULL;
  }
  
  Nb_Points = Nb_Pts( Range );

  Res = (double*) calloc( (size_t)Nb_Points, sizeof(double) );
  if ( (double*)NULL != Res ) {

    /* Two separate loops for Cumulative / Non Cumulative to sort out */
    /* the mess */

    if ( Cumulative ) {
      /* initialization */
      if ( Upwards )
	Res[0] = CDF( The_Law, Range.Start - Range.Step );

      for ( i=0; i<Nb_Points-1; i++ ) {
	x = Range.Start + i*Range.Step;
	v = CDF( The_Law, x );
	if ( Upwards )
	  Res[i+1] = v;
	else
	  Res[i] = v;
      }
      /* Final values. */
      if ( !Upwards )
	Res[Nb_Points-1] = CDF( The_Law, Range.Stop );
    }
    else {
      /* Not cumulative */

      /* initialization */
      v0 = CDF( The_Law, Range.Start - Range.Step );

      for ( i=0; i<Nb_Points; i++ ) {
	x = Range.Start + i*Range.Step;
	v1 = CDF( The_Law, x );
	delta = v0 - v1;
	v0 = v1;
	/* Delta contains P( i-1 < x <= i ) */
	if ( Upwards )
	  Res[i] = delta;
	else if ( i >= 1 )
	  Res[i-1] = delta;
      }

      /* Final values. */
      if ( !Upwards ) {
	Res[Nb_Points-1] = v1;
      }
    }

    /* trace the values */
    if ( ShowDists ) {
      printf("%swards ", ( Upwards ? "Up" : "Down" ) );
      printf("Discretization of law ");
      Write_Law( stdout, The_Law, 0 );
      if ( Cumulative ) 
	printf(" (cdf)");
      printf("\n");
      for ( i=0; i<Nb_Points; i++ ) {
	printf("%f %f\n", Range.Start + i*Range.Step, Res[i] );
      }
    }
  }

  return Res;

}

DISTRIB Prepare_Distribs( Law_Desc 	Arrival_Law,
			  Law_Desc 	Service_Law,
			  RANGE_INFO 	W_Range,
			  RANGE_INFO	S_Range )
{
  int		K;
  int		N;
  int		u;
  int		s;
  double	Res;
  double	*Sigma;
  double	*Tau;
  double	*Tau_Cum;
  double	*U;
  double	*U_Cum;
  RANGE_INFO	T_Range;	/* Range for Tau */
  DISTRIB	D_Res;

  /* Sigma ranges between 0 and N 
     W     ranges between 0 and K
     U     ranges between -K and K
     Tau   ranges between 0 and N+K+1 (needed when computing U_Cum */
  T_Range = S_Range;
  T_Range.Stop += W_Range.Stop + W_Range.Step;

  N = (int) S_Range.Stop / S_Range.Step;
  K = (int) W_Range.Stop / W_Range.Step;

  Sigma = Discretize_Law( Service_Law, S_Range, 
			  /* Upwards */ FALSE, /* CDF */ FALSE );
  Tau   = Discretize_Law( Arrival_Law, T_Range, 
			  /* Upwards */ TRUE, /* CDF */ FALSE );
  Tau_Cum = Discretize_Law( Arrival_Law, T_Range, 
			  /* Upwards */ TRUE, /* CDF */ TRUE );

  U = (double*) calloc( (size_t)(2*K+1), sizeof(double) );
  U_Cum = (double*) calloc( (size_t)(2*K+1), sizeof(double) );
  
  for ( u=-K; u <= K; u++ ) {
    /* Calculating P[ U = u ] */
    Res = 0.0;
    s = ( u > 0 ? u : 0 );
    for ( ; s <= N; s++ ) {
      Res += Sigma[s] * Tau[s-u];
    }
    U[u+K] = Res;
    /* Calculating P[ U >= u ] */
    Res = 0.0;
    for ( s = 0; ( s <= u ) && ( s <= N ); s++ )
      Res += Sigma[s];
    for ( ; s <= N; s++ ) {
      Res += Sigma[s] * Tau_Cum[s-u+1];
    }
    U_Cum[u+K] = 1.0 - Res;
  }

  if ( ShowDists ) {
    printf("Distribution of U (cdf)\n");
    for ( u=-K; u <= K; u++ ) {
      printf("%f %f %f\n", u*W_Range.Step, U[K+u], U_Cum[K+u] );
    }
  }

  /* A little cleanup before asking for more memory */
  free( Sigma );
  free( Tau );
  free( Tau_Cum );

  D_Res.Proba = U;
  D_Res.CDF = U_Cum;

  return D_Res;
}

double *Compute_Distrib ( Law_Desc 	Arrival_Law,
			  Law_Desc 	Service_Law,
			  RANGE_INFO 	W_Range,
			  RANGE_INFO	S_Range,
			  int		Cust_Nb )
{
  int		Iteration;
  int		N;
  int		K;
  int		k;
  int		u;
  boolean	Continue;
  double	EW;
  double	New_EW;
  double 	Res;
  DISTRIB	U_Dist;
  double	*U;
  double	*U_Cum;
  double	*W_Dist;
  double	*New_W_Dist;
  double	*Tmp;		/* for swapping arrays */

  U_Dist = Prepare_Distribs( Arrival_Law, Service_Law, W_Range, S_Range );

  U = U_Dist.Proba;
  U_Cum = U_Dist.CDF;

  N = (int) S_Range.Stop / S_Range.Step;
  K = (int) W_Range.Stop / W_Range.Step;

  /* Initialization of the distribution of W to that of  min( max( U,0), K ) */
  W_Dist = (double*) calloc( (size_t)(1+K), sizeof(double) );
  New_W_Dist = (double*) calloc( (size_t)(1+K), sizeof(double) );
  W_Dist[0] = 1.0;
  for ( u=1; u <= K; u++ ) {
    if ( u <= N )
      W_Dist[u] = U_Cum[K+u];
    else 
      W_Dist[u] = 0.0;
  }

  /* Starting iterations on W */
  EW = Compute_Moment( W_Dist, W_Range, 1, /* CDF? */ TRUE );
  Iteration = 1;
  Continue = ( Cust_Nb != 1 );
  if ( Debug ) {
    printf("# Iteration %4d:  EW = %f\n", Iteration, EW );
  }
  if ( ShowDists ) {
    printf("Distribution of W (cdf)\n");
    for ( u=0; u <= K; u++ ) {
      printf("%f %f\n", u*W_Range.Step, W_Dist[u] );
    }
  }
  
  do {
    Iteration++;

    New_W_Dist[0] = 1.0;

    for ( k=1; k <= K; k++ ) {
      Res = U_Cum[K+k];
      for ( u=1; u <= K; u++ ) {
	Res += W_Dist[u] * U[K+k-u];
      }
      New_W_Dist[k] = Res;
    }

    /* swapping old and new */
    Tmp = W_Dist; W_Dist = New_W_Dist; New_W_Dist = Tmp;

    /* Determining if it's over... */
    if ( Cust_Nb > 0 )
      Continue = ( Iteration < Cust_Nb );
    else {
      /* performing convergence test */
      New_EW = Compute_Moment( W_Dist, W_Range, 1, TRUE );
      Continue = !( fabs( New_EW - EW ) < 1e-8 );
      EW = New_EW;

      if ( Debug ) {
	printf("# Iteration %4d:  EW = %f\n", Iteration, EW );
      }
    }

    if ( ShowDists ) {
      printf("Distribution of W (cdf)\n");
      for ( u=0; u <= K; u++ ) {
	printf("%f %f\n", u*W_Range.Step, W_Dist[u] );
      }
    }
  } while ( Continue );

  return W_Dist;
}

int Nb_Pts( RANGE_INFO Range )
{
  if ( Range.Step < 0.0 )
    return 0;
  else if ( Range.Stop == Range.Start )
    return 1;
  else if ( Range.Step == 0.0 )
    return 0;
  else {
    return 1 + (int)floor( ( Range.Stop - Range.Start ) / Range.Step );
  }
}

/* procedure pour imprimer les valeurs de la distribution */
void	Print_Range( RANGE_INFO	P_Range,
		     int	Cust_Nb,
		     RANGE_INFO W_Range,
		     double	*Distrib )
{
  int		p;
  int		Nb_Points;
  int		k;
  double  	Val;
  double	Fbar;

  if ( P_Range.Start > P_Range.Stop ) {
    fprintf( stderr, "%s: bad range\n", TOOLNAME );
    exit( 1 );
  }
  else if ( P_Range.Step < 0.0 ) {
    fprintf( stderr, "%s: bad step value\n", TOOLNAME );
    exit( 1 );
  }

  Nb_Points = Nb_Pts( P_Range );

  for ( p=0; p<Nb_Points; p++ ) {

    Val = P_Range.Start + p*P_Range.Step;

    if ( Val < W_Range.Start )
      Fbar = 1.0;
    else if ( Val >= W_Range.Stop )
      Fbar = 0.0;
    else {
      /* find the interval in W_Range */
      k = (int) ( ( Val - W_Range.Start ) / W_Range.Step );
      Fbar = Distrib[k];
    }

    if ( !P_Range.Queue )
      printf("%10.6f %12.8f", Val, 1.0-Fbar );
    else {
      if ( !P_Range.Log && !P_Range.Log10 )
	printf("%10.6f %14.10e", Val, Fbar );
      else if ( P_Range.Log )
	printf("%10.6f %12.8e", Val, log(Fbar) );
      else /* Log10 == TRUE */
	printf("%10.6f %12.8e", Val, log10(Fbar) );
    }
    if ( NewLine )
      printf("\n");
  }
}

/************************************************************************/
/* Procedure to compute the moments of the stationary distribution	*/
/* It assumes stability.						*/
/************************************************************************/
double	Compute_Moment( double *Distrib, 
			RANGE_INFO Range,
			int Order,
			boolean	Cumulative )
{
  int		i;
  int 		N;
  double	Proba;
  double	Val;
  double	Res;

  if ( Order == 0 )
    return 1.0;
  else {
    Res = 0.0;

    N = Nb_Pts( Range );

    if ( Cumulative ) {
      if ( Order == 1 ) {
	for ( i=1; i<N; i++ ) {
	  Res += Distrib[i];
	}
	Res *= Range.Step;
      }
      else if ( Order == 2 ) {
	for ( i=1; i<N; i++ ) {
	  Res += Distrib[i] * (2*i-1);
	}
	Res *= Range.Step * Range.Step;
      }
      if ( Order >= 3 ) {
	for ( i=1; i<N; i++ ) {
	  Proba = Distrib[i]-Distrib[i-1];
	  Val = Range.Start + i*Range.Step;
	  Res += Proba + pow( Val, (double)Order );
	}
      }
    }
    else {
      for ( i=0; i<N; i++ ) {
	Val = Range.Start + i*Range.Step;
	Res += Distrib[i] + ( Order == 1 ? 
			      Val :
			      pow( Val, (double)Order ) );
      }
    }
  }

  return Res;
}

double Load_Factor( Law_Desc Arrival_Law,
		    Law_Desc Service_Law )
{
  double Res;

  Res = Mean( Service_Law ) / Mean( Arrival_Law );

  return Res;
}
		       
int Parse_Law_From_Args( char **argv, int index, int argc,
			   Law_Desc The_Law )
{
  int		i;
  int		j;
  int		Nb_Params;
  boolean 	Error;
  boolean	Short;
  boolean 	Ok;
  double	Foo;
  
  i = index;
  Error = FALSE;
  Short = FALSE;

  The_Law->Name = argv[i][0];
  i++;

  if ( i < argc ) {
    switch( The_Law->Name ) {
    case 'C':
    case 'E':
    case 'B':
    case 'g':
    case 'G':
      if ( 1 != sscanf( argv[i], "%lf", &(The_Law->Parameters[0] ) ) ) {
	Error = TRUE;
      }
      else
	i++;
      break;
    case 'R':
    case 'U':
    case 'W':
    case 'p':
    case 'N':
    case 'L':
      The_Law->Parameters = (double*) 
	realloc ( The_Law->Parameters, 2*sizeof(double) );
      if ( 1 != sscanf( argv[i], "%lf", &(The_Law->Parameters[0] ) ) ) {
	Error = TRUE;
      } 
      else {
	i++;
	if ( i < argc ) {
	  if ( 1 != sscanf( argv[i], "%lf", &(The_Law->Parameters[1] ) ) ) {
	    Error = TRUE;
	  }
	  else 
	    i++;
	} 
	else {
	  Short = TRUE;
	}
      }
      break;
    case 'O':
      The_Law->Parameters = (double*) 
	realloc ( The_Law->Parameters, 3*sizeof(double) );
      if ( 1 != sscanf( argv[i], "%lf", &(The_Law->Parameters[0] ) ) ) {
	Error = TRUE;
      }
      else {
	i++;
	if ( i < argc ) {
	  if ( 1 != sscanf( argv[i], "%lf", &(The_Law->Parameters[1] ) ) ) {
	    Error = TRUE;
	  } 
	  else {
	    i++;
	    if ( i < argc ) {
	      if ( 1 != sscanf( argv[i], "%lf", &(The_Law->Parameters[2] ) ) ) {
		Error = TRUE;
	      } 
	    }
	  }
	}
	else 
	  Error = TRUE;
      }
      break;
    case 'H':
      /* the number of parameters is not known in advance... have to 	*/
      /* find it first.							*/
      Ok = TRUE;
      for ( j=i; Ok && ( j < argc ); j++ ) {
	Ok = ( 1 == sscanf( argv[j], "%lf", &Foo ) );
      }
      /* counting the number of floats read */
      if ( j < argc )
	j--;
      Nb_Params = (j-i);
      if ( Nb_Params %2 != 0 ) {
	Error = TRUE;
      } else {
	The_Law->Parameters = (double*) 
	  realloc ( The_Law->Parameters, (1+Nb_Params)*sizeof(double) );
	The_Law->Parameters[0] = (double)(Nb_Params);
	for( j=i; j<i+Nb_Params; j++ ) {
	  sscanf( argv[j], "%lf", &(The_Law->Parameters[1+j-i]) );
	}
	i = i+Nb_Params;
      }
      break;
    case 'D':
      /* skip the (a priori) "[" argument */
      i++;
      /* the number of parameters is not known in advance... have to 	*/
      /* find it first, looking for the next "]".			*/
      Ok = TRUE;
      for ( j=i; Ok && ( j < argc ); j++ ) {
	Ok = ( 0 != strcmp( argv[j], "]" ) );
      }
      /* counting the number of floats read */
      if ( j >= argc ) {
	Short = TRUE;
      }
      else {
	j--;
	Nb_Params = (j-i);
	The_Law->Parameters = (double*) 
	  realloc ( The_Law->Parameters, (1+2*Nb_Params)*sizeof(double) );
	The_Law->Parameters[0] = (double)(Nb_Params);
	for( j=0; j<Nb_Params; j++ ) {
	  sscanf( argv[i+j], "%lf", &(The_Law->Parameters[1+j]) );
	}
	/* skip "]" and "[" */
	i += Nb_Params+2;
	/* continue reading. Have to protect against short arg list */
	Ok = TRUE;
	for( j=0; Ok && (j<Nb_Params) && (i+j<argc); j++ ) {
	  Ok = ( 1 == sscanf( argv[i+j], "%lf", 
			      &(The_Law->Parameters[1+Nb_Params+j]) ) );
	}
	if ( !Ok ) {
	  Error = TRUE;
	  i = i+j-1;
	} 
	else if ( i+j >= argc ) {
	  Short = TRUE;
	}
	else {
	  /* skip final "]", hoping it is actually there */
	  i = i + Nb_Params + 1;
	}
      }
      break;
    case 'P':
      /* read first parameter */
      if ( 1 != sscanf( argv[i], "%lf", &(The_Law->Parameters[0] ) ) ) {
	Error = TRUE;
      }
      else {
	i++;
	/* recursively read two laws */
	The_Law->Law_Parameters = (Law_Desc*) calloc( 2, sizeof(Law_Desc) );
	The_Law->Law_Parameters[0] = New_Law();
	The_Law->Law_Parameters[1] = New_Law();
	i = Parse_Law_From_Args( argv, i, argc, The_Law->Law_Parameters[0] );
	i = Parse_Law_From_Args( argv, i, argc, The_Law->Law_Parameters[1] );
      }
    default:
      printf("%s: unknown law id (%c)\n", TOOLNAME, The_Law->Name );
    }
  }
  else {
    Short = TRUE;
  }

  if ( Short ) {
    printf("%s: error parsing law %c: not enough arguments.\n", 
	   TOOLNAME, The_Law->Name );
    i = index;
  }
  else if ( Error ) {
    printf("%s: error parsing law %c at argument %s.\n", 
	   TOOLNAME, The_Law->Name, argv[i] );
    i = index;
  }

  return i;
}
