#include <stdio.h>
#include <malloc.h>
#include <math.h>

#include "stats.h"

#define	TOOLNAME	"distrib"

#define	MAX(x,y)	( (x>y) ? (x) : (y) )

boolean	Debug;

typedef struct {
  double	Start;
  double	Stop;
  double	Step;
  boolean	Queue;
  boolean	Log;
  boolean	Log10;
} RANGE_INFO;
#define DEFAULT_RANGE { 0.0, 10.0, 0.0, FALSE, FALSE, FALSE }

typedef struct NODE {
  double 	Val;
  int		Weight;
  struct NODE	*Left;
  struct NODE	*Right;
} *Tree;

typedef struct TREE_RANGE {
  Tree	Values;
  int   Nb_Values;
  int	Left_Samples;	/* Nb of samples "at the left" of the range */
} *Tree_Range;

typedef struct LINE_RANGE {
  int	Nb_Samples;
  int	Nb_Values;
  int	Left_Samples;	/* Nb of samples "at the left" of the range */
  int	*Values;
  RANGE_INFO Range;
} *Line_Range;

/* forward declaration */
Tree_Range	Build_Sorted_Range( double From, double To );
Line_Range	Build_Histogram( RANGE_INFO );
void	Print_Tree_Range( Tree_Range, RANGE_INFO, boolean Steps );
void	Print_Histogram( Line_Range, RANGE_INFO, boolean Steps );
int	Tree_Size( Tree T );

int main(int argc,
	 char *argv[] )
{
  static char		Sample_File_Name[256] = "";
  static char		Result_File_Name[256] = "";
  int			i;
  boolean		Help = FALSE;
  boolean		Args_Ok = TRUE;
  boolean		Value = FALSE;
  boolean		Steps = FALSE;
  double		x_Val;
  RANGE_INFO		P_Range = DEFAULT_RANGE;
  Tree_Range		The_Range;
  Line_Range		The_Histo;

  i = 1;

  if (argc>1)
    do {
      if ( 0 == strcmp( "-f", argv[i] ) ) {
	i++;
	if ( i < argc ) {
	  strcpy( Sample_File_Name, argv[i] );
	  i++;
	}
	else {
	  printf("%s: filename expected\n", TOOLNAME);
	  Args_Ok = FALSE;
	}
      }
      else if ( 0 == strcmp( "-o", argv[i] ) ) {
	i++;
	if ( i < argc ) {
	  strcpy( Result_File_Name, argv[i] );
	  i++;
	}
	else {
	  printf("%s: filename expected\n", TOOLNAME);
	  Args_Ok = FALSE;
	}
      }
      else if ( 0 == strcmp( "-xfrom", argv[i] ) ) {
	i++;
	if ( ( i < argc ) && ( 1 == sscanf( argv[i], "%lf",
					    &(P_Range.Start) ) ) )
	  i++;
	else {
	  printf("%s: real expected after %s\n", TOOLNAME,argv[i-1]);
	  Args_Ok = FALSE;
	}
      }
      else if ( 0 == strcmp( "-xto", argv[i] ) ) {
	i++;
	if ( ( i < argc ) && ( 1 == sscanf( argv[i], "%lf", 
					    &(P_Range.Stop) ) ) )
	  i++;
	else {
	  printf("%s: real expected after %s\n", TOOLNAME,argv[i-1]);
	  Args_Ok = FALSE;
	}
      }
      else if ( 0 == strcmp( "-xstep", argv[i] ) ) {
	i++;
	if ( ( i < argc ) && ( 1 == sscanf( argv[i], "%lf", 
					    &(P_Range.Step) ) ) )
	  i++;
	else {
	  printf("%s: real expected after %s\n", TOOLNAME,argv[i-1]);
	  Args_Ok = FALSE;
	}
      }
      else if ( 0 == strcmp( "-xval", argv[i] ) ) {
	i++;
	if ( ( i < argc ) && ( 1 == sscanf( argv[i], "%lf", &x_Val ) ) ) {
	  i++;
	  Value = TRUE;
	}
	else {
	  printf("%s: real expected after %s\n", TOOLNAME,argv[i-1]);
	  Args_Ok = FALSE;
	}
      }
      else if ( 0 == strcmp( "-debug", argv[i] ) ) {
	Debug = TRUE;
	i++;
      }
      else if ( 0 == strcmp( "-steps", argv[i] ) ) {
	Steps = TRUE;
	i++;
      }
      else if ( 0 == strcmp( "-queue", argv[i] ) ) {
	P_Range.Queue = TRUE;
	i++;
      }
      else if ( 0 == strcmp( "-log", argv[i] ) ) {
	P_Range.Log = TRUE;
	i++;
      }
      else if ( 0 == strcmp( "-log10", argv[i] ) ) {
	P_Range.Log10 = TRUE;
	i++;
      }
      else if ( ( 0 == strcmp( "-h", argv[i] ) )
		|| ( 0 == strcmp( "-?", argv[i] ) ) ) {
	printf("Usage: %s [-f <file>] [-o <file>] ", TOOLNAME);
	printf("[-xfrom <real>] [-xto <real>]\n\t[-xval <real>] ");
	printf("[-xstep <real>] [-queue] [-log] [-log10]");
	printf("\n\t[-steps] [-debug] [-h] [-?]\n");
	Args_Ok = FALSE;
	Help = TRUE;
	i++;
      }
      else if ( 0 == strcmp( "-v", argv[i] ) ) {
	printf("%s (Version %s)\n", TOOLNAME, VERSION);
	Args_Ok = FALSE;
	i++;
      }
      else
	i++; /* skip unknown argument */
    }
    while ( Args_Ok && ( i < argc ) );

  if (Args_Ok) {

    if ( strlen(Sample_File_Name) != 0 )
      freopen( Sample_File_Name, "r", stdin );

    if ( Value ) {
      P_Range.Start = x_Val;
      P_Range.Stop = x_Val;
    }
    if ( P_Range.Step == 0.0 ) {
      /* The detailed empirical distribution is constructed */
      The_Range = Build_Sorted_Range( P_Range.Start, P_Range.Stop );
      if ( The_Range != NULL ) {
	if ( strlen(Result_File_Name) != 0 )
	  freopen( Result_File_Name, "w", stdout );

	Print_Tree_Range( The_Range, P_Range, Steps );

	/* no cleanup (shame on me) to save a little time */
      }
    }
    else {
      The_Histo = Build_Histogram( P_Range );
      if ( The_Histo != (Line_Range)NULL ) {
	
	if ( strlen(Result_File_Name) != 0 )
	  freopen( Result_File_Name, "w", stdout );

	Print_Histogram( The_Histo, P_Range, Steps );

	/* cleanup */
	free( The_Histo->Values );
	free( The_Histo );
      }
    }
 
    return 0;
  }
  else if (Help)
    return 0;
  else
    return 1;
}

/* Construction of a histogram */
Line_Range Build_Histogram( RANGE_INFO Range )
{
  Line_Range	The_Range;
  int		Box;
  double	Sample;

  if ( Range.Start > Range.Stop ) {
    fprintf( stderr, "%s: bad range\n", TOOLNAME );
    return NULL;
  }
    
  The_Range = (Line_Range) malloc( (size_t)sizeof( struct LINE_RANGE ) );
  The_Range->Nb_Values = 
    1 + (int) floor( (Range.Stop-Range.Start)/Range.Step );
  The_Range->Values = (int*) calloc( The_Range->Nb_Values, sizeof(int) );
  The_Range->Nb_Samples = 0;
  The_Range->Left_Samples = 0;
    
  while ( EOF != scanf( "%lf\n", &Sample ) ) {
    if ( Sample <= Range.Start ) {
      The_Range->Left_Samples++;
    }
    else if ( Sample <= Range.Stop ) {
      Box = (int) floor( (Sample-Range.Start)/Range.Step );
      The_Range->Values[Box]++;
    }
    The_Range->Nb_Samples++;
  }
    
  if ( The_Range->Nb_Samples == 0 ) {
    fprintf(stderr, "%s: no samples read\n", TOOLNAME );
    return (Line_Range)NULL;
  }
  else {
    if ( Debug ) {
      printf("# Samples read: %d\n", The_Range->Nb_Samples );
      printf("# Samples at the left: %d\n", The_Range->Left_Samples );
    }

    return The_Range;
  }
}

/* Prints a value according to the range specification. */
/* Value is assumed to contains a __quantile__ P(X>x)   */
void Print_Val( double Value, RANGE_INFO P_Range )
{
  if ( !P_Range.Queue )
    printf("%12.8f", Value );
  else {
    if ( !P_Range.Log && !P_Range.Log10 )
      printf("%14.10e", 1.0-Value );
    else if ( P_Range.Log )
      printf("%12.8e", log(1.0-Value) );
    else /* Log10 == TRUE */
      printf("%12.8e", log10(1.0-Value) );
  }
}

/* printing a histogram */
void Print_Histogram( Line_Range The_Histo,
		      RANGE_INFO P_Range,
		      boolean	Steps )
{
  int		i;
  double	Val;

  /* value at the left of the range */
  printf("%f ", P_Range.Start );
  Val = (double) The_Histo->Left_Samples / (double) The_Histo->Nb_Samples;
  Print_Val( Val, P_Range );
  printf("\n");

  for ( i=0; i<The_Histo->Nb_Values; i++ ) {
    printf("%f ", P_Range.Start + (i+1)*P_Range.Step );
    Val += (double) The_Histo->Values[i] / (double) The_Histo->Nb_Samples;
    Print_Val( Val, P_Range );
    printf("\n");
  }
}

/* construction du tri dans un arbre binaire de recherche */
Tree	Make_New_Node( double Val );
void	Insert_Val( double Val, Tree T );
int     Tree_Depth( Tree T );

Tree_Range	Build_Sorted_Range( double	Start,
			    double	Stop )
{
  Tree	The_Tree;
  Tree_Range	The_Range;
    
  int 		Nb_Samples;
  double	Sample;
  int		Nb_Left = 0;

  if ( Start > Stop )
    {
      fprintf( stderr, "%s: bad range\n", TOOLNAME );
      return NULL;
    }
    
  The_Range = (Tree_Range) malloc( (short)sizeof( struct TREE_RANGE ) );
  The_Range->Values = NULL;
    
  Nb_Samples = 0;
    
  The_Tree = NULL;
    
  /* creation premier noeud */
  while ( ( The_Tree == NULL ) && ( EOF != scanf( "%lf\n", &Sample ) ) ) {
    if ( Sample <= Start )
      Nb_Left++;
    else if ( Sample <= Stop )
      The_Tree = Make_New_Node( Sample );
	    
    Nb_Samples++;
  }
  while ( EOF != scanf( "%lf\n", &Sample ) ) {
    if ( Sample <= Start )
      Nb_Left++;
    else if ( Sample <= Stop )
      Insert_Val( Sample, The_Tree );
	    
    Nb_Samples++;
  }
    
  if ( Nb_Samples == 0 ) {
    fprintf(stderr, "%s: no samples read\n", TOOLNAME );
    return NULL;
  }
  else {
    if ( Debug ) {
      printf("Samples read: %d\n", Nb_Samples );
      printf("Samples below %f: %d\n", Start, Nb_Left );
      printf("Tree Size: %d\n", Tree_Size( The_Tree) );
      printf("Tree depth: %d\n", Tree_Depth( The_Tree) );
    }
	
    The_Range->Values = The_Tree;
    The_Range->Nb_Values = Nb_Samples;
    The_Range->Left_Samples = Nb_Left;
	
    return The_Range;
  }
}

void	Insert_Val(double	The_Val,
		   Tree		The_Tree )
{
  boolean	Continue = TRUE;
  Tree		Aux_Tree;

  Aux_Tree = The_Tree;

  while ( Continue ) {

    if ( The_Val < Aux_Tree->Val )
      {
	if ( Aux_Tree->Left == NULL )
	  {
	    Aux_Tree->Left = Make_New_Node( The_Val );

	    Continue = FALSE;
	  }
	else {
	  Aux_Tree = Aux_Tree->Left;
	}
      }
    else if ( The_Val > Aux_Tree->Val )
      {
	if ( Aux_Tree->Right == NULL )
	  {
	    Aux_Tree->Right = Make_New_Node( The_Val );

	    Continue = FALSE;
	  }
	else {
	  Aux_Tree = Aux_Tree->Right;
	}
      }
    else /* ( The_Val == Aux_Tree->Val ) */
      {
	/* un des deux poids: par exemple le gauche */
	Aux_Tree->Weight++;		
	Continue = FALSE;
      }
  }
}

Tree	Make_New_Node( double	The_Val )
{
  Tree	The_Node;

  The_Node = (Tree) malloc( (short)sizeof( struct NODE ) );
  The_Node->Val = The_Val;
  The_Node->Weight = 1;
  The_Node->Left = NULL;
  The_Node->Right = NULL;

  return The_Node;
}

int	Print_Tree(Tree		The_Tree,
		   int		Nb_Start,
		   boolean	Steps,
		   int     	Nb_Values )
{
  int	Weight;

  if ( The_Tree->Left != NULL )
    Weight = Print_Tree( The_Tree->Left, Nb_Start, Steps, Nb_Values );
  else
    Weight = Nb_Start;

  if ( Steps )
    printf("%f %f\n", The_Tree->Val, Weight / (double)Nb_Values );
  Weight += The_Tree->Weight;
  printf("%f %f\n", The_Tree->Val, Weight / (double)Nb_Values );

  if ( The_Tree->Right != NULL )
    Weight = Print_Tree( The_Tree->Right, Weight, Steps, Nb_Values );

  return	Weight;
}

void	Print_Tree_Range(Tree_Range		The_Range,
			 RANGE_INFO	P_Range,
			 boolean	Steps )
{
  int	Total_Nb;

  printf("%f %f\n", 
	 P_Range.Start, 
	 The_Range->Left_Samples / (double)The_Range->Nb_Values );

  if ( NULL != The_Range->Values ) {
    Total_Nb = Print_Tree( The_Range->Values, The_Range->Left_Samples, 
			   Steps, The_Range->Nb_Values );

    printf("%f %f\n", 
	   P_Range.Stop, 
	   Total_Nb / (double)The_Range->Nb_Values );
  }
}

int     Tree_Size( Tree	The_Tree )
{
  if ( NULL == The_Tree )
    return 0;
  else if ( NULL == The_Tree->Left )
    return 1 + Tree_Size( The_Tree->Right );
  else if ( NULL == The_Tree->Right )
    return 1 + Tree_Size( The_Tree->Left );
  else
    return 1 + Tree_Size( The_Tree->Left ) +
      Tree_Size( The_Tree->Right );
}

int     Tree_Depth( Tree	The_Tree )
{
  int D_Left;
  int D_Right;

  if ( NULL == The_Tree )
    return 0;
  else if ( NULL == The_Tree->Left )
    return 1 + Tree_Depth( The_Tree->Right );
  else if ( NULL == The_Tree->Right )
    return 1 + Tree_Depth( The_Tree->Left );
  else {
    D_Left = Tree_Depth( The_Tree->Left );
    D_Right = Tree_Depth( The_Tree->Right );
    return 1 + MAX( D_Left, D_Right );
  }
}

