/* programme/module pour la simulation equationelle d'un Graphe 	*/
/* d'evenements.							*/

#include <stddef.h>
#include <malloc.h>
#include <math.h>		/* for random number generation	*/
#include <stdio.h>

#include "taskgraph.h"

#define	TOOLNAME	"tg_sim"

extern Graph		*Read_Network(char *File_Name);
extern void		dumpe_network(Graph*);
extern boolean		topological_analysis(Graph*);
extern double		draw_iid_table(double *,int,Law_Desc);

#define INFINI		1e12

#define max(x,y)        ( (x>y) ? x : y )
#define min(x,y)        ( (x<y) ? x : y )

int	Sim_Length = 1000;	/* default number of iterations */
double	**Y;			/* table of finish times 	*/
double	T_Tot = 0.0;		/* for non repetitive graphs 	*/
boolean *T_Dont_Do_It;		/* for saving time when drawing tables */
boolean **L_Dont_Do_It;		/* for saving time when drawing tables */

void Fill_Tables(
    Graph	*Reseau,
    double	**Exec_Times,
    double	***Comm_Times,
    int		Buf_Size )
{
  int		i,j;
  Link_Desc	The_Link;
  Liste_Link	L_Scan;

  for (i=0; i<Reseau->Size; i++) {
    if ( !T_Dont_Do_It[i] )
      draw_iid_table(Exec_Times[i], Buf_Size, Reseau->T[i].Task_Law );

    SCAN( L_Scan, Reseau->T[i].Neighbors ) {
      The_Link = L_Scan->Val;
      j = The_Link->Task_To;
      if ( ( The_Link->Link_Type == PRECEDENCE )
	   && !L_Dont_Do_It[i][j] )
	draw_iid_table( &( Comm_Times[i][j][0] ),
			Buf_Size, The_Link->Link_Law );
    }
  }
}

boolean init_simulation(Graph 		*Reseau,
			double		**Exec_Times,
			double		***Comm_Times,
			int		Buf_Size )
{
    int		i,j;
    int		The_Size;
    Link_Desc	The_Link;
    boolean	Alloc_Ok = TRUE;
    Liste_Link	L_Scan;

    The_Size = Reseau->Size;

    Y = (double**) calloc( The_Size, sizeof (double*) );
    T_Dont_Do_It = (boolean*) calloc( The_Size, sizeof( boolean ) );
    L_Dont_Do_It = (boolean**) calloc( The_Size, sizeof( boolean* ) );

    /* Initialize date tables */
    for (i=0; (i<Reseau->Size) && Alloc_Ok ; i++) {

	Exec_Times[i] = (double*) calloc( Buf_Size, sizeof( double ) );
	Alloc_Ok = ( Exec_Times[i] != NULL );

	Comm_Times[i] = (double**) calloc( The_Size, sizeof( double* ) );
	L_Dont_Do_It[i] = (boolean*) calloc( The_Size, sizeof( boolean ) );

	Y[i] = (double*) calloc( Buf_Size, sizeof( double ) );
	Alloc_Ok = ( Y[i] != NULL );

	if ( Alloc_Ok )
	    SCAN( L_Scan, Reseau->T[i].Neighbors ) {

		The_Link = L_Scan->Val;
		j = The_Link->Task_To;
		if ( The_Link->Link_Type == PRECEDENCE ) {
		    Comm_Times[i][j] = (double*) 
			calloc( Buf_Size, sizeof( double ) );

		    Alloc_Ok = ( Comm_Times[i][j] != NULL );

		}
	    }
    }

    /* make the first draw */
    if ( Alloc_Ok ) {
	for ( i=0; i<The_Size; i++ )
	    T_Dont_Do_It[i] = FALSE;

	Fill_Tables( Reseau, Exec_Times, Comm_Times, Buf_Size );

	/* position swith not to redraw constant times anymore */
	for ( i=0; i<The_Size; i++ ) {
	    T_Dont_Do_It[i] = ( Reseau->T[i].Task_Law->Name == 'C' );
	    if ( Debug && T_Dont_Do_It[i] )
		printf( "Won't draw tables again for task %d\n", i );

	    SCAN( L_Scan, Reseau->T[i].Neighbors ) {
		The_Link = L_Scan->Val;
		j = The_Link->Task_To;
		if ( The_Link->Link_Type == PRECEDENCE )
		    L_Dont_Do_It[i][j] = ( The_Link->Link_Law->Name == 'C' );

		if ( Debug && L_Dont_Do_It[i][j] ) {
		    printf( "Won't draw tables again for link" );
		    printf(" (%1d->%1d).\n", i, j );
		}
	    }
	}
    }

    return Alloc_Ok;
}

Link_Desc get_link( Graph	*Reseau,
		    int		Orig,
		    int		Dest )
{
  Link_Desc	Resultat;
  Liste_Link	L_Scan;

  for ( L_Scan = Reseau->T[Orig].Neighbors; L_Scan != NULL; 
	L_Scan = L_Scan->Next )
    if ( L_Scan->Val->Task_To == Dest )
      Resultat = L_Scan->Val;

  return Resultat;
}
	
void simulate(Graph *Reseau,
	      int Buf_Size )
{	
  int		i;
  int		j;
  int		k;
  int		Time;
  int		The_Size;
  double	x;
  double	y;
  int		Task_Id, Prev_Id;
  int		Max_Depth, Mu;
  Link_Desc	The_Link;
  double	**Exec_Times;
  double	**YY;
  double	***Comm_Times;
	
  The_Size = Reseau->Size;

  Exec_Times = (double**) calloc( The_Size, sizeof (double*) );
  Comm_Times = (double***) calloc( The_Size, sizeof (double**) );

  Max_Depth = Reseau->Max_Dependency;

  if (! init_simulation( Reseau, Exec_Times, Comm_Times, Buf_Size ) ) {
    printf("Could not allocate space. Simulation not launched.\n");
  }
  else {

    YY = Y;

    /* Compute Max_Depth first firing dates. Do not compute modulo  */
    /* Buf_Size, because Buf_Size is always > Max_Depth		*/
    for (Time=0; (Time<Sim_Length) && (Time<Max_Depth) ; Time++) {
      for( i=0; i<Reseau->Nb_Layers; i++) 
	for( j=0; j<Reseau->Level[i]->Nb_Tasks; j++) {
	  Task_Id = Reseau->Level[i]->Task_Id[j];

	  if ( Reseau->T[Task_Id].Prev->Nb_Tasks == 0 ) {
	    /* tasks without precedence execute only once */
	    YY[Task_Id][Time] = ( Time == 0 ? 0.0 : INFINI );
	  }
	  else {
	    /* for the other, compute precedence */
	    for( k=0; k<Reseau->T[Task_Id].Prev->Nb_Tasks; k++) {
	      Prev_Id = Reseau->T[Task_Id].Prev->Task_Id[k];
	      The_Link = get_link( Reseau, Prev_Id, Task_Id );
	      switch( The_Link->Link_Type ) {
	      case PRECEDENCE:
		y = YY[Prev_Id][Time] +
		  Comm_Times[Prev_Id][Task_Id][Time];
		break;
	      case ITERATION:
		Mu = The_Link->Depth;
		if ( Mu <= Time )
		  y = YY[Prev_Id][Time-Mu];
		/* no comm times (yet) */
		else
		  y = 0.0;
		/* and not y = -INFINI; */
		break;
	      }
	      if ( k == 0 )
		x = y;
	      else
		switch( The_Link->Synchro_Type ) {
		case AND:
		  x = max( x, y );
		  break;
		case OR:
		  x = min( x, y );
		  break;
		}
	    }
	    /* here, Time < Buf_Size */
	    YY[Task_Id][Time] = x + Exec_Times[Task_Id][Time];
	  }
	}
    }

    /* compute the rest... */
    for ( ; Time<Sim_Length; Time++) {
      for( i=0; i<Reseau->Nb_Layers; i++)
	for( j=0; j<Reseau->Level[i]->Nb_Tasks; j++) {
	  Task_Id = Reseau->Level[i]->Task_Id[j];

	  /* tasks without precedence execute only once */
	  if ( Reseau->T[Task_Id].Prev->Nb_Tasks == 0 ) {
	    YY[Task_Id][Time % Buf_Size] = ( Time == 0 ? 0.0 : INFINI );
	  }
	  else {
	    for( k=0; k<Reseau->T[Task_Id].Prev->Nb_Tasks; k++) {
	      Prev_Id = Reseau->T[Task_Id].Prev->Task_Id[k];
	      The_Link = get_link( Reseau, Prev_Id, Task_Id );
	      switch( The_Link->Link_Type ) {
	      case PRECEDENCE:
		y = YY[Prev_Id][Time % Buf_Size] +
		  Comm_Times[Prev_Id][Task_Id][Time % Buf_Size];
		break;
	      case ITERATION:
		Mu = The_Link->Depth;
		y = YY[Prev_Id][(Time-Mu) % Buf_Size];
		/* no comm times yet */
		break;
	      }
	      if ( k == 0 )
		x = y;
	      else
		switch( The_Link->Synchro_Type ) {
		case AND:
		  x = max( x, y );
		  break;
		case OR:
		  x = min( x, y );
		  break;
		}
	    }
	    YY[Task_Id][Time % Buf_Size] = x + 
	      Exec_Times[Task_Id][ Time % Buf_Size];
	  }
	}

      if ( ( (Time+1) % Buf_Size ) == 0 ) {
	/* time to draw new variables */
	Fill_Tables( Reseau, Exec_Times, Comm_Times, Buf_Size );
      }

    }

  } /* fin du "else" */

}

void dumpe_simulation(Graph *Reseau, int Buf_Size)
{
  int	i;
  int	j;
  int	Start;
  int   Stop;

  if ( Sim_Length < Buf_Size ) {
    Start = 0;
    Stop = Sim_Length;
  }
  else {
    Start = Sim_Length - Buf_Size;
    Stop = Sim_Length;
  }

  for (i=0; i<Reseau->Size; i++) {
    printf("Task %d ...\n",i);
    for ( j=Start; j<Stop; j++ )
      if ( Y[i][j % Buf_Size] > 100000.0 )
	printf("oo ");
      else	
	printf("%8.3f ",Y[i][j % Buf_Size] );
    printf("\n");
  }
}

void write_stat_results(Graph 	*Reseau,
			int	Buf_Size )
{
  int		i;
  double	Gamma;

  for (i=0; i<Reseau->Size; i++) {
    printf("Y_%1d[%1d] = %f  ", i, Sim_Length, 
	   Y[i][(Sim_Length-1) % Buf_Size] );

    printf("--> cycle time ");
    Gamma = Y[i][(Sim_Length-1) % Buf_Size] / Sim_Length;

    if (  Y[i][(Sim_Length-1) % Buf_Size] >= INFINI ) {
      printf("oo\n");
    }
    else {
      printf("%f\n", Gamma );
    }
  }
}

/************************************************************************/
/* Procedures pour la simulation de graphes non repetitifs.		*/
/************************************************************************/
void simulate_PERT(Graph *Reseau, int Buf_Size)
{
  int		i;
  int		j;
  int		k;
  int		Time;
  int		The_Size;
  double	x;
  double	y;
  int		Task_Id;
  int		Prev_Id;
  Link_Desc	The_Link;
  double	**Exec_Times;
  double	*YY;
  double	*Sum_Y;
  double	***Comm_Times;
  double	End_Time;

  The_Size = Reseau->Size;

  Exec_Times = (double**) calloc( The_Size, sizeof (double*) );
  Comm_Times = (double***) calloc( The_Size, sizeof (double**) );

  if (! init_simulation(Reseau, Exec_Times, Comm_Times, Buf_Size ) ) {
    printf("Could not allocate space. Simulation not launched.\n");
  }
  else {
    YY = (double*) calloc( The_Size, sizeof( double ) );

    Sum_Y = Y[0];
    for ( i=0; i<The_Size; i++ )
      Sum_Y[i] = 0.0;

    for ( Time=0; Time<Sim_Length; Time++) {
      for ( i=0; i<The_Size; i++ )
	YY[i] = 0.0;
      End_Time = 0.0;

      for( i=0; i<Reseau->Nb_Layers; i++)
	for( j=0; j<Reseau->Level[i]->Nb_Tasks; j++) {
	  Task_Id = Reseau->Level[i]->Task_Id[j];

	  if ( Reseau->T[Task_Id].Prev->Nb_Tasks == 0 )
	    x = 0;
	  else {
	    for( k=0; k<Reseau->T[Task_Id].Prev->Nb_Tasks; k++) {
	      Prev_Id = Reseau->T[Task_Id].Prev->Task_Id[k];
	      The_Link = get_link( Reseau, Prev_Id, Task_Id );
	      if( The_Link->Link_Type == PRECEDENCE ) {
		y = YY[Prev_Id] +
		  Comm_Times[Prev_Id][Task_Id][Time % Buf_Size];
	      }
	      else {
		fprintf( stderr, "%s: error: non precedence link ", 
			 TOOLNAME );
		fprintf( stderr, "encountered.\n" );
	      }
	      if ( k == 0 )
		x = y;
	      else
		switch( The_Link->Synchro_Type ) {
		case AND:
		  x = max( x, y );
		  break;
		case OR:
		  x = min( x, y );
		  break;
		}
	    }
	  }
	  YY[Task_Id] = x + Exec_Times[Task_Id][ Time % Buf_Size ];
	  End_Time = max( End_Time, YY[Task_Id] );
	}

      for ( i=0; i<The_Size; i++ )
	Sum_Y[i] += YY[i];
      T_Tot += End_Time;

      if ( ( (Time+1) % Buf_Size ) == 0 ) {
	/* time to draw new variables */
	Fill_Tables( Reseau, Exec_Times, Comm_Times, Buf_Size );
      }
    }

  } /* fin du "else" */

}

void write_stat_PERT(Graph 	*Reseau)
{
  int	i;

  for (i=0; i<Reseau->Size; i++) {
    printf("Time of execution end of Y_%1d = %f\n",
	   i, Y[0][i] / Sim_Length );
  }
  printf("Total execution time: ET = %f\n", T_Tot / Sim_Length );
}

int main(int argc, char *argv[])
{	
  int		i;
  static char	File_Name[128] = "";
  Graph 	*Reseau;
  boolean	Analyze_Alone = FALSE;
  boolean	Args_Ok = TRUE;
  boolean	Help = FALSE;
  int		Buf_Size = 1024;	/* taille buffer par defaut */

  i = 1;

  if (argc>1)
    do {
      if GET_STRING( "-f", File_Name )
      else if GET_INT( "-bs", Buf_Size )

      else if SWITCH( "-debug", Debug, TRUE )
      else if SWITCH( "-a", Analyze_Alone, TRUE )

      else if IS_VERSION

      else if ( ( 0 == strcmp( "-h", argv[i] ) )
		|| ( 0 == strcmp( "-?", argv[i] ) ) ) {
	printf("Usage: %s [-f <file>] [-nevt <int>]", TOOLNAME );
	printf(" [-bs <int>] [-debug] [-a] [-h] [-?]\n");
	Args_Ok = FALSE;
	Help = TRUE;
	i++;
      }
      else if ( 0 == strcmp( "-nevt", argv[i] ) ) {
	i++;
	if ( !( ( i < argc )
		&& ( 0 != sscanf( argv[i], "%d", &(Sim_Length) ) )
		&& ( Sim_Length > 0 ) ) ) {
	  Args_Ok = FALSE;
	  printf("%s: positive integer expected ", argv[0]);
	  printf("after %s\n",argv[i-1]);
	}
	else 
	  i++;
      }
      else
	i++; /* skip unknown argument */
    }
    while ( Args_Ok && ( i < argc ) );

  if (Args_Ok) {
    Reseau = Read_Network( File_Name );

    if ( Reseau != NULL ) {
      if ( topological_analysis(Reseau) ) {
	    
	if (Debug)
	  dumpe_network(Reseau);

	if (!Analyze_Alone) {

	  if ( Reseau->Max_Dependency == 0 ) {
	    if ( Debug ) {
	      printf("Non-repetitive network: estimating completion times\n.");
	    }
	    Buf_Size = max( Buf_Size, 1 );

	    simulate_PERT( Reseau, Buf_Size );
	    write_stat_PERT( Reseau );
	  }
	  else {
	    if ( Debug ) {
	      printf("Repetitive network: estimating cycle times\n.");
	    }
	    /* check if buffer will be sufficient */
	    Buf_Size = max( Buf_Size, 1 + Reseau->Max_Dependency );

	    simulate(Reseau, Buf_Size);

	    if (Debug)
	      dumpe_simulation(Reseau, Buf_Size );
	    write_stat_results(Reseau, Buf_Size );
	  }
	}
	else
	  dumpe_network(Reseau);

      }

      return 0;
    }
    else {
      printf("%s: error: could not open file  %s for reading.\n",
	     TOOLNAME, File_Name);
      return -1;
    }
  }
  else if (Help)
    return 0;
  else 
    return -1;

}
