#
# Various utilities for mtool
#
######################################################################
# DisplayInfo -
# Display information string in text-variable 
#
# Parameters:
# - v		global variable
# - str		information string
#
proc  DisplayInfo {v str} {
    upvar \#0 $v tmp
    
    set tmp $str
    update idletasks
}

######################################################################
# Quit -
# User dialog before quitting.
#
# Parameters: none
#
proc  Quit {} {
 
    switch [tk_dialog .end "Quit Confirm" \
		"Are you sure you want to quit?" warning 0 "Quit" "Cancel"] {
		    0 { destroy . }
		    1 { return }
		}
}


######################################################################
# Detex -
# A filter function to eliminate TeX-type expressions such as "\...",
# from a given sting. Returns the cleaned string.
#
# Parameters:
# - line	the string to process.
#
proc  Detex { line } {

    set t [TranslateText $line]
    if [file executable detex] {
	set text [exec echo $t | detex]
    } else {
	# No detex. Rudimentary processing 
	regsub -all {\\[a-zA-z_0-9]+[ ]*} $t {} text
	regsub -all {[\{\}]*} $text {} text
    }

    return  $text
}

######################################################################
# InsertWithTags -
# A procedure that inserts a string with the good tag in a specific window.
#
# Parameters:
# - w		text widget in which to insert the string
# - where	index where to insert
# - text	the string to insert
# - args	tags
#
proc  InsertWithTags { w where text args } {

    set start [$w index insert]
    $w insert $where $text
    foreach tag [$w tag names $start] {
        $w tag remove $tag $start insert
    }
    foreach i $args {
        $w tag add $i $start insert
    }
}

######################################################################
# ChangeCursor -
# Change the form of the cursor.
#
# Parameters:
# - w		widget 
# - cursor	the new cursor
#
proc  ChangeCursor { w cursor } {
    $w configure -cursor $cursor
    update idletask
}

######################################################################
# write -
# Boolean function to write the contents of a text widget to a file.
# Returns the success of the operation.
#
# Parameters:
# - w		the text widget to write
# - file	the name of the file in which to write.
#
proc  write { w file } {
    
    if [file exists $file] {
	if [tk_dialog $w.error "Existing File" "This file already exists." {warning} -1 {Overwrite} {Cancel}] {
	    return 0
	}
    }
    set file_id [open $file w]
    set text [$w get 0.0 end]
    puts $file_id $text
    close $file_id
    return 1
}

######################################################################
# Save -
# User interface for specifying the name of the file in which to save
# the "error" window.
#
# Parameters:
# - w		prefix name of the toplevel window
#
proc Save { w } {
    
    toplevel $w.name

    frame $w.name.top -relief raised -bd 1
    pack $w.name.top
    
    label $w.name.top.label -text "Enter the name of the file" -height 3 -width 30
    pack $w.name.top.label -fill x

    entry $w.name.top.entry -relief raised -bd 2 -width 20 -textvariable save_in
    pack $w.name.top.entry -pady 7
    bind $w.name.top.entry <Return> "
	           $w.name.bot.ok invoke "
    
    frame $w.name.bot -relief raised -bd 1
    pack $w.name.bot -fill x

    button $w.name.bot.ok -text OK -relief raised -bd 2\
	-command "
                        if \[write $w \$save_in\] {
                           grab release $w.name
                           destroy $w.name
                        } else {
                           grab $w.name
                        }"
    button $w.name.bot.cancel -text CANCEL -relief raised -bd 2 \
	-command "grab release $w.name ; destroy $w.name"
    pack $w.name.bot.ok $w.name.bot.cancel -expand yes -side left -pady 7 -ipadx 5 -ipady 3
    
    grab $w.name
}

######################################################################
# LoadRef -
# A Procedure that loads a bibliography (reference) file in bibtex format.
# The result is put in the global array ref(key) with "key" the 
# bibliographic key, and with contents a list [Author Title Year].
#
# Parameters:
# -r 		name of the reference file
#
proc  LoadRef { r } {
    global ref env

    set in_ref 0

    if ![file exists $r] {
	# perhaps it is relative to ERSHOME?
	if [info exists env(ERSHOME)] {
	    set fullname "$env(ERSHOME)/$r"
	    puts -nonewline "trying with $fullname "
	    if [file exists $fullname] {
		set r $fullname
	    } else {
		puts "Can't find reference file $r"
		return
	    }
	} else {
	    puts "Can't find reference file $r"
	    return
	}
    }

    set idref [open $r r]
    while { [gets $idref line]!=-1 } {
	if ![string compare [string index $line 0] "@"] {
	    set n [lindex [split $line \{,] 1]
	    set ref($n) ""
	    set in_ref 1
	} elseif { $in_ref } {
	    case [lindex $line 0] {
		AUTHOR {
		    append ref($n) "[join [FindPart $line] \"\"]. "
		}
		
		TITLE {
		    append ref($n) "[join [FindPart $line] \"\"]. "
		}
		
		YEAR {
		    append ref($n) "[join [FindPart $line] \"\"]. "
		}
	    }
	}
    }
    close $idref
}

proc  ToLower { p } {

    if [regexp {[A-Z]} $p] {
	set p_tolower "upper_"
	append p_tolower [string tolower $p]
	return $p_tolower
    } 
    return $p
}

