#
# Module with procedures useful for playing the token game, either
# by hand, or randomly.
#
######################################################################
# Fire_Node -
# Procedure to fire a transition or a source a certain number of times.
# If the number is negative, the transition is fired backwards.
#
# Parameters:
# - win		the canvas in which this happens
# - node_id	id of the transition
# - number	number of firings
#
proc Fire_Node { win node_id number } {

    if { [Is_Type $node_id TRANS] || [Is_Type $node_id SOURCE] } {
	if { $number > 0 } {
	    for { set i 0 } { $i < $number } { incr i 1 } {
		Fire_Transition $win $node_id 0
	    }
	} elseif { $number < 0 } {
	    for { set i $number } { $i < 0 } { incr i 1 } {
		UnFire_Transition $win $node_id 0
	    }
	}
    }
}

######################################################################
# Fire_Current_Node -
# Procedure to fire the transition currently pointed at.
# If the number is negative, the transition is fired backwards.
#
# Parameters:
# - win		the canvas in which this happens
# - number	number of firings. Defaults to 1.
#
proc Fire_Current_Node { win { number 1 } } {

    set cur_id [$win find withtag current]

    if { [llength $cur_id] > 0 } { 
	Fire_Node $win $cur_id $number
    }
}

######################################################################
# Fire_Transition -
# UnFire_Transition -
# Procedures to fire a transition or a source once.
# If the transition is not enabled, the bell is rang unless specified
# Return 1 if firing was successful, 0 otherwise.
#
# Parameters:
# - win		the canvas in which this happens
# - node_id	id of the transition
# - silent	a boolean saying if impossible firings should cause a beep
#               (defaults to false)
#
proc Fire_Transition { win node_id { silent 0 } } {
    global mark platra_list trapla_list link_weight

    # compute if transition is enabled.
    set ok 1
    foreach link $platra_list {
	if { [lindex $link 2] == $node_id } {
	    set place [lindex $link 1]
	    if { $mark($place) < $link_weight([lindex $link 0]) } {
		set ok 0
	    }
	}
    }
    if $ok {
	# remove input tokens
	foreach link $platra_list {
	    if { [lindex $link 2] == $node_id } {
		set place [lindex $link 1]
		incr mark($place) -$link_weight([lindex $link 0])
		Draw_Marking $win $place
	    }
	}
	foreach link $trapla_list {
	    if { [lindex $link 1] == $node_id } {
		set place [lindex $link 2]
		incr mark($place) $link_weight([lindex $link 0])
		Draw_Marking $win $place
	    }
	}
    } elseif !$silent {
	bell 
    }

    return $ok
}

proc UnFire_Transition { win node_id { silent 0 } } {
    global mark platra_list trapla_list link_weight

    # compute if transition is un-enabled.
    set ok 1
    foreach link $trapla_list {
	if { [lindex $link 1] == $node_id } {
	    set place [lindex $link 2]
	    if { $mark($place) < $link_weight([lindex $link 0]) } {
		set ok 0
	    }
	}
    }

    if $ok {
	# remove output tokens
	foreach link $trapla_list {
	    if { [lindex $link 1] == $node_id } {
		set place [lindex $link 2]
		incr mark($place) -$link_weight([lindex $link 0])
		Draw_Marking $win $place
	    }
	}
	foreach link $platra_list {
	    if { [lindex $link 2] == $node_id } {
		set place [lindex $link 1]
		incr mark($place) $link_weight([lindex $link 0])
		Draw_Marking $win $place
	    }
	}
    } elseif !$silent {
	bell 
    }

    return $ok
}

######################################################################
# Play_Token_Game -
# Runs the random token game. The rules of the game are to pick a 
# transition (or a source) at random, and fire it if possible. 
# This is not a simulation in that timing information is ignored.
# The game is controled by global variables that can be modified
# through the interface (Token_Game_Interface).
#
# Parameters:
# - win		the canvas to animate
# - direction	the direction (forward|backwards) of the firings
#
proc Play_Token_Game { win direction } {
    global trans_list source_list nbtry speed game
    global DEBUG

    bind $win <KeyPress> {set game STOP}
    bind $win <1> {set game STOP}

    set game GO
    set interval [expr 1000/$speed]

    randomInit [pid]

    # merge transition and source list
    set global_list [concat $trans_list $source_list]
    set nb_trans [llength $global_list]

    for { } { $game != "STOP" } { } {
	set idx [expr int([random]*$nb_trans)]
	set trans [lindex $global_list $idx]
	
	Highlight_Objects $win [lindex $trans 0]
	update
	after [expr (8*$interval)/10]

	if { $direction == "forward" } {
	    Fire_Transition $win [lindex $trans 0] 1
	} else {
	    UnFire_Transition $win [lindex $trans 0] 1
	}

	update idletasks
	after [expr (2*$interval)/10]

	UnHighlight_Objects $win [lindex $trans 0]

	if $DEBUG {
	    puts $game
	}

	incr nbtry -1
	if { $nbtry < 0 } {
	    set game STOP
	}
    }

    if { $nbtry < 0 } {
	set nbtry 0
    }

    Reset_Standard_Bindings $win
}

proc Token_Game_Interface { win } {
    global nbtry game

    if [Is_Empty_Graph] { return }

    if [catch {toplevel .params}] { return }

    wm title .params "Token Game Parameters"

    frame .params.nbevt
    label .params.nbevt.label -text "Number of Firings:" -width 15  -anchor w
    entry .params.nbevt.field -textvar nbtry -width 3 -relief sunken

    frame .params.dir
    label .params.dir.label -text "Direction:" -width 15  -anchor w
    radiobutton .params.dir.forw  -text "Forward"  -variable direction \
	-value forward
    radiobutton .params.dir.backw  -text "Backwards" -variable direction \
	-value backwards

    frame .params.speed
    label .params.speed.label -text "Speed (evt/s):" -width 15 -anchor w
    scale .params.speed.field -variable speed -length 100 \
	-orient horizontal -from 1 -to 100 -relief sunken

    pack .params.nbevt.label .params.nbevt.field -side left
    pack .params.dir.label .params.dir.forw .params.dir.backw -side left
    pack .params.speed.label .params.speed.field -side left
    pack .params.nbevt .params.dir .params.speed -side top -fill x

    frame .params.actions
    button .params.actions.run -text "Run" \
	-command "Play_Token_Game $win \$direction"
    button .params.actions.stop -text "Stop" \
	-command "set game STOP"
    button .params.actions.exit -text "Exit" \
	-command "set game  STOP\ndestroy .params"

    pack .params.actions.run .params.actions.stop .params.actions.exit \
	-side left
    pack .params.actions -side top


}