#
# Module containing the procedures to read and write files into pned
# - save/load in the "pned" format
# - save in EGL, PNL etc
#
######################################################################
# Check_Network - 
# A procedure for checking that the various attributes of the objects
# are ok. Prompts the user for a correction if the boolean PROMPT is true.
# Actually not used at the moment!
#
# Parameters:
# - prompt	a boolean telling if the uses should be prompted for 
#		corrections
#
proc Check_Network { prompt } {
    global trans_list place_list source_list platra_list trapla_list 
    global mark TransTime PlaceTime source_rate name link_weight
    global DEBUG

    set OK 1

    foreach trans [concat $source_list $trans_list] {
	set id [lindex $trans 0]
	if ![Is_Numeric_Law $TransTime($id)] {
	    switch [tk_dialog .export "Bad Value" "Transition $id has an improper timing\nDo you wish to edit it?" {warning} 0 {Yes} {No} {Cancel Save}] {
		0 {
		    Edit_Trans_Attributes .graph.canvas 0 0 $id 
		    tkwait window .att
		    set OK [expr $OK && [Is_Numeric_Law $TransTime($id)]]
		}
		1 { 
		    set OK 0
		}
		2 {
		    return 0
		}
	    }
	}
    }
    return $OK
}

######################################################################
# write_graph -
# The procedure that writes a Petri net into a file. Returns a boolean
# value assessing the success of the save.
#
# Parameters:
# - win		the canvas containing the graph to save
# - file_name	name of the file in which to write
#
proc write_graph { win file_name } {
    global trans_list place_list source_list platra_list trapla_list 
    global mark TransTime PlaceTime source_rate name link_weight
    global curzoomvalue
    global DEBUG

    if [catch {set outfile [open $file_name w]} err] {
	ReportError "Write Error" "Cannot write to file $file_name"
	return 0
    }

    if $DEBUG {
	puts -nonewline "Saving graph in file $file_name..."
    }

    if ![Check_Network 1] { return 0 }

    # computing the correction factor
    set scale [expr 100.0/$curzoomvalue]
    # renumbering all places and transitions
    set curtrans 1
    foreach trans [concat $source_list $trans_list] {
	set tr_number([lindex $trans 0]) $curtrans
	incr curtrans 1
    }
    set curplace 1
    foreach place $place_list {
	set pl_number([lindex $place 0]) $curplace
	incr curplace 1
    }

    foreach trans [concat $source_list $trans_list] {
	set trans_id [lindex $trans 0]
	if [Is_Type $trans_id SOURCE] {
	    puts -nonewline $outfile "Src $tr_number($trans_id) "
	} else {
	    puts -nonewline $outfile "Tr $tr_number($trans_id) "
	}
	puts -nonewline $outfile \
	    [Scale_Line_Coords [$win coords $trans_id] $scale]
	regsub -all {[ ]} $name($trans_id) "_WHITE_" the_name
	puts -nonewline $outfile " \"$the_name\""
	if [Is_Type $trans_id SOURCE] {
	    puts $outfile " $source_rate($trans_id)"
	} else {
	    puts $outfile " \{ $TransTime($trans_id) \}"
	}
    }
    foreach place $place_list {
	set place_id [lindex $place 0]
	puts -nonewline $outfile "Pl $pl_number($place_id) "
	puts -nonewline $outfile \
	    [Scale_Line_Coords [$win coords $place_id] $scale]
	regsub -all {[ ]} $name($place_id) "_WHITE_" the_name
	puts -nonewline $outfile " \"$the_name\""
	puts -nonewline $outfile " \{ $PlaceTime($place_id) \} "
	puts $outfile $mark($place_id)
    }

    foreach link $trapla_list {
	set link_id [lindex $link 0]
	set trans_id [lindex $link 1]
	set place_id [lindex $link 2]
	puts -nonewline $outfile "TrPl $tr_number($trans_id) $pl_number($place_id) $link_weight($link_id) "
	puts $outfile [Scale_Line_Coords [$win coords $link_id] $scale]
    }
    foreach link $platra_list {
	set link_id [lindex $link 0]
	set trans_id [lindex $link 2]
	set place_id [lindex $link 1]
	puts -nonewline $outfile "PlTr $pl_number($place_id) $tr_number($trans_id) $link_weight($link_id) "
	puts $outfile [Scale_Line_Coords [$win coords $link_id] $scale]
    }

    close $outfile
    if $DEBUG {
	puts " ...Done"
    }

    # success!
    return 1
}

######################################################################
# read_graph -
# The procedure that reads a Petri net from a file. It draws the objects
# on the canvas, and instantiates the attributes.
# The main file name (visible on the display) is updated only when there
# is no merging.
#
# Parameters:
# - win		the canvas in which to draw 
# - file_name	name of the file in which to read
# - merge	boolean telling if the net read is merged with the current
# - Xoff	(optional) offset in the x direction (in mm)
# - Yoff	(optional) offset in the y direction (in mm)
#
proc read_graph { win file_name merge { Xoff 0.0 } { Yoff 0.0 } } {
    global mark TransTime PlaceTime link_weight 
    global curzoomvalue object_distance
    global state Main_File_Name
    global DEBUG

    if [catch {open $file_name r} infile] {
	puts "Could not open file $file_name for reading"
	return
    }

    if !$merge {
	Clear_Graph $win
    }

    if $DEBUG {
	puts -nonewline "Reading graph from $file_name..."
    }

    # set the boolean useful in case the file is in old format
    set WARNED 0
    set old 0

    # convert offset into pixels
    set xoff [winfo pixels . ${Xoff}m]
    set yoff [winfo pixels . ${Yoff}m]

    # compute current scaling factor
    set scale [expr $curzoomvalue/100.0]

    # read file and place objects
    while { [gets $infile line] >= 0 } {
	# removing multiple whilespaces. This is normally harmless, 
	# because there are NO white spaces in (string) names.
	regsub -all {[ ]+} $line " " line
	set args [split $line]
	set type [lindex $args 0]
	switch $type {
	    Src {
		set xpos \
		    [expr $xoff+$scale*([lindex $args 2]+[lindex $args 4])/2]
		set ypos \
		    [expr $yoff+$scale*([lindex $args 3]+[lindex $args 5])/2]
		set the_name [string trim [lindex $args 6] \"]
		regsub {_WHITE_} $the_name " " $the_name
		set the_rate [lindex $args 7]
		set newid [Make_Source $win \
			       [list \
				    -x $xpos \
				    -y $ypos \
				    -name $the_name \
				    -time $the_rate \
				   ] ]
		set tr_id([lindex $args 1]) $newid
		Display_Name $win $newid
	    }
	    Tr {
		# trying to guess if old or new format
		if { -1 == [lsearch -exact $args "name" ] } {
		    set old 1
		    if !$WARNED {
			tk_dialog .warning "Read Warning" \
			    "This looks like an old format.\nYou should re-save this file." {warning} 0 {Ok, Go on}
			set WARNED 1
		    }
		}
		set xpos \
		    [expr $xoff+$scale*([lindex $args 2]+[lindex $args 4])/2]
		set ypos \
		    [expr $yoff+$scale*([lindex $args 3]+[lindex $args 5])/2]
		set the_name [string trim [lindex $args 6] \"]
		regsub {_WHITE_} $the_name " " $the_name
		if $old {
		    set the_law "name constant value [lindex $args 7]"
		} else {
		    set the_law [lrange $args 7 end]
		    regsub {\\\{} $the_law "{" the_law
		    regsub {\\\}} $the_law "}" the_law
		    set the_law [Extract_Braces $the_law]
		}
		set newid [Make_Trans $win \
			       [list \
				    -x $xpos \
				    -y $ypos \
				    -name $the_name \
				    -lawparams $the_law \
				   ] ]
		set tr_id([lindex $args 1]) $newid
	    }
	    Pl { 
		# trying to guess if old or new format
		if { -1 == [lsearch -exact $args "name"] } {
		    set old 1
		    if !$WARNED {
			tk_dialog .warning "Read Warning" \
			    "This looks like an old format.\nYou should re-save this file." {warning} 0 {Ok, Go on}
			set WARNED 1
		    }
		}
		set xpos \
		    [expr $xoff+$scale*([lindex $args 2]+[lindex $args 4])/2]
		set ypos \
		    [expr $yoff+$scale*([lindex $args 3]+[lindex $args 5])/2]
		set the_name [string trim [lindex $args 6] \"]
		regsub {_WHITE_} $the_name " " $the_name
		set the_law [lrange $args 7 end]
		regsub {\\\{} $the_law "{" the_law
		regsub {\\\}} $the_law "}" the_law
		if $old {
		    set the_law "name constant value [lindex $args 8]"
		} else {
		    set the_law [Extract_Braces $the_law]
		}
		set newid [Make_Place $win \
			       [list \
				    -x $xpos \
				    -y $ypos \
				    -name $the_name \
				    -lawparams $the_law \
				    -marking [lindex $args end] \
				   ] ]
		set pl_id([lindex $args 1]) $newid
	    }
	    TrPl -
	    PlTr {
		# trying to guess if old or new format
		set points [lrange $args 4 [llength $args]]
		if { 1 == [expr [llength $points] % 2] } {
		    if !$WARNED {
			tk_dialog .warning "Read Warning" \
			    "This looks like an old format.\nYou should re-save this file." {warning} 0 {Ok, Go on}
			set WARNED 1
		    }
		    set points [lrange $args 3 [llength $args]]
		    set weight 1
		} else {
		    set weight [lindex $args 3]
		}

		# chop off coord list, and shift it
		set coords [lrange $points 2 end]
		set idx [expr [llength $coords]-3]
		set coords [lrange $coords 0 $idx]
		set coords [Shift_Line_Coords $coords $xoff $yoff]
		set coords [Scale_Line_Coords $coords $scale]

		# create link
		if { $type == "TrPl" } {
		    set newid [Put_TraPla_Link $win \
				   $tr_id([lindex $args 1]) \
				   $pl_id([lindex $args 2]) \
				   $coords]
		} else {
		    set newid [Put_PlaTra_Link $win \
				   $pl_id([lindex $args 1]) \
				   $tr_id([lindex $args 2]) \
				   $coords]
		}

		# update attributes
		set link_weight($newid) $weight
		Display_Weight $win $newid
	    }
	}
    }
    
    # adjusting the size of the canvas
    set bbox [$win bbox TRANS PLACE LINK MARK TEXT SOURCE]
    set OD [expr $curzoomvalue/100.0*$object_distance]
    Resize_Graph $win [expr [lindex $bbox 2]+$OD] [expr [lindex $bbox 3]+$OD]

    close $infile

    # overriding the modification flag.
    if !$merge {
	set state($win) saved
	set Main_File_Name $file_name
    }

    if $DEBUG {
	puts " ...Done"
    }

}

######################################################################
# Save_Graph -
# User Inteface procedure to prompt for the name of a file into which
# to save the graph. If successful, updates the main file name 
# (visible on the display).
#
# Parameters:
# - win		the canvas to save to the file
#
proc Save_Graph { win } {
    global Main_File_Name state

    set types {
	{{Petri Net File}       {.pned}  }
	{{All Files}        *             }
    }

#     set file_name [tk_getSaveFile -defaultextension pned \
# 		       -initialfile [Make_File_Name $file_name pned] \
# 		       -filetypes $types -parent .]
    set options "-defaultextension pned -initialfile {$Main_File_Name} -filetypes {$types} -parent ."

    set file_name [GetSaveFile $options]

    if { $file_name != "" } {
	if [write_graph $win $file_name] {
	    set state($win) saved
	    set Main_File_Name $file_name
	}
    }
}

proc Browse_File { } {
    global file_name 

    set types {
	{{Petri Net File}       {.pned}  }
	{{All Files}        *             }
    }

#     set file_name [tk_getOpenFile -defaultextension pned \
# 		       -initialfile [file tail $file_name] \
# 		       -filetypes $types -parent .]
    set options "-defaultextension pned -initialfile {[file tail $file_name]} -filetypes {$types} -parent ."
    set file_name [GetOpenFile $options]

    return $file_name
		  
}


proc Browse_Sequence_File { } {
    global file_name

    set types {
	{{Sequence File}       {.teted .tet}  }
	{{All Files}        *             }
    }

#     set new_file_name [tk_getOpenFile -defaultextension teted \
# 			   -initialfile [file tail $file_name] \
# 			   -filetypes $types -parent .]
    set options "-defaultextension tet -initialfile {[file tail $file_name]} -filetypes {$types} -parent ."

    set new_file_name [GetOpenFile $options]
    
    if { $new_file_name != "" } {
	set file_name $new_file_name
    }
}

######################################################################
# Load_Graph -
# A user dialog interface for specifying a sequence file to load into
# the Petri Net editor. 
# 
# Parameters:
# - win		the canvas widget into which to load
#
proc Load_Graph { win } {
    global file_name Main_File_Name

    set file_name $Main_File_Name

    if [catch {toplevel .params}] { return }

    wm title .params "Load"

    frame .params.filename
    label .params.filename.label -text "File Name" -width 15 -anchor w
    entry .params.filename.field -textvar file_name -width 40 \
	-relief sunken

    frame .params.offset
    label .params.offset.label -text "Offset (mm)" -width 20 -anchor w
    label .params.offset.xlabel -text "X=" -width 3 -anchor w
    entry .params.offset.xoffset -textvar create_props(xoffset) \
	-width 4 -relief sunken
    label .params.offset.ylabel -text "Y=" -width 3 -anchor w
    entry .params.offset.yoffset -textvar create_props(yoffset) \
	-width 4 -relief sunken
    pack .params.offset.label .params.offset.xlabel .params.offset.xoffset \
	.params.offset.ylabel .params.offset.yoffset -side left 
    
    frame .params.actions
    button .params.actions.browse -text Browse \
	-underline 0 \
	-command "Browse_File"
    button .params.actions.ok -text "Load" \
	-underline 0 \
	-command "destroy .params\n read_graph $win \$file_name 0 \$create_props(xoffset) \$create_props(yoffset)"
    button .params.actions.merge -text "Merge" \
	-underline 0 \
	-command "destroy .params\n read_graph $win \$file_name 1 \$create_props(xoffset) \$create_props(yoffset)"
    button .params.actions.cancel -text "Cancel" \
	-underline 0 \
	-command "destroy .params"
    pack .params.actions.browse .params.actions.ok \
	.params.actions.merge .params.actions.cancel \
	-side left

    bind .params <Return> {.params.actions.ok invoke}
    bind .params <Escape> {.params.actions.cancel invoke}

    pack .params.filename.label -side left
    pack .params.filename.field -side left -expand true -fill x
    pack .params.filename -side top -expand true -fill x
    pack .params.offset -side top -fill x
    pack .params.actions -side top

    grab .params

}

######################################################################
#
# Procedures for exporting in various languages for SEDs
#
######################################################################
# write_EGL -
# Write a description of a Petri Net of the event graph family into a
# file in the EGL format. 
# The network may be saved with or without its sources, which allows to
# study closed and open networks.
# A boolean value reporting the success of the save is returned.
#
# Parameters:
# - egl_file_name	name of the file into which to write
# - with_sources	boolean telling whether sources should be exported
#
proc write_EGL { egl_file_name with_sources } {
    global trans_list place_list source_list platra_list trapla_list 
    global mark TransTime PlaceTime source_rate name number2id
    global file_name
    global DEBUG

    set file_name $egl_file_name
    set outfile [open $file_name w]

    if $DEBUG {
	puts -nonewline "Generating EGL code in file $file_name..."
    }

    if ![Check_Network 1] { return 0 }

    # renumbering all transitions
    set curtrans 1
    if { $with_sources } {
	foreach trans $source_list {
	    set number([lindex $trans 0]) $curtrans
	    set number2id($curtrans) [lindex $trans 0]
	    incr curtrans 1
	}
    } 	
    foreach trans $trans_list {
	set number([lindex $trans 0]) $curtrans
	set number2id($curtrans) [lindex $trans 0]
	incr curtrans 1
    }
    puts $outfile "Size <- [expr $curtrans-1]"

    if { $with_sources } {
	set the_list [concat $source_list $trans_list]
    } else {
	set the_list $trans_list
    }

    foreach trans $the_list {
	set trans_id [lindex $trans 0]
	puts -nonewline $outfile "$number($trans_id) "
	if { $name($trans_id) != "" } {
	    puts -nonewline $outfile "\"$name($trans_id)\" "
	}
	if [Is_Type $trans_id SOURCE] {
	    puts -nonewline $outfile "E [format %f $source_rate($trans_id)]"
	} else {
	    write_law $outfile $TransTime($trans_id)
	}

	foreach link $trapla_list {
	    if { [lindex $link 1] == $trans_id } {
		set place_id [lindex $link 2] 

		foreach link $platra_list {
		    if { [lindex $link 1 ] == $place_id } {
			set dest_id [lindex $link 2]
			puts -nonewline $outfile [format "\n\t, %d " $number($dest_id)]
			if { $name($place_id) != "" } {
			    puts -nonewline $outfile "\"$name($place_id)\" "
			}
			write_law $outfile $PlaceTime($place_id)

			puts -nonewline $outfile " $mark($place_id) "
			for { set i 0 } { $i < $mark($place_id) } { incr i 1 } {
			    puts -nonewline $outfile "0.0 "
			}
		    }
		}
	    }
	}
	puts $outfile ";"
    }

    close $outfile

    if $DEBUG {
	puts " ...Done"
    }

    return 1
}

proc gen_EGL { egl_file_name } {
    
    set types {
	{{EGL file}       {.egl}  }
	{{All Files}        *             }
    }

#     set file_name [tk_getSaveFile -defaultextension egl \
# 		       -initialfile [file tail $file_name] \
# 		       -filetypes $types -parent .]
    set options "-defaultextension egl -initialdir [file dirname $egl_file_name] -initialfile {$egl_file_name} -filetypes {$types} -parent ."

    set file_name [GetSaveFile $options]

    if { $file_name != "" } {
	# write egl code with sources (the "1" below)
	write_EGL $file_name 1
    }
}

######################################################################
# write_PNL -
# Write a description of a Petri Net of the event graph family into a
# file in the PNL format. 
# A boolean value reporting the success of the save is returned.
#
# Parameters:
# - pnl_file_name	name of the file into which to write
#
proc write_PNL { pnl_file_name } {
    global trans_list place_list platra_list trapla_list 
    global mark TransTime PlaceTime name link_weight
    global file_namep
    global DEBUG

    set file_name $pnl_file_name
    set outfile [open $file_name w]

    if $DEBUG {
	puts -nonewline "Generating PNL code in file $file_name..."
    }

    if ![Check_Network 1] { return 0 }

    # renumbering all transitions and places
    set curtrans 1
    foreach trans $trans_list {
	set number([lindex $trans 0]) $curtrans
	incr curtrans 1
    }
    set curplace 1
    foreach place $place_list {
	set number([lindex $place 0]) $curplace
	incr curplace 1
    }
    puts $outfile "Transitions <- [expr $curtrans-1]"
    puts $outfile "Places <- [expr $curplace-1]"

    foreach trans $trans_list {
	set trans_id [lindex $trans 0]
	puts -nonewline $outfile "T $number($trans_id) "
	if { $name($trans_id) != "" } {
	    puts -nonewline $outfile "\"$name($trans_id)\" "
	}
	write_law $outfile $TransTime($trans_id)

	foreach link $trapla_list {
	    if { [lindex $link 1] == $trans_id } {

		set place_id [lindex $link 2] 
		puts -nonewline $outfile [format ", %d %d " \
					      $number($place_id) \
					      $link_weight([lindex $link 0])]
	    }
	}
	puts $outfile ";"
    }

    foreach place $place_list {
	set place_id [lindex $place 0]
	puts -nonewline $outfile "P $number($place_id) "
	if { $name($place_id) != "" } {
	    puts -nonewline $outfile "\"$name($place_id)\" "
	}
	write_law $outfile $PlaceTime($place_id)

	puts -nonewline $outfile " $mark($place_id) "
	for { set i 0 } { $i < $mark($place_id) } { incr i 1 } {
	    puts -nonewline $outfile "0.0 "
	}
	foreach link $platra_list {
	    if { [lindex $link 1 ] == $place_id } {
		set dest_id [lindex $link 2]
		puts -nonewline $outfile [format ", %d %d " \
					      $number($dest_id) \
					      $link_weight([lindex $link 0])]
	    }
	}
	puts $outfile ";"
    }

    close $outfile
    puts " ...Done"
}

proc gen_PNL { pnl_file_name } {

    set types {
	{{PNL file}       {.pnl}  }
	{{All Files}        *             }
    }

#     set file_name [tk_getSaveFile -defaultextension pnl \
# 		       -initialfile [file tail $file_name] \
# 		       -filetypes $types -parent .]
    set options "-defaultextension pnl -initialdir [file dirname $pnl_file_name] -initialfile {$pnl_file_name} -filetypes {$types} -parent ."

    set output_name [GetSaveFile $options]

    if { $output_name != "" } {
	write_PNL $output_name
    }
}

######################################################################
# write_matrix -
# Procedure to write the (max,+) matrix corresponding to some event graph
# into a file. Handles the various modes and styles.
# Two possibilities are provided: whether to generate a unique matrix
# or just write the set of matrices.
#
# Parameters:
# - mat_file_name	the name of the file into which to write
# - mode	the matrix language to use
# - prefix	a string used as matrix name/prefix for symbolic languages
# - format	full or sparse
# - astar	yes or no
# - flagged_trans	a transition number used for X(0)
#
proc write_matrix { mat_file_name mode prefix { format full } { astar yes } \
			{ flagged_trans 0 } } {
    global KEEP

    switch $astar {
	"no" {
	    Write_MaxPlus_Sequence $mat_file_name $mode $prefix $format 
	}
	"yes" {
	    Write_MaxPlus_Matrix $mat_file_name $mode $prefix $format \
		$flagged_trans
	}
    }

}

######################################################################
# Write_MaxPlus_Matrix -
# Procedure to write a single (max,+) matrix corresponding to some event graph
# into a file. Handles the various modes and styles.
# This is an interface to the application eg2matrix and eg2eg for flattening.
# The matrix is generated with an "initial vector" X(0) which is useful
# for some algorithms. Normally, this is (0,...,0) but it may be 
# (eps,..,0,eps,...) if the parameter flagged_trans is not zero.
# Two possibilities are provided: whether to generate a unique matrix
# or just write the set of matrices.
#
# Parameters:
# - mat_file_name	the name of the file into which to write
# - mode	the matrix language to use
# - prefix	a string used as matrix name/prefix for symbolic languages
# - format	full or sparse
# - flagged_trans	a transition number used for X(0)
#
proc Write_MaxPlus_Matrix { mat_file_name mode prefix { format full } \
				{ flagged_trans 0 } } {
    global KEEP

    set TMP_EGL_FILE1 "/tmp/pned[pid]_1.egl"
    set TMP_EGL_FILE2 "/tmp/pned[pid]_2.egl"

    # generate EGL network
    write_EGL $TMP_EGL_FILE1 0
    # flatten it...
    exec eg2eg -flatten -f $TMP_EGL_FILE1 -o $TMP_EGL_FILE2
    switch $mode {
	bruno -
	scilab {
	    if { $flagged_trans > 0 } {
		catch [exec eg2matrix -f $TMP_EGL_FILE2 -o $mat_file_name \
			   -$mode -$format -prefix $prefix \
			   -flag_trans $flagged_trans]
	    } else {
		catch [exec eg2matrix -f $TMP_EGL_FILE2 -o $mat_file_name \
			   -$mode -$format -prefix $prefix ]
	    }
	}
	sven {
	    exec eg2matrix -f $TMP_EGL_FILE2 -o $mat_file_name -sven
	}
	default {
	    exec eg2matrix -f $TMP_EGL_FILE2 -o $mat_file_name \
		-format $format -prefix $prefix
	}
    }

    if !$KEEP {
	RemoveFile $TMP_EGL_FILE1
	RemoveFile $TMP_EGL_FILE2
    }
}

######################################################################
# Write_MaxPlus_Sequence -
# Procedure to write the sequence of (max,+) matrices corresponding to 
# some event graph into a file. Handles the various modes and styles.
# This is an interface to the application eg2matrix and eg2eg for flattening.
#
# Parameters:
# - mat_file_name	the name of the file into which to write
# - mode	the matrix language to use
# - prefix	a string used as matrix name/prefix for symbolic languages
# - format	full or sparse
#
proc Write_MaxPlus_Sequence { mat_file_name mode prefix { format full } } {
    global KEEP

    set TMP_EGL_FILE "/tmp/pned[pid].egl"

    # generate EGL network
    write_EGL $TMP_EGL_FILE 0

    # make matrices
    switch $mode {
	bruno -
	scilab {
	    set command "eg2matrix -f $TMP_EGL_FILE -o $mat_file_name -sequence -$mode -prefix \"$prefix\" -$format"
	    if [catch "exec $command" problem] {
		LogResult "Export Error" "" "$problem"
	    }
	}
	sven {
	    ReportWarning "Export Error" "Operation meaningless for this format"
	}
	default {
	    exec eg2matrix -f $TMP_EGL_FILE -o $mat_file_name \
		-prefix $prefix
	}
    }

    # cleanup
    if !$KEEP {
	RemoveFile $TMP_EGL_FILE
    }
}

proc gen_MAXPLUS { mat_file_name } {
    global gen_file_name Maxplus

    set types {
	{{MaxPlus matrix}       {.mat}  }
	{{All Files}        *           }
    }

    set options "-defaultextension mat -initialfile {$mat_file_name} -filetypes {$types} -parent ."

    set gen_file_name $mat_file_name

    if [catch {toplevel .params}] { return }

    wm title .params "Generation Parameters"

    frame .params.filename
    label .params.filename.label -text "File Name" -width 15 -anchor w
    entry .params.filename.field -textvar gen_file_name -width 40 \
	-relief sunken

    frame .params.style 
    label .params.style.label -text "Matrix language" -width 15 -anchor w
    radiobutton .params.style.standard -text "Standard" \
	-variable Maxplus(mode) -value bruno
    radiobutton .params.style.scilab -text "Scilab" \
	-variable Maxplus(mode) -value scilab
#    radiobutton .params.style.sven   -text "Sven"   -variable Maxplus(mode) -value sven

    set f [frame .params.format]
    label $f.label -text "Format" -width 15 -anchor w
    radiobutton $f.sparse -text "Sparse" \
	-variable Maxplus(matrix_format) -value sparse
    radiobutton $f.full -text "Full" \
	-variable Maxplus(matrix_format) -value full
    pack $f.label $f.sparse $f.full -side left -fill x

    frame .params.prefix
    label .params.prefix.label -text "Matrix Name" -width 15 -anchor w
    entry .params.prefix.field -textvar Maxplus(prefix) \
	-width 20 -relief sunken

    frame .params.astar
    label .params.astar.label -text "Matrices..." -width 15 -anchor w
    radiobutton .params.astar.yes -text "One (A1.A0*)" \
	-variable Maxplus(astar) -value yes
    radiobutton .params.astar.no -text "All (A0,A1,...)" \
	-variable Maxplus(astar) -value no

    frame .params.actions
    button .params.actions.browse -text Browse \
	-underline 0 \
	-command "set gen_file_name \[GetOpenFile \"$options\"\]"
    button .params.actions.ok -text "Save" \
	-underline 0 \
	-command "destroy .params\n\write_matrix \$gen_file_name \$Maxplus(mode) \$Maxplus(prefix) \$Maxplus(matrix_format) \$Maxplus(astar) 0"
    button .params.actions.cancel -text "Cancel" \
	-underline 0 \
	-command "destroy .params"
    
    pack .params.filename.label -side left -anchor w 
    pack .params.filename.field -side left -expand true -fill x
    pack .params.style.label .params.style.standard .params.style.scilab \
	-side left -fill x
    pack .params.prefix.label .params.prefix.field -side left
    pack .params.astar.label .params.astar.yes .params.astar.no \
	-side left -fill x
    pack .params.actions.browse .params.actions.ok .params.actions.cancel \
	-side left
    pack .params.filename .params.style .params.format .params.prefix \
	.params.astar \
	-side top -expand true -fill x
    pack .params.actions -side top

    bind .params <Return> {.params.actions.ok invoke}
    bind .params <Escape> {.params.actions.cancel invoke}

}

proc write_QNAP { qnap_file_name sim_length } {
    global file_name
    global KEEP

    set file_name $qnap_file_name

    set TMP_EGL_FILE "/tmp/pned[pid].egl"

    write_EGL $TMP_EGL_FILE 1
    exec eg2qnap -f $TMP_EGL_FILE -o $file_name -nevt $sim_length

    if !$KEEP {
	RemoveFile $TMP_EGL_FILE
    }
}

proc gen_QNAP { new_file_name } {
    global gen_file_name

    set types {
	{{QNAP program}       {.qnp}  }
	{{All Files}        *           }
    }

    set options "-defaultextension mat -initialfile {$new_file_name} -filetypes {$types} -parent ."

    set gen_file_name $new_file_name

    if [catch {toplevel .params}] { return }

    wm title .params "Generation Parameters"

    frame .params.filename
    label .params.filename.label -text "File Name" -width 15 -anchor w
    entry .params.filename.field -textvar gen_file_name -width 40 \
	-relief sunken

    frame .params.simlen
    label .params.simlen.label -text "Simulation Length" -width 15 -anchor w
    entry .params.simlen.field -textvar sim_length -width 10 \
	-relief sunken
    
    frame .params.actions
    button .params.actions.browse -text Browse \
	-underline 0 \
	-command "set gen_file_name \[GetOpenFile \"$options\"\]"
    button .params.actions.ok -text "Save" \
	-underline 0 \
	-command "destroy .params\n\write_QNAP \$gen_file_name \$sim_length"
    button .params.actions.cancel -text "Cancel" \
	-underline 0 \
	-command "destroy .params"
    
    pack .params.filename.label -side left 
    pack .params.filename.field -side left -expand true -fill x 
    pack .params.simlen.label .params.simlen.field -side left
    pack .params.actions.browse .params.actions.ok .params.actions.cancel \
	-side left
    pack .params.filename .params.simlen -side top -fill x
    pack .params.actions -side top

}

proc gen_postscript { win ps_file_name } {

    set types {
	{{PostScript files}       {.ps .eps .epsi}  }
	{{All Files}        *           }
    }

    set options "-defaultextension ps -initialfile {$ps_file_name} -filetypes {$types} -parent ."

    set gen_file_name [GetSaveFile $options]

    if { $gen_file_name != "" } {
	$win postscript -file $gen_file_name
    }
}

######################################################################
# Export_Graph -
# The general dispatching procedure for exporting the Petri Net
# in whatever is not its native format "pned".
# Before calling the proper procedure, it fabricates a tentative file
# name from the current one using a type-specific extension.
#
# Parameters:
# - type	a string representing the type of object to obtain
# - extension	the extension to be given to the file name
#
proc Export_Graph { win type extension } {
    global Main_File_Name

    set name [Make_File_Name $Main_File_Name $extension]

    if { $type == "EGL" } {
	if [Is_Event_Graph] {
	    gen_EGL $name
	} else {
	    ReportError "Export Error" "This Petri Net is not an Event Graph."
	}
    } elseif { $type == "PNL" } {
	gen_PNL $name
    } elseif { $type == "QNAP" } {
	if [Is_Event_Graph] {
	    gen_QNAP $name
	} else {
	    ReportError "Export Error" "This Petri Net is not an Event Graph."
	}
    } elseif { $type == "MAXPLUS" } {
	if [Is_Event_Graph] {
	    if [Is_Deterministic] {
		gen_MAXPLUS $name
	    } else {
		ReportError "Export Error" "This Petri Net is not Deterministic."
	    }
	} else {
	    ReportError "Export Error" "This Petri Net is not an Event Graph."
	}
    } elseif { $type == "PS" } {
	gen_postscript $win $name
    } {
	ReportError "Export Error" "Exporting in $type not available yet. Sorry!"
    }
}    
