#
# "methods" attached to pieces.
# A piece is originally a list of squares.
# A normalized piece is a pair low profile / high profile
#
######################################################################
# low_point -
# Function returning the lowest point of a NON NORMALIZED piece.
#
# Parameters:
# - piece	the piece as a list of squares, i.e. [id x y]
#
proc low_point { piece } {

    set low 100000
    foreach elem $piece {
	set y [lindex $elem 2]
	if { $low > $y } { set low $y }
    }

    return $low
}

######################################################################
# high_point -
# Function returning the highest point of a NORMALIZED piece.
#
# Parameters:
# - piece	the piece as a list of hi/lo profiles.
#
proc high_point { piece } {
    global cur_width
    global DEBUG

    if $DEBUG {
	puts "Looking for altitude of $piece"
    }

    array set hi_prof [lindex $piece 1]
    set high 0
    for { set i 0 } { $i < $cur_width } { incr i 1 } {
	if { [string compare $hi_prof($i) eps] != 0 } {
	    if { $high < $hi_prof($i) } {
		set high $hi_prof($i)
	    }
	}
    }

    if $DEBUG {
	puts "...found $high"
    }

    return $high
}

######################################################################
# Bounding_Box -
# Function returning the bounding box of a NON NORMALIZED piece.
# The result is a list of 4 elements: the NW and SE corners of the smallest
# box containing the piece, in the coordinates of the store.
#
# Parameters:
# - piece	the piece as a list squares [id x y]
#
proc Bounding_Box { piece } {

    set xmin 10000
    set xmax 0
    set ymin 10000
    set ymax 0

    foreach elem $piece {
	set x [lindex $elem 1]
	set y [lindex $elem 2]
	if { $xmin > $x } { set xmin $x }
	if { $xmax < $x } { set xmax $x }
	if { $ymin > $y } { set ymin $y }
	if { $ymax < $y } { set ymax $y }
    }

    return [list $xmin $ymin [expr 1+$xmax] [expr 1+$ymax]]
}

######################################################################
# Norm_Bounding_Box -
# Function returning the bounding box of a NORMALIZED piece.
# The result is a list of 4 elements: the NW and SE corners of the smallest
# box containing the piece, in the coordinates of the store.
#
# Parameters:
# - piece	the piece as a list of hi/lo profiles.
#
proc Norm_Bounding_Box { piece } {

    array set lo_prof [lindex $piece 0]
    array set hi_prof [lindex $piece 1]
    set width [llength [array names lo_prof]]

    for { set xmin 0 } \
	{ [expr { $xmin < $width } && { 0 == [string compare $lo_prof($xmin) eps] } ] } \
	{ incr xmin 1 } {
	    # nil
	}

    for { set xmax [expr $width-1] } \
	{ [expr { $xmax >= 0 } && { 0 == [string compare $lo_prof($xmax) eps] } ] } \
	{ incr xmax -1 } {
	    # nil aussi
	}

    set ymin $lo_prof($xmin)
    set ymax $lo_prof($xmin)

    for { set i $xmin } { $i <= $xmax } { incr i 1 } {
	set y0 $lo_prof($i)
	set y1 $hi_prof($i)
	if { $ymin > $y0 } { set ymin $y0 }
	if { $ymax < $y1 } { set ymax $y1 }
    }

    return [list $xmin $ymin [expr 1+$xmax] [expr 1+$ymax]]
}

######################################################################
# Area =
# Function returning the surface (number of squares) of a NORMALIZED piece.
#
# Parameters:
# - piece	the piece as a list of hi/lo profiles.
#
proc Area { piece } {
    global cur_width

    array set lo_prof [lindex $piece 0]
    array set hi_prof [lindex $piece 1]

    set area 0

    for { set i 0 } { $i < $cur_width } { incr i 1 } {
	if { 0 != [string compare $lo_prof($i) eps] } {
	    set area [expr $area+$hi_prof($i)-$lo_prof($i)]
	}
    }

    return $area
}

######################################################################
# Normalize_Piece -
# This procedure takes a set of square locations and returns a 
# Tetris piece in the form of a lower/upper profile pair.
# The square locations are triples [id x y] where x,y are integers.
#
# Parameters:
# - newpiece	the list of squares occupied by the piece
#
proc Normalize_Piece { newpiece } {
    global cur_width 
    global DEBUG

    # initialize profiles to "empty" (epsilon)
    for { set i 0 } { $i < $cur_width } { incr i 1 } {
	set low_prof($i) eps
	set hi_prof($i) eps
    }

    # fill in profiles, keeping lowest point in memory for normalization
    set low eps
    foreach elem $newpiece {
	# eliminating "empty" squares
	if { [llength $elem] == 3 } {
	    set col [lindex $elem 1]
	    set height [lindex $elem 2]

	    if [expr { [string compare $low_prof($col) eps] == 0 } \
		    || { $low_prof($col) > $height } ] {
		set low_prof($col) $height
		if [expr { [string compare $low eps] == 0 } \
			|| { $height < $low } ] {
		    set low $height
		}
	    }
	    if [expr { [string compare $hi_prof($col) eps] == 0 } \
		    || { $hi_prof($col) < [expr $height+1] } ] {
		set hi_prof($col) [expr $height+1]
	    }
	    if $DEBUG {
		puts "c $col / h $height \[l/h = $low_prof($col)/$hi_prof($col)\]"
	    }
	}
    }

    # normalize profiles
    set low [expr -$low]
    if $DEBUG {
	puts "Adding $low to all heights"
    }
    for { set i 0 } { $i < $cur_width } { incr i 1 } {
	if { [string compare $low_prof($i) eps] != 0 } {
	    incr low_prof($i) $low
	}
	if { [string compare $hi_prof($i) eps] != 0 } {
	    incr hi_prof($i) $low
	}
    }
    
    set result [list [array get low_prof] [array get hi_prof]]
    if $DEBUG {
	puts "Normalisee = $result"
    }
    return $result
}

######################################################################
# New_Piece -
# Procedure preparing the piece editing area for a new piece.
#
# Parameters: none
#
proc New_Piece { } {
    global cur_piece_id cur_piece_nb ison
    global cur_piece_props default_piece_props

    # resetting piece default
    array set cur_piece_props [array get default_piece_props]

    set cur_piece_id [expr $cur_piece_nb+1]
    set cur_piece_props(id) $cur_piece_id

    if [info exists ison] {
	unset ison
    }

    Empty_Graph .piece_edit.graph
}
