#
# The I/O module of tged
#
# This module contains the procedures to read and write files
# - save/load in the graph format
# - save in EGL, PNL etc
#
######################################################################
# write_graph -
# The procedure that writes a Task graph into a file.
#
# Parameters:
# - win		the canvas containing the graph to save
# - file_name	name of the file in which to write
#
proc write_graph { win file_name } {
    global task_list dep_link_list prog_link_list 
    global name TaskTime LinkTime link_lag
    global curzoomvalue
    global DEBUG

    set outfile [open $file_name w]

    if $DEBUG {
	puts -nonewline "Saving graph in file $file_name..."
    }

    # computing the correction factor
    set scale [expr 100.0/$curzoomvalue]
    # renumbering all tasks 
    set curtask 1
    foreach task $task_list {
	set ta_number([lindex $task 0]) $curtask
	incr curtask 1
    }

    foreach task $task_list {
	set task_id [lindex $task 0]
	puts -nonewline $outfile "Task $ta_number($task_id) "
	puts -nonewline $outfile \
	    [Scale_Line_Coords [$win coords $task_id] $scale]
	regsub -all {[ ]} $name($task_id) "_WHITE_" the_name
	puts -nonewline $outfile " \"$the_name\""
	puts $outfile " \{ $TaskTime($task_id) \}"
    }

    foreach link $dep_link_list {
	set link_id [lindex $link 0]
	set orig_id [lindex $link 1]
	set dest_id [lindex $link 2]
	puts -nonewline $outfile "Dep "
	puts -nonewline $outfile "$ta_number($orig_id) $ta_number($dest_id) "
	puts -nonewline $outfile \
	    [Scale_Line_Coords [$win coords $link_id] $scale]
	puts $outfile " \{ $LinkTime($link_id) \}"
    }
    foreach link $prog_link_list {
	set link_id [lindex $link 0]
	set orig_id [lindex $link 1]
	set dest_id [lindex $link 2]
	puts -nonewline $outfile "Prog "
	puts -nonewline $outfile "$ta_number($orig_id) $ta_number($dest_id)"
	puts -nonewline $outfile " $link_lag($link_id) "
	puts $outfile \
	    [Scale_Line_Coords [$win coords $link_id] $scale]
    }

    close $outfile
    if $DEBUG {
	puts " ...Done"
    }
}

######################################################################
# read_graph -
# The procedure that reads a Task Graph from a file. It draws the objects
# on the canvas, and instantiates the attributes.
#
# Parameters:
# - win		the canvas in which to draw 
# - file_name	name of the file in which to read
# - merge	boolean telling if the net read is merged with the current
# - Xoff	(optional) offset in the x direction (in mm)
# - Yoff	(optional) offset in the y direction (in mm)
#
proc read_graph { win file_name { merge 0 } { Xoff 0.0 } { Yoff 0.0 } } {
    global LinkTime link_lag TaskTime name link_width 
    global curzoomvalue object_distance
    global prog_link_list dep_link_list 
    global state
    global DEBUG

    if [catch {open $file_name r} infile] {
	puts "Could not open file $file_name for reading"
	return 
    }
	
    if !$merge {
	Clear_Graph $win
    }

    if $DEBUG {
	puts -nonewline "Reading graph from $file_name..."
    }

    # convert offset into pixels
    set xoff [winfo pixels . ${Xoff}m]
    set yoff [winfo pixels . ${Yoff}m]

    # compute current scaling factor
    set scale [expr $curzoomvalue/100.0]

    while { [gets $infile line] >= 0 } {
	# removing multiple whilespaces. This is normally harmless, 
	# because there are NO white spaces in (string) names.
	regsub -all {[ ]+} $line " " line
	set args [split $line]
	set type [lindex $args 0]
	switch $type {
	    Task { 
		set xpos \
		    [expr $xoff+$scale*([lindex $args 2]+[lindex $args 4])/2]
		set ypos \
		    [expr $yoff+$scale*([lindex $args 3]+[lindex $args 5])/2]
		set the_name [string trim [lindex $args 6] \"]
		regsub {_WHITE_} $the_name " " $the_name
		set the_law [lrange $args 7 end]
		regsub {\\\{} $the_law "{" the_law
		regsub {\\\}} $the_law "}" the_law
		set the_law [Extract_Braces $the_law]
		set newid [Make_Task $win \
			       [list \
				    -x $xpos \
				    -y $ypos \
				    -name $the_name \
				    -lawparams $the_law \
				   ] ]
		set ta_id([lindex $args 1]) $newid
	    }
	    Dep {
		set idx [lsearch $args \{]
		set coords [lrange $args 3 [expr $idx - 1 ]]
		# puts "CO -> $coords"
		set coords [Shift_Line_Coords $coords $xoff $yoff]
		set coords [Scale_Line_Coords $coords $scale]

		set tlist [list LINK DEP]
		set newid [eval {$win create line} $coords \
			       -tag {$tlist} -arrow last \
			       -width $link_width(DEP) ]
		set the_law [lrange $args $idx end]
		regsub {\\\{} $the_law "{" the_law
		regsub {\\\}} $the_law "}" the_law
		set the_law [Extract_Braces $the_law]
		set newlink [list $newid $ta_id([lindex $args 1]) \
				 $ta_id([lindex $args 2])]
		set LinkTime($newid) $the_law
		set dep_link_list [lappend dep_link_list $newlink]

		Display_Timing $win $newid
	    }
	    Prog { 
		set coords [lrange $args 4 [llength $args]]
		set coords [Shift_Line_Coords $coords $xoff $yoff]
		set coords [Scale_Line_Coords $coords $scale]

		set tlist [list LINK PROG]
		set newid [eval {$win create line} $coords \
			       -tag {$tlist} -arrow last \
			       -width $link_width(PROG) ]
		set link_lag($newid) [lindex $args 3]
		set newlink [list $newid $ta_id([lindex $args 1]) \
				 $ta_id([lindex $args 2])]
		set prog_link_list [lappend prog_link_list $newlink]
	    }
	}
    }

    # adjusting the size of the canvas
    set bbox [$win bbox TASK LINK]
    set OD [expr $curzoomvalue/100.0*$object_distance]
    Resize_Graph $win [expr [lindex $bbox 2]+$OD] [expr [lindex $bbox 3]+$OD]
    
    # overriding the modification flag.
    if !$merge {
	set state($win) saved
    }

    close $infile
    if $DEBUG {
	puts " ...Done"
    }
}

######################################################################
# Save_Graph - 
# User dialog interface for choosing a file name in which to save the
# graph, and call the output procedure.
#
# Parameters:
# - win		the graph canvas to be saved
#
proc Save_Graph { win } {
    global file_name state

    set types {
	{{Task Graph File}       {.tged}  }
	{{All Files}        *             }
    }

#     set file_name [tk_getSaveFile -defaultextension tged \
# 		       -initialfile [file tail $file_name] \
# 		       -filetypes $types -parent .]
    set options "-defaultextension tged -initialfile {[file tail $file_name]} -filetypes {$types} -parent ."

    set file_name [GetSaveFile $options]

    if { $file_name != "" } {
	write_graph $win $file_name
	set state($win) saved
    }

}

######################################################################
# Save_Graph - 
# User dialog interface for choosing a file name to load into the editor,
# and call the input procedure.
#
# Parameters:
# - win		the graph canvas in which to load the file
#
proc Load_Graph { win } {
    global file_name

    set types {
	{{Task Graph File}       {.tged}  }
	{{All Files}        *             }
    }

#     set file_name [tk_getOpenFile -defaultextension tged \
# 		       -initialfile [file tail $file_name] \
# 		       -filetypes $types -parent .]
    set options "-defaultextension tged -initialfile {[file tail $file_name]} -filetypes {$types} -parent ."

    set file_name [GetOpenFile $options]

    if { $file_name != "" } {
	read_graph $win $file_name
    }

}

######################################################################
#
# Procedures for exporting in various languages for SEDs
#
######################################################################
# Write_Law_For_TGL -
# Procedure to write a distribution using the convention of TGL.
#
# Parameters:
# - outfile	the stream into which to write
# - lawdesc	the list of parameters of the distribution
#
proc Write_Law_For_TGL { outfile lawdesc } {

    array set TheLaw $lawdesc

    switch $TheLaw(name) {
	bernoulli {
	    puts -nonewline $outfile "BER( $TheLaw(proba) )"
	}
	constant {
	    puts -nonewline $outfile "CST( $TheLaw(value) )"
	}
	erlang {
	    puts -nonewline $outfile "ERL( $TheLaw(phases), $TheLaw(mean) )"
	}
	exponential {
	    puts -nonewline $outfile "EXP( $TheLaw(mean) )"
	}
	geometric {
	    puts -nonewline $outfile "GEOM( $TheLaw(proba) )"
	}
	pareto {
	    puts -nonewline $outfile "PARETO( $TheLaw(alpha), $TheLaw(beta) )"
	}
	uniform {
	    puts -nonewline $outfile "UNIF( $TheLaw(low), $TheLaw(high) )"
	}
	default {
	    ReportError "Export Error" "Don't know distribution $TheLaw(name)"
	}
    }
}

######################################################################
# write_TGL -
# Procedure to write the task graph in a file in the TGL language.
#
# Parameters:
# - file_name	the name of the file in which to write
#
proc write_TGL { file_name } {
    global dep_link_list prog_link_list task_list 
    global LinkTime link_lag TaskTime name
    global DEBUG

    set outfile [open $file_name w]

    if $DEBUG {
	puts -nonewline "Generating TGL code..."
    }

    # renumbering all tasks 
    set curtask 1
    foreach task $task_list {
	set ta_number([lindex $task 0]) $curtask
	incr curtask 1
    }

    foreach task $task_list {
	set task_id [lindex $task 0]
	puts -nonewline $outfile "TASK task.$task_id"
	regsub -all {[ ]} $name($task_id) "_WHITE_" the_name
	puts $outfile " $the_name"
	puts -nonewline $outfile "  DURATION "
	Write_Law_For_TGL $outfile $TaskTime($task_id)
	puts $outfile "\nEND\n"
    }

    foreach link $dep_link_list {
	set link_id [lindex $link 0]
	set orig_id [lindex $link 1]
	set dest_id [lindex $link 2]
	puts $outfile "LINK link.$link_id"
	puts $outfile "  TYPE DATA"
	puts $outfile "  IS NECESSARY"
	puts $outfile "  FROM task.$orig_id"
	puts $outfile "  TO   task.$dest_id"
	puts -nonewline $outfile "  DURATION "
	Write_Law_For_TGL $outfile $LinkTime($link_id)
	puts $outfile "\nEND\n"
    }
    foreach link $prog_link_list {
	set link_id [lindex $link 0]
	set orig_id [lindex $link 1]
	set dest_id [lindex $link 2]
	puts $outfile "LINK link.$link_id"
	puts $outfile "  TYPE PROGRAM"
	puts $outfile "  IS NECESSARY"
	puts $outfile "  FROM task.$orig_id"
	puts $outfile "  TO   task.$dest_id"
	puts $outfile "  DEPTH $link_lag($link_id)"
	puts $outfile "END\n"
    }

    close $outfile
    if $DEBUG {
	puts " ...Done"
    }
}

proc gen_TGL { new_file_name } {
    global file_name

    set types {
	{{TGL File}       {.tgl}  }
	{{All Files}        *      }
    }

#     set gen_file_name [tk_getSaveFile -defaultextension tgl \
# 			   -initialfile [file tail $gen_file_name] \
# 		       -filetypes $types -parent .]
    set options "-defaultextension tgl -initialfile {[Make_File_Name $file_name tgl]} -filetypes {$types} -parent ."

    set gen_file_name [GetSaveFile $options]

    if { $gen_file_name != "" } {
	write_TGL $gen_file_name
    }
}

proc write_EGL { file_name } {
    global KEEP

    set TMP_TGL_FILE "/tmp/tged[pid].tgl"

    write_TGL $TMP_TGL_FILE
    exec tg2eg -f $TMP_TGL_FILE -o $file_name

    if !$KEEP {
	RemoveFile $TMP_TGL_FILE
    }
}

proc gen_EGL { new_file_name } {
    global file_name

    set types {
	{{EGL File}       {.egl}  }
	{{All Files}        *      }
    }

#     set gen_file_name [tk_getSaveFile -defaultextension tgl \
# 			   -initialfile [file tail $gen_file_name] \
# 		       -filetypes $types -parent .]
    set options "-defaultextension egl -initialfile {[Make_File_Name $file_name egl]} -filetypes {$types} -parent ."

    set gen_file_name [GetSaveFile $options]

    if { $gen_file_name != "" } {
	write_EGL $gen_file_name
    }
    
}

proc write_PNL { file_name } {
    global KEEP DEBUG

    if $DEBUG {
	puts -nonewline "Generating PNL code..."
    }

    set TMP_TGL_FILE "/tmp/tged[pid].tgl"

    write_TGL $TMP_TGL_FILE
    exec tg2eg -f $TMP_TGL_FILE | eg2pn -o $file_name

    if !$KEEP {
	RemoveFile $TMP_TGL_FILE
    }

    if $DEBUG {
	puts " ...Done"
    }
}

proc gen_PNL { new_file_name } {

    global gen_file_name

    set gen_file_name $new_file_name
    
    if [catch {toplevel .params}] { return }

    wm title .params "Generation Parameters"

    frame .params.filename
    label .params.filename.label -text "File Name" -width 15 -anchor w
    entry .params.filename.field -textvar gen_file_name -width 20 \
	-relief sunken
    
    frame .params.actions
    button .params.actions.ok -text "Ok" \
	-command "destroy .params\n\write_PNL \$gen_file_name"
    button .params.actions.cancel -text "Cancel" \
	-command "destroy .params"
    
    pack .params.filename.label .params.filename.field -side left
    pack .params.actions.ok .params.actions.cancel -side left
    pack .params.filename -side top -fill x
    pack .params.actions -side top

}

proc write_matrix { file_name mode flagged_trans} {
    global KEEP

    set TMP_TGL_FILE  "/tmp/tged[pid].tgl"
    set TMP_EGL_FILE1 "/tmp/tged[pid]_1.egl"
    set TMP_EGL_FILE2 "/tmp/tged[pid]_2.egl"

    # generate Event Graph
    write_TGL $TMP_TGL_FILE
    exec tg2eg -f $TMP_TGL_FILE -o $TMP_EGL_FILE1
    # flatten it...
    exec eg2eg -flatten -f $TMP_EGL_FILE1 -o $TMP_EGL_FILE2
    switch $mode {
	bruno {
	    if { $flagged_trans > 0 } {
		catch [exec eg2matrix -f $TMP_EGL_FILE2 -o $file_name -bruno\
			  -flag_trans $flagged_trans]
	    } else {
		catch [exec eg2matrix -f $TMP_EGL_FILE2 -o $file_name -bruno]
	    }
	}
	sven {
	    exec eg2matrix -f $TMP_EGL_FILE2 -o $file_name -sven
	}
	default {
	    exec eg2matrix -f $TMP_EGL_FILE2 -o $file_name
	}
    }

    if !$KEEP {
	RemoveFile $TMP_TGL_FILE
	RemoveFile $TMP_EGL_FILE1
	RemoveFile $TMP_EGL_FILE2
    }
}

proc gen_MAXPLUS { new_file_name } {
    global gen_file_name mode

    set gen_file_name $new_file_name

    if [catch {toplevel .params}] { return }

    wm title .params "Generation Parameters"

    frame .params.filename
    label .params.filename.label -text "File Name" -width 15 -anchor w
    entry .params.filename.field -textvar gen_file_name -width 20 \
	-relief sunken

    frame .params.style 
    label .params.style.label -text "A la mode de" -width 15 -anchor w
    set mode bruno
    radiobutton .params.style.bruno -text "Bruno" -variable mode -value bruno
    radiobutton .params.style.alain -text "Alain" -variable mode -value alain
    radiobutton .params.style.sven  -text "Sven" -variable mode  -value sven

    frame .params.actions
    button .params.actions.ok -text "Ok" \
	-command "destroy .params\n\write_matrix \$gen_file_name \$mode 0"
    button .params.actions.cancel -text "Cancel" \
	-command "destroy .params"
    
    pack .params.filename.label .params.filename.field -side left
    pack .params.style.label .params.style.bruno .params.style.alain \
	.params.style.sven -side left -fill x
    pack .params.actions.ok .params.actions.cancel -side left
    pack .params.filename .params.style -side top -fill x
    pack .params.actions -side top

}

proc write_QNAP { file_name sim_length } {
    global KEEP

    set TMP_TGL_FILE "/tmp/tged[pid].tgl"

    write_TGL $TMP_TGL_FILE
    exec tg2qnap -f $TMP_TGL_FILE -o $file_name -tmax $sim_length

    if !$KEEP {
	RemoveFile $TMP_TGL_FILE
    }
}

proc gen_QNAP { new_file_name } {

    global gen_file_name

    set gen_file_name $new_file_name

    if [catch {toplevel .params}] { return }

    wm title .params "Generation Parameters"

    frame .params.filename
    label .params.filename.label -text "File Name" -width 15 -anchor w
    entry .params.filename.field -textvar gen_file_name -width 20 \
	-relief sunken

    frame .params.simlen
    label .params.simlen.label -text "Simulation Length" -width 15 -anchor w
    entry .params.simlen.field -textvar sim_length -width 10 \
	-relief sunken
    
    frame .params.actions
    button .params.actions.ok -text "Ok" \
	-command "destroy .params\n\write_QNAP \$gen_file_name \$sim_length"
    button .params.actions.cancel -text "Cancel" \
	-command "destroy .params"
    
    pack .params.filename.label .params.filename.field -side left
    pack .params.simlen.label .params.simlen.field -side left
    pack .params.actions.ok .params.actions.cancel -side left
    pack .params.filename .params.simlen -side top -fill x
    pack .params.actions -side top

}

proc gen_postscript { win new_file_name } {

    global gen_file_name

    set gen_file_name $new_file_name 

    if [catch {toplevel .params}] { return }

    wm title .params "Postscript Parameters"

    frame .params.filename
    label .params.filename.label -text "File Name" -width 15 -anchor w
    entry .params.filename.field -textvar gen_file_name -width 20 \
	-relief sunken
    
    frame .params.actions
    button .params.actions.ok -text "Ok" \
	-command "destroy .params\n\$win postscript -file \$gen_file_name"
    button .params.actions.cancel -text "Cancel" \
	-command "destroy .params"
    
    pack .params.filename.label .params.filename.field -side left
    pack .params.actions.ok .params.actions.cancel -side left
    pack .params.filename -side top -fill x
    pack .params.actions -side top

}

######################################################################
# Export_Graph -
# The general dispatching procedure for exporting the task graph
# in whatever is not its native format "tged".
# Before calling the proper procedure, it fabricates a tentative file
# name from the current one using a type-specific extension.
#
# Parameters:
# - type	a string representing the type of object to obtain
# - extension	the extension to be given to the file name
#
proc Export_Graph { win type extension } {
    global gen_file_name

    set name [Make_File_Name $gen_file_name $extension]

    if { $type == "TGL" } {
	gen_TGL $name
    } elseif { $type == "EGL" } {
	gen_EGL $name
    } elseif { $type == "PNL" } {
	gen_PNL $name
    } elseif { $type == "QNAP" } {
	gen_QNAP $name
    } elseif { $type == "MAXPLUS" } {
	gen_MAXPLUS $name
    } elseif { $type == "PS" } {
	gen_postscript $name
    } {
	ReportError "Export Error" "Exporting in $type not available yet. Sorry!"
    }
}    
