#
# Module for the management of edit/select/delete modes (and others)
# 
######################################################################
# Enter_Mode -
# Procedure handling the start of a mode: text written in the help mode
# area, appearance of buttons, cursors and so on.
# If the current mode is not "none", then it is exited before by a call
# to Exit_Mode.
#
# Parameters:
# - mode	the new mode in which to enter
#
proc Enter_Mode { mode } {
    global cur_mode Information Mode img_dir link_type
    global DEBUG

    if $DEBUG {
	puts "Entering mode $mode"
    }

    # win not passed as parameter for now. But might be one day...
    set win .graph.canvas

    if { 0 != [string compare $cur_mode "none"] } {
	Exit_Mode $cur_mode
    }

    # operations common to all changes of modes
    if [winfo exists .edit.$mode] {
	.edit.$mode configure -relief sunken
    }
    # default rebinding.
    bind $win <3> "Exit_Mode $mode"
    bind $win <2> "Exit_Mode $mode"
    bind $win <Escape> "Exit_Mode $mode"

    # operations that are mode-specific.
    switch $mode {
	task {
	    $win configure -cursor "@$img_dir/task.xbm black"
	    bind $win <1> { Put_Task %W [%W canvasx %x] [%W canvasy %y] }
	    set Mode "Task"
	    set Information "\[l\] Put Task \[c\] Quit mode \[r\] Quit mode"
	}
	deplink {
	    $win configure -cursor ""
	    set link_type DEP
	    bind $win <1> { Start_Link %W [%W canvasx %x] [%W canvasy %y] }
	    bind $win <2> { Abort_Link %W }
	    set Mode "Dependence Link"
	    set Information "\[l\] Start/Stop Link \[c\] Abort Link \[r\] Quit mode"
	}
	proglink {
	    $win configure -cursor ""
	    set link_type PROG
	    bind $win <1> { Start_Link %W [%W canvasx %x] [%W canvasy %y] }
	    bind $win <2> { Abort_Link %W }
	    set Mode "Program Link"
	    set Information "\[l\] Start/Stop Link \[c\] Abort Link \[r\] Quit mode"
	}
	move {
	    $win configure -cursor target
	    bind $win <1> { Pick_Object_Or_Zone %W } 
	    bind $win <2> { Drop_Object_Or_Zone %W [%W canvasx %x] [%W canvasy %y] } 
	    bind $win <Right> { Move_Object_Or_Zone %W 5 0}
	    bind $win <Left>  { Move_Object_Or_Zone %W -5 0}
	    bind $win <Up>    { Move_Object_Or_Zone %W 0 -5}
	    bind $win <Down>  { Move_Object_Or_Zone %W 0 5}
	    bind $win <Shift-Right> { Move_Object_Or_Zone %W 20 0}
	    bind $win <Shift-Left>  { Move_Object_Or_Zone %W -20 0}
	    bind $win <Shift-Up>    { Move_Object_Or_Zone %W 0 -20}
	    bind $win <Shift-Down>  { Move_Object_Or_Zone %W 0 20}
	    bind $win <Control-Right> { Move_Object_Or_Zone %W 1 0}
	    bind $win <Control-Left>  { Move_Object_Or_Zone %W -1 0}
	    bind $win <Control-Up>    { Move_Object_Or_Zone %W 0 -1}
	    bind $win <Control-Down>  { Move_Object_Or_Zone %W 0 1}

	    set Mode "Move"
	    set Information "\[l\] Pick \[c\] Put \[r\] Quit mode"
	}
	copy {
	    $win configure -cursor target
	    bind $win <1> { Pick_Object_Or_Zone %W } 
	    bind $win <2> { Copy_Object_Or_Zone %W [%W canvasx %x] [%W canvasy %y] } 
	    set Mode "Copy"
	    set Information "\[l\] Pick \[c\] Put \[r\] Quit mode"
	}
	del {
	    $win configure -cursor pirate
	    bind $win <1> { Delete_Current_Object %W }
	    set Mode "Delete"
	    set Information "\[l\] Delete \[c\] Quit mode \[r\] Quit mode"
	}
	sel {
	    .graph configure -cursor target
	    bind $win <1> { Select_Current_Object %W }
	    bind $win <2> { Select_Zone %W [%W canvasx %x] [%W canvasy %y]}
	    set Mode "Select"
	    set Information "\[l\] Select Object \[c\] Select Zone \[r\] Quit mode"
	}
	info {
	    $win configure -cursor question_arrow
	    bind $win <1> { Edit_Current_Object_Attributes %W %X %Y }
	    set Information "\[l\] Inspect Object Attributes \[c\] Quit mode \[r\] Quit mode"
	}
	default {
	    ReportError "Mode Error" "unregistered mode: $mode"
	    return
	}
    }

    set cur_mode $mode
}

######################################################################
# Exit_Mode -
# Procedure handling the exit of a mode: text written in the help mode
# area, appearance of buttons, cursors and so on.
#
# Parameters:
# - mode	the mode to exit.
#
proc Exit_Mode { mode } {
    global cur_mode Information Mode curlink
    global cur_link cur_sel_zone cur_picked_obj
    global DEBUG

    # win not passed as parameter for now. But might be one day...
    set win .graph.canvas

    if $DEBUG {
	puts "Exiting mode $mode"
    }

    # operations common to all modes
    Reset_Standard_Bindings $win
    set Information ""
    set Mode ""
    .edit.$mode configure -relief raised

    # operations that are mode-specific.
    switch $mode {
	deplink -
	proglink {
	    if [info exists curlink] {
		$win delete $curlink
	    }
	}
	copy -
	move {
	    if [info exists cur_picked_obj] {
		$win itemconfigure $cur_picked_obj -outline black
		unset cur_picked_obj
	    }
	    $win delete [$win find withtag SELZONE]
	}
	sel { 
	    if { [llength $cur_sel_zone] != 4 } {
		set cur_sel_zone {}
	    }
	}
    }
   
    # finish
    set cur_mode none
}
