#
# Quantitative analysis module for tged
#
# Methods:
# - deterministic (max,+) methods
#    + cycle time (via EGL)
#    + critical paths
# - simulation
#
######################################################################
# Cycle_Time -
# Interface procedure to scscyc. It is applicable when the Task Graph
# is equivalent to a deterministic Event Graph, that is, possesses 
# program dependency links.
#
# Parameters: none
#
proc Cycle_Time { } {
    global KEEP

    set TMP_TGL_FILE "/tmp/tged[pid].tgl"
    set TMP_EGL_FILE "/tmp/tged[pid].egl"
    set TMP_RES_FILE "/tmp/tged[pid].res"
    set TMP_MAT_FILE "/tmp/tged[pid].mat"

    if ![Has_Program_Links] {
	ReportError "Analysis Error" "Not a repetitive Task Graph."
	return
    } elseif ![Is_Deterministic] {
	ReportError "Analysis Error" "Not a deterministic Task Graph."
	return
    }

    write_TGL $TMP_TGL_FILE
    exec tg2eg -f $TMP_TGL_FILE | eg2eg -flatten | \
	eg2matrix -bruno -o $TMP_MAT_FILE
    
    if [catch {exec scscyc -f $TMP_MAT_FILE -precision 1000 \
		   -o $TMP_RES_FILE} message] {
	LogResult "Execution error/warning messages:\n$message"
    }
    # Even if an error has been encountered, we have not exited yet.
    # This is because some errors of scscyc are just warnings, and
    # a result may have been produced anyway.
    # So proceed with displaying the result... if any.
    if [catch {set resfile [open $TMP_RES_FILE r]}] { return }

    # There is indeed. Displaying the result 
    set result [read $resfile]
    close $resfile 
    
    set match [regsub {.*cycle mean: ([0-9]*)[ ]*/[ ]*([0-9]*).*} \
		   $result {\1/\2} lyapu]
    set match [regsub {.*cycle mean: ([0-9]*)[ ]*/[ ]*([0-9]*).*} \
		   $result {\1.0/\2.0} fract]
    set result "Cycle time = $lyapu ([expr $fract])\n"

    LogResult "Analysis Results" "$result"
    
    if !$KEEP {
	RemoveFile $TMP_TGL_FILE
	RemoveFile $TMP_EGL_FILE
	RemoveFile $TMP_RES_FILE
	RemoveFile $TMP_MAT_FILE
    }
}

proc Critical_Paths { win } {
    global KEEP

    set TMP_TGL_FILE "/tmp/tged[pid].tgl"
    set TMP_RES_FILE "/tmp/tged[pid].res"
    
    if [Has_Program_Links] {
	ReportError "Analysis Error" \
	    "Cannot handle graphs\nwith program links (yet!)."
	return 
    } else {

	write_TGL $TMP_TGL_FILE 

	set seed [expr 60*[string trimleft [exec date +%H] 0]+[exec date +%S]]
	if [catch "exec tg_crit_path -f $TMP_TGL_FILE -paths -o $TMP_RES_FILE -seed $seed" problem] {
	    LogResult "Analysis Error" "" "$problem"
	}

	# getting the result in a list of "objects"
	set resfile [open $TMP_RES_FILE r]
	set objs {}
	while {[gets $resfile line] >= 0} {
	    if [expr { -1 == [string first "_" $line] } \
		    && { -1 == [string first "\#" $line] } \
		    && { 0 != [string length $line] } ] {
		regsub -all {\"} $line {} line
		regsub -all {task.|link.} $line {} line
		lappend objs $line
	    }
	}
	close $resfile 

	Highlight_Objects $win $objs

	switch [tk_dialog .graph.property "Highlighting" "Reset original colors now?" {} 0 {Yes} {No}] {
	    0 { UnHighlight_Objects $win $objs }
	    1 { return }
	}

	if !$KEEP {
	    RemoveFile $TMP_TGL_FILE
	    RemoveFile $TMP_RES_FILE
	}
    }
}

######################################################################
# Makespan -
# Interface procedure to tg_crit_path with the function of calculating
# the makespan (that is, a sample of it). 
#
# Parameters: none
#
proc Makespan { } {
    global KEEP

    set TMP_TGL_FILE "/tmp/tged[pid].tgl"
    set TMP_RES_FILE "/tmp/tged[pid].res"
    
    write_TGL $TMP_TGL_FILE 

    set seed [expr 60*[exec date +%H]+[exec date +%S]]

    if [catch "exec tg_crit_path -f $TMP_TGL_FILE -seed $seed > $TMP_RES_FILE" problem] {
	LogResult "Analysis Error" "" "$problem"
    }

    # getting the result in a string
    set resfile [open $TMP_RES_FILE r]
    set result [read $resfile]
    close $resfile

    LogResult "Critical Path Analysis" "$result"
}

proc Simulate { } {

    global sim_length gen_file_name

    if [catch {toplevel .params}] {return}

    wm title .params "Simulation Parameters"

    frame .params.simlength
    label .params.simlength.label -text "Simulation Length" \
	-width 15 -anchor w
    entry .params.simlength.field -textvar sim_length -width 8 \
	-relief sunken
    
    frame .params.actions
    button .params.actions.ok -text "Ok" \
	-command "destroy .params\n Run_Simulation \$sim_length"
    button .params.actions.cancel -text "Cancel" \
	-command "destroy .params"
    
    pack .params.simlength.label .params.simlength.field -side left \
	-fill x
    pack .params.actions.ok .params.actions.cancel -side left
    pack .params.simlength -side top -fill x
    pack .params.actions -side top

    bind .params <Return> {.params.actions.ok invoke}
    bind .params <Escape> {.params.actions.cancel invoke}
    pack .params.actions -side top

    grab .params
}

######################################################################
# Run_Simulation -
# Interface to the call of "tg_sim" which simulates the Task Graph.
#
# Parameters: 
# - sim_length	the number firings to compute
# - with	a flag telling whether sources should be included (value
#		"with). They are by default. 
#
proc Run_Simulation { sim_length } {
    global KEEP

    set TMP_TGL_FILE "/tmp/tged[pid].tgl"
    set TMP_RES_FILE "/tmp/tged[pid].res"

    write_TGL $TMP_TGL_FILE
    if [catch "exec tg_sim -f $TMP_TGL_FILE -tmax $sim_length > $TMP_RES_FILE" problem] {
	LogResult "Analysis Error" "" "$problem"
	return
    }

    set resfile [open $TMP_RES_FILE r]
    set result [read $resfile]
    close $resfile

    LogResult "Simulation Result" "$result"

    if !$KEEP {

	RemoveFile $TMP_TGL_FILE
	RemoveFile $TMP_RES_FILE
    }	
}
